###############################################################################
# autoPlayList Package 1.0
#
# Auto Playlist functions for snackAmp Player in Tcl/Tk
#
# Copyright (C) 2001 Tom Wilkason
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
#
# Please send any comments/bug reports to
# tom.wilkason@cox.net  (Tom Wilkason)
#
###############################################################################
set [file root [file tail [info script]]]_History {
   $Header: /cvsroot/snackamp/snackamp/lib/autoPlayList.tcl,v 1.61 2003/05/17 14:28:00 wilkason Exp $
}
package require snID3

namespace eval APL {
   variable folderList {}
   variable folderPos -1
   variable folderNext
   variable folderPrev
   variable icons
   variable Table
   variable TableCols
   variable paneOrient
   if {[info exists ::snackAmpSettings(folderList)]} {
     set  folderList $::snackAmpSettings(folderList)
   }

   set icons [list "Tree::Folder" "Tree::ROM"]
   set ::TA(Title,$::RMT) "/Removable Media"
   variable rmName $::TA(Title,$::RMT)
   option add *FolderTree*Font            treeFont
   ;# TFW: This linkage to other namespaces could be cleaned up
   foreach {col} $id3Tag::TableCols {
      if {[catch {set TableCols($col) $::snackAmpSettings(column,$col)}]} {
         set TableCols($col) 0
         set ::snackAmpSettings(column,$col) 0
      }
   } ;# end foreach
   ;# Trace variables used during settings when adding/removing
   ;# a column from the table.
   foreach {key} [array names ::snackAmpSettings column,*] {
      trace variable ::snackAmpSettings($key) w [list APL::rebuildTable .wFC]
   } ;# end foreach
}
###############################################################################
# Function   : APL::Title
# Description: Change the title to the last file
# Author     : Tom Wilkason
###############################################################################
proc APL::Title {args} {
   catch {
      wm title .wFC "SnackAmp Auto-Playlist Manager - [file root [file tail $::snackAmpSettings(lastFolderFile)]]"
   }
}
###############################################################################
# Function   : saveSize
# Description: Save the APL window sizes
# Author     : Tom Wilkason
# Date       : 1/9/2001
###############################################################################
proc APL::saveSize {args} {
   if {![catch {winfo toplevel $::TREE} tl]} then {
      set ::snackAmpSettings(treeSize) [wm geometry $tl]
      catch {set ::snackAmpSettings(FileListSize) [lindex [pane $tl.sv.lb -size] 0]}
   }
}
###############################################################################
# Function   : folderTree
# Description: Toss up a folder tree linked to a variable
# Author     : Tom Wilkason
# Date       : 1/9/2001
###############################################################################
proc APL::FolderTree {args} {
   package require saTree
   global TA snackAmpSettings
   variable folderNext
   variable folderPrev
   variable paneOrient
   set W .wFC
   if {[winfo exists $W]} {
      wm deiconify $W
      #raise $W      ;# This causes a noticible delay on KDE
      focus $W
      return
   }

   toplevel $W -cursor $::TA(arrow) -class FolderTree
   trace variable ::snackAmpSettings(lastFolderFile) w APL::Title
   APL::Title
   wm protocol $W WM_DELETE_WINDOW {
      APL::close
   }

   ;##
   ;# Button bar
   ;#
   if {$::tcl_platform(platform)=="windows"} {
      frame $W.mb -relief groove -bd 2
   } else {
      frame $W.mb
   }
   pack $W.mb -side top -fill x -pady 3

   ;# Save Button
   pack [button $W.mb.save -image Save -command {macros::saveCurrentAPL 0}]\
      -padx 3 -side left
   balloon::button $W.mb.save "Save the current settings"

   ;# Save As button
   pack [button $W.mb.saveas -image SaveAs -command {macros::SaveFolderFile}]\
    -padx 3 -side left
   balloon::button $W.mb.saveas "Save Auto-Playlist under a new file name"

   ;# Apply button
   pack [button $W.mb.apply -image Apply -command {macros::applyCurrentAPL 0}]\
     -padx 3 -side left
   balloon::button $W.mb.apply "Update the playlist with the current settings"

   menubutton $W.mb.files -relief flat -image apllist -menu $W.mb.files.list

   ;# Refresh button
   pack [button $W.mb.refresh -image refresh -command {aplLB::UpdateList $::TA(highlightPath) 1}]\
     -padx 3 -side left
   balloon::button $W.mb.refresh "Refresh the file view"

   ;# Filelist menu
   pack [button $W.mb.togpl -image FileList \
      -command "APL::setFileListToShow $W.sv.lb $W.mb.togpl"] \
      -padx 3 -side left
   balloon::button $W.mb.togpl "Depress to show files under the selected folder"

   ;# Initially, we toggle the setting to start the statemachine
   set ::snackAmpSettings(hideFileList) [expr {! ($::snackAmpSettings(hideFileList)&0x1)}]

   ;# Play button
   pack [button $W.mb.play -image noatun2-16 -command {aplLB::playSelectedFiles}] \
      -padx 3 -side left
   balloon::button $W.mb.play "Play Selection(s)"

   ;# Level Buttons
   foreach {level} {0 1 2 3 4 5} {
      pack [radiobutton $W.mb.level$level -width 2 -text [expr {$level+1}] -value $level \
         -font menu -padx 0 -pady 0 -bd 1 -variable ::snackAmpSettings(treeDepth) -indicatoron 0 -relief flat \
         -command {trackList::populateHighighted 1 1}] \
         -padx 0 -side left
      balloon::button $W.mb.level$level "Show an additional $level levels below the selected folder"
   } ;# end foreach

   ;# Edit cell button
   pack [button $W.mb.edit -image edit-16 \
      -command trackList::changeEdit] \
      -padx 3 -side left
   balloon::button $W.mb.edit "Toggle Cell Editing"
   set trackList::editCellButton $W.mb.edit

   ;# Back button
   pack [button $W.mb.prev -image NavBack -command APL::prevSelection] \
      -padx 3 -side left
   balloon::button $W.mb.prev "Change to prior selected folder"


   ;# Dropdown history button (redundant with menu)
   pack [menubutton $W.mb.prevHistory -relief flat -image dropdownImage -menu $W.mb.prevHistory.list] -side left
   balloon::button $W.mb.prevHistory  "List of previous locations"
   ;# History menu, one for button and one for RMB
   menu $W.mb.prevHistory.list -font menu -postcommand [list APL::cbHistoryMenu $W.mb.prevHistory.list]

   ;# Forward Button
   pack [button $W.mb.next -image NavForward -command APL::nextSelection] \
      -padx 3 -side left
   balloon::button $W.mb.next "Change to next selected folder"

   ;##
   ;# Find next entry
   ;#
   pack [button $W.mb.findFile -image findfile \
      -command {APL::findNextSong $::TA(SAFind)}] -side left
   balloon::button $W.mb.findFile "Press to search for next file containing text to the right."

   pack [button $W.mb.findLab -image findfolder \
      -command {APL::findNextFolder $::TA(SAFind)}] -side left
   balloon::button $W.mb.findLab "Press to search for next folder containing text to the right."

   set ::SAFind [entry $W.mb.find -textvariable ::TA(SAFind) -width 10 -font entry \
   -background $::snackAmpSettings(ebg) -foreground $::snackAmpSettings(efg)]
   pack $::SAFind -side left -expand yes -fill x
   settings::regColor $::SAFind -background ::snackAmpSettings(ebg) -foreground ::snackAmpSettings(efg)

   pack [button $W.mb.ckfilter -image filter \
      -command [list APL::toggleFilter $W.mb.ckfilter]] \
      -padx 3 -side left
   balloon::button $W.mb.ckfilter "Depress to restrict files to those containing text to the right"

   set ::SAFilter [entry $W.mb.filter -textvariable ::TA(SAFilter) -width 10 -font entry \
   -background $::snackAmpSettings(ebg) -foreground $::snackAmpSettings(efg)]
   pack $::SAFilter -side left -expand yes -fill x
   settings::regColor $::SAFilter -background ::snackAmpSettings(ebg) -foreground ::snackAmpSettings(efg)

   balloon::define $::SAFind "Enter text then hit return to search for next folder with matching text. Enter Ctrl-Return to search for next song with matching text. You may use wildcards for searching."
   balloon::define $::SAFilter "Enter text then depress the filter button to the left to restrict track names that contain this text."
   ;# Save the button names
   set folderNext $W.mb.next
   set folderPrev $W.mb.prev

   set Folders [menu $W.mb.files.list -postcommand "[list APLFiles $W.mb.files.list]"  -font menu]
   set Removal [menu $W.mb.files.list.remove -tearoff 0 -postcommand "[list APLFilesRemove $W.mb.files.list.remove]"  -font menu]

   $Folders add command -label "Save Auto-Playlist As..." -command SaveFolderFile
   $Folders add cascade -label "Remove..." -menu $Removal
   $Folders add separator

#   pack [button $W.mb.close -image Close -font menu -command  {APL::close} \
#      -borderwidth 1] -side left
#   balloon::button $W.mb.close "Close this window"

   ;##
   ;# Label and File List check button


   ;##
   ;# Frame to hold the tree/filelist
   ;#
   pack [frame $W.sv -bd 2] -side top -fill both -expand true
   ;#
   set TA(ck,selected)   "Checked: 0"
   set TA(ck,unselected) "Unchecked: 0"
   set TA(ck,total)      "Total: 0"
   set TA(ck,totalFolders)  "Files: 0"
   set TA(ck,totalTime)  "Time: 00:00:00"
   set TA(ck,totalSize)  "Size: 0"

   pack [frame $W.file -relief flat] -fill x
   set i 0
   ;# TFW: Can't get grid to force them to the right like I would think -sticky e would
   set statusWidgets {
      selected   12 left
      unselected 14 left
      total      10 left
   }
   lappend statusWidgets totalFolders 10 right totalSize 16 right totalTime 14 right
   foreach {item wid where} $statusWidgets {
      pack [label $W.file.[incr i] \
         -textvariable TA(ck,$item) -width $wid  -bd 1 -font menu -relief sunken -padx 2 -pady 1] \
         -side $where -padx 1
   } ;# end foreach

   ;##
   ;# Frame to hold the tree and scrollbars, pane packs it
   ;#
   frame $W.sv.tree -bd 1 -relief sunken ;# "placed" by pane'r

   set ::TREE $W.sv.tree.w  ;# Canvas
   scrollbar $W.sv.tree.sb -orient vertical -command [list $::TREE yview]

   pack $W.sv.tree.sb -side right -fill y
   ;##
   ;# Build the guts of the tree (menus are created here)
   ;#
   APL::buildAPL $::TREE
   mscroll $::TREE

   ;##
   ;# File listbox/Table
   ;#
   set ::PLB [trackList::buildTable $W.sv.lb ::snackAmpSettings]
   ;##
   ;# Create the appropriate bindings
   ;#
   aplLB::bindings
   bindings::assignHotKeys
   APL::treeBind
   ;# Keys for the APL search box
   APLFilterAssign <Return> {
      APL::findNextFolder $::TA(SAFind)
   }
   APLFilterAssign <Control-Return> {
      APL::findNextSong $::TA(SAFind)
   }
   ;##
   ;# Top Menus
   ;#
   APL::aplMenu $W
   ;##
   ;# Pane them
   ;#
   package require saPane
   pane $W.sv.tree -orient $paneOrient -dynamic 0
   APL::setFileListToShow $W.sv.lb $W.mb.togpl
   ;##
   ;# Table/Listbox Menus
   ;#
   aplLB::menus
   ;# Expand initially to folder level on menu
   if {[info exists ::snackAmpSettings(treeSize)]} {
      wm geometry $W $::snackAmpSettings(treeSize)
   } ;# end if
   ;# Open from the top 1 layer down
   Tree::open $::TREE / 1
   focus $W
   update
   myIcon $W
}
###############################################################################
# Function   : APL::rebuildTable
# Description: Rebuild the table when snackAmpSettings columns change, called
#              from a trace writing to one of the column settings.
#              Rebuild if the col changed from the current value.
# Author     : Tom Wilkason
# Date       : 5/27/2002
###############################################################################
proc APL::rebuildTable {W _variable index mode} {
   Trace
   if {[catch {
      variable paneOrient
      upvar #0 $_variable settings
      set what $W.sv.lb
      set wasOpen [winfo ismapped $what]
      foreach {xx col} [split $index ,] {break} ;# end foreach
      if {[info exists APL::TableCols($col)] && ($::snackAmpSettings($index) != $APL::TableCols($col)) && [winfo exists $what]} {
         #puts "rebuildTable Trace $index:"
         set APL::TableCols($col) $::snackAmpSettings($index)
         #foreach {x y X Y} [split [wm geometry .wFC] +x] {break}
         set geom [wm geometry $W]
         set ::snackAmpSettings(FileListSize) [lindex [pane $what -size] 0]
         destroy $what
         set trackList::TABLEDATA [list]
         set ::PLB [trackList::buildTable $what ::snackAmpSettings]
         if {$trackList::editCellState} {
            trackList::changeEdit
         } ;# end if
         ;##
         ;# Table/Listbox Menus
         ;#
         aplLB::menus
         ;##
         ;# Create the appropriate bindings
         ;#
         aplLB::bindings
         if {$wasOpen} {
            pane $what -orient $paneOrient -dynamic 0 \
               -size $::snackAmpSettings(FileListSize)
            aplLB::UpdateList $::TA(highlightPath) 1
         } ;# end if
         wm geometry $W $geom
      } ;# end if
   } result] } then {
      saLog "$::errorInfo"
   }
}
###############################################################################
# Function   : APL::close
# Description: Close the window asking to save changes if needed
# Author     : Tom Wilkason
###############################################################################
proc APL::close {args} {
   $::TA(treeButton) configure -relief flat
   if {$::TA(APLDirty)} {
      if {[tk_messageBox -type yesno -icon question -message "Do you want to save your current Auto-playlist?"] == "yes"} {
         saveCurrentAPL 0
      }
   }
   wm withdraw .wFC
   APL::saveSize
   catalog::SaveDirtyCache
   trackList::saveColumnWidths
   #destroy .wFC
}
###############################################################################
# Function   : APL::cbDndPLBToTree
# Description: Callback on drag-n-drop target from PLB to Tree
# Author     : Tom Wilkason
###############################################################################
proc APL::cbDndPLBToTree {X Y} {
   Trace
   set x [expr {$X - [winfo rootx $::TREE]}]
   set y [expr {$Y - [winfo rooty $::TREE]}]
   set lbl [Tree::labelat $::TREE $x $y]
   if {$lbl != {}} {
      if {![catch {APL::GetTreeFolder $::TREE $lbl} folderName]} {
         ;# We let go over a node, ask to copy files
         if {[catalog::moveClipboardFiles $folderName] > 0} {
            catch {
               aplLB::UpdateList $::TA(highlightPath) [winfo ismapped $::PLB]
               APL::treeSelect $::TA(highlightPath)
            }
         }
      }
   } ;# end if
}
###############################################################################
# Function   : APL::cbDndPLBToLB
# Description: Callback on drag-n-drop target from PLB to LB
# Author     : Tom Wilkason
###############################################################################
proc APL::cbDndPLBToLB {x y} {
   Trace
   if {$x>0} {
      set y [expr {$y - [winfo rooty $::LB]}]
      set new [$::LB nearest $y]
   } else {
      set new end
   }
   if {[catch {selection get -selection CLIPBOARD} tryFiles]} {
      return
   }
   clipboard clear
   set files [list]
   ;# Ignore errors due to funky clipboard data
   foreach {file} $tryFiles {
      if {[lsearch -exact $files $file] < 0} {
         set keyName [LB::getListName $file]
         lappend files $keyName
         set ::TF($keyName) $file
      } ;# end if
   } ;# end foreach
   LB::insertList $new $files
   updateFileDisplay
}
###############################################################################
# Function   : APL::cbDndTreeToClipboard
# Description: Callback on drag-n-drop target from Tree to PlayList
# Author     : Tom Wilkason
###############################################################################
proc APL::cbDndTreeToClipboard {X Y} {
   Trace
   set src $::TA(aplDNDSource)
   clipboard clear
   set data [list]
   foreach {entry} [catalog::filesMatchingPattern "$src*"] {
      lappend data $entry
   } ;# end foreach
   clipboard append [lsort -dictionary $data]
}
###############################################################################
# Function   : APL::cbDndTreeToTree
# Description: Callback when moving folders in the APL tree
# Author     : Tom Wilkason
# Date       : 6/3/2002
###############################################################################
proc APL::cbDndTreeToTree {W lbl} {
   if {![catch {APL::GetTreeFolder $W $lbl} folderName]} {
      set newName [file join $folderName [file tail $::TA(aplDNDSource)]]
      if {![string equal $::TA(aplDNDSource) $folderName] && ![string equal $::TA(aplDNDSource) $newName]} {
         #puts "Release: $::TA(aplDNDSource) over $folderName"
         set ans [tk_messageBox -type yesno -icon question \
            -message "Are you sure you want to move \n'$::TA(aplDNDSource)'\n  to\n'$newName'"]
         if {$ans=="yes"} {
           if {[catch {
              catalog::moveFolder $::TA(aplDNDSource) $newName
              if {$::snackAmpSettings(dndExpand)} {
                 ;# Highlight and expand the new target
                 APL::treeOpen [file dirname $newName]
                 APL::treeSelect $newName
                 APL::viewHighlighted $::TREE
              } ;# end if
           } result] \
           } then {
              tk_messageBox -type ok -icon warning \
                 -message "Failed to move $::TA(aplDNDSource) under $folderName\n$result\n$::errorInfo"
           }
        } else {
           APL::treeSelect $::TA(aplDNDSource)
        }
        set ::TA(aplDND) 0
      } ;# end if
   }
}
###############################################################################
# Function   : APL::findNextSong
# Description: Highlight the next song in tree with matching text
# Author     : Tom Wilkason
# Date       : 3/28/2002
###############################################################################
proc APL::findNextSong {text} {
   if {[catch {aplLB::getActiveFullName} fullName]} {
      set pos 0
   } else {
      ;# Locate where song is now, and search from there
      set pos [catalog::whereInDb $fullName]
      if {$pos < 0} {
         set pos 0
      } ;# end if
   }
   ;# Search from current place down
   foreach {file} [lrange [catalog::sortedList] [incr pos] end]  {
      if {[string match -nocase "*$text*" [file tail $file]]} {
         SelectFile $file 1
         return 1
      } ;# end if
   } ;# end foreach
   ;##
   ;# Didn't find from current pos, start from top (wrap)
   ;#
   foreach {file} [lrange [catalog::sortedList] 0 $pos] {
      if {[string match -nocase "*$text*" [file tail $file]]} {
         SelectFile $file 1
         return 1
      } ;# end if
   } ;# end foreach
}
###############################################################################
# Function   : APL::positionFolder
# Description: Bound to cursor keys, this moves the selection within the tree
# Author     : Tom Wilkason
# Date       : 3/28/2002
###############################################################################
proc APL::positionFolder {how} {
   set folders [lsort [array names ::teFolders]]
   set current $::TA(highlightPath)
   set currentLen [llength [file split $current]]
   if {[catch {lsearch $folders $::TA(highlightPath)} pos]} {
      set pos 0
   }
   switch -- $how {
      prev {
         set new [lindex $folders [incr pos -1]]
         SelectFile $new 1
      }
      next {
         set new [lindex $folders [incr pos 1]]
         SelectFile $new 1
      }
      left {
         ;# Go up one
         SelectFile [file dirname $current] 1
      }
      right {
         set new [lindex $folders [incr pos 1]]
         SelectFile $new 1
      }
      default {
      }
   } ;# End Switch
}
###############################################################################
# Function   : APL::findNextFolder
# Description: Highlight the next folder in tree with matching text
# Author     : Tom Wilkason
# Date       : 3/28/2002
###############################################################################
proc APL::findNextFolder {text} {
   set folders [lsort [array names ::teFolders]]
   if {[catch {lsearch -exact $folders $::TA(highlightPath)} pos]} {
      set pos 0
   }
   ;# Now search from current place
   foreach {folder} [lrange $folders [incr pos] end] {
      if {[string match -nocase "*$text*" [file tail $folder]]} {
         SelectFile $folder 1
         #puts "Found $folder"
         return 1
      } ;# end if
   } ;# end foreach
   ;##
   ;# Didn't find from current pos, start from top (wrap)
   ;#
   foreach {folder} [lrange $folders 0 $pos] {
      if {[string match -nocase "*$text*" [file tail $folder]]} {
         SelectFile $folder 1
         #puts "Found/wrap $folder"
         return 1
      } ;# end if
   } ;# end foreach
}
###############################################################################
# Function   : APL::prevSelection
# Description: Move to the last know selected folder
# Author     : Tom Wilkason
# Date       : 3/28/2002
###############################################################################
proc APL::prevSelection {args} {
   variable folderList
   variable folderPos
   variable folderNext
   variable folderPrev
   if {$folderPos > 0} {
      incr folderPos -1
      set folder [lindex $folderList $folderPos]
      SelectFile $folder 0
      #puts "set $folder at $folderPos"
   }
   APL::setNavButtons $folderNext $folderPrev
}
###############################################################################
# Function   : APL::prevSelection
# Description: Move to the next know selected folder
# Author     : Tom Wilkason
# Date       : 3/28/2002
###############################################################################
proc APL::nextSelection {args} {
   variable folderList
   variable folderPos
   variable folderNext
   variable folderPrev
   if {($folderPos+1)  < [llength $folderList] && $folderPos >= 0} {
      incr folderPos 1
      set folder [lindex $folderList $folderPos]
      SelectFile $folder 0
      #puts "set $folder at $folderPos"
   }
   APL::setNavButtons $folderNext $folderPrev
}
###############################################################################
# Function   : insertUnique
# Description: Insert an item in the of a list if it is unique
#              folderPos points to location in list of current selection
#              if list is empty, it will be -1 (initial condition)
#              When called, the next entry will be inserted into the list
#              after the current item and the pointer incremented.
#              This only will occur if the item at the current position
#              does not equal the item to be inserted
# Author     : Tom Wilkason
# Date       : 3/28/2002
###############################################################################
proc APL::insertUnique {item} {
   variable folderList
   variable folderPos
   variable folderNext
   variable folderPrev
   set curItem [lindex $folderList $folderPos]
   set pos [lsearch $folderList $item]
   if {![string equal $curItem $item]} {
      incr folderPos  ;# Point to next item
      set folderList [linsert $folderList $folderPos $item]
   } else {
      set folderPos $pos
      #puts "\nalready have : $folderPos in $folderList"
   }
   APL::setNavButtons $folderNext $folderPrev
}
###############################################################################
# Function   : APL::historyMenu
# Description: Display a history menu
# Author     : Tom Wilkason
# Date       : 3/5/2002
###############################################################################
proc APL::cbHistoryMenu {menu} {
   variable folderList
   variable folderPos
   variable folderNext
   variable folderPrev

   $menu delete 0 end
   ;# Clamp menu size to 20 high
   set mh 20
   set pos $APL::folderPos
   set end [expr {[llength $folderList]-1}]
   set min [expr {$pos - $mh/2}]
   ;# If closer than mh/2 from end, offset min back by delta
   if {($end-$pos) < $mh/2} {
      incr min [expr {($end-$pos) - $mh/2}]
   }
   ;# Clamp min
   if {$end < $mh || $min < 0} {
      set min 0
   }
   ;# Display up to $mh menus
   set i 0
   set cnt 0
   foreach item $folderList {
      if {$i >= $min && $cnt <= $mh} {
         $menu add radiobutton -variable APL::folderPos -value $i \
            -label "$i  $item" -command [list APL::positionHistory $i]
         incr cnt
      }
      incr i
   } ;# end foreach
}
###############################################################################
# Function   : APL::positionHistory
# Description: Position the folder at some location and update the buttons
# Author     : Tom Wilkason
# Date       : 3/5/2002
###############################################################################
proc APL::positionHistory {location} {
   variable folderList
   variable folderPos
   variable folderNext
   variable folderPrev
   set folder [lindex $folderList $folderPos]
   SelectFile $folder 0
   APL::setNavButtons $folderNext $folderPrev

}
###############################################################################
# Function   : APL::setNavButtons
# Description: Set the state of the navigation buttons
# Author     : Tom Wilkason
# Date       : 3/5/2002
###############################################################################
proc APL::setNavButtons {next prev} {
   variable folderList
   variable folderPos
   ;# Configure previous button
   if {$folderPos > 0} {
      $prev config -state normal
   } else {
      $prev config -state disabled
   }
   ;# configuration next button
   if {$folderPos < ([llength $folderList]-1)} {
      $next config -state normal
   } else {
      $next config -state disabled
   }
}
###############################################################################
# Function   : APL::popWatch
# Description: Tempoaraly put a busy watch up
# Author     : Tom Wilkason
# Date       : 3/5/2002
###############################################################################
proc APL::popWatch {{time 300}} {
   APL::busyCursor 1
   update idletasks
   after 250 APL::busyCursor 0
}
;##
;# Toggle the filter button and update the track listing
;#
proc APL::toggleFilter {W} {
   set state [$W cget -relief]
   switch -- $state {
      sunken {
         $W configure -relief raised
         set ::TA(useSAFilter) 0
      }
      raised {
         $W configure -relief sunken
         set ::TA(useSAFilter) 1
      }
   } ;# End Switch
   aplLB::UpdateList $::TA(highlightPath) 1
}
###############################################################################
# Function   : APL::aplMenu
# Description: Create the main menu for the APL Window, reuse
#              pop-up menus for the tree and table.
# Author     : Tom Wilkason
# Date       : 5/27/2002
###############################################################################
proc APL::aplMenu {W} {
   set M $W.aplMenu
   bindings::init $W menuBindings
   menu $M -font menu
   # attach it to the main window
   $W config -menu $M
   # Create more cascade menus
   foreach m {File Edit Folders Tracks AutoPlaylists Columns Help} {
      switch -- $m {
         Folders {
            ;# Use existing menu
            $M add cascade -label $m -menu $::TREE.popup
         }
         Tracks {
            ;# Use existing menu
            $M add cascade -label $m -menu $::PLB.popup
         }
         AutoPlaylists {
            ;# Use existing menu
            $M add cascade -label $m -menu $W.mb.files.list
         }
         Columns {
            set $m [menu $M.m$m -tearoff 1 -font menu]
            $M add cascade -label $m -menu $M.m$m

            foreach {item} [lrange $id3Tag::TableCols 1 end] {
               $Columns add checkbutton -label $item -variable ::snackAmpSettings(column,$item)
            } ;# end foreach
         }
         Subfolders {
            set $m [menu $M.m$m -tearoff 0 -font menu]
            $M add cascade -label $m -menu $M.m$m
            foreach {subfolders} {0 1 2 3 4 5 6 7 8} {
               $Subfolders add radiobutton -label "Show [expr {$subfolders+1}] levels" -value $subfolders -variable ::snackAmpSettings(treeDepth)
            } ;# end foreach
         }
         History {
            ;# Use existing menu
            $M add cascade -label $m -menu $W.mb.prevHistory.list
         }
         default {
            set $m [menu $M.m$m -tearoff 0 -font menu]
            $M add cascade -label $m -menu $M.m$m
         }
      } ;# End Switch
   } ;#end foreach
   ;##
   ;# Here is how to add a menu to the "system" menu
   ;# Use it to put "pinned" command
   ;#
   ontop::pinOnSystemMenu $M .wFC
   ;##
   ;# File Menu
   ;#
   bindings::addCommand $File "Catalog Folders" {catalog::RescanAskFirst}
   bindings::addCommand $File "Read Catalogs" {catalog::ReReadCatalogs}
   $File add separator
   bindings::addCommand $File "Catalog Removable Media" catalog::CatalogMedia
   bindings::addCommand $File "Mount Removable Media"  {catalog::MountMedia "Removable"}
   $File add separator
   bindings::addCommand $File "Manage Media Folders..."  {catalog::manageMediaFolders}
   bindings::addCommand $File "Manage Removable Folders..." {
      package require saSettings
      folderManager::launch RemovableFolders
   }
   $File add separator
   bindings::addCommand $File "Save Auto-Playlist" {saveCurrentAPL 0}
   bindings::addCommand $File "Save Auto-Playlist As..." {SaveFolderFile}
   bindings::addCommand $File "Apply Auto-Playlist" {applyCurrentAPL 0}
   $File add separator
   bindings::addCommand $File "Sync to Playing" {SelectFile 1}
   bindings::addCommand $File "Sync to Selected" {SelectFile 0}
   $File add separator
   bindings::addCommand $File "Close" {APL::close}
   bindings::addCommand $Edit "Copy" {APL::cut}
   bindings::addCommand $Edit "Paste" {APL::PasteFolderToTree}
   $Edit add separator
   $Edit add checkbutton -label "Edit Cell Mode" -variable trackList::editCellCheckState \
      -command {trackList::changeEdit}
   $Help add command -label "About" -command \
   {
      tk_messageBox -type ok -icon info -title "Help" -message \
"The Autoplaylist Manager allow allows you to set/save/restore preferred folders.
You can also rename, remove and move files. With the mouse you can:
- Select a folder to include (checked) or exclude (unchecked)
- Control Click: Check selected and all subfolders
- Alt Click: Uncheck selected and all subfolders
- Double Click: Toggle expand/collapse selected folder
- Right MB Click: Menu allowing you do above actions, apply and/or
  save the settings.
   - Rescan folder, sync internal list with any changes that has taken place
     on the disk.
   - Add folder to playlist, copy all of this folders files to the main
     playlist.
   - Move files copied to clipboard to highlighted folder
   - Create a new folder or rename and existing folder
   - Remove a folder and all of its subfolders and files.
- Using drag and drop you can move an entire tree limb from one
  folder to another.

From the menu and menu bar you can:
- Apply the settings to the playlist (does not save to file)
- Save, applies and saves the settings to the current Auto-playlist file
- Save as, save the settings to a new file name and make it current.
- File List checkbox, toggle a view of the files contained within the
  selected folder at the bottom.
- Close the window

You can drag highlighted files from the playlist to a folder
in the tree view to move them.
"
   }

}
proc APL::SelectedFolderRenameInPlace {{select 0}} {
   set W $::TREE
   ;# Note access into tree global here
   set tag [Tree::activetag $::TREE]
   if {[string length $tag] == 0} {
      return
   }
   set w  .canRndialog
   destroy $w
   toplevel $w
   set oldName [file tail $::TA(highlightPath)]
   set folder $::TA(highlightPath)
   set base [file dirname $folder]
   foreach {x y - -} [$W bbox $tag] {break}
   ;# Need to account for scrolled region
   ;# bu subtracting out canvas displayed 0 value
   set rx [winfo rootx $W]
   set ry [winfo rooty $W]
   set X [expr {$rx+$x-round([$W canvasx 0])}]
   set Y [expr {$ry+$y-round([$W canvasy 0])}]
   set font [$W itemcget $tag -font]
   set bg [$W cget -background]
   set fg $Tree::fg
   set ::TA(rename) $oldName
   set wid [expr {[string length $oldName]+5}]
   pack [entry $w.rn -bd 2 -relief groove -font $font -insertbackground $fg \
      -width $wid -bg $bg -fg $fg -textvariable ::TA(rename)]
   wm withdraw $w
   wm overrideredirect $w 1
   wm transient $w [winfo toplevel [focus]]
   if {$select} {
      $w.rn selection range 0 end
   }
   focus -force $w.rn
   # with it hiden, let the layout take place so we know its geometry
   update idletasks
   wm geometry $w "+$X+$Y"
   bind $w <FocusOut> {
      destroy .canRndialog
      set ::TA(rename) ""
   }
   bind $w <Escape> {
      destroy .canRndialog
      set ::TA(rename) ""
   }
   bind $w <Return> [list destroy .canRndialog]
   wm deiconify $w
   ;# Force on top if possible
   update idletasks
   if {$ontop::enabled && [ontop::state $W]} {
      ontop::control $w 1
   }
   ;##
   ;# Wait for window to close
   ;#
   tkwait window $w
   ;##
   ;# Back from window closure
   ;#
   if {[string length $::TA(rename)]>0 && $::TA(rename) != $oldName} {
      #puts "Renaming '$oldName' to '$::TA(rename)'"
      if {[catch {
         set newFullName [file join $base $::TA(rename)]
         catalog::moveFolder $folder $newFullName
         APL::treeSelect $newFullName
         APL::viewHighlighted $::TREE
         catalog::FlagDirtyCache $newFullName
         set ::TA(highlightPath) $newFullName
         set res 1
      } result] \
      } then {
         tk_messageBox -type ok -icon warning -message "Could not rename folder \n$folder to \n$newFullName\n$result"
         set res 0
      }
   }
}
###############################################################################
# Function   : APL::treeBind
# Description: Add bindings to the autoplay list tree
# Author     : Tom Wilkason
# Date       : 12/23/2001
###############################################################################
proc APL::treeBind {args} {
   Trace
   ;##
   ;# Pop-up menu, save some data for use by callbacks
   ;#
   bind $::TREE <ButtonPress-3> {
      set lbl [Tree::labelat %W %x %y]
      Tree::setselection %W $lbl
      if {$lbl != ""} {
         if {[catch {APL::GetTreeFolder %W $lbl} folderName]} {
            return
         }
         APL::insertUnique $folderName
         set TA(highlightPath) $folderName
         #set TA(highlightEntry) $lbl
         #set TA(treeID) %W
         tk_popup %W.popup %X %Y
      }
      catch {unset lbl}
   }
   set ::TA(aplDND) 0
   ;##
   ;# Auto scroll the canvas on a drag
   ;#
#   bind [winfo parent $::TREE] <B1-Leave> {
#      puts "Leaving"
#      set TA(cy) %y
#      APL::canvasScroll $::TREE
#   }
#   bind $::TREE <B1-Leave> {
#      puts "Leaving Tree %y"
#      set TA(cy) %y
#      APL::canvasScroll $::TREE
#   }
#   bind $::TREE <ButtonRelease-1> {
#      catch {after cancel $::TA(afterID)}
#   }
   ;##
   ;# During a drag, highlight the current folder under the cursor
   ;#
   APL_TreeNodeAssign <B1-Motion> {
      set lbl [Tree::labelat %W %x %y]
      if {![catch {APL::GetTreeFolder %W $lbl} folderName]} {
         if {![string match $folderName $::TA(aplDragTarget)]} {
            Tree::setselection %W $lbl
            set ::TA(aplDragTarget) $folderName
         }
      }
      $::TREE config -cursor hand2
      set ::TA(aplDND) 1
      catch {unset lbl}
   }
   ;##
   ;# macros::Highlight the current folder, save away the selected folder
   ;#
   APL_TreeNodeAssign <ButtonPress-1> {
      catch {
         focus [winfo parent %W]
         set lbl [Tree::labelat %W %x %y]
         Tree::setselection %W $lbl
         if {![catch {APL::GetTreeFolder %W $lbl} folderName]} {
            set ::TA(highlightPath) $folderName
            set ::TA(aplDNDSource) $folderName
            set ::TA(aplDragTarget) $folderName
            APL::insertUnique $folderName
            #puts "Press: over $folderName"
            after idle [list aplLB::UpdateList $folderName [winfo ismapped $::PLB]]
         }
         ;# Update file listbox if it exists.
      }
   }
   ;##
   ;# Used for DND drop
   ;#
   APL_TreeNodeAssign <ButtonRelease-1> {
      APL::busyCursor 0
      set lbl [Tree::labelat %W %x %y]
      set targetWin [winfo containing %X %Y]
      if {[winfo exists $targetWin] && $::TA(aplDND)} {
         APL::busyCursor 1
         if {$targetWin == $::TREE} {   ;# move the folder
            APL::cbDndTreeToTree %W $lbl
         } elseif {$targetWin == $::LB} {
            APL::cbDndTreeToClipboard %X %Y   ;# get to clipboard
            APL::cbDndPLBToLB %X %Y    ;# now works like PLB to LB
         } elseif {[winfo toplevel $targetWin] == [winfo toplevel $::LB]} {
            ;# Improve to append to bottom here
            APL::cbDndTreeToClipboard %X %Y   ;# get to clipboard
            APL::cbDndPLBToLB 0 0    ;# now works like PLB to LB
         }
         APL::busyCursor 0
      }
      catch {unset lbl}
   }
   ;##
   ;# Paste Files/folders (not on canvas)
   ;#
   APL_TreeAssign <<Paste>> {APL::PasteFolderToTree}
   ;##
   ;# filtering::Append selected folder to clipboard
   ;#
   APL_TreeAssign <<Copy>> {APL::popWatch;clipboard append "[list $TA(highlightPath)] "}
   APL_TreeAssign <<Cut>> {APL::popWatch;clipboard append "[list $TA(highlightPath)] "}
   ;##
   ;# Cursor movement keys
   ;#
   APL_TreeAssign <Left> {APL::positionFolder left}
   APL_TreeAssign <Right> {APL::positionFolder right}
   APL_TreeAssign <Up>   {APL::positionFolder prev}
   APL_TreeAssign <Down> {APL::positionFolder next}
   ;##
   ;# Assign hotkey to rename
   ;#
   APL_TreeAssign <F2> {APL::SelectedFolderRenameInPlace}
   APL_TreeAssign <Delete> {::catalog::removeFolder $::TA(highlightPath)}
   ;# Toggle open/close on Return key
   APL_TreeAssign <Return> {Tree::toggleExpand $::TREE [APL::pathToTreeName $::TA(highlightPath)]}
   APL_ListAssign <Escape> {
      if {$trackList::editCellState} {
         trackList::changeEdit
      } ;# end if
   }
   #APL_TreeAssign <F2> {::catalog::renameFolder $::TA(highlightPath)}
   ;##
   ;# Set this and all child checkboxes
   ;#
   APL_TreeNodeAssign <Control-1> {
      set lbl [Tree::labelat %W %x %y]
      Tree::setselection %W $lbl
      set ::TA(APLDirty) 1
      if {![catch {APL::GetTreeFolder %W $lbl} folderName]} {
         APL::setSomeFolders $folderName 1
      }
      APL::insertUnique $folderName
   }
   ;##
   ;# Clear this and all child checkboxes
   ;#
   APL_TreeNodeAssign <Alt-1> {
      set lbl [Tree::labelat %W %x %y]
      set ::TA(APLDirty) 1
      Tree::setselection %W $lbl
      if {![catch {APL::GetTreeFolder %W $lbl} folderName]} {
         APL::setSomeFolders $folderName 0
      }
      APL::insertUnique $folderName
   }
   ;##
   ;# Recurse expand on double click
   ;#
   APL_TreeNodeAssign <Double-1> {
      APL::busyCursor 1
      Tree::toggleExpand %W [Tree::labelat %W %x %y] 10
      APL::busyCursor 0
   }
   ;##
   ;# Now build the pop-up menu for the tree
   ;#
   set m $::TREE.popup
   menu $m -tearoff 0 -font menu
   set P [winfo parent $::TREE]
   bindings::init $P menuBindings
   bindings::addCommand $m "Create Folder" {APL::createFolder $TA(highlightPath)}
   bindings::addCommand $m "Create Stream Shortcut" {
      if {[playurl::checkuskfolder]} {
         playurl::addPromptUrl $TA(highlightPath)
         aplLB::UpdateList $::TA(highlightPath) 1
      }
   }

   #bindings::addCommand $W $File
   bindings::addCommand $m "Rename Folder" {APL::SelectedFolderRenameInPlace}
   bindings::addCommand $m "Remove Folder" {::catalog::removeFolder $::TA(highlightPath)}
   $m add separator
   bindings::addCommand $m "Move Clipboard Files Here" {APL::PasteFolderToTree}
   bindings::addCommand $m "Copy Folder to Clipboard" \
      {APL::popWatch;clipboard append "[list $TA(highlightPath)] "}
   $m add separator
   bindings::addCommand $m "Set This and Children"  {APL::setSomeFolders $TA(highlightPath) 1}
   bindings::addCommand $m "Clear This and Children" {APL::setSomeFolders $TA(highlightPath) 0}
   $m add separator
   bindings::addCommand $m "Expand Children"  {APL::expandChildren 1}
   bindings::addCommand $m "Collapse Children" {APL::expandChildren 0}
   $m add separator
   bindings::addCommand $m "Catalog Folder" {APL::catalogFolder $TA(highlightPath)}
   bindings::addCommand $m "Catalog Folder and Children" {
      APL::catalogFolderChildren $TA(highlightPath)
   }
   bindings::addCommand $m "Cache Track Details" {
      aplLB::BuildCaches $TA(highlightPath)
   }
   $m add separator
   bindings::addCommand $m "Correct Track Names" {fixplus::correctCase $TA(highlightPath)}

   $m add separator
   bindings::addCommand $m "Append Folder Tracks to Playlist" {APL::AddToPlayList $TA(highlightPath)}
   bindings::addCommand $m "Queue Folder Tracks to Playlist" {APL::InsertToPlayList $TA(highlightPath)}
   bindings::addCommand $m "Append then Play Folder Tracks" {APL::AddThenPlayFolder $TA(highlightPath)}
   $m add separator
   bindings::addCommand $m "Apply Auto-Playlist" {
      macros::filterMusic replace $::TA(matchAgainstAll)
   }
   bindings::addCommand $m "Save Auto-Playlist" {
      macros::saveCurrentAPL 0
   }
   $m add separator
   bindings::addCommand $m "ID3 Table Editor"  {::APL::treeNameToTag 0}
   bindings::addCommand $m "ID3 Tag To Name" {::APL::treeTagToName}
   $m add separator
   bindings::addCommand $m "Properties" {
      APL::treeProperties $::TA(highlightPath)
   }

}
###############################################################################
# Function   : APL::canvasScroll
# Description: Callback to scroll the canvas when the button is down and a drag
#              is done. Based on the tk listbox feature.
# Author     : Tom Wilkason
# Date       : 3/8/2002
###############################################################################
# proc APL::canvasScroll {w} {
#    if {![winfo exists $w]} return
#    set y $::TA(cy)
#    puts $y
#    if {$y >= [winfo height $w]} {
#       $w yview scroll 1 units
#    } elseif {$y < 0} {
#       $w yview scroll -1 units
#    } else {
#       return
#    }
#    set ::TA(afterID) [after 50 [list APL::canvasScroll $w]]
# }
proc APL::cut {args} {
   set here [focus]
   if {$here == [winfo parent $::TREE]} {
      APL::popWatch
      clipboard append "[list $::TA(highlightPath)] "
   } elseif {[winfo parent $here] == $::PLB} {
      APL::popWatch
      clipboardCopy $::PLB 1
   } else {
      tk_messageBox -type ok -icon warning \
         -message "You must have a selection in either the tree of file list area"
   }
}
###############################################################################
# Function   : APL::treeProperties
# Description: Show the stats of the tree under the folder
# Author     : Tom Wilkason
# Date       : 6/22/2002
###############################################################################
proc APL::treeProperties {selectedFolder} {
   ;##
   ;# Total up stats for all files within a tree
   ;#
   set totalSize 0.0
   set totalFiles 0
   foreach {folder} [array names ::teFolders $selectedFolder*] {
      if {[clock clicks -milliseconds]%10 == 0} {
         waitWindow::show "Checking $folder"
      }
      update   ;# keep playing
      set files [list]
      foreach {file} [glob -nocomplain -directory $folder *] {
         if {[info exists ::taExtensions([file extension $file])]} {
            set totalSize [expr {$totalSize+double([file size $file])}]
            incr totalFiles
         }
      } ;# end foreach
   }
   waitWindow::close
   set totalSize [expr {round($totalSize/1048576.)}]
   tk_messageBox -type ok -icon info -title "Properties for $selectedFolder" \
      -message "A total of [commify $totalFiles] media files containing [commify $totalSize] Mbytes were found."
}
###############################################################################
# Function   : APL::PasteFolderToTree
# Description: Move the files in the clipboard to the selected folder
# Author     : Tom Wilkason
# Date       : 6/18/2002
###############################################################################
proc APL::PasteFolderToTree {args} {
   APL::popWatch 100
   if {[catalog::moveClipboardFiles $::TA(highlightPath)] > 0} {
      aplLB::UpdateList $::TA(highlightPath) [winfo ismapped $::PLB]
      APL::treeSelect $::TA(highlightPath)
   } ;# end if
}
###############################################################################
# Function   : APL::createFolder
# Description: Prompt then create a subfolder under some level
# Author     : Tom Wilkason
###############################################################################
proc APL::createFolder {folder} {
   set newName "                               "
   set res 0

   ;##
   ;# TFW: ToDo: Need to create a new folder called "New Folder"
   ;# highlight it
   ;# bring up the rename dialog on it  SelectedFolderRenameInPlace
   ;# If it didn't change then remove it

   if {[getEntryDialog "Enter Folder Name" newName]=="ok"} {
      if {[catch {
         set newName [file join $folder [string trim $newName]]
         file mkdir $newName
         set ::teFolders($newName) $::teFolders([file dirname $newName])
         newItem $newName
         APL::treeOpen [file dirname $newName]
         APL::treeSelect [file dirname $newName]
         update idletasks
         APL::viewHighlighted $::TREE
         set res 1
      } result] \
      } then {
         tk_messageBox -type ok -icon warning -message "Could not create folder $newName\n$result"
         set res 0
      }
   } ;# end if
   return $res
}
###############################################################################
# Function   : APL::chooseIcon
# Description: Select the appropropiate icon based on file name
# Author     : Tom Wilkason
###############################################################################
proc APL::chooseIcon {Name} {
   variable icons
   variable rmName
   set icon [lindex $icons [string match "$rmName*" $Name]]
}
###############################################################################
# Function   : APL::treeSelect
# Description: macros::Highlight a node
# Author     : Tom Wilkason
# Date       : 3/4/2002
###############################################################################
proc APL::treeSelect {folder} {
   variable folderList
   Tree::setselection $::TREE [APL::pathToTreeName $folder]
}
###############################################################################
# Function   : APL::treeOpen
# Description: Open a node
# Author     : Tom Wilkason
# Date       : 3/4/2002
###############################################################################
proc APL::treeOpen {folder {recurse 0}} {
   Tree::open $::TREE [APL::pathToTreeName $folder] $recurse
}
###############################################################################
# Function   : APL::busyCursor
# Description: Turn on/off a busy cursor
# Author     : Tom Wilkason
# Date       : 3/9/2002
###############################################################################
proc APL::busyCursor {busy} {
   if {$busy} {
      .wFC config -cursor watch
      $::TREE config -cursor watch
   } else {
      .wFC config -cursor $::TA(arrow)
      $::TREE config -cursor $::TA(arrow)
   }

}
proc APL::expandChildren {how} {
   if {$how} {
      Tree::close $::TREE [APL::pathToTreeName $::TA(highlightPath)] 10
      Tree::open $::TREE [APL::pathToTreeName $::TA(highlightPath)] 10
   } else {
      Tree::close $::TREE [APL::pathToTreeName $::TA(highlightPath)] 10
   }
}
###############################################################################
# Function   : APL::catalogFolderChildren
# Description: Catalog a folder tree, called after it has been modified.
#              Assign to hotkeys
# Author     : Tom Wilkason
###############################################################################
proc APL::catalogFolderChildren {folder} {
   catalog::Rescan $folder
   # Need to reselect the folder, make a dummy file so the folder is found in APL::SelectFile
   APL::treeOpen $folder
   APL::SelectFile [file join $folder xx]  ;# need to get file name selected in listbox, not the folder name
}
###############################################################################
# Function   : APL::catalogFolder
# Description: Catalog a single folder, called after it has been modified.
#              Assign to hotkeys
# Author     : Tom Wilkason
###############################################################################
proc APL::catalogFolder {folder} {
   set ::snackAmpSettings(diff) 0
   ;# Remove any cache file if one exists.
   set cName [file join $folder ".saDetails"]
   if {[file exists $cName]} {
      if {[catch {file delete -force $cName} result] } then {
         # what to do, what to do...
         saLog $result
      }
   } ;# end if
   catalog::rescanFolder $folder
   if {$::snackAmpSettings(diff) > 0} {
      catalog::FlagDirtyCache $folder
   } ;# end if
   aplLB::UpdateList $folder [winfo ismapped $::PLB]
}
###############################################################################
# Function   : APL::GetTreeFolder
# Description: Return the folder name corresponding to the label
#              Throws an error if can find corresponding folder name
# Author     : Tom Wilkason
# Date       : 12/27/2001
###############################################################################
proc APL::GetTreeFolder {W lbl} {
   upvar #0 $W myTree
   regexp {^.+?\((.+?)\)$} $myTree($lbl:variable) x Var
   return $Var
}
###############################################################################
# Function   : APL::buildAPL
# Description: Rebuild the Auto Play List tree with new data.
# Author     : Tom Wilkason
# Date       : 11/4/2001
###############################################################################
proc APL::buildAPL {tree} {
   variable rmName
   global TA snackAmpSettings
   set base [join [lrange [split $tree .] 0 end-1] .]
   destroy $tree
   set ::Tree::font treeFont
   if {$snackAmpSettings(useAPLColors)} {
      set Tree::fg $snackAmpSettings(fg)
      set Tree::bg $snackAmpSettings(bg)
   } else {
      set Tree::fg black
      set Tree::bg white
   } ;# end if
   set Tree::highlight $snackAmpSettings(highlight)
   ;# To work
   ;# 1. Text Tree selection outline box color can't be black (saTree ln 544)
   ;# 2. Text on tablelist for highlighted item can't be white
   Tree::create $tree -bg $Tree::bg -width 300 -height 400 -yscrollcommand [list $base.sb set]
   #Tree::create $tree -width 300 -height 400 -yscrollcommand \
   #   [list scrollCmd $base.sb [list pack $base.sb -side right -fill y]]
   set i 0
   ;##
   ;# Build up list of media paths including the RMC removable virtual path
   ;#
   set TA(mediaFolders) [list]
   foreach {Path} [concat $snackAmpSettings(mediaDirs) $::RMT] {
      ;# Folders won't have a / on end, so we don't want our pattern match to either
      lappend TA(mediaFolders) $Path [string trimright $Path /]
      ;# Title for tree just has the leading slash, nothing else
      set TA(Title,$Path) "/[string trim [string map {: {} / { }} $Path]]"
   } ;# end foreach
   ;# Replace /RMC with A little better name
   set TA(Title,$::RMT) $rmName
   ;##
   ;# Map each real folder name into one based from each media folder
   ;# to show it in the tree. Each media folder is considered one
   ;# level regardless of how may real directory levels it is on the disk.
   ;#
   ;# F:/Toms/Stuff     -> /F Toms/Stuff
   ;# //Toms/Toms/Stuff -> /Toms Toms/Stuff
   ;# Set the two types of icons, then need to have .open/.close types for each
   if {[catch {
      foreach folder [lsort [array names ::teFolders]] {
         newItem $folder
      } ;# end foreach
   } result]} {
      saLog $result
   }
   set ::TA(APLDirty) 0
   pack $tree -side left -fill both -expand 1 -padx 0 -pady 0
}
###############################################################################
# Function   : APL::delItem
# Description: Remove a new node from the tree
# Author     : Tom Wilkason
# Date       : 5/8/2002
###############################################################################
proc APL::delItem {folder} {
   Tree::delitem $::TREE [APL::pathToTreeName $folder]
}
###############################################################################
# Function   : APL::newItem
# Description: Add a new node to the tree
# Author     : Tom Wilkason
# Date       : 5/8/2002
###############################################################################
proc APL::newItem {folder} {
   set name [APL::pathToTreeName $folder]
   if {[string length $name] > 0} {
      ;# TFW: would be nice to find if it already exists
      if {[catch {Tree::newitem $::TREE $name -variable ::teFolders($folder) -image [APL::chooseIcon $name]} result] } then {
         saLog $result
      }
   }
}
###############################################################################
# Function   : APL::EditId3, NameToTag, TagToName
# Description: Operate on ID3 tags for multiple files
# Author     : Tom Wilkason
# Date       : 10/30/2001
###############################################################################
proc APL::EditId3 {args} {
   id3Tag::id3Editor [aplLB::getSelectedFullNames]
}
proc APL::NameToTag {smart} {
   id3Tag::id3NameToTag [aplLB::getSelectedFullNames] $smart
}
proc APL::treeNameToTag {smart} {
   ;# Limit to 200 files
   set files [trackList::getFiles [trackList::getFolders $::TA(highlightPath) $::snackAmpSettings(treeDepth)]]
   if {[llength $files] > $::snackAmpSettings(ID3editLimit)} {
      tk_messageBox -type ok -icon warning \
         -message "Limiting operation to first $::snackAmpSettings(ID3editLimit) files"
   }
   id3Tag::id3NameToTag [lrange $files 0 $::snackAmpSettings(ID3editLimit)]
}
proc APL::treeEditId3 {args} {
   ;# Limit to 200 files
   set files [trackList::getFiles [trackList::getFolders $::TA(highlightPath) $::snackAmpSettings(treeDepth)]]
   if {[llength $files]>$::snackAmpSettings(ID3editLimit)} {
      tk_messageBox -type ok -icon warning \
         -message "Limiting operation to first $::snackAmpSettings(ID3editLimit) files"
   }
   id3Tag::id3Editor [lrange $files 0 $::snackAmpSettings(ID3editLimit)]
}
proc APL::TagToName {args} {
   id3Tag::id3TagToName [aplLB::getSelectedFullNames]
}
proc APL::treeTagToName {args} {
   ;# Limit to 200 files
   set files [trackList::getFiles [trackList::getFolders $::TA(highlightPath) $::snackAmpSettings(treeDepth)]]
   if {[llength $files]>$::snackAmpSettings(ID3editLimit)} {
      tk_messageBox -type ok -icon warning \
         -message "Limiting operation to first $::snackAmpSettings(ID3editLimit) files"
   }
   id3Tag::id3TagToName [lrange $files 0 $::snackAmpSettings(ID3editLimit)]
}
proc APL::removeTag {args} {
   id3Tag::removeTag [aplLB::getSelectedFullNames]
}
;##
;# Rename files by splitting name and moving into subfolders
;# so each folder contains one album.
;# Note: Only handles Artist - Album - xxx formatted files
#        Adding Album - xxx format
;# No error checking for shorter formats.
;#
proc APL::SplitFile {args} {
   set dirName ""
   set Artist ""
   set Album ""
   set sep $::snackAmpSettings(splitChar)
   set files [aplLB::getSelectedFullNames]
   foreach file $files {
      if {[string length $dirName] == 0} {
         set dirName [file dirname $file]
      }
      set parts [split [file tail $file] $sep]
      ;# Skip short names
      if {[llength $parts] < 2} {
         ;# This doesn't skip everything
         continue
      } ;# end if
      ;# Assume Artist - Album format
      #saLog "'$parts'"
      foreach {Artist Album} $parts {
         set Artist [string trim $Artist]
         set Album [string trim $Album]
         set F(Artist,$Artist) 1
         set F(Album,$Album) 1
         break
      }
      #saLog "'$Artist' '$Album'"
   }
   ;# Param 1 check for uniqueness
   if {[llength [array names F Artist,*]] > 1} {
      tk_messageBox -message  "Artist Names must be unique"
      return
   }
   ;##
   ;# Create Artist folder and create from there
   ;#
   if {![string equal -nocase $Artist [file tail $dirName]]} {
      ;# If Artist Folder doesnt already exist, need to create it
      set base [file join [file dirname $file] $Artist]
      #saLog "Creating '$Artist' on '[file tail $dirName]'"
   } else {
      ;# Artist folder already exists, so use it
      set base [file dirname $file]
      #saLog "Using $Artist"
   }
   ;# Param 2 check for uniqueness
   ;# Assume format of Artist/Album - xxx
   if {[llength [array names F Album,*]] > 1} {
      set index 1
      set message  "Do you want to split [llength $files] files into folder '$base'?"
   } else {
      set index 2
      set message "Do you want to split [llength $files] files for '$Artist' into folder '[file join $base $Album]'?"
   }
   ;##
   ;# Move the files
   ;#
   set result [tk_messageBox -type yesno -icon question -message $message]
   if {$result == "yes"} {
      foreach file $files {
         switch -- $index {
            1 {set newFile [file join $base [string trim [join [lrange [split $file $sep] $index end] $sep]]]}
            2 {set newFile [file join $base $Album [string trim [join [lrange [split $file $sep] $index end] $sep]]]}
            default {}
         }
         if {[catch {catalog::RenameFile $file $newFile} result]} {
            tk_messageBox -icon warning -message "Could not rename\n$file to \n$newFile\n$result"
         } else {
            catalog::RenameEntry $file $newFile
         }
      }
      ;# Catalog this tree to see the new folders
      APL::catalogFolderChildren $::TA(highlightPath)
   } ;# end if
}
;##
;# Rename the files based on the folder names beneath them
;#
proc APL::CombineFile {level} {
   set sep $::snackAmpSettings(splitChar)
   set files [aplLB::getSelectedFullNames]
   foreach file $files {
      set folder [file dirname $file]
      set tail [file tail $file]
      set newTail [list]
      for {set i $level} {$i > 0} {incr i -1} {
         set newTail [linsert $newTail 0 [file tail $folder]]
         set folder [file dirname $folder]
      } ;# end for
      set newTail [join $newTail " $sep "]
      set tail "$newTail $sep $tail"
      #set newFile [file join $folder $tail]
      set newFile [file join [file dirname $file] $tail] ;# leave in same folder
      if {[catch {catalog::RenameFile $file $newFile} result]} {
         tk_messageBox -icon warning -message "Could not rename\n$file to \n$newFile\n$result"
      } else {
         catalog::RenameEntry $file $newFile
      }
   }
   ;# Catalog this tree to see the new folders
   APL::catalogFolderChildren $::TA(highlightPath)
}
###############################################################################
# Function   : APL::DeleteFiles
# Description: Remove some files from the auto playlist list box
# Author     : Tom Wilkason
# Date       : 10/30/2001
###############################################################################
proc APL::DeleteFiles {args} {
   set toDelete [aplLB::getSelectedFullNames 0]
   if {[llength $toDelete] > 0} {
      if {$::snackAmpSettings(dontMoveFiles)} {
         tk_messageBox -icon warning -type ok \
            -message "You don't have permission to remove files"
         return
      }
      set result [tk_messageBox -type yesno -title "Please Answer" \
         -icon question -message "Are you sure you want to delete:\n[join $toDelete \n]"]
      if {$result == "yes"} {
         foreach {File} $toDelete {
            catalog::DeleteFile $File
         }
         ;# Schedule a rescan when all is said and done
         ;# to reflect the new configuration
         trackList::populateHighighted 1 1
         updateFileDisplay
      } ;# end if
   } ;# end if
}
###############################################################################
# Function   : AddThenPlayFolder
# Description: Add selected folder to playlist then play the first track
# Author     : Tom Wilkason
# Date       : 6/18/2002
###############################################################################
proc APL::AddThenPlayFolder {folder} {
   Stop
   APL::AddToPlayList $folder
   Play active
}
###############################################################################
# Function   : APL::AddToPlayList
# Description: Add the entire set of songs in this folder to the playlist
# Author     : Tom Wilkason
# Date       : 10/28/2001
###############################################################################
proc APL::AddToPlayList {index} {
   ;##
   ;# Now list the files out
   ;#
   set Cursor [LB::end]
   foreach {File} [aplLB::getAllNames 1] {
      filtering::Append $File
   } ;# end foreach
   LB::activate $Cursor
   updateFileDisplay
}
###############################################################################
# Function   : APL::InsertToPlayList
# Description: Insert the entire set of songs in this folder to the playlist
# Author     : Tom Wilkason
# Date       : 10/28/2001
###############################################################################
proc APL::InsertToPlayList {index} {
   ;##
   ;# Now list the files out
   ;#
   set cp [LB::playing]
   foreach {File} [aplLB::getAllNames 1] {
      filtering::Insert [incr cp] $File
   } ;# end foreach
   updateFileDisplay
}
###############################################################################
# Function   : APL::SelectFile
# Description: Given some file, select its folder in the tree if possible
# Author     : Tom Wilkason
# Date       : 6/6/2001
###############################################################################
proc APL::SelectFile {Path {pushPath 1}} {
   Trace
   variable folderList
   set W $::TREE
   if {[winfo exists $W] && ([wm state .wFC] == "normal")} {
      if {[info exists ::teFolders($Path)] || [file isdirectory $Path]} {
         set file {}
         set Dir $Path
      } else {
         set file $Path
         set Dir [file dirname $Path]
      }
      ;# Save the path in the navagiation history
      if {$pushPath} {
         insertUnique $Dir
      }
      set showDir $Dir
      set ::TA(highlightPath) $Dir
      set Item [APL::pathToTreeName $Dir]
      set Dir [file dirname $Dir]
      ;# Expand containing folder and all parents too
      while {[file dirname $Dir] != $Dir} {
         Tree::open $W [APL::pathToTreeName $Dir]
         set Dir [file dirname $Dir]
      }
      ;# Extract the index number from the Tree variable
      Tree::setselection $W $Item
      upvar #0 $W myTree
      if {[catch {set index [lindex [split $myTree($Item:variable) "()"] 1]}]} {
         return
      }
      ;##
      ;# Update file listbox if it exists.
      ;# Note, must be drawn
      if {[winfo ismapped $::PLB]} {
         aplLB::UpdateList $showDir
         ;##
         ;# macros::Highlight the item in the listbox
         ;#
         if {[string length $file] > 0} {
            set contents [aplLB::getAllNames]
            set index [lsearch -exact $contents [file tail $Path]]
            aplLB::selectionSet $index
         }
      } ;# end if
      ;##
      ;# Scroll the canvas so the highlighted item is visible
      ;# (if not all of data shown)
      APL::viewHighlighted $W
      ;# Give tree the focus to allow pasting to work
      focus [winfo parent $W]
   } ;# end if
}
###############################################################################
# Function   : APL::viewHighlighted
# Description: Scroll the canvas so the highlighted item is visible
# Author     : Tom Wilkason
# Date       : 3/5/2002
###############################################################################
proc APL::viewHighlighted {W} {
   upvar #0 $W myTree
   set OnScreenHeight [winfo height $W]
   set EntireHeight  [lindex [$W cget -scrollregion] 3]
   if {$EntireHeight > $OnScreenHeight} {
      set CursorY [lindex [$W bbox $myTree(selidx)] 3]
      if {[string length $CursorY] > 0} {
         set MoveTo [expr {double($CursorY-$OnScreenHeight+50)/$EntireHeight}]
         $W yview moveto $MoveTo
      } ;# end if
   } ;# end if
}
###############################################################################
# Function   : APL::pathToTreeName
# Description: Convert a File Path into a Tree Name
# Author     : Tom Wilkason
# Date       : 6/6/2001
###############################################################################
proc APL::pathToTreeName {folder} {
   global TA
   foreach {Path TPath} $TA(mediaFolders) {
      ;# folder names won't end in a / (except drive vols) so our match patten can't either
      ;# The titles are formed to not have path separators within them
      if {[regsub -- "^$TPath" $folder $TA(Title,$Path) name]} {
         set name [string trimright $name /]    ;# Vol names will have a trailing / in linkArray
         return $name
      } ;# end if
   } ;# end foreach
   return {}
}
###############################################################################
# Function   : APL::setSomeFolders
# Description: Set/clear folders after base to some value
# Author     : Tom Wilkason
###############################################################################
proc APL::setSomeFolders {base value} {
   global teFolders
   foreach Path [array names teFolders "$base*"] {
      set teFolders($Path) $value
   } ;# end foreach
   APL::refreshTree
}
###############################################################################
# Function   : APL::refreshTree
# Description: Refresh the tree such that any variables updated are relfected
#              in the checkboxes
# Author     : Tom Wilkason
# Date       : 1/11/2001
###############################################################################
proc APL::refreshTree {args} {
   global snackAmpSettings TA
   set W $::TREE
   ;# Save scroll % and restore it when done
   set scroll [lindex [$W yview] 0]
   foreach {Path} $snackAmpSettings(mediaDirs) {
      Tree::close $W $TA(Title,$Path)
      Tree::open $W $TA(Title,$Path)
   }
   $W yview moveto $scroll
}
###############################################################################
# Function   : APL::setFileListToShow
# Description: Show or hide the filelist
# Author     : Tom Wilkason
# Date       : 3/9/2001
###############################################################################
proc APL::setFileListToShow {what widget} {
   variable paneOrient
   if {$::snackAmpSettings(hideFileList)} {
      pane $what -orient $paneOrient -dynamic 0 \
         -size $::snackAmpSettings(FileListSize)
      $widget configure -relief sunken
      set ::snackAmpSettings(hideFileList) 0
   } else {
      $widget configure -relief flat   ;# button
      set ::snackAmpSettings(hideFileList) 1
      set curSize [pane $what -size]
      if {$curSize != {}} {
         set ::snackAmpSettings(FileListSize) $curSize
         pane forget [winfo parent $what] $what
      } ;# end if
   }
}
###############################################################################
# Function   : APL::Table
# Description: Create the Table to show the song details
# Author     : Tom Wilkason
# Date       : 3/9/2001
###############################################################################
#proc APL::Table {args} {
#   package require Tablelist
#   destroy .tblHolder
#   toplevel .tblHolder
#   set APL::Table [trackList::buildTable .tblHolder.tf ::snackAmpSettings]
#   pack .tblHolder.tf -fill both -expand true
#}

###############################################################################
# Function   : aplLB Namespace
# Description: Create bindings for the APL tree
# Author     : Tom Wilkason
# Date       : 3/9/2001
###############################################################################
namespace eval aplLB {
   proc commonBindings {args} {
      bind .wFC <F5>         {SelectFile 0}
      bind .wFC <Control-F5> {SelectFile 1}
      APL_ListAssign <F2> {::APL::RenameFile}
      APL_ListAssign <Control-F2> ::APL::RenameFile
      APL_ListAssign <F3> ::APL::EditId3
      APL_ListAssign <F4> {::APL::NameToTag 0}
      APL_TreeAssign <F4> {::APL::treeNameToTag 0}
      APL_TreeAssign <F3> {::APL::treeEditId3}
      APL_ListAssign <F6> {::APL::TagToName}
      APL_TreeAssign <F6> {::APL::treeTagToName}
      APL_ListAssign <Delete> ::APL::DeleteFiles
      APL_ListAssign <B1-Motion> {
         ;# Prevent the select-drag action
         ;# It would mess up the DND
         $::PLB config -cursor hand2
         break
      }
      APL_ListAssign <B1-Leave> {
         ;# Leaving the listbox, snapshot the files
         clipboardCopy $::PLB 0
         break   ;# Prevent the drag select
      }
      APL_ListAssign <ButtonPress-3> {
         tk_popup $::PLB.popup %X %Y
      }
      APL_ListAssign <ButtonPress-1> {
         ;# Allow key bindings to work
         focus %W
      }
      APL_ListAssign <<Copy>> {APL::popWatch;clipboardCopy $::PLB 1;break}
      APL_ListAssign <ButtonRelease-1> {
         ;# See if we are over a node on the tree
         set target [winfo containing %X %Y]
         if {[winfo exists $target]} {
            if {[string equal $target $::TREE]} {
               APL::cbDndPLBToTree %X %Y
            } elseif {[string equal $target $::LB]} {
               APL::cbDndPLBToLB %X %Y
            } elseif {[winfo toplevel $target] == [winfo toplevel $::LB]} {
               APL::cbDndPLBToLB 0 0
            }
         } ;# end if
         $::PLB config -cursor $::TA(arrow)
         catch {unset target}
      }
   }
   proc appendSelectedFiles {args} {
      if {[array size ::trackList::Index2File] > 0} {
         foreach {File} [aplLB::getSelectedFullNames 0] {
            filtering::Append $File
         } ;# end foreach
         updateFileDisplay
      }
   }
   proc insertSelectedFiles {args} {
      if {[array size ::trackList::Index2File] > 0} {
         set cp [LB::playing]
         foreach {File} [aplLB::getSelectedFullNames 0] {
            filtering::Insert [incr cp] $File
         } ;# end foreach
         updateFileDisplay
      }
   }
   ;##
   ;# Play the highlighted files or folder
   ;#
   proc playSelectedFiles {args} {
      set added 0
      if {[winfo parent $::TREE] == [focus]} {
         ;# Folder
         set end [LB::end]
         APL::AddToPlayList $::TA(highlightPath)
         LB::activate $end
         Play active
      } else {
         ;# Selected Files
         set end [LB::end]
         set selected [aplLB::getSelectedFullNames 0]
         if {[llength $selected] == 1} {
            filtering::InsertPlay [lindex $selected 0]
         } else {
            foreach {file} $selected {
               incr added
               filtering::Append $file
            } ;# end foreach
            if {$added > 0} {
               LB::inverseVideoFile xx 0
               LB::activate $end
               Play active
               updateFileDisplay
            }
         }
      }
   }
   ;##
   ;# Menus bound to listbox
   ;#
   proc menus {args} {
      ;##
      ;# File list box bindings
      ;#
      set plb [$::PLB bodypath]
      bindings::init $plb menuBindings
      set m $::PLB.popup
      menu $m -tearoff 0 -font menu

      bindings::addCommand $m "Play Tracks(s)" {aplLB::playSelectedFiles}
      ;##
      ;# Copy the files to the playlist
      ;#
      bindings::addCommand $m "Queue Tracks(s) to Play Next" {aplLB::insertSelectedFiles}
      bindings::addCommand $m "Append Tracks(s) to Playlist" {aplLB::appendSelectedFiles}
      $m add separator
      bindings::addCommand $m "Save All to Playlist File"  {macros::SavePlaylistFile $::PLB replace all}
      bindings::addCommand $m "Append All to Playlist File"  {macros::SavePlaylistFile $::PLB append all}
      $m add separator
      bindings::addCommand $m "Save Selected to Playlist File"  {macros::SavePlaylistFile $::PLB replace selected}
      bindings::addCommand $m "Append Selected to Playlist File"  {macros::SavePlaylistFile $::PLB append selected}
      ;##
      ;# Rename the active selection
      ;#
      $m add separator
      bindings::addCommand $m "Copy Tracks(s) to clipboard" {APL::popWatch;clipboardCopy $::PLB 1}
      bindings::addCommand $m "Delete Tracks(s)"  {::APL::DeleteFiles}
      bindings::addCommand $m "Rename Track"  {::APL::RenameFile}
      $m add separator
      bindings::addCommand $m "Correct Track Name(s)" {fixplus::correctCase [aplLB::getSelectedFullNames]}
      bindings::addCommand $m "Split Into Folders" {::APL::SplitFile}
      bindings::addCommand $m "Rename Based on Folder" {::APL::CombineFile 1}
      bindings::addCommand $m "Rename Based on Two Folders" {::APL::CombineFile 2}
      $m add separator
      bindings::addCommand $m "Create Stream Shortcut" {
         if {[playurl::checkuskfolder]} {
            playurl::addPromptUrl $TA(highlightPath)
            aplLB::UpdateList $::TA(highlightPath) 1
         }
      }
      bindings::addCommand $m "Edit Track ID3 Info"  {::APL::EditId3}
      bindings::addCommand $m "ID3 Table Editor"  {::APL::NameToTag 0}
      bindings::addCommand $m "ID3 Tag To Name" {::APL::TagToName}
   }
   proc selected {args} {
      return [$::PLB curselection]
   }
   proc selectionSet {index} {
      $::PLB selection clear 0 end
      $::PLB selection set $index
      $::PLB activate $index
      $::PLB see $index
   }
   ;##
   ;# Table Stuff
   ;#
   set APL::paneOrient "horizontal"
   package require Tablelist
   ;##
   ;# Build a set of caches for a folder and sub-folders
   ;#
   proc BuildCaches {path} {
      foreach {folder} [array names ::teFolders $path*] {
         if {[catch {
            update
            trackList::populate $folder $::PLB APL::TableCols 0 0
         } result] } then {
            saLog $::errorInfo
         }
      } ;# end foreach
   }
   ;##
   ;# Update the listbox based on incoming folder
   ;#
   proc UpdateList {index {insert 1}} {
      Trace
      ;##
      ;# New Table
      ;#
      if {$insert} {
         if {[catch {
            trackList::populate $index $::PLB APL::TableCols 1 1
         } result] } then {
            saLog $::errorInfo
         }
      } ;# end if
      ;##
      ;# Update the selected counts
      ;#
      set checked 0
      set unchecked 0
      set pattern "[string trimright $index /]/*"
      foreach {File} [catalog::filesMatchingPattern $pattern] {
         set dirName [file dirname $File]
         if {[info exists ::teFolders($dirName)] && ($::teFolders($dirName) == 1)} {
            incr checked
         } else {
            incr unchecked
         } ;# end if
      }
      if {($checked+$unchecked) > 0} {
         set ::TA(ck,selected)   "Checked $checked"
         set ::TA(ck,unselected) "Unchecked $unchecked"
         set ::TA(ck,total)      "Total [expr {$checked+$unchecked}]"
      }
      ;# Signal done
   }
   proc rowIndex2file {index} {
      return [trackList::rowIndex2file $index]
   }
   proc getAllNames {{long 0}} {
      set files [list]
      ;# Get file name for each row
      set end [$::PLB index end]
      for {set i 0} {$i < $end } {incr i} {
         if {$long} {
            lappend files [catalog::TranslateMount [trackList::rowIndex2file $i]]
         }  else {
            lappend files [file tail [trackList::rowIndex2file $i]]
         }
      } ;# end for
      return $files
   }
#      proc getFullName {index} {
#         return [catalog::TranslateMount [trackList::rowIndex2file $index]]
#      }
   proc getActiveFullName {args} {
      set active  [$::PLB index active]
      return [catalog::TranslateMount [trackList::rowIndex2file $active]]
   }
   proc getSelectedFullNames {{translate 1}} {
      set fullNames [list]
      if {$translate} {
         foreach {index} [selected] {
            lappend fullNames [catalog::TranslateMount [trackList::rowIndex2file $index]]
         } ;# end foreach
      } else {
         foreach {index} [selected] {
            lappend fullNames [trackList::rowIndex2file $index]
         } ;# end foreach
      }
      return $fullNames
   }
   proc RemoveFile {File} {
      if {[winfo exists $::PLB]} {
         set lnFile [file tail $File]
         set index [lsearch -exact [getAllNames] $lnFile]
         if {$index > -1} {
            $::PLB delete $index
         }
      } ;# end if
   }
   ;##
   ;# Key bindings for table
   ;#
   proc bindings {args} {
      ;# Common Keys
      commonBindings
      ;# Couple hot keys

      APL_ListAssign <Double-ButtonPress-1> {
         if {[array size ::trackList::Index2File] > 0} {
            if {[catch {trackList::row2file %W %y} File] } then {
               saLog $File
            } else {
               filtering::InsertPlay $File
            }
         }
      }
   }
}
###############################################################################
# Function   : APL::RenameFile
# Description: Rename a file using a dialog
# Author     : Tom Wilkason
# Date       : 10/30/2001
###############################################################################
proc APL::RenameFileDialog {args} {
   if {[catch {aplLB::getActiveFullName} fullName]} {
      saLog "APL::RenameFile $fullName\n$::errorInfo"
      return
   }
   set oldName [file tail $fullName]
   set dirName [file dirname $fullName]
   set newName $oldName
   set answer [getEntryDialog "$oldName" newName]
   if {$answer == "ok" && $newName != $oldName} {
      set newFullName [file join $dirName $newName]
      if {[catch {catalog::RenameFile $fullName $newFullName} result]} {
         tk_messageBox -icon warning -message "Could not rename\n$fullName to \n$newFullName\n$result"
      } else {
         catalog::RenameEntry $fullName $newFullName
      }
   } ;# end if
}
####################################################################################
# Function   : RenameFile
# Description: Rename a file directly from the listbox
# Author     : Tom Wilkason
# Date       : 6/19/2002
###############################################################################
proc APL::RenameFile {args} {
   set W $::PLB
   if {[catch {aplLB::getActiveFullName} fullName]} {
      saLog "APL::RenameFile $fullName\n$::errorInfo"
      return
   }
   set oldName [file tail $fullName]
   set dirName [file dirname $fullName]
   set newName $oldName
   ;##
   ;# Block on dialog
   ;#
   set name [catalog::renameInPlaceDialog $W 1 $oldName]
   ;##
   ;# Back from window closure
   ;#
   if {[string length $::TA(rename)]>0 && $::TA(rename) != $oldName} {
      #debug "Renaming '$oldName' to '$::TA(rename)'"
      set newFullName [file join $dirName $::TA(rename)]
      if {[catch {catalog::RenameFile $fullName $newFullName} result]} {
         tk_messageBox -icon warning -message "Could not rename\n$fullName to \n$newFullName\n$result"
      } else {
         catalog::RenameEntry $fullName $newFullName
         #debug "catalog::RenameEntry $fullName $newFullName"
      }
   }
}
;##
;# trackList namespace: Used for table widget
;#
namespace eval trackList {
   variable tableSound
   variable col2Use [list]
   variable Index2File
   variable lastFolder ""
   variable checkTime
   variable editCellButton
   variable editCellState 0
   variable editCellCheckState 0
   variable TABLEDATA [list]
   set tableSound [snack::sound table]
   switch -- $tcl_platform(platform) {
      "windows" {
         if {$tcl_platform(os) == "Windows NT"} {
            set checkTime "mtime"   ;# atime works somewhat
         } else {
            set checkTime "mtime"
         }
      }
      default {
         set checkTime "mtime"
      }
   } ;# End Switch
}
;##
;# Schedule a rescan when all is said and done
;# to reflect the new configuration
proc trackList::populateHighighted {sort insert} {
   if {![afterExists trackList::populate]} {
      #puts "Scheduling a scan on $::TA(highlightPath)"
      after idle [list trackList::populate $::TA(highlightPath) $::PLB APL::TableCols $sort $insert]
   }
}
;##
;# Return a file name based on row index
;# Col 0 has a hidden column with original key values
;# which are maintained with the table is sorted.
;# use the value to lookup the correct file name
proc trackList::rowIndex2file {index} {
   variable Index2File
   set rowIndex [$::PLB cellcget $index,0 -text]
   return $Index2File($rowIndex)
}
;##
;# Map selected row to full file name
;#
proc trackList::row2file {W y} {
   set File {}
   set y [expr {$y + [winfo y $W]}]
   set index [$::PLB nearest $y]
   return [rowIndex2file $index]
}
#------------------------------------------------------------------------------
# trackList::editCellStart
#
# When a cell is first edited, this callback fires
#
#------------------------------------------------------------------------------
proc trackList::editCellStart {W row col newVal} {
   variable oCellValue
   set oCellValue $newVal
   return $newVal
}
#------------------------------------------------------------------------------
# trackList::editCellEnd
#
# When a cell is edited, this callback fires
#
#------------------------------------------------------------------------------
proc trackList::editCellEnd {W row col newVal} {
   variable TABLEDATA
   variable Index2File
   variable oCellValue
   ;#$W rejectinput ;# To do nothing leaving editing open
   #$W cancelediting ;# To do nothing leaving editing open
   #.wFC.sv.lb.tbl cancelediting
   set col [$W columncget $col -title]
   set file $Index2File([lindex [lindex $TABLEDATA $row] 0])
   switch -exact -- $col {
      Folder {
         ;# This can be dangerous, disable Folder from editing
         set newFullName [file join $newVal [file tail $file]]
         #puts "Rename '$file' to '$newFullName'"
         if {[catch {catalog::RenameFile $file $newFullName} result]} {
            tk_messageBox -icon warning -message "Could not rename\n$file to \n$newFullName\n$result"
            return $oCellValue
         } else {
            catalog::RenameEntry $file $newFullName
            #debug "catalog::RenameEntry $fullName $newFullName"
         }
      }
      File {
         set newFullName [file join [file dirname $file] $newVal]
         #puts "Rename '$file' to '$newFullName'"
         if {[catch {catalog::RenameFile $file $newFullName} result]} {
            tk_messageBox -icon warning -message "Could not rename\n$file to \n$newFullName\n$result"
            return $oCellValue
         } else {
            catalog::RenameEntry $file $newFullName
            #debug "catalog::RenameEntry $fullName $newFullName"
         }
      }
      Title -
      Artist -
      Album -
      Track -
      Media -
      Year -
      Genre -
      Note -
      Genre -
      Comp -
      Desc -
      Comp -
      Comp {
         if {$::snackAmpSettings(ID3ReadOnly)==0} {
            if {[catch {
               id3Tag::id3Label $file fnameData V1
               set fnameData($col) $newVal
               id3Tag::id3V1Modify $file fnameData
            } result] } then {
               saLog $result
               return $oCellValue
            }
         } else {
            return $oCellValue
         }
      }
      Gain {
         ;# This needs a conversion first
         gainControl::add $file [lindex $newVal 0]
      }
      default {saLog "Unknown option '$col'"}
   } ;# End Switch
   return $newVal
}
#------------------------------------------------------------------------------
# trackList::changeEdit
#
# Change the editing mode of a table
#------------------------------------------------------------------------------
proc trackList::changeEdit {args} {
   variable editCol
   variable editCellButton
   variable editCellCheckState
   variable editCellState
   catch {tablelist::canceleditingSubCmd $::PLB}
   ;##
   ;# set or clear the editable columns
   ;#
   if {$editCellState==0} {
      $editCellButton configure -relief sunken
      foreach {index} [array names editCol] {
         $::PLB columnconfigure $index -editable $editCol($index)
      } ;# end foreach
      set editCellState 1
   } else {
      $editCellButton configure -relief flat
      foreach {index} [array names editCol] {
         $::PLB columnconfigure $index -editable 0
      } ;# end foreach
      set editCellState 0
   }
   set editCellCheckState $editCellState
}
#------------------------------------------------------------------------------
# trackList::buildTable
#
# Builds the widget w as a tablelist widget
# contained in a newly created top-level widget.  Returns the name of the
# tablelist widget.
#------------------------------------------------------------------------------
proc trackList::buildTable {top _config} {
   variable tableSound
   variable col2Use
   variable editCol
   variable TABLEDATA
   upvar $_config config
   ;# List of possible columns
   set col2Use [list]
   set i 0
   set j 0

   set sortMode [list]
   ;# Use saved table widths if possible, if new ones added make those 0
   if {[info exists ::snackAmpSettings(TableWidths)]} {
      set TableWidths $::snackAmpSettings(TableWidths)
      while {[llength $TableWidths] < [llength $id3Tag::TableWidths]} {
         lappend TableWidths 0
      } ;# end while
   } else {
      set TableWidths $id3Tag::TableWidths
   }
   ;# Build up col info
   catch {unset editCol}
   foreach {col} $id3Tag::TableCols {just} $id3Tag::TableJust {type} $id3Tag::TableType {width} $TableWidths {edit} $id3Tag::TableEdit {
      if {[info exists config(column,$col)] && $config(column,$col)} {
         lappend col2Use $width $col $just
         lappend sortMode $j $type
         set editCol($j) $edit
         incr j
      } ;# end if
      incr i
   } ;# end foreach
   #
   # Create a scrolled tablelist widget with 5 dynamic-width
   # columns and interactive sort capability within the top-level
   #
   set tf [frame $top]     ;# User should pack in frame
   set tbl $tf.tbl
   set vsb $tf.vsb
   set hsb $tf.hsb
   tablelist::tablelist $tbl \
         -columns $col2Use \
         -labelcommand tablelist::sortByColumn \
         -xscrollcommand [list scrollCmd $hsb  [list grid $hsb -row 1 -column 0 -sticky ew]] \
         -yscrollcommand [list scrollCmd $vsb  [list place $vsb -relx 1.0 -rely 0.0 -anchor ne -relheight 1.0]] \
         -selectbackground $::snackAmpSettings(highlight) \
         -selectforeground $::snackAmpSettings(fg) \
         -foreground       $Tree::fg \
         -background       $Tree::bg   \
         -showseparators   no \
         -borderwidth      1 \
         -selectmode       extended \
         -setgrid          no \
         -listvariable   trackList::TABLEDATA \
         -editstartcommand {trackList::editCellStart} \
         -editendcommand {trackList::editCellEnd}
#         -movablecolumns 1

   ;# Define editiable columns
   ;##
   ;# Set the options
   ;#
   if {$::snackAmpSettings(useTableStripes)} {
      settings::regColor $tbl -stripebackground ::snackAmpSettings(tsbg)
      $tbl config -stripebackground $::snackAmpSettings(tsbg)
   }
   if {$::snackAmpSettings(useTableBars)} {
      $tbl config -showseparators yes
   }
   if {$::snackAmpSettings(autoTableStretch)} {
      $tbl config -stretch all
   }
   ;# Define how each col is sorted.
         # -yscrollcommand [list scrollCmd $vsb  [list grid $vsb -row 0 -column 1 -sticky ns]]
   foreach {col mode} $sortMode {
      switch -- $mode {
         "commify" {
            $tbl columnconfigure $col -sortmode integer -formatcommand commify
         }
         "timify" {
            $tbl columnconfigure $col -sortmode real -formatcommand timify
         }
         default {
            $tbl columnconfigure $col -sortmode $mode
         }
      } ;# End Switch
   }
   ;# Hide column used for the key
   $tbl columnconfigure 0 -hide 1
   scrollbar $vsb -orient vertical   -command [list $tbl yview]
   scrollbar $hsb -orient horizontal -command [list $tbl xview]
   #
   # Manage the widgets
   #
   grid $tbl -row 0 -column 0 -sticky news
   #grid $vsb -row 0 -column 1 -sticky ns
   place $vsb -relx 1.0 -rely 0.0 -anchor ne -relheight 1.0
   grid $hsb -row 1 -column 0 -sticky ew
   grid rowconfigure    $tf 0 -weight 1
   grid columnconfigure $tf 0 -weight 1
   ;# Below has hard reference to table .body
   mscroll $tbl.body
   return $tbl
}
#------------------------------------------------------------------------------
# trackList::saveColumnWidths
#
# Save the current column widths for later reuse
#
#
#------------------------------------------------------------------------------
proc trackList::saveColumnWidths {} {
   variable Index2File
   variable tableSound
   variable checkTime
   variable lastFolder
   variable TABLEDATA
   Trace
   ;# Get displayed col widths
   foreach {width name just} [$::PLB cget -columns] {
      set data($name) $width
   } ;# end foreach
   set colData [list]
   ;# Save real or if not default widths
   foreach {col} $id3Tag::TableCols {width} $id3Tag::TableWidths {
      if {[info exists data($col)]} {
         lappend colData $data($col)
      } else {
         lappend colData $width
      }
   } ;# end foreach
   set ::snackAmpSettings(TableWidths) $colData
   set ::snackAmpSettings(table,sortcol) [$::PLB sortcolumn]
   set ::snackAmpSettings(table,sortorder) [$::PLB sortorder]
}
#------------------------------------------------------------------------------
# trackList::populate
#
# Populate the table widget with track information
# if 'insert' is not true, the cache files are built
# but no attempt is made to populate the listbox
#------------------------------------------------------------------------------
proc trackList::populate {baseIndex tbl _config sort insert} {
   variable Index2File
   variable tableSound
   variable checkTime
   variable lastFolder
   variable TABLEDATA
   upvar $_config config
   set llen [llength $TABLEDATA]
   set col2Use [list]
   set isDirty 0
   catch {unset ::shortLong}
   # --------------------------------------------------------------
   foreach {col} $id3Tag::TableCols {
      if {[info exists config($col)] && $config($col)} {
         lappend col2Use $col
      } ;# end if
   } ;# end foreach
   catch {tablelist::canceleditingSubCmd $::PLB}

   set totalSize 0
   set totalTime 0
   set tableData [list]
   ;##
   ;# Go deep
   ;#
   set i 0
   foreach {index} [getFolders $baseIndex $::snackAmpSettings(treeDepth)] {
      update
      APL::busyCursor 1
      catch {unset Tags}
      catch {unset Cache}
      # Try to open cache file for reading, if it doesn't
      # exist flag it as dirty
      #
      set folder [catalog::TranslateMount $index]
      ;##
      ;# If a translate failed (e.g. remov media not there), don't bother to read/glob anything
      ;#
      set files [list]
      if {[string length $folder] > 0} {
         set dataFile [file join $folder .saDetails]
         if {[catch {open $dataFile "r"} fid] } then {
            set isDirty 1
         } else {
            if {! [eof $fid]} {
               #puts "Found Data"
               if {[catch {array set Tags [gets $fid]} result] } then {
                  saLog "Tags:$result"
                  set isDirty 1
               }
               set row [gets $fid]
               if {[catch {array set Cache $row} result] } then {
                  saLog "Cache:$result"
                  set isDirty 1
               }
            } else {
               #puts "New Cache"
               set isDirty 1
            }
            close $fid
         }
         #
         # For each file retrive and display the details
         #
         foreach {file} [glob -nocomplain -types {f l} -directory $folder *] {
            if {[info exists ::taExtensions([file extension $file])]} {
               if {$::TA(useSAFilter)==0 || ($::TA(useSAFilter)==1 && [string match -nocase "*$::TA(SAFilter)*" $file])} {
                  lappend files $file
               }
            }
         } ;# end foreach
      }
      ;##
      ;# If RMT in folder and [llength $files] == 0 then assume we have
      ;# to use the file only information contained in TFList (remov. media).
      ;#
      if {[llength $files] == 0 && [string match "${::RMT}*" $index]} {
         set pattern "[string trimright $index /]"
         foreach {File} [catalog::sortedList] {
            if {[string equal [file dirname $File] $pattern]} {
               set data [list]
               foreach {col} $col2Use {
                  switch -- $col {
                     Key     {lappend data $i}
                     File    {lappend data [file tail $File]}
                     default {lappend data 0}
                  } ;# End Switch
               } ;# end foreach
               lappend tableData $data
               set _Index2File($i) $File
               incr i
            } ;# end foreach
         }
      } else {
         ;##
         ;# Handle real files here
         ;#
         foreach {file} $files {
            catch {unset Details}
            set fName [file tail $file]
            file stat $file Stat        ;#mtime
            set ttag "$Stat(size):$Stat($checkTime)"
            ;##
            ;# Get cached copy of data if it exists
            ;#
            ;# Check time/size tag on cached file
            if {[info exists Tags($fName)] && [string equal $Tags($fName) $ttag]} {
               if {[catch {array set Details $Cache($fName)} result]} {
                  saLog $::errorInfo
                  continue
               }

               ;# Use fresh folder name
               set Details(Folder) [file dirname $file]
               ;# Use fresh gain values
               if {[catch {gainControl::get $file} gain]} {
                  set Details(Gain) ""
               } else {
                  set Details(Gain) [gainControl::levelText $gain]
               }

            } else {
               waitWindow::show "Caching [file tail $file]"
#saLog "[file tail $file]: $Tags($fName) vs. $ttag"
               if {$::tcl_version >= 8.4} {
                  ;# This is very fast with 8.4
                  catalog::addToDbIfNeeded $file
               } ;# end if
               id3Tag::songData $file $tableSound Details
               update ;# Make sure song continues to play, why does cursor resort back?
               set Details(Size) $Stat(size)
               set Details(Date) [clock format $Stat($checkTime) -format "%Y-%m-%d %H:%M:%S"]
               ;# Save data for this file
               set Cache($fName) [array get Details]
               set Tags($fName) $ttag
               set isDirty 1
            }
            catch {
               set totalSize [expr {$totalSize+$Details(Size)/1024}]
               set totalTime [expr {$totalTime+$Details(Duration)}]
            }
            set _Index2File($i) $file
            set data [list]
            set Details(Key) $i
            foreach {col} $col2Use {
               if {[info exists Details($col)]} {
                  lappend data $Details($col)
               } else {
                  lappend data ""
               }
            } ;# end foreach
            lappend tableData $data
            incr i
         } ;# end foreach
      }

      #
      # Update cache if needed, ro media will get an error
      #
      if {$isDirty} {
         if {![catch {open $dataFile "w"} fid] } then {
            puts $fid [array get Tags]
            puts $fid [array get Cache]
            close $fid
         }
      } ;# end if
   }
   #-----------------------------------------
   # Insert the list configSet into the tablelist widget
   # Only if data changed or files added/removed
   #puts "$insert && $isDirty && $llen != [llength $tableData]"
   if {$insert && ($isDirty || $llen != [llength $tableData] || ![string equal $baseIndex $lastFolder])} {
      if {![string equal $baseIndex $lastFolder]} {
         $::PLB selection clear 0 end
      } ;# end if
      ;# TABLEDATA is linked such that changing it changes
      ;# the displayed data.
      set trackList::TABLEDATA $tableData
      catch {unset Index2File}
      array set Index2File [array get _Index2File]
      set ::TA(ck,totalSize) "Size: [commify $totalSize] KB"
      set ::TA(ck,totalFolders) "Files: [commify $i]"
      set days [expr {int($totalTime)/86400}]
      if {$days == 0} {
         set days ""
      } ;# end if
      set ::TA(ck,totalTime) "Time:$days [timify $totalTime]"
      if {$sort} {
         ;##
         ;# Sort by last used or by col 1 by default (col 0 is hidden)
         ;#
         set lastSort [$::PLB sortcolumn]
         set lastSortOrder [$::PLB sortorder]
         if {$lastSort >= 0} {
            $tbl sortbycolumn $lastSort -$lastSortOrder
         } elseif {$::snackAmpSettings(table,sortcol) >= 0} {
            if {[catch {
               $tbl sortbycolumn $::snackAmpSettings(table,sortcol) \
                  -$::snackAmpSettings(table,sortorder)
            } result] } then {
               saLog $result
            }
         }
      } ;# end if
   } ;# end if
   APL::busyCursor 0
   waitWindow::close
   set lastFolder $baseIndex
}
###############################################################################
# Function   : trackList::getFolders
# Description: Return a list of some folders and subfolders of subfolders 'subfolders'
# Author     : Tom Wilkason
# Date       : 2/16/2003
###############################################################################
proc trackList::getFolders {selectedFolder {subfolders 0}} {
   lappend folders $selectedFolder
   ;# Find any remaining subfolders
   if {$subfolders>0} {
      set baseDepth [llength [file split $selectedFolder]]  ;#0 is just this folder
      foreach {index} [array names ::teFolders $selectedFolder/*] {
         set thisDepth [llength [file split $index]]
         if {$thisDepth <= ($baseDepth+$subfolders)} {
            lappend folders $index
         } ;# end if
      }
   }
   return [lsort $folders]
}
;##
;# Return a list of files within some folders
;#
proc trackList::getFiles {folders} {
   set result [list]
   foreach folder $folders {
      foreach candidate [catalog::filesMatchingPattern $folder*] {
         if {[string equal $folder [file dirname $candidate]]} {
            lappend result $candidate
         }
      }
   } ;# end foreach
   return $result
}
package provide autoPlayList 1.0
