##################################################################################
# snackAmp - Catalog package
#
# Music File Player in Tcl/Tk
#
# Copyright (C) 2001 Tom Wilkason
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
#
# Please send any comments/bug reports to
# tom.wilkason@cox.net  (Tom Wilkason)
##################################################################################
set [file root [file tail [info script]]]_History {
   $Header: /cvsroot/snackamp/snackamp/lib/catalog.tcl,v 1.53 2003/05/17 14:30:06 wilkason Exp $
}
namespace eval catalog {
   variable isSorted 0
}
;##
;# Create a csv file to contain the listing
;#
proc catalog::documentCache {args} {
   global TA TF snackAmpSettings
   ;##
   ;# Restore the cached files from each media Dir
   ;#
   # catch {unset teFolders}
   set catFiles [list]
   foreach {Path} $snackAmpSettings(mediaDirs) {
      catch {unset dirListing}
      if {[catch {set fid [open [file join $Path $TA(cacheFile)] r]}]} {
         return
      }
      fconfigure $fid -buffersize 100000
      catch {unset TaFiles}
      set TaFiles [gets $fid]
      close $fid
      ;##
      ;# Cached files are store relative to this dir so join with mediaDir to get abs file name
      ;#
      foreach {File} $TaFiles {
         lappend dirListing([file dirname $File]) [file tail $File]
      } ;# end foreach
      ;# Update latest cache time to current file
      if {[catch {set fid [open [file join $Path listing.csv] w]}]} {
         return -code error
      }
      lappend catFiles [file join $Path listing.csv]
      puts $fid [subst {"Folder","Title"}]
      foreach {Dir} [lsort [array names dirListing]] {
         update
         set Songs $dirListing($Dir)
         foreach {Song} [lsort $Songs] {
            puts $fid [subst {"$Dir","$Song"}]
         } ;# end foreach
      } ;# end foreach
      close $fid
   }
   tk_messageBox -icon info -type ok -message "csv list of files created\n[join $catFiles \n]"
}
###############################################################################
# Function   : catalog::addToDbIfNeeded
# Description: Add a file to TFList if it is not already there
# Author     : Tom Wilkason
# Date       : 12/15/2002
###############################################################################
proc catalog::addToDbIfNeeded {file} {
   global TFList
   set pos [catalog::whereInDb $file]
   if {$pos < 0} {
      lappend TFList $file
      catalog::dirtyList
   }
   return $pos
}
###############################################################################
# Function   : catalog::pushList/popList
# Description: Save/return the current master list order
# Author     : Tom Wilkason
# Date       : 12/18/2002
###############################################################################
proc catalog::pushList {lst} {
   set ::TA(lastList) $lst
   return 1
}
proc catalog::popList {args} {
   variable isSorted
   ;# If the master list is dirty, regenerate
   ;# the shuffled/sorted so we don't miss
   ;# any changes
   if {$isSorted==0} {
      catalog::makeSorted
      return [filtering::sortMasterList]
   } elseif {[info exists ::TA(lastList)]} {
      return $::TA(lastList)
   } else {
      return [filtering::sortMasterList]
   }
}
###############################################################################
# Function   : catalog::filesMatchingPattern
# Description: Return a list of files that match a pattern
# Author     : Tom Wilkason
# Date       : 12/15/2002
###############################################################################
if {$tcl_version < 8.4} {
   proc catalog::filesMatchingPattern {pattern} {
      global TFList
      set results [list]
      catalog::makeSorted
      foreach {File} $TFList {
         if {[string match $pattern $File]} {
            lappend results $File
         }
      }
      return $results
   }
   proc catalog::filesNotMatchingPattern {pattern} {
      global TFList
      set results [list]
      catalog::makeSorted
      foreach {File} $TFList {
         if {![string match $pattern $File]} {
            lappend results $File
         }
      }
      return $results
   }
   proc catalog::indexMatchingPattern {pattern} {
      global TFList
      set results [list]
      catalog::makeSorted
      set i 0
      foreach {File} $TFList {
         if {[string match $pattern $File]} {
            lappend results $i
         }
         incr i
      }
      return $results
   }
} else {
   proc catalog::filesMatchingPattern {pattern} {
      catalog::makeSorted
      return [lsearch -all -inline -glob $::TFList $pattern]
   }
   proc catalog::filesNotMatchingPattern {pattern} {
      catalog::makeSorted
      return [lsearch -not -all -inline -glob $::TFList $pattern]
   }
   proc catalog::indexMatchingPattern {pattern} {
      catalog::makeSorted
      return [lsearch -all -glob $::TFList $pattern]
   }
} ;# end if
###############################################################################
# Function   : catalog::whereInDb
# Description: Return the location in the file list of an item
# Author     : Tom Wilkason
# Date       : 12/15/2002
###############################################################################
if {$tcl_version < 8.4} {
   proc catalog::whereInDb {item} {
      catalog::makeSorted
      return [lsearch -exact $::TFList $item]
   }
} else {
   proc catalog::whereInDb {item} {
      catalog::makeSorted
      return [lsearch -exact -sorted $::TFList $item]
   }
} ;# end if
###############################################################################
# Function   : catalog::saveAs
# Description: Save the current file as some format
# Author     : Tom Wilkason
###############################################################################
proc catalog::saveAs {args} {
   set file [snack::getSaveFile]
   if {[string length $file]==0} {
      return
   }
   Stop
   ;# BUG: Conflict with s play -command and -progress callbacks, the
   ;# last one set is current.
   ;# Clone current sound
   set snd [snack::sound -file [soundControl::fileName]]
   ;# Get extension
   set ext [string tolower [string range [file extension $file] 1 end]]
   ;##
   ;# Status Window
   ;#
   waitWindow::show "Saving ... 0% complete"
   ;# Save the file
   if {[catch {
      switch -- $ext {
         "ogg" {
            #puts "saving ogg file"
            ;# TFW: Prompt for -nominalbitrate, -maxbitrate, -minbitrate or use existing
            switch -- [$snd config -fileformat] {
               MP3 {set rate [$snd config -bitrate]}
               OGG {set rate [$snd config -nominalbitrate]}
               default {set rate 0}
            } ;# End Switch
            if {$rate>0} {
               $snd write $file -fileformat $ext -progress catalog::waiter
            } else {
               $snd write $file -fileformat $ext -nominalbitrate $rate -progress catalog::waiter
            }
         }
         default {
            $snd write $file -fileformat $ext -progress catalog::waiter
         }
      } ;# End Switch
   } result]} {
      tk_messageBox -icon warning -type ok -title "Save Error" -message $result
   }
   ;# Make sure it is in the master list
   catalog::addToDbIfNeeded $file
   waitWindow::close
   $snd destroy
}
###############################################################################
# Function   : catalog::waiter
# Description: Update global used for wait box
# Author     : Tom Wilkason
###############################################################################
proc catalog::waiter {cmd percent} {
   set percent [expr {round($percent*1000)/10.0}]
   waitWindow::show  "Saving ... $percent% complete"
   update   ;# allow music to play
}
###############################################################################
# Function   : catalog::MonitorScanFolders
# Description: Monitor the base folders for incoming data, useful
#              for napaster adding files.
# Author     : Tom Wilkason
# Date       : 12/30/2000
###############################################################################
proc catalog::MonitorScanFolders {args} {
   global snackAmpSettings TA TF teFolders
   if {$snackAmpSettings(MonitorScanFolders)} {
      foreach {Folder} $snackAmpSettings(IncomingDirs) {
         ;# First time will be empty
         set TA($Folder,thisData) [list]
         foreach {file} [glob -nocomplain -directory $Folder *] {
            if {[info exists ::taExtensions([file extension $file])]} {
               lappend TA($Folder,thisData) $file
            }
         } ;# end foreach

         if {![info exists TA($Folder,lastData)]} {
            set TA($Folder,lastData) $TA($Folder,thisData)
         } elseif {[string match -nocase $TA($Folder,lastData) $TA($Folder,thisData)] == 0} {
            ;##
            ;# Detected added files
            ;#
            foreach {item} $TA($Folder,thisData) {
               # See if this is a new file, if so append it [lsearch -exact $TA($Folder,lastData) $item]
               if {[catalog::whereInDb $item] < 0} {
                  filtering::Append $item  ;# This will also add it to the db
                  set teFolders([file dirname $item])  1   ;# Sometime case is different on directory
                  saLog "Detected new file $item"
                  catalog::FlagDirtyCache $item
               } ;# end if
            } ;# end foreach
            ;##
            ;# Detected removed or moved files
            ;# Check each old file to see if it is in the recent list (which may be shorter)
            foreach {oldfile} $TA($Folder,lastData) {
               ;# If old files not found in this list
               if {[lsearch -exact $TA($Folder,thisData) $oldfile] < 0} {
                  set index [catalog::whereInDb $oldfile]
                  if {$index >= 0} {
                     puts "Detected removal of $oldfile"
                     ldelete ::TFList $index
                     ;# Remove any entries in the playlist
                     removeFile $oldfile
                     catalog::FlagDirtyCache $oldfile
                  }
               } ;# end if
            } ;# end foreach
            set TA($Folder,lastData) $TA($Folder,thisData)
         } ;# end if
      } ;# end foreach
   }
   after cancel catalog::MonitorScanFolders
   after 10000 catalog::MonitorScanFolders
}

###############################################################################
# Function   : catalog::fixMediaPath
# Description: Make sure the directories in the media path don't have trailing slashes
#              needed by tree matching functions
# Author     : Tom Wilkason
# Date       : 1/11/2001
###############################################################################
proc catalog::fixMediaPath {args} {
   cleanPathList ::snackAmpSettings(IncomingDirs) 1
   cleanPathList ::snackAmpSettings(mediaDirs) 1
   cleanPathList ::snackAmpSettings(RemovableFolders) 0
   foreach {folder} $::snackAmpSettings(mediaDirs) {
      if {![info exists ::TA(Dirty,$folder)]} {
         set ::TA(Dirty,$folder) 0
      } ;# end if
   } ;# end foreach
}
;##
;# Function to routinely monitor the Cache files and read them in if they change
;#
proc catalog::MonitorCacheFolders {args} {
   global snackAmpSettings TA TF
   if {$snackAmpSettings(MonitorCacheFolders)} {
      foreach {Folder} $snackAmpSettings(mediaDirs) {
         if {[catch {
            file stat [file join $Folder $TA(cacheFile)] Stat
            if {![info exists TA(CheckedCache,$Folder)]} {
               set TA(CheckedCache,$Folder) $Stat(mtime)
            } else {
               if {$Stat(mtime) > $TA(CheckedCache,$Folder)} {
                  ReadCacheFiles
                  saLog "[clock format [clock seconds]]: Read Cache $Stat(mtime) vs. $TA(CheckedCache,$Folder)"
                  break
               } ;# end if
            }
         } result]} {
            saLog "MonitorCacheFolders: $result"
         }
      }
   } ;# end if
   ;# Check in two minutes
   after cancel catalog::MonitorCacheFolders
   after 120000 catalog::MonitorCacheFolders
}
###############################################################################
# Function   : catalog::Rescan
# Description: Reread the directories for files, include subdirectories
#              For each folder, add a menu entry
# Author     : Tom Wilkason
# Date       : 12/26/2000
###############################################################################
proc catalog::RescanAskFirst {args} {
   if {[tk_messageBox -icon question -type yesno -title "Please Respond" \
      -message "Are you sure you want to catalog to following folders?\n$::snackAmpSettings(mediaDirs)"]=="yes"} {
      catalog::Rescan
   } ;# end if
}
proc catalog::Rescan {{Folder ""}} {
   global TF TFList TA LB snackAmpSettings teFolders
   Wait 1
   set here [pwd]
   ;# If now folder specifed, catalog all media dirs, otherwise just
   ;# catalog the specified folder retaining information on other folders.
   if {$Folder == ""} {
      set TFList [list]   ;# //Lauras/Media/New Age/Various Artists/Celtic Voices - Women Of Song.mp3
      ;# save folder settings
      catch {array set saveTeFolders [array get teFolders]}
      catch {unset teFolders}                      ;#//Lauras/Media/New Age
      ;# Get a list of all the files into TFList
      foreach dir $snackAmpSettings(mediaDirs) {
         RescanGetRecurseFiles $dir
      } ;# end foreach
   } else {
      ;# Only remove items from the folder to be scanned
      set newTFList [list]
      foreach {item} [catalog::filesNotMatchingPattern "$Folder*"] {
         lappend newTFList $item
      } ;# end foreach
      set TFList $newTFList
      ;# Remove auto-play list folders withing this folder
      foreach {item} [array names teFolders "$Folder*"] {
         unset teFolders($item)
      } ;# end foreach
      ;# Rescanning will re-insert the files in the Folder.
      RescanGetRecurseFiles $Folder
   }
   catalog::dirtyList
   ;# Restore exclusion settings, teFolders will now only contain
   ;# folders just found, and if they were there before the value
   ;# will be restored, otherwise they will be disabled.
   ;##
   ;# Build the folder list for the menu
   ;#
   waitWindow::show "Caching Data ... Almost There"
   update
   ;# Save the files
   RescanSaveCache $Folder
   ;# Reread cache to set teFolders and sort the music
   ;# (only if all media dirs were scanned)
   waitWindow::close
   if {$Folder == ""} {
      ReReadCatalogs
   } else {
      ;##
      ;# If the auto play list tree exists, then replace it
      ;#
      if {[APLExists]} {
         APL::buildAPL $::TREE
         APL::treeBind
         Tree::open $::TREE / 1
      } ;# end if
   }
   ;##
   ;# Reapply the APL file
   ;#
   readAPLFile $snackAmpSettings(lastFolderFile)
   Wait 0
}
;##
;# Recursively scan a directory and build a list of all files
;#
proc catalog::RescanGetRecurseFiles {dir} {
   global TF TA LB teFolders TFList
   set dirList [list]
   ;# Show banner every once in a while
   if {[clock clicks -milliseconds]%10 == 0} {
      waitWindow::show "Scanning $dir"
   } ;# end if
   update       ;# Allow music to play and update display
   set hasSome 0
   if {$::snackAmpSettings(fastscan)} {
   ;# This is the fastest recursive globbing method I can come up with
      foreach file [glob -nocomplain [file join $dir *]] {
         ;# This is problematic if folder ending in .mp3
         if {[info exists ::taExtensions([file extension $file])]} {
            lappend ::TFList $file
            incr hasSome
         } elseif {[file isdirectory $file]} {
            lappend dirList $file
            incr hasSome
         }
      }
   } else {
      ;# This is a bit slower due to additional directory checks
      ;# but will see ALL directories
      foreach file [glob -nocomplain [file join $dir *]] {
         if {[file isdirectory $file]} {
            lappend dirList $file
            incr hasSome
         } elseif {[info exists ::taExtensions([file extension $file])]} {
            lappend ::TFList $file
            incr hasSome
         }
      }
   }
   ;# Only set the folder (as existing) for the tree if
   ;# it was non-empty
   if {$hasSome} {
      set teFolders($dir) 0     ;# Add this dir, if new will be turned off
   }
   ;# process each directory recursively
   foreach {folder} $dirList {
      RescanGetRecurseFiles $folder
   } ;# end foreach
}
###############################################################################
# Function   : RescanSaveCache
# Description: Save the cached data
# Author     : Tom Wilkason
# Date       : 1/7/2001
###############################################################################
proc catalog::RescanSaveCache {{File {}}} {
   global TA TF teFolders snackAmpSettings TFList
   ;##
   ;# Save the settings
   ;#
   fixMediaPath
   ;##
   ;# Cache the data in each mediaDir for shared use
   ;# If a file was specifed, only do the folder that
   ;# it is contained in
   set Folders [list]
   if {$File == {}} {
      set Folders $snackAmpSettings(mediaDirs)
   } else {
      foreach Path $snackAmpSettings(mediaDirs) {
         if {[string match "$Path*" $File] } {
            lappend Folders $Path
         }
      } ;# end foreach
   }
   ;##
   ;# For each path in the mediaFolders or user specified folder
   ;# save the catalog to that path root folder.
   ;#
   foreach {Path} $Folders {
      ;# For each media path, strip if off and save data there, array names
      set PathDat [list]
      set lPath [llength [file split $Path]]
      foreach {File} [catalog::filesMatchingPattern "$Path*"] {
         ;# Get path name with front path part removed.
         lappend PathDat [eval file join [lrange [file split $File] $lPath end]]
      } ;# end foreach
      saLog "Saving [llength $PathDat] files for $Path..."

      if {[catch {safeWrite [file join $Path $TA(cacheFile)] $PathDat} result] } then {
         waitWindow::close
         tk_messageBox -icon warning -type ok -title "Access Problem" \
         -message "Could not gain write access to catalog file [file join $Path $TA(cacheFile)]\n$result"
         continue
      }
      set ::TA(Dirty,$Path) 0
      ;# Update latest cache time to current file
      if {[catch {
         file stat [file join $Path $TA(cacheFile)] Stat
      } result] \
      } then {
         saLog $result
      } else {
         set TA(CheckedCache,$Path) $Stat(mtime)
      }
   } ;# end foreach
}
###############################################################################
# Function   : SaveDirtyCache
# Description: Save Dirty Caches, call before exit and such
# Author     : Tom Wilkason
###############################################################################
proc catalog::SaveDirtyCache {args} {
   foreach {folder} $::snackAmpSettings(mediaDirs) {
      if {[info exists ::TA(Dirty,$folder)] && $::TA(Dirty,$folder) > 0} {
         RescanSaveCache $folder
      } ;# end if
   } ;# end foreach
}
###############################################################################
# Function   : catalog::FlagDirtyCache
# Description: Flag one of the media dir caches as dirty, so we can know to
#              write it later.
# Author     : Tom Wilkason
###############################################################################
proc catalog::FlagDirtyCache {{File {}}} {
   global TA TF teFolders snackAmpSettings TFList
   ;##
   ;# Save the settings
   ;#
   fixMediaPath
   ;##
   ;# Cache the data in each mediaDir for shared use
   ;# If a file was specifed, only do the folder that
   ;# it is contained in
   set Folders [list]
   if {$File == {}} {
      set Folders $snackAmpSettings(mediaDirs)
   } else {
      foreach Path $snackAmpSettings(mediaDirs) {
         if {[string match "$Path*" $File] } {
            set Folders [list $Path]
            break
         }
      } ;# end foreach
   }
   ;# Increment dirty counter
   foreach {folder} $Folders {
      incr TA(Dirty,$folder)
   } ;# end foreach
}
;##
;# Re-read all of the catalog files
;#
proc catalog::ReReadCatalogs {args} {
   catch {unset ::teFolders}
   soundControl::latency long
   catalog::ReadCacheFiles
   if {[APLExists]} {
      APL::buildAPL $::TREE
      APL::treeBind
      Tree::open $::TREE / 1
   } ;# end if
   soundControl::latency normal
}
;##
;# Read the cached files into memory
;# Set up the default teFolders
;# Returns a list of paths not found
proc catalog::ReadCacheFiles {args} {
   global TA TF TFList teFolders snackAmpSettings
   ;##
   ;# Restore the cached files from each media Dir
   ;# Support either new type (a list) or old type (Array)
   set result [list]
   set TFList [list]
   ;# Move list into a hash table for faster access
   ;# and to allow replacing parts by single path
   foreach {Path} [lsort $snackAmpSettings(mediaDirs)] {
      waitWindow::show "Reading Directory $Path - Please Wait"
      update
      ;# Read either v1 or v2 cache files
      if {![file exists $Path]} {
         waitWindow::close
         tk_messageBox -icon warning -type ok -title "Warning" \
            -message "Media Folder '$Path' does not exist,\nit may be named improperly or on \nan unavailable network computer."

      } else {
         if {[catch {set fid [open [file join $Path $TA(cacheFile)] r]}]} {
            lappend result $Path
         } else {
            ;# V2 files
            fconfigure $fid -buffersize 100000
            gets $fid tLlist
            close $fid
            saLog "Reading $Path (v2)"
            # add the new ones for this path back
            foreach {File} $tLlist {
               lappend TFList [file join $Path $File]
            }
            ;# Update latest cache time to current file
            file stat [file join $Path $TA(cacheFile)] Stat
            set ::TA(Dirty,$Path) 0
            set TA(CheckedCache,$Path) $Stat(mtime)
         }
      }
   }
   ;# Read Removable Media Catalogs, merge in file names here
   ;# to prevent duplicates.
   if {$::snackAmpSettings(UseRemovableFolders)} {
      saLog "Reading Removable Folders"
      waitWindow::show "Reading Removable Folders"
      update
      foreach {file} [readMediaCatalogs] {
         lappend TFList $file
      } ;# end foreach
   }
   ;#
   catalog::dirtyList
   waitWindow::show "Building Folder List - Please Wait"
   update
   ;# Need to make sure teFolder entries all the way back
   ;# to the root exist. Since the files are sorted, the
   ;# first ones in the list will be seen first so they
   ;# will typically make sure the path to the root is
   ;# set. This is very fast.
   foreach {Index} [catalog::sortedList] {
      set folder [file dirname $Index]
      while {! [info exists teFolders($folder)]} {
         set teFolders($folder) 0
         ;# Backup one
         set root [file dirname $folder]
         if {[string equal $folder $root]} {
            break
         } else {
            set folder $root
         }
      }
   }
   waitWindow::close
   return $result
}
###############################################################################
# Function   : catalog::manageMediaFolders
# Description: Bring up dialog then update folder caches if needed
# Author     : Tom Wilkason
# Date       : 6/5/2002
###############################################################################
proc catalog::manageMediaFolders {args} {
   package require saSettings
   global snackAmpSettings
   set oldMedia $::snackAmpSettings(mediaDirs)
   folderManager::launch mediaDirs
   ;# Blocked here until GUI closes
   catalog::fixMediaPath  ;# Remove braces if added pulling data off listbox
   if {$oldMedia != $::snackAmpSettings(mediaDirs)} {
      $::TA(treeButton) configure -relief flat
      destroy .wFC
      ;# Re-read/scan catalogs now that something changed
      set result [catalog::ReadCacheFiles]
      if {[llength $result] > 0} {
         catalog::checkRescanCaches $result
      }
   } ;# end if
}
###############################################################################
# Function   : checkRescanCaches
# Description: Ask use if he wants to rescan folders
# Author     : Tom Wilkason
# Date       : 6/5/2002
###############################################################################
proc catalog::checkRescanCaches {folderList} {
   global snackAmpSettings
   set ans [tk_messageBox -type yesno -title "Catalogs not found" \
      -icon question -message "Do you want to catalog the following music folders now?\n[join $folderList \n]"]
   if {$ans == "yes"} {
      foreach {folder} $folderList {
         catalog::Rescan $folder
      } ;# end foreach
      readAPLFile $snackAmpSettings(lastFolderFile)
   } else {
      tk_messageBox -title "SnackAmp - Please Respond" \
         -icon info -message "Don't forget to catalog [join $folderList ,] Before you try to use them"
   }
}
###############################################################################
# Function   : initCacheFiles
# Description: Rescan or read the existing cache files
# Author     : Tom Wilkason
# Date       : 6/5/2002
###############################################################################
proc catalog::initCacheFiles {args} {
   global snackAmpSettings TA
   ;##
   ;# Rescan if set true or the cache read failed
   ;#
   if {$snackAmpSettings(AutoRescan)} {
      catalog::Rescan
   }
   ;# Try to read caches
   set result [catalog::ReadCacheFiles]
   if {[llength $result] > 0} {
      catalog::checkRescanCaches $result
   } else {
      ;# Read cache files, now restore playlist if required
      ;# otherwise just toss up music based on default auto-playlist file
      if {$snackAmpSettings(restorePlayList)} {
         waitWindow::show "Reading Playlist Cache"
         if {[catch {catalog::ReadPlaylistCache} result]} {
            waitWindow::show "Reading $snackAmpSettings(lastFolderFile)"
            readAPLFile $snackAmpSettings(lastFolderFile)
            waitWindow::show "Filtering Music"
            macros::filterMusic replace $::TA(matchAgainstAll)
            saLog "No play cache file found\n$result"
         } else {
            waitWindow::show "Reading $snackAmpSettings(lastFolderFile)"
            readAPLFile $snackAmpSettings(lastFolderFile)
         }
      } else {
         waitWindow::show "Reading $snackAmpSettings(lastFolderFile)"
         readAPLFile $snackAmpSettings(lastFolderFile)
         waitWindow::show "Filtering Music"
         macros::filterMusic replace $::TA(matchAgainstAll)
         filtering::sortMusic
      }
   }
   waitWindow::close
}

###############################################################################
#            R E M O V A B L E   M E D I A   S U P P O R T
###############################################################################

###############################################################################
# Function   : catalog::TagMedia
# Description: Return a hopefully unique tag for some removable media, note
#              if the media changes then the tags won't match.
# Author     : Tom Wilkason
###############################################################################
proc catalog::TagMedia {vol} {
   if {[catch {set first [lindex [lsort [glob [file join $vol *]]] 0]} result]} {
      return -code error $result
   }
   return [file tail "$first-[file mtime $first]"]
}
###############################################################################
# Function   : catalog::CatalogMedia
# Description: Reread the directories for files, include subdirectories
#              For each folder, add a menu entry
# Author     : Tom Wilkason
# Date       : 12/26/2000
###############################################################################
proc catalog::CatalogMedia {{vol ""}} {
   global TF TA LB snackAmpSettings teFolders
   if {$snackAmpSettings(UseRemovableFolders) == 0} {
      set ans [tk_messageBox -icon question -type yesno -title "Warning" \
         -message "You have Removable Media disabled, do you want to enable it?"]
      if {$ans == "yes"} {
         set snackAmpSettings(UseRemovableFolders) 1
         catalog::ReadCacheFiles
      } ;# end if
   }
   if {[llength $snackAmpSettings(RemovableFolders)] == 0} {
      set ans [tk_messageBox -icon question -type yesno -title "Warning" \
         -message "You need to setup your removable media first,\ndo you want to set them up now?"]
      if {$ans=="yes"} {
         package require saSettings
         after idle folderManager::launch RemovableFolders
         return
      } else {
         return
      }
   }
   if {$vol == ""} {
      set ::useFolder [lindex $snackAmpSettings(RemovableFolders) 0]
   } else {
      set ::useFolder $vol
   }
   toplevel .p
   wm transient .p .saTop
   wm title .p "Removable Media Selector"
   pack [message .p.l -width 300 \
      -text "Select Removable Media folder, Enter Label for it, Insert Media into drive, then press OK to create a catalog for it"] \
      -side top -expand true -fill both
   eval tk_optionMenu .p.vol ::useFolder $snackAmpSettings(RemovableFolders)
   .p.vol configure -width 20
   pack .p.vol -side top
   pack [entry .p.en -font entry -textvariable ::MediaLabel -background $::snackAmpSettings(ebg) -foreground $::snackAmpSettings(efg)] \
      -side top -fill x -expand true
   settings::regColor .p.en -background ::snackAmpSettings(ebg) -foreground ::snackAmpSettings(efg)
   pack [frame .p.b -height 2 -borderwidth 2 -relief sunken] \
      -fill x -expand true -side top
   pack [frame .p.bb] -side top
   pack [button .p.bb.ok -width 8 -borderwidth 1  -relief raised -text OK -command {
      destroy .p
      set ::TA(continue) 1
   }] -side left -padx 6
   pack [button .p.bb.cancel -width 8 -borderwidth 1 -relief raised -text Cancel -command {
      destroy .p
      set ::TA(continue) 0
   }] -side left -padx 6
   focus .p.en
   bind .p <Return> [list .p.bb.ok invoke]
   bind .p <Escape> [list .p.bb.cancel invoke]
   centerWindow .p
   myIcon .p
   ;# Wait for user to close window
   grab .p
   tkwait visibility .p
   raise .p
   set ::TA(continue) 0
#  ----------------------------------------------------------------------------
#  Blocking
#  ----------------------------------------------------------------------------
   vwait ::TA(continue)
   ;# Cancel key hit
   if {$::TA(continue) == 0} {
      return
   } ;# end if
   ;# No label
   set ::MediaLabel [string trim $::MediaLabel]
   if {[string length $::MediaLabel]==0} {
      set result [tk_messageBox -icon question -type yesno -title "No Label" \
         -message "You did not specify a catalog label, do you want to start over?"]
      if {$result == "yes"} {
         after 100 catalog::CatalogMedia $::useFolder
      } else {
         return
      }
   }
   ;##
   ;# Now catalog the media
   ;#
   set ::mediaList [list]
   ;# save folder settings
   set Tag [file join $::HOME $::MediaLabel.rmc]
   if {[file exists $Tag]} {
      set result [tk_messageBox -icon question -type yesno -title "File Exists" \
         -message "Select Yes to overwrite file $Tag. No to to try again"]
      if {$result == "no"} {
         after 100 catalog::CatalogMedia
         Wait 0
         return
      } ;# end if

   } ;# end if
   ;# Prompt user for Name (TBD)
   if {[catch {catalog::TagMedia $::useFolder} result]} {
      tk_messageBox -icon warning -type ok -title "Tag Problem" \
         -message "Could not create Tag for $::useFolder\nIt may be blank\n$result"
      return
   }
   Wait 1
   waitWindow::show "Cataloging $::useFolder ...Please Wait"
   update
   ;# useFolder has to end in / for this to work (they should though)
   CatalogMediaRecurse $::useFolder [llength [file split $::useFolder]]
   if {[catch {set fid [open $Tag w]} result]} {
      waitWindow::close
      tk_messageBox -icon warning -type ok -title "Access Problem" \
         -message "Could not gain write access to $Tag"
      Wait 0
      waitWindow::close
      return
   }
   fconfigure $fid -buffersize 100000

   puts $fid [catalog::TagMedia $::useFolder]
   puts $fid $::mediaList
   close $fid
   unset ::mediaList
   ReadCacheFiles
   Wait 0
   waitWindow::close
   update
   catch {destroy .wFC}
   tk_messageBox -icon info -type ok -title "Catalog" \
      -message "Catalog of '$::MediaLabel' Complete"

}
###############################################################################
# Function   : CatalogMediaRecurse
# Description: Recursively scan a directory and build a list of all files
# Author     : Tom Wilkason
###############################################################################
proc catalog::CatalogMediaRecurse {dir strip} {
   global TF TA LB teFolders TFList
   set dirList [list]
   update       ;# Allow music to play and update display
   # saLog "Scanning: $dir"
   foreach file [glob -nocomplain [file join $dir *]] {
      if {[file isdirectory $file]} {
         lappend dirList $file
      } elseif {[info exists ::taExtensions([file extension $file])]} {
         set appendFile [eval file join [lrange [file split $file] $strip end]]
         lappend ::mediaList $appendFile
      }
   }
   ;# process each directory recursively
   foreach {file} $dirList {
      CatalogMediaRecurse [file join $dir $file] $strip
   } ;# end foreach
}
###############################################################################
# Function   : readMediaCatalogs
# Description: Read the disk based media catalogs for the previously indexed
#              media folders. Build up proper name
# Author     : Tom Wilkason
# Date       : 3/5/2001
###############################################################################
proc catalog::readMediaCatalogs {args} {
   set catalogList [list]
   foreach {File} [glob -nocomplain [file join $::HOME *.rmc]] {
      if {[catch {set fid [open $File r]} result]} {
         waitWindow::close
         tk_messageBox -icon warning -type ok -title "$File Open Error" -message $result
         continue
      }
      ;# Save the tag for this description
      set mediaTag [gets $fid]
      set Entries [gets $fid]
      close $fid
      set Description [file rootname [file tail $File]]
      set ::MountTagToDescription($mediaTag) $Description
      foreach {Item} $Entries {
         lappend catalogList [file join $::RMT $Description $Item]
      } ;# end foreach
   } ;# end foreach
   return [lsort $catalogList]
}
###############################################################################
# Function   : findIfExists
# Description: Return the file name of a song if it really exists,
#              attempt to translate the name into a removeable media mount
#              even possibly asking the user to mount the volume.
# Author     : Tom Wilkason
###############################################################################
proc catalog::findIfExists {fullName} {
   if {[file exists $fullName] || ([set fullName [TranslateMount $fullName]]!={})} {
      return $fullName
   } else {
      return {}
   }
}
###############################################################################
# Function   : isARemMedia
# Description: Return true if the file is on a removable media
#              find if it is in a mounted folder. If it is not, then prompt the
#              user to mount it.
# Author     : Tom Wilkason
###############################################################################
proc catalog::isARemMedia {File} {
   return [string match "$::RMT*" $File]
}
###############################################################################
# Function   : TranslateMount
# Description: During a play, determine the true filename if it doesn't exist
#              find if it is in a mounted folder. If it is not, then prompt the
#              user to mount it.
# Author     : Tom Wilkason
###############################################################################
proc catalog::TranslateMount {File} {
   ;# Get lower two parts of name
   if {[string match [file join $::RMT *] $File]} {
      set parts [file split $File]
      ;# Rip off the RMC and Description parts
      ;# so we can replace them from the
      ;# mount lookups
      set base [eval file join [lrange $parts 0 2]]
      set passes 0
      while {1} {
         if {[info exists ::MountedMedia($base)]} {
            set result [eval file join [list $::MountedMedia($base)] [lrange $parts 3 end]]
            if {[file exists $result]} {
               ;# All is well
               return $result
            } else {
               ;# Prompt for mount if enabled, otherwise
               ;# try to silently mount once.
               ;# This shouldn't happen unless the the media
               ;# was swapped from the CD.
               if {$::snackAmpSettings(PromptRemovableFolders)} {
                  MountMedia "'[lindex $parts 2]'" 0
               } else {
                  if {$passes==0} {
                     MountMedia "'[lindex $parts 2]'" 1
                     incr passes
                  } else {
                     return {}
                  }
               }
            }
         } else {
            ;# Prompt for mount
            ;# This will happen if not yet mounted.
            if {$::snackAmpSettings(PromptRemovableFolders)} {
               MountMedia "'[lindex $parts 2]'" 0
            } else {
               if {$passes==0} {
                  MountMedia "'[lindex $parts 2]'" 1
                  incr passes
               } else {
                  return {}
               }
            }
         }
      }
   } else {
      if {[file exists $File]} {
         return $File
      } else {
         return {}
      }
   }
}
###############################################################################
# Function   : MountMedia
# Description: Go through all of the mountable volumes and determine
#              what is each of them.
# Author     : Tom Wilkason
###############################################################################
proc catalog::MountMedia {Message {quiet 0}} {
   global snackAmpSettings
   if {! $quiet} {
      if {$::snackAmpSettings(UseRemovableFolders) == 0} {
         set ans [tk_messageBox -icon question -type yesno -title "Warning" \
            -message "You have Removable Media disabled, do you want to enable it?"]
         if {$ans == "yes"} {
            set ::snackAmpSettings(UseRemovableFolders) 1
            catalog::ReadCacheFiles
         } else {
            return
         }
      }
      if {[llength $::snackAmpSettings(RemovableFolders)] == 0} {
         set ans [tk_messageBox -icon question -type yesno -title "Warning" \
            -message "You need to setup your removable media first,\ndo you want to set them up now?"]
         if {$ans=="yes"} {
            package require saSettings
            after idle folderManager::launch RemovableFolders
            return
         } else {
            return
         }
      }
      set result [tk_messageBox -icon question -type okcancel \
         -message "Please insert $Message media into one or more of\n$::snackAmpSettings(RemovableFolders)" ]
      if {$result == "cancel" && $::snackAmpSettings(PromptRemovableFolders)} {
         set ans [tk_messageBox -icon question -type yesno \
            -message "To prevent furthur prompts can I turn off\n'Allow Removable Media Prompts'" ]
         if {$ans=="yes"} {
            set ::snackAmpSettings(PromptRemovableFolders) 0
         } ;# end if
         return
      } ;# end if
   } ;# end if
   set Mounted {}
   set Empty {}
   catch {unset ::MountedMedia}
   foreach {vol} $::snackAmpSettings(RemovableFolders) {
      ;##
      ;# Foreach media that has data in it.
      ;#
      if {![catch {set Tag [catalog::TagMedia $vol]} result]} {
         ;# Save a mapping from the tag to the vol name
         ;# Skip if find media that hasn't been indexed
         if {[catch {set ::MountedMedia([file join $::RMT $::MountTagToDescription($Tag)]) $vol} result]} {
            if {! $quiet} {
               saLog "Unknown Media Found:$Tag not in '[array names ::MountTagToDescription]'"
               set result [tk_messageBox -icon question -type yesno \
                  -message "Unknown Media Found:$Tag in $vol\ndo you want to catalog it?'" ]
               if {$result=="yes"} {
                  set ::useFolder $vol
                  after idle catalog::CatalogMedia
                  return
               } ;# end if
               ;# Prompt user if they want to catalog it.
            }
         } else {
            append Mounted "'$::MountTagToDescription($Tag)' mounted on '$vol'\n"
         }
         ;# Lookup in Play is if file doesn't exist, then
         ;# Look for it in MountedMedia, if exists then map
         ;# to the vol name. Then correctCase if it exists again.
      } else {
         append Empty "\nMedia '$vol' is empty"
      }
   }
   if {! $quiet} {
      if {$Mounted != {}} {
         tk_messageBox -type ok -icon info \
         -message "The following volumes were mounted\n$Mounted$Empty"

      } else {
         tk_messageBox -type ok -icon warning \
         -message "No media was found in one of \n$::snackAmpSettings(RemovableFolders)$Empty"
      }
   } ;# end if
}
###############################################################################
# Function   : catalog::songInfo
# Description: Return info on the song
# Author     : Tom Wilkason
# Date       : 4/23/2002
###############################################################################
proc catalog::channelInfo {sInfo _Details} {
   Trace
   upvar $_Details Details
   ;# Get the bitrate info
   if {[catch {
      switch -- [$sInfo config -fileformat] {
         MP3 {set rate [expr {[$sInfo config -bitrate]/1000}]}
         OGG {set rate [expr {[$sInfo config -nominalbitrate]/1000}]}
         WAV {set rate [expr {[$sInfo config -rate]/1000}]}
         default {set rate 0}
      } ;# End Switch
   }]} then {
      set rate "Unknown"
   }
   set Details(rate) $rate
   set Details(sampRate) [expr {double([$sInfo cget -frequency])/1000.}]
   set len [$sInfo length -units sec]
   set Details(length) $len
   set Details(duration) [clock format [expr {int($len)}] -format "%M:%S"]
   return 1
}
###############################################################################
# Function   : catalog::songInfo
# Description: Return info on the song
# Author     : Tom Wilkason
# Date       : 4/23/2002
###############################################################################
proc catalog::songInfo {_Details fileName} {
   Trace
   upvar $_Details Details
   set type [string trim [file extension $fileName] .]
   set Details(size) [file size $fileName]
   if {[catch {snack::sound -file $fileName -fileformat $type} sInfo] } then {
      set Details(rate)     0
      set Details(sampRate) 0.0
      set Details(length)   0
      set Details(duration) "00:00"
      saLog $sInfo
      return 0
   } else {
      channelInfo $sInfo Details
      $sInfo destroy
      return 1
   }
}
###############################################################################
# Function   : catalog::RenameFile
# Description: Try to rename/move a file, handle sound objects if they are playing
#              resume play if just renamed a playing file
# Author     : Tom Wilkason
# Date       : 3/22/2002
###############################################################################
proc catalog::RenameFile {oldName newName} {
   Trace
   set wasPlaying 0
   ;##
   ;# Can't rename files on mounted media
   ;#
   if {[isARemMedia $oldName] || [isARemMedia $newName] || $::snackAmpSettings(dontMoveFiles)} {
      return -code error "\nYou don't have permission to rename or move $oldName"
   }
   ;##
   ;# If the current song playing is the one to be renamed
   ;# we must unlink it from snack then later relink it
   ;# TFW: Need to add code to prevent writing to RMC files
   set justCaseChange [expr {[string equal -nocase $oldName $newName] && ($::tcl_platform(platform) == "windows")}]
   if {[file exists $newName] && !$justCaseChange} {
      catalog::songInfo oldInfo $oldName
      catalog::songInfo newInfo $newName
      ;# oldName is really latest file, newName will be existing file
      set textInfo "
     \tNew ->\tExisting
Rate:\t$oldInfo(rate)\t$newInfo(rate)
Len: \t$oldInfo(duration)\t$newInfo(duration)
Size:\t$oldInfo(size)\t$newInfo(size)
"
      if {[file isdirectory $newName]} {
         set res [tk_messageBox -icon question -type yesno -title "Target Exists" \
            -message "Warning: Folder $newName exists\nis it OK to overwrite any duplicate files?"]
      } elseif {($oldInfo(size) == $newInfo(size))} {
         set res [tk_messageBox -icon question -type yesno -title "File Exists" \
            -message "Warning: $newName exists\nand appears to be identical, do you still want to overwrite it?"]
      } else {
         set res [tk_messageBox -icon question -type yesno -title "File Exists" \
            -message "Warning: $newName\nexists, do you want to overwrite it?\n$textInfo"]
      }
      if {$res != "yes"} {
         return -code error "$newName File Exists"
      } ;# end if
   } ;# end if
   set currentSound [soundControl::fileName]
   if {[string equal $currentSound $oldName]} {
      ;# This is a trick to see if the song is actually playing
      set wasPlaying [soundControl::isPlaying]
      set location [soundControl::samplePosition]
      soundControl::stop
      soundControl::reset
      set resumePlay 1
   } else {
      set resumePlay 0
   }
   ;##
   ;# Rename the file, if on windows then allow of a case insensitive name
   ;# Note: -forces allows overwritting of files but supresses errors if
   ;# the overwrite failed. This messes up upper error handling.
   if {$justCaseChange} {
      if {[catch {file rename $oldName $oldName.tmp} result]} {
         return -code error $result
      } else {
         #debug "0 renamed \n $oldName \n $oldName.tmp"
         if {[catch {file rename $oldName.tmp $newName} result]} {
            return -code error $result
         }
         FlagDirtyCache $newName
         FlagDirtyCache $oldName
         #debug "1 renamed \n $oldName.tmp \n $newName"
      }
   } else {
      ;# User may have inserted path separators to move into new folder
      set dirPart [file dirname $newName]
      if {![file exists $dirPart]} {
         if {[catch {file mkdir $dirPart} result]} {
            return -code error $result
         }
      } ;# end if
      if {[file isdirectory $oldName] && [file isdirectory $newName]} {
#         saLog "Moving folders '$oldName' to existing folder '$newName'"
         set srcs [glob -nocomplain -directory $oldName *]
         set errors 0
         set Result ""
         foreach {file} $srcs {
            if {[catch {file rename -force -- $file $newName} result]} {
               append Result "\n$result"
               incr errors
            } else {
#               saLog "moved '$file' to '$newName'"
            }
            update
         }
         if {$errors > 0} {
            return -code error " ($errors) Move Errors $Result"
         }
         ;# Old folder will be left, can remove it if empty (if all went well)
         if {[file exists $oldName] && [llength [glob -nocomplain -directory $oldName *]]==0} {
            #debug "Removing '$oldName'"
            catch {file delete $oldName}
         }
       } else {
         # The target folder is new
         if {[catch {file rename -force -- $oldName $newName} result]} {
            return -code error $result
         }
      }
      FlagDirtyCache $newName
      FlagDirtyCache $oldName
      #debug "2 renamed \n $oldName \n $newName"
   }
   ;##
   ;# Resume playing if active song was renamed
   ;#
   if {$resumePlay} {
      soundControl::fileName $newName
      if {$wasPlaying} {
         ;# Resume play after calling routines finish
         after 100 Play current $location
      }
   } ;# end if
}
###############################################################################
# Function   : DeleteFile
# Description: Delete a file, if it is playing then handle it
# Author     : Tom Wilkason
# Date       : 3/24/2002
###############################################################################
proc catalog::DeleteFile {fileName} {
   set currentSound [soundControl::fileName]
   set goNext 0
   if {[string equal $currentSound $fileName]} {
      ;# This is a trick to see if the song is actually playing
      if {[soundControl::isPlaying]} {
         soundControl::stop
         set goNext 1
      } ;# end if
      soundControl::reset
   }
   ;# Now try to delete it
   if {[catch {
      file delete -force $fileName
   } result] \
   } then {
      tk_messageBox -icon warning -type ok -message "Error removing $fileName\n$result"
   } else {
     catalog::removeFile $fileName
     FlagDirtyCache $fileName
   }
   return $goNext
}

###############################################################################
# Function   : TFRemove
# Description: Remove a file from the master and playlists list if it exists
# Author     : Tom Wilkason
###############################################################################
proc catalog::removeFile {File} {
   global TFList TF LB PLB
   ;# Remove from the master list
   set index [catalog::whereInDb $File]
   if {$index >= 0} {
      ;##
      ;# Remove from the master list
      ;#
      ldelete TFList $index
      ;##
      ;# Remove from the Playlist
      ;#
      set i 0
      ;##
      ;# Since the playlist can have multiple entries for the same file, we have
      ;# to take some care and remove each of them from the bottom up.
      ;#
      set lnFile [LB::getListName $File]
      set end [expr {[LB::end]-1}]
      for {set i $end} {$i >= 0} {incr i -1} {
         set Entry [LB::get $i]
         if {[string equal $Entry $lnFile]}  {
            LB::delete $i
         } ;# end if
      } ;# end for
      if {[info exists TF($lnFile)]} {
         catch unset TF($lnFile)
      }
      ;##
      ;# Remove from the APL File List (in case it is there)
      ;#
      catch {aplLB::RemoveFile $File}
   } ;# end if
}
##########################################################################
# Function   : renameInPlaceDialog
# Description: Bring up a in place editing window on a listbox
# Author     : Tom Wilkason
# Date       : 6/19/2002
###############################################################################
proc catalog::renameInPlaceDialog {W table {oldName "xxx"}} {
   if {![winfo ismapped $W]} {
      return
   } ;# end if
   set w  .rndialog
   destroy $w
   toplevel $w
   ;##
   ;# Need to compute location for dialog
   ;#

   set active [$W index active]
   foreach {x y - -} [$W bbox $active] {break}
   ;# If table, adjust for heading
   if {$table} {
      set y [expr {$y - [winfo y $W]}]
   }
   ;# Left Hand alignment
   set rx [winfo rootx $W]
   set ry [winfo rooty $W]
   set X [expr {$rx+$x}]
   set Y [expr {$ry+$y}]
   ;# Make it look like the text underneath
   set font [$W cget -font]
   set bg [$W cget -background]
   set fg [$W cget -foreground]
   set ::TA(rename) $oldName
   ;# Set size available width and linespacing+2
   set wid [expr {[string length $oldName]+5}]
   set Wwid [winfo width $W]
   set Wheight [expr {[font metrics $font -linespace]+2}]
   pack [entry $w.rn -bd 2 -relief groove -font $font -insertbackground $fg \
      -width $wid -bg $bg -fg $fg -textvariable ::TA(rename)] -expand true -fill x
   #$w.rn selection range 0 end
   wm withdraw $w
   wm overrideredirect $w 1
   if {[string length [focus]]} {
      wm transient $w [winfo toplevel [focus]]
   }
   focus -force $w.rn
   wm geometry $w "${Wwid}x$Wheight+$X+$Y"
   bind $w <Escape> {
      destroy .rndialog
      set ::TA(rename) ""
   }
   bind $w <FocusOut> {
      destroy .rndialog
      set ::TA(rename) ""
   }
   bind $w <Return> [list destroy .rndialog]
   wm deiconify $w
   ;# Force on top if possible
   update idletasks
   if {$ontop::enabled && [ontop::state $W]} {
      ontop::control $w 1
   }
   tkwait window $w
   ;# Waiting for window to go away
   return $::TA(rename)
}

###############################################################################
# Function   : catalog::RenameEntry
# Description: Replace a file in the master list with another one only if it exists
# Author     : Tom Wilkason
###############################################################################
proc catalog::RenameEntry {oldFile newFile} {
   global TFList TF LB PLB
   ;# If it doesn exist, such as incoming data
   ;# then go ahead and add it to master list
   set index [catalog::addToDbIfNeeded $oldFile]
   if {$index < 0} {
      set index [catalog::whereInDb $oldFile]
      #saLog "Adding new file $oldFile"
   }
   if {$index >= 0} {
      set tail [file tail $oldFile]
      ;##
      ;# Update Main Play List
      ;#
      set active [LB::active]
      set data [LB::selected]
      changePlaylistFiles $oldFile $newFile
      ;# Restore selection
      foreach {position} $data {
         LB::selectionSet $position
      } ;# end foreach
      LB::activate $active
      ;##
      ;# Update APL oldFile List (if visible in list)
      ;#
      if {[APLExists]} {
         ;##
         ;# Schedule a rescan to be done at the next event, if one is pending
         ;# then don't schedule another. Not very clean but it works.
         ;# TFW: Move this to the trackList namespace as a function
         ;# trackList::changeTable $old $new
         set newTail [file tail $newFile]
         set newData [list]
         set changed 0
         if {[info exists trackList::TABLEDATA]} {
            foreach row $trackList::TABLEDATA {
               set cc 0
               set key [lindex $row 0]  ;# Used to lookup full file name, key is hidden column
               foreach col $row {
                  ;# If col is old file name, replace it with the new one
                  if {[string equal $col $tail]} {
                     lset row $cc $newTail
                     #debug "Changing '$tail' to '$newTail' for $key"
                     ;# Replace name entry for this row key
                     set trackList::Index2File($key) $newFile
                     incr changed
                  }
                  incr cc
               } ;# end foreach
               lappend newData $row
            } ;# end foreach
            ;# If one or more files renamed, then re-insert the data by set the TABLEDATA
            ;# lined variable to the new table data.
            if {$changed} {
               set trackList::TABLEDATA $newData
               ;# Schedule a rescan when all is said and done, don't re-populate
               ;# since we did in place above, just update the cache file.
               trackList::populateHighighted 0 0
            }
         }
      }
      ;##
      ;# Replace equalization if tail of file was renamed
      ;#
      gainControl::move $oldFile $newFile
      ;##
      ;# Replace full name in master list
      ;#
      lset TFList $index $newFile
      catalog::FlagDirtyCache $newFile
      catalog::dirtyList
   } else {
      saLog "Could not find $oldFile in list to replace"
   }
}
###############################################################################
# Function   : catalog::moveFolder
# Description: Move a folder (and its files) to some other place
#              and adjust the APL tree accordingly.
#              folder is existing full name
#              target is new full name
#              only called by APL code
# Author     : Tom Wilkason
# Date       : 3/3/2002
###############################################################################
proc catalog::moveFolder {folder target} {
   Trace
   global LB TF TFList
   ;##
   ;# Don't allow moving of removeable media folder
   ;#
   if {[isARemMedia $folder] || [isARemMedia $target]} {
      return
   }
   ;# filtering::Append / to folder so we don't match
   ;# Other folders with same base name
   ;# eg. Classic vs. Classic Rock
   set folder [string trimright $folder /]
   set target [string trimright $target /]
   if {[string equal $folder $target]} {
      return
   } ;# end if
   APL::busyCursor 1
   update idletasks
   if {[catch {catalog::RenameFile $folder $target} result]} {
      saLog $result
      APL::busyCursor 0
      return -code error $result
   }
   ;# Remove old folder name
   APL::delItem $folder
   ;##
   ;# Replace teFolder entries with new name
   ;#
   set subLen [expr {[string length $folder]-1}]
   set i -1
   set pattern "$folder/*"
   foreach path [lsort [array names ::teFolders]] {
      incr i
      if {[string match $pattern $path] || [string equal $folder $path]} {
         set newName [string replace $path 0 $subLen $target]
         set value $::teFolders($path)
         unset ::teFolders($path)
         set ::teFolders($newName) $value
         APL::newItem $newName
      } ;# end if
   } ;# end foreach
   ;##
   ;# Replace each matching entry in TFList and update playlist and APL file list
   ;# if affected songs were shown in them.
   ;#
   set i -1
   ;# Append to folder for pattern matching
   set PlayListFiles [LB::getAllNames]
   set active [LB::active]
   set data [LB::selected]
   set changed 0
   ;# TFW: Function to return indicies of match
   foreach {file} $TFList {
      incr i
      if {[string match $pattern $file]} {
         set newName [string replace $file 0 $subLen $target]
         incr changed
         lset TFList $i $newName
         ;# TFW: This could be improved to pass in a paired list so we
         ;# don't have to call this function for each file.
         changePlaylistFiles $file $newName PlayListFiles
      } ;# end if
   } ;# end foreach
   ;# Restore selection
   foreach {position} $data {
      LB::selectionSet $position
   } ;# end foreach
   LB::activate $active
   if {$changed > 0} {
      catalog::dirtyList
   } ;# end if
   APL::busyCursor 0
}
###############################################################################
# Function   : catalog::dirtyList
# Description: Flag the master track list as unsorted
# Author     : Tom Wilkason
# Date       : 12/19/2002
###############################################################################
proc catalog::dirtyList {args} {
   variable isSorted
   set isSorted 0
}
###############################################################################
# Function   : catalog::sortedList
# Description: Return the master track list sorted
# Author     : Tom Wilkason
# Date       : 12/19/2002
###############################################################################
proc catalog::sortedList {args} {
   variable isSorted
   if {$isSorted==0} {
      if {$::tcl_version >= 8.3} {
         set ::TFList [lsort -unique $::TFList]
      } else {
         set ::TFList [lsort $::TFList]
      }
      set isSorted 1
   } ;# end if
   return $::TFList
}
###############################################################################
# Function   : catalog::makeSorted
# Description: Make the master track list sorted if necessary
# Author     : Tom Wilkason
# Date       : 12/19/2002
###############################################################################
proc catalog::makeSorted {args} {
   variable isSorted
   if {$isSorted==0} {
      set ::TFList [sortedList]
   }
   return 1
}
###############################################################################
# Function   : catalog::changePlaylistFiles
# Description: Change all of the instances of a file lookup in the main
#              playlist with another
# Author     : Tom Wilkason
###############################################################################
proc catalog::changePlaylistFiles {oldFile newFile {_lbArray {}}} {
   global TF LB
   if {[string length $_lbArray] > 0} {
      upvar $_lbArray lbArray
   } else {
      set lbArray [LB::getAllNames]
   }
   set oldKey [LB::getListName $oldFile]
   set newKey [LB::getListName $newFile]
   #puts "newkey: $newKey"
   ;# It it is designated as being in the playlist
   ;# Check each entry and rename it
   if {[info exists TF($oldKey)]} {
      set i 0
      ;# If the keys changed, update the list
      if {![string equal $newKey $oldKey]} {
         foreach element $lbArray {
            if {[string equal $element $oldKey]} {
               LB::replace $i $newKey
               #puts "LB::replace $i $newKey"
               ;# TFW: check here if files not being renamed in playlist
            }
            incr i
         } ;# end foreach
      }
      ;# Update the full file name
      unset TF($oldKey)
      set TF($newKey) $newFile
   }
}
###############################################################################
# Function   : catalog::moveClipboardFiles
# Description: Move the files in the clipboard to some new folder
# Author     : Tom Wilkason
###############################################################################
proc catalog::moveClipboardFiles {target} {
   if {[catch {selection get -selection CLIPBOARD} tryFiles]} {
      return ""
   }
   clipboard clear
   set files [list]
   ;# Ignore clibboard entries that aren't files
   ;# Ignore clibboard entries are duplicates
   ;# Ignore errors due to funky clipboard data
   catch {
      foreach {file} $tryFiles {
         if {[lsearch -exact $files $file] < 0 && [file exists $file]} {
            lappend files $file
         } ;# end if
      } ;# end foreach
   }
   #puts $files
   set moved 0
   soundControl::latency long
   if {[llength $files] > 0} {
      if {$::snackAmpSettings(dontMoveFiles)} {
         tk_messageBox -icon warning -type ok \
            -message "You don't have permission to move files or folders"
         return
      }
      set result [tk_messageBox -type yesno -title "Please Answer" \
         -icon question -message "Are you sure you want to move:\n[join $files \n]\n to\n$target"]
      if {$result == "yes"} {
         update
         set errors [list]
         ;##
         ;# For each file to be moved, move it
         ;#
         foreach file $files {
            set tail [file tail $file]
            set newName [file join $target $tail]
            waitWindow::show "Moving $tail ... Please Wait"
            update
            ;##
            ;# If source is a folder, move it otherwise
            ;# move the file
            ;#
            if {[file isdirectory $file]} {
               if {[catch {
                  if {![string equal $file $target]} {
                     catalog::moveFolder $file $newName
                  }
               } result] \
               } then {
                  saLog $result
                  lappend errors $file
               }
            } else {
               if {![string equal $file $newName]} {
                  if {[catch {catalog::RenameFile $file $newName} result]} then {
                     saLog $result
                     lappend errors $file
                  } else {
                     ;# TFW: This can be slow on big moves, you have a full LB lookup
                     ;# for each file moved.
                     catalog::RenameEntry $file $newName
                     incr moved
                  }
               } ;# end if
            }
         } ;# end foreach
         waitWindow::close
         ;# Flag these two areas as dirty
         catalog::FlagDirtyCache $target
         catalog::FlagDirtyCache $newName
         ;# Handle any errors
         if {[llength $errors] > 0} {
            tk_messageBox -type ok -icon warning -message "The following files were not moved\n[join $errors \n]"
            #puts $::errorInfo
         } ;# end if
      }
   } ;# end if
   if {$moved > 0} {
      catalog::dirtyList
   } ;# end if
   soundControl::latency normal
   return $moved
}

###############################################################################
# Function   : renameFolder
# Description: Prompt then rename a subfolder under some level
# Author     : Tom Wilkason
###############################################################################
proc catalog::renameFolder {folder} {
   set tail [file tail $folder]
   set base [file dirname $folder]
   set newName $tail
   set res 0
   if {[getEntryDialog $newName newName]=="ok" && ![string equal $tail $newName]} {
      if {[catch {
         set newFullName [file join $base $newName]
         moveFolder $folder $newFullName
         if {[APLExists]} {
            APL::treeSelect $newFullName
            APL::viewHighlighted $::TREE
         }
         catalog::FlagDirtyCache $newFullName
         set res 1
      } result] \
      } then {
         tk_messageBox -type ok -icon warning -message "Could not rename folder $newName\n$result"
         set res 0
      }
   } ;# end if
   return $res
}
###############################################################################
# Function   : catalog::rescanFolder
# Description: Rescan the files in a folder to make sure they are up to date
#              Will work recursively on folders only if they are new.
# Author     : Tom Wilkason
# Date       : 10/18/2001
###############################################################################
proc catalog::rescanFolder {index} {
   set Files [glob -nocomplain -dir $index *]
   ;##
   ;# Verify that all sub-folders still exist, if not.
   ;# Then remove them from the tree.
   ;#
   foreach {folder} [array names ::teFolders "$index/*"] {
      if {![file exists $folder]} {
         ;##
         ;# Remove the files & icon
         ;#
         APL::delItem $folder
         unset ::teFolders($folder)
      } ;# end if
   }
   ;##
   ;# Second see if there are new folders, add them if necessary
   ;#
   foreach {folder} $Files {
      if {[file isdirectory $folder]} {
         if {[string length $folder] > 0 && ![info exists ::teFolders($folder)]} {
            incr ::snackAmpSettings(diff)
            set ::teFolders($folder) $::teFolders([file dirname $folder])
            APL::newItem $folder
            ;# New folder, go ahead and scan it too
            ###rescanFolder $folder
         } ;# end if
      } ;# end if
   } ;# end foreach
   ;##
   ;# Now do the files, first add unique files
   ;#
   foreach {file} $Files {
      if {[info exists ::taExtensions([file extension $file])]} {
         if {[catalog::addToDbIfNeeded $file] < 0} {
            incr ::snackAmpSettings(diff)
         }
      } ;# end if
   } ;# end foreach
   ;##
   ;# Now, make sure any files that should be on this folder
   ;# are indeed on this folder. Remove the ones that are not.
   ;#
   set newFiles $::TFList
   catch {unset ::TFList}
   foreach {File} $newFiles {
      ;# If file is in selected tree, check it
      if {[string equal $index [file dirname $File]]} {
         if {[lsearch -exact $Files $File] < 0} {
            ;# File was removed, to add to new list
            incr ::snackAmpSettings(diff)
         } else {
            lappend ::TFList $File
         }
      } else {
         lappend ::TFList $File
      }
   }
   catalog::dirtyList
}

###############################################################################
# Function   : removeFolder
# Description: Remove a folder and all of its children files
#              Only used by APL tree
# Author     : Tom Wilkason
###############################################################################
proc catalog::removeFolder {folder} {
   set i 0
   if {$::snackAmpSettings(dontMoveFiles)} {
      tk_messageBox -icon warning -type ok \
         -message "You don't have permission to remove folders"
      return
   }

   set pattern [file join $folder *]
   ;# If a removeable file in name then check if at base level and if so
   ;# verify before removing the catalog file from disk
   set i [llength [catalog::filesMatchingPattern $pattern]]
   set parts [file split $folder]
   if {[isARemMedia $folder] && [llength $parts] == 3} {
      set message "Are you sure you want to remove the catalog $folder"
      set catalog 1
   } else {
      set message "Are you sure you want to remove $folder and its decendents, it contains $i tracks"
      set catalog 0
   }
   if {$i == 0 || [tk_messageBox -icon question -type yesno -message $message] == "yes"} {
      ;##
      ;# Remove the files & icon
      ;#
      APL::delItem $folder
      ;##
      ;# Handle the files
      ;#
      foreach {file} [catalog::filesMatchingPattern $pattern] {
         if {[catch {
            DeleteFile $file
         } result] \
         } then {
            saLog $result
         }
      } ;# end foreach
      catalog::FlagDirtyCache $folder
      ;##
      ;# Handle the folders
      ;#
      foreach path [array names ::teFolders $folder*] {
         if {[string match $pattern $path] || [string equal $folder $path]} {
            if {[catch {
               file delete -force $path
               unset ::teFolders($path)
            } result] \
            } then {
               saLog $result
            }
         } ;# end if
      } ;# end foreach
      APL::treeSelect [file dirname $folder]
      aplLB::UpdateList [file dirname $folder]
      FlagDirtyCache $folder
      updateFileDisplay
      if {$catalog} {
         ;# Remove the RM catalog
         set fname [file join $::HOME [lindex $parts 2].rmc]
         file delete $fname
         #puts $fname
      } ;# end if
   } ;# end if
   return
}
###############################################################################
# Function   : catalog::SavePlaylistCache
# Description: Cache the current playlist for later restoration
# Author     : Tom Wilkason
###############################################################################
proc catalog::SavePlaylistCache {args} {
   global LB TA TF
   set sList [LB::getAllNames 1]
   set loc $LB::indexPlaying
   if {$loc < 0} {
      set loc [LB::active]
   } ;# end if
   append sList "\n$loc"
   append sList "\n[soundControl::samplePosition]"
   append sList "\n$soundControl::playState"
   if {[catch {safeWrite $TA(playCacheFile) $sList} result]} {
      return -code error "Save Error: Unable to open '$TA(playCacheFile)' for writing\n"
   }
}
###############################################################################
# Function   : catalog::ReadPlaylistCache
# Description: Restore the playlist from a cached file (during initialization)
# Author     : Tom Wilkason
###############################################################################
proc catalog::ReadPlaylistCache {args} {
   global TF TA LB TFList
   if {[catch {open $TA(playCacheFile) r} InFile]} {
      return -code error "Save Error: Unable to open '$TA(playCacheFile)' for reading\n"
   }
   fconfigure $InFile -buffersize 100000
   set sList  [gets $InFile]
   set cIndex [gets $InFile]
   ;# Restore position within the file
   if {[eof $InFile]} {
      set cPosition -1
   } else {
      set cPosition [gets $InFile]
   } ;# end if
   if {[eof $InFile]} {
      set playState "Stopped"
   } else {
      set playState [gets $InFile]
   } ;# end if

   close $InFile
   ;# Replace current playlist very fast
   catch {unset TF}
   LB::deleteAll
   if {[llength $sList] > 0} {
      foreach {file} $sList {
         set keyName [LB::getListName $file]
         lappend insPL $keyName
         set TF($keyName) $file
      } ;# end foreach
      LB::insertList end $insPL
      unset insPL
      ;# Restore current file and highlight it
      ;# Optionally resume play/pause
      set TA(currentMF) [lindex $sList $cIndex]
      updateFileDisplay
      macros::Highlight $cIndex
      if {$::snackAmpSettings(resumePlay)} {
         if {$playState=="Playing"} {
            after 300 Play active $cPosition
#            saLog "Playing at $cPosition"
         } elseif {$playState=="Paused"} {
            after 300 Play active $cPosition
            after 300 Pause
#            saLog "Pausing at $cPosition"
         }
      } ;# end if
      saLog "Restored playlist cache of [llength $sList] files"
   }
}
###############################################################################
# Function   : catalog::ShowDups
# Description: Command line routine to show potential duplicate files
# Author     : Tom Wilkason
# Date       : 3/28/2002
###############################################################################
proc catalog::ShowDups {{pattern *}} {
   foreach {file} [catalog::filesMatchingPattern $pattern] {
      set dir [file dirname $file]
      set tail "$dir/[lrange [split $file -] end-1 end]"
      set fileTail [lrange [file split $file] end-1 end]
      if {![info exists tailHash($tail)]} {
         set tailHash($tail) $file
      } else {
         update
         saLog "Duplicate $tail\n   '$file'\n   '$tailHash($tail)'\n"
      }
   } ;# end foreach
}
;##
;# Return a list of all playlist files
;#
proc catalog::getAutoPlaylists {args} {
   return [lsort -dictionary [glob -nocomplain -directory $::HOME *.tafs]]
}
proc catalog::getPlaylists {args} {
   return [lsort -dictionary [glob -nocomplain -directory $::HOME *.taf *.m3u]]
}

###############################################################################
# Function   : catalog::DupFiles
# Description: Command line routine to show potential duplicate files
# Author     : Tom Wilkason
# Date       : 3/28/2002
###############################################################################
proc catalog::DupFiles {{pattern *}} {
   foreach file $::TFList {
      if {[string match $pattern $file] && ![string match "/RMC/*" $file]} {
         set tail [lindex [file split $file] end]
         if {![info exists tailHash($tail)]} {
            set tailHash($tail) $file
         } else {
            update
            saLog "Duplicate $tail\n   '$file - [file size $file]'\n   '$tailHash($tail) - [file size $tailHash($tail)]'\n"
         }
      }
   } ;# end foreach
}
###############################################################################
# Function   : catalog::DupSizes
# Description: Command line routine to show potential duplicate files
#              based on like file sizes. Good for finding renamed files.
# Author     : Tom Wilkason
# Date       : 3/28/2002
###############################################################################
proc catalog::DupSizes {{pattern *} {threshold 2}} {
   foreach {file} [catalog::filesMatchingPattern $pattern] {
      if {![catch {set tail [file size $file]} result]} then {
         lappend tailHash($tail) $file
      }
   } ;# end foreach

   foreach entry [array names tailHash] {
      set files  $tailHash($entry)
      if {[llength $files] > 1} {
         set last [expr {[llength $files]-1}]
         set matches 0
         ;# Pass through each set of duplicate files
         for {set i 0} {$i < $last} {incr i} {
            set base [split [string map {" " ""} [string tolower [file tail [lindex $files $i]]]] -]
            set next [split [string map {" " ""} [string tolower [file tail [lindex $files [incr i]]]]] -]
            #puts "$base + $next"
            foreach part $base {
               if {[lsearch -exact $next $part] > -1} {
                  incr matches
               }
            } ;# end foreach
         } ;# end for
         if {$matches > $threshold} {
            puts "Duplicates found for size $entry ($matches)"
            foreach file $files {
               puts "   $file"
            }
         }
      }
   } ;# end foreach
}

###############################################################################
# Function   : fixplus::correctCase
# Description: Replace + in file names (due to http download), also fix other
#              case problems and clean up spacing.
# Author     : Tom Wilkason
# Date       : 1/8/2001
###############################################################################
namespace eval fixplus  {}
proc fixplus::correctCase {folder} {
   set changed 0
   if {[file isdirectory $folder]} {
      set Files [glob -nocomplain -directory $folder -types {f l} *]
   } else {
      set Files $folder
   }
   foreach File $Files {
      set folder [file dirname $File]
      set File   [file tail $File]
      ;##
      ;# Clean up the file name
      ;#
      set ext [file extension $File]
      set root [file rootname $File]   ;# base part of name
      if {[info exists ::taExtensions($ext)]} {
         ;# Clean out funny stuff
         set newRoot [cleanName $root]
         set ofn [file join $folder $File]
         set fnf [file join $folder "$newRoot$ext"] ;#full new file
         ;##
         ;# Have new file name
         ;#
         if {![string equal $newRoot $root]} {
            ;##
            ;# Try to rename the file here
            ;#
            if {[catch {catalog::RenameFile $ofn $fnf} result]} {
               saLog "correctCase Error:$result"
            } else {
               catalog::RenameEntry $ofn $fnf
               #saLog "correctCase:\nWas '$ofn' \n Is '$fnf'"
               incr changed
            }
         } ;# end if
      } ;# end if
   } ;# end foreach
   if {$changed > 0} {
      catalog::dirtyList
   } ;# end if
   return $changed
}
###############################################################################
# Function   : cleanName
# Description: Clean up a file name to use - as separator
# Author     : Tom Wilkason
# Date       : 3/19/2002
###############################################################################
proc fixplus::cleanName {root} {
   ;# sp is user supplied split char, "-" is recommended
   set sp [string trim $::snackAmpSettings(splitChar)]
   set root [string map {[ ( ] )} $root]
   regsub -all -- {[+_ ]+} $root { } root
   regsub -all -- {( *- *)|( *\+ *)} $root " $sp " root
   set root [string trim $root]
   ;# Fix first part if a number
   set reExp [subst -nocommand {^([0-9]+)[$sp ]*(.*?)}]
   if {[regexp $reExp $root -> num rest]} {
      set root "$num $sp $rest"
   }
   ;# Cap each token
   set root [stringTitle $root]
   return $root
}
###############################################################################
# Function   : gainControl Namespace
# Description: Functions to access the auto-level gain control
# Author     : Tom Wilkason
# Date       : 2/27/2003
###############################################################################
namespace eval gainControl {
   variable es ""
   variable dirty 0
}
proc gainControl::levelText {value} {
   if {$::snackAmpSettings(autoLevel) && [string is double $value]} {
      if {[catch {toDb $value} db] } then {
         set db 0.0
      } else {
         if {$db > $::snackAmpSettings(maxGain)} {
            set db $::snackAmpSettings(maxGain)
         } elseif {$db < $::snackAmpSettings(minGain)} {
            set db $::snackAmpSettings(minGain)
         }
      }
      set result "[string trim [format %-5.2f $db]] dB"
   } else {
      set result "--"
   }
}
;##
;# Remove all stale equalization settings
;#
proc gainControl::cleanStaleEqualizations {args} {
   variable dirty
   foreach item $::TFList {
      set hash([getkey $item]) 1
   } ;# end foreach
   set count 0
   foreach item [array names ::TEqual] {
      if {![info exists hash($item)]} {
         unset ::TEqual($item)
         #saLog "Removing Stale '$item'"
         incr dirty
         incr count
      }
   } ;# end foreach
   if {$count > 0} {
      tk_messageBox -icon info -type ok -message "Cleaned out a total of $count stale records"
   }
}
proc gainControl::gs {args} {
   variable es
   if {[string length $es]==0} {
      set es [snack::sound]
   }
   return $es
}
proc gainControl::create {fname {snd ""}} {
   variable dirty
   if {[string length $snd]==0} {
      set snd [gs]
   }
   $snd configure -file $fname
   set gain [soundControl::logPower $snd $::snackAmpSettings(autoLevelSeconds)]
   set ::TEqual([getkey $fname]) $gain
   incr dirty
   return $gain
}
proc gainControl::createIfNeeded {fname {snd ""}} {
   if {![exists $fname]} {
      return [create $fname $snd]
   } else {
      return [get $fname]
   }
}
proc gainControl::get {fname} {
   set key [getkey $fname]
   ;# Can throw an error
   return $::TEqual($key)
}
proc gainControl::add {fname amount} {
   variable dirty
   set key [getkey $fname]
   set ::TEqual($key) $amount
   incr dirty
}
proc gainControl::remove {fname} {
   variable dirty
   set key [getkey $fname]
   if {[info exists ::TEqual($key)]} {
      unset ::TEqual($key)
      incr dirty
      return 1
   }
   return 0
}
proc gainControl::exists {fname} {
   set key [getkey $fname]
   return [info exists ::TEqual($key)]
}
proc gainControl::move {from to} {
   variable dirty
   set fkey [getkey $from]
   set tkey [getkey $to]
   if {![string equal $fkey $tkey]} {
      if {[info exists ::TEqual($fkey)]} {
         set ::TEqual($tkey) $::TEqual($fkey)
         unset ::TEqual($fkey)
         incr dirty
      }
   }
}
proc gainControl::saveSettings {args} {
   variable dirty
   if {$dirty} {
      if {[catch {safeWrite [file join $::HOME GainControl.dat] [array get ::TEqual]} result] } then {
         tk_messageBox -icon warning -type ok -title "Open Error" -message $result
      }
   }
}
proc gainControl::restoreSettings {args} {
   variable dirty
   catch {
      set fid [open [file join $::HOME GainControl.dat] r]
      array set ::TEqual [read $fid]
      close $fid
   }
   set dirty 0
}
proc gainControl::resetAll {args} {
   variable dirty
   set gf [file join $::HOME GainControl.dat]
   set num [array size ::TEqual]
   if {$num > 0 || [file exists $gf]} {
      if {[tk_messageBox -icon question -type yesno \
         -title "Please Verify" -message "Are you sure you want to remove $num gain entries from $gf"]=="yes"} {
         catch {unset ::TEqual}
         catch {file delete -force $gf}
         incr dirty
      } ;# end if
   } ;# end if
}
proc gainControl::resetSome {{fileNames {}}} {
   global TF TA LB
   variable dirty
   if {[llength $fileNames] == 0} {
      set fileNames [LB::getSelectedFullNames]
   }
   foreach {file} $fileNames {
      incr dirty
      remove $file
   } ;# end foreach
}
proc gainControl::getkey {fname} {
   ;# Use key containing folder, prevents duplicates
   return [file join [file tail [file dirname $fname]] [file tail $fname]]
   #return [file tail $fname]
}
package provide catalog 1.0
