;##
;# Associate registry entry for music files for snackAmp
;# Note: Windows only, will not be called on Unix/Mac
;# Proudly stolen from WaveSurfer a very cool sound editor
;# Tom Wilkason
;#
package require registry
set [file root [file tail [info script]]]_History {
   $Header: /cvsroot/snackamp/snackamp/lib/fileAssociation.tcl,v 1.7 2003/03/28 01:25:15 wilkason Exp $
}
namespace eval fileAssociation {
# These currently get temp saved to a file first, so no good
   variable mimeTypes {
      .m3u audio/x-mpegurl
      .pls audio/x-scpls
      .mp3 audio/x-mpeg
      .mp3 audio/mpeg
      .ogg audio/x-ogg
      .ogg audio/ogg
      .wav audio/x-wav
   }

}
###############################################################################
# Function   : isRegisteredMF
# Description: Determine if a file is registered to SnackAmp
###############################################################################
proc fileAssociation::isRegisteredMF {{ftype .mp3} _var} {
   upvar $_var var
   if {[catch {registry get [join [list HKEY_CLASSES_ROOT $ftype] \\] ""} class]} {
      set var 0
      return 0
   }
   set resKey "HKEY_CURRENT_USER\\Software\\SnackAmp\\Restore\\$class\\DefaultIcon"
   if {[catch {registry get $resKey ""} defIcon]} {
      set var 0
      return 0
   }
   set var 1
   return 1
}

###############################################################################
# Function   : registerMF
# Description: Register some file type on windows for launching snackamp
# need to a setting for mime types for m3u/pls file
# HKEY_CLASSES_ROOT\MIME\Database\Content Type\audio/x-scpls -> Extension=.pls
# Will have to rename and CSLID items to CSLID.bak to allow the file association
# to be used.
###############################################################################
proc fileAssociation::registerMF {{ftype .mp3}} {
   variable mimeTypes
   if {[tk_messageBox -icon question -type yesno -title "Please Respond" \
      -message "Are you sure you want to associate $ftype files with SnackAmp"]!="yes"} {
      return
   } ;# end if

   package require registry
   set conf($ftype,default) 1
   set conf($ftype,dde) 1
   set conf($ftype,assoc) $ftype
   set resKey "HKEY_CURRENT_USER\\Software\\SnackAmp\\Restore"
   foreach key [array names conf *,assoc] {
      set ext [lindex [split $key ,] 0]
      ;# Set the class name of the file type
      if {[catch {registry get [join [list HKEY_CLASSES_ROOT $ftype] \\] ""} class]} {
         set class "[string trimleft $ext .]file"
         registry set [join [list HKEY_CLASSES_ROOT $ext] \\] "" $class
         registry set [join [list HKEY_CLASSES_ROOT $class] \\] "" "[string toupper [string trimleft $ext .]] sound format"
      }
      ;# Save the existing icon
      if {![catch {registry get HKEY_CLASSES_ROOT\\$class\\DefaultIcon ""} file]} {
         registry set $resKey\\$class\\DefaultIcon "" $file
      }
      if {![catch {registry get HKEY_CLASSES_ROOT\\$class\\shell ""} value]} {
         registry set $resKey\\$class\\shell "" $value
      }
      ;# Register the default icon for this class to SnackAmp (only gets icon if wrapped)
      registry set HKEY_CLASSES_ROOT\\$class\\DefaultIcon "" "$::TA(appname)"
      set base HKEY_CLASSES_ROOT\\$class\\shell
      ;##
      ;# Add menu handlers (These can just be removed later)
      ;#
      ;# Add son(s) to playlist
      if {$conf($ext,default)} {
         registry set $base "" snAmp.append
      }
      registry set $base\\snAmp.append "" "&Append to SnackAmp Playlist"
      registry set $base\\snAmp.append\\command "" "$::TA(appname)"
      if {$conf($ext,dde)} {
         registry set $base\\snAmp.append\\ddeexec "" "Append {%1}"
         registry set $base\\snAmp.append\\ddeexec\\Application "" TclEval
         registry set $base\\snAmp.append\\ddeexec\\Topic "" snackAmp
      }
      ;# Play the file
      if {$conf($ext,default)} {
         registry set $base "" snAmp.play
      }
      registry set $base\\snAmp.play "" "&Play using SnackAmp"
      registry set $base\\snAmp.play\\command "" "$::TA(appname)"
      if {$conf($ext,dde)} {
         registry set $base\\snAmp.play\\ddeexec "" "PlayFile {%1}"
         registry set $base\\snAmp.play\\ddeexec\\Application "" TclEval
         registry set $base\\snAmp.play\\ddeexec\\Topic "" snackAmp
      }
      foreach {Ext type} $mimeTypes {
         if {$ext==$Ext} {
            if {[catch {registerMIME $ext $type} result]} {
               saLog "registerMIME:$result"
            }
         }
      } ;# end foreach
   }
}
;##
;# registerMIME: Register a mime type for IE handling
;#
proc fileAssociation::registerMIME {ext type} {
   set loc "HKEY_CLASSES_ROOT\\MIME\\Database\\Content Type\\$type"
   ;# Rename any CLSID handlers to allow extensions to be used
   if {[catch {registry set $loc "Extension" $ext} result]} {
      saLog "registerMIME: $result"
   } else {
      saLog "Registered '$ext' as type '$type'"
   }
   if {![catch {registry get $loc "CLSID"} value]} {
      registry set $loc "CLSID.sa" $value
      registry delete $loc "CLSID"
   }

}
###############################################################################
# Function   : unRegisterMF
# Description: UnRegister some file type on windows for launching snackamp
###############################################################################
proc fileAssociation::unRegisterMF {{ftype .mp3}} {
   if {[catch {registry get [join [list HKEY_CLASSES_ROOT $ftype] \\] ""} class]} {
      puts "Not registered to snackAmp"
      return
   }
   if {[tk_messageBox -icon question -type yesno -title "Please Respond" \
      -message "Are you sure you want to restpre $ftype associations to their previous settings?"]!="yes"} {
      return
   } ;# end if
   set resKey "HKEY_CURRENT_USER\\Software\\SnackAmp\\Restore"
   ;# Restore single entries for
   #HKEY_CLASSES_ROOT .mp3
   #HKEY_CLASSES_ROOT mp3file
   ;# Remove snAmp.append handlers
   registry delete HKEY_CLASSES_ROOT\\$class\\shell\\snAmp.append
   registry delete HKEY_CLASSES_ROOT\\$class\\shell\\snAmp.play
   ;# Restore old Icon
   if {![catch {registry get $resKey\\$class\\DefaultIcon ""} file]} {
      registry set HKEY_CLASSES_ROOT\\$class\\DefaultIcon "" $file
      registry delete $resKey\\$class\\DefaultIcon
   }
   if {![catch {registry get $resKey\\$class\\shell ""} value]} {
      registry set HKEY_CLASSES_ROOT\\$class\\shell "" $value
      registry delete $resKey\\$class\\shell
   }
   foreach {ext type} $mimeTypes {
      if {$ext==$ftype} {
         if {[catch {unregisterMIME $ext $type} result]} {
            saLog "unregisterMIME:$result"
         }
      }
   } ;# end foreach
}
;##
;# unregisterMIME: Restore any previous CLSID values
;#
proc fileAssociation::unregisterMIME {ext type} {
   set loc "HKEY_CLASSES_ROOT\\MIME\\Database\\Content Type\\$type"
   ;# Rename any CLSID handlers to allow extensions to be used
   if {![catch {registry get $loc "CLSID.sa"} value]} {
      registry set $loc "CLSID" $value
      registry delete $loc "CLSID.sa"
   }

}
package provide fileAssociation 1.0
