###############################################################################
# gpWidgets 1.0
#
# General Purpose Widget Utilities for snackAmp Player in Tcl/Tk
#
# Copyright (C) 2001 Tom Wilkason
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
#
# Please send any comments/bug reports to
# tom.wilkason@cox.net  (Tom Wilkason)
#
###############################################################################
set [file root [file tail [info script]]]_History {
   $Header: /cvsroot/snackamp/snackamp/lib/gpWidgets.tcl,v 1.38 2003/05/17 14:26:12 wilkason Exp $
}
#//////////////////////////////////////////////////////////////////////////////
#                   B A L L O O N    H E L P
#//////////////////////////////////////////////////////////////////////////////
namespace eval balloon {
}
###############################################################################
# Function   : balloon
# Description: Bind balloon help to some widget
# Author     : Tom Wilkason
# Date       : 1/2/2000
###############################################################################
proc balloon::button {w help} {
   bind $w <Any-Enter> "balloon::_bubbleIt %W 1;after $::snackAmpSettings(balloonDelay) [list balloon::show %W %X %Y [list $help]]"
   bind $w <Any-Leave> "balloon::_bubbleIt %W 0;destroy %W.balloon"
   bind $w <Any-Motion> [list destroy %W.balloon]
}
###############################################################################
# Function   : bubbleIt
# Description: Raise the bubble on the button if it is not depressed (like pause)
# Author     : Tom Wilkason
###############################################################################
proc balloon::_bubbleIt {W Enter} {
   if {[$W cget -relief] != "sunken"} {
      if {$Enter} {
         $W configure -relief raised
         #$W configure -relief ridge
      } else {
         $W configure -relief flat
      }
   }
}
###############################################################################
# Function   : balloon
# Description: Bind some widget to balloon help
# Author     : Tom Wilkason
###############################################################################
proc balloon::define {w help} {
    bind $w <Any-Enter> "after $::snackAmpSettings(balloonDelay) [list balloon::show %W %X %Y [list $help]]"
    bind $w <Any-Leave> "destroy %W.balloon"
    #bind $w <Any-Motion> "destroy %W.balloon"
}
###############################################################################
# Function   : balloon::aplDefine
# Description: Bind balloon details on a song when moving mouse over it
#              For the auto play list box
# Author     : Tom Wilkason
###############################################################################
;# proc balloon::aplDefine {args} {
;#    package require snID3
;#    set ::TA(IDautoPlayListCursor) 0
;#    bind $::PLB <Any-Motion> {
;#       if {$::snackAmpSettings(ID3ballon)} {
;#          balloon::_autoPlayListName %W %y
;#          destroy %W.balloon
;#          after cancel $::TA(IDautoPlayListCursor)
;#          set ::TA(IDautoPlayListCursor) [after $::snackAmpSettings(balloonDelay) [list balloonSongInfo %W $::TA(autoPlayListCursor)]]
;#       } ;# end if
;#    }
;#    bind $::PLB <Any-Leave> "destroy %W.balloon"
;# }
###############################################################################
# Function   : balloon::plDefine
# Description: Bind balloon details on a song when moving mouse over it
#              for the play list box
# Author     : Tom Wilkason
###############################################################################
proc balloon::plDefine {args} {
   package require snID3
   set ::TA(IDplayListCursor) 0
   bind $::LB <Any-Motion> {
      if {$::snackAmpSettings(ID3ballon)} {
         balloon::_PlayListName %W %y
         destroy %W.balloon
         after cancel $::TA(IDplayListCursor)
         set ::TA(IDplayListCursor) [after $::snackAmpSettings(balloonDelay) [list balloonSongInfo %W $::TA(playListCursor)]]
      } ;# end if
   }
   bind $::LB <Any-Leave> "destroy %W.balloon"
}
###############################################################################
# Function   : autoPlayListName
# Description: Save away the song name for the song under the mouse cursor
#             in the main playlist
# Author     : Tom Wilkason
# Date       : 11/11/2001
###############################################################################
;# proc balloon::_autoPlayListName {W y} {
;#    set closest [$W nearest $y]
;#    if {[catch {
;#       set ::TA(autoPlayListCursor) $::shortLong([$W get $closest])
;#    }]} {
;#       set ::TA(autoPlayListCursor) {unknown}
;#    }
;# }
###############################################################################
# Function   : playListName
# Description: Save away the song name for the song under the mouse cursor
#             in the main playlist
# Author     : Tom Wilkason
# Date       : 11/11/2001
###############################################################################
proc balloon::_PlayListName {W y} {
   set closest [$W nearest $y]
   if {[catch {
      set ::TA(playListCursor) $::TF([$W get $closest])
   }]} {
      set ::TA(playListCursor) {}
   }
}
###############################################################################
# Function   : balloon::show
# Description: Pop-up the ballon help based on a callback for the widget enter
# Author     : Tom Wilkason
# Date       : 1/2/2000
###############################################################################
proc balloon::show {w x y arg} {
   ;# Don't bother if the mouse has left the widget
   ;# The other approach is the track afters and cancel
   ;# them on leaving, this is easier.
   if {[eval winfo containing  [winfo pointerxy .]]!=$w} {
      return
   }
   set top $w.balloon
   catch {destroy $top}
   toplevel $top -bd 1 -bg black -cursor $::TA(arrow)
   wm overrideredirect $top 1
   pack [message $top.txt -aspect 1000 -bd 2 -bg lightyellow -fg black -font balloon -text $arg]
   ;# if 0 for pointer x/y issued, then figure out proper location
   ;# Use current pointer always, not what was passed in in after call
   set x [winfo pointerx $w]
   set y [winfo pointery $w]
   set tWidth [winfo reqwidth $top.txt]
   set tHeight [winfo reqheight $top.txt]
   ;# Keep the balloon in the visible screen, flip as necessary
   ;# Don't colocate with pointer location otherwise
   ;# flashing will occur (bug?)
   if {($x+$tWidth) > $::TA(screen_right)} {
      set x [expr {$x-$tWidth-1}]
   } else {
      incr y
   }
   if {($y+$tHeight) > $::TA(screen_bottom)} {
      set y [expr {$y-$tHeight-1}]
   } else {
      incr x
   }
   ;# Offset by one pixel so balloon is off cursor
   wm geometry $top ${tWidth}x${tHeight}+$x+$y
   raise $top
   ;# Force on top if possible
   if {$ontop::enabled && [ontop::state $w]} {
      ontop::control $top 1
   }

}
###############################################################################
# Function   : Scrolled_Listbox
# Description: Utility function to create a scrolled listbox
# Author     : Tom Wilkason
###############################################################################
proc Scrolled_Listbox { f args } {
   frame $f
   listbox $f.list \
      -xscrollcommand [list scrollCmd $f.xscroll  [list grid $f.xscroll -row 1 -column 0 -sticky ew]] \
      -yscrollcommand [list scrollCmd $f.yscroll  [list grid $f.yscroll -row 0 -column 1 -sticky ns]]
   eval {$f.list configure} $args
   scrollbar $f.xscroll -orient horizontal -command [list $f.list xview]
   scrollbar $f.yscroll -orient vertical   -command [list $f.list yview]
   grid $f.list -sticky news
   ;##
   ;# Only the data should expand, not the scroll bars
   ;#
   grid rowconfigure    $f 0 -weight 1
   grid columnconfigure $f 0 -weight 1
   ;##
   ;# Modify to not resize when the scroll bars go away (make a minimum size)
   ;#
   grid columnconfigure $f 1 -minsize 0 ;#1
   grid rowconfigure    $f 1 -minsize 0 ;#1

   ;# For *nix mousewheels
   mscroll $f.list

   return $f.list
}
##
# Binds the *nix mousewheel events to scroll a scrollbar
#
proc mscroll {bindtag} {
   if {$::tcl_platform(platform) == "unix"} {
      bind $bindtag <Button-5> {%W yview scroll 5 units;break}
      bind $bindtag <Button-4> {%W yview scroll -5 units;break}
      bind $bindtag <Shift-Button-5> [list %W yview scroll 1 units]
      bind $bindtag <Shift-Button-4> [list %W yview scroll -1 units]
      bind $bindtag <Control-Button-5> [list %W yview scroll 1 pages]
      bind $bindtag <Control-Button-4> [list %W yview scroll -1 pages]
   }
}

###############################################################################
# Function   : getEntryDialog
# Description: Generic dialog to ask user for some text
# Author     : Tom Wilkason
# Date       : 3/12/2001
;# Todo: Add a dir selector
###############################################################################
proc getEntryDialog {message _variable} {
   set i 0
   upvar $_variable variable
   while {[info command .ged$i] != {}} {incr i}
   set T .ged$i
   toplevel $T
   wm title $T "Please Correct '$variable'"
   set ::TA(dialogEntry) $variable
   wm protocol $T WM_DELETE_WINDOW {set ::TA(continue) 1}
   # wm overrideredirect $T 1
   set len [expr {round([string length $variable]*1.25)}]
   if {$len < 20} {
      set len 20
   } ;# end if
   pack [label $T.l -text $message -anchor w] -side top -fill x -anchor w
   pack [entry $T.e -width $len -textvariable ::TA(dialogEntry) -exportselection false] -side top -fill x
   $T.e config -background $::snackAmpSettings(ebg) -foreground $::snackAmpSettings(efg)
   settings::regColor $T.e -background ::snackAmpSettings(ebg) -foreground ::snackAmpSettings(efg)
   focus $T.e
   pack [frame $T.s -height 2 -borderwidth 2 -relief sunken] -side top -fill x -pady 3
   pack [frame $T.f] -side top -fill x -expand true
   pack [button $T.f.o -width 8 -text OK -command "set ::TA(continue) ok" -borderwidth 1 -relief raised] -side left
   pack [button $T.f.c -width 8 -text Cancel -command "set ::TA(continue) cancel" -borderwidth 1 -relief raised] -side right
   bind $T <Return> [list $T.f.o invoke]
   bind $T <Escape> [list $T.f.c invoke]
   grab set $T
   catch {wm transient $T [winfo toplevel [focus]]}
   centerWindow $T
   tkwait visibility $T
   focus $T
   set ::TA(continue) 0
   #-------WAIT---------
   vwait ::TA(continue)
   grab release $T
   destroy $T
   set variable $::TA(dialogEntry)
   return $::TA(continue)
}
###############################################################################
# Function   : myIcon
# Description: Try to place an icon for this window
# Author     : Tom Wilkason
###############################################################################
proc myIcon {w} {
   catch {
      switch -- $::tcl_platform(platform) {
         "unix" {wm iconbitmap $w snackPause}
         "windows" {winico_seticon $w $::TA(IconFile)}
         default {}
      } ;# End Switch
   } result
}
###############################################################################
# Function   : wmIconImage
# Description: This only works from some window managers
#              Not currently used
# Author     : Tom Wilkason
###############################################################################
proc wmIconImage {toplevel image} {
    variable _imgIcon_count
    if {![info exists _imgIcon_count]} {
        set _imgIcon_count 0
    }
    set iconTop [toplevel ._iconImage[incr _imgIcon_count]]
    pack [label $iconTop.l -image $image]
    wm iconwindow $toplevel $iconTop
}
###############################################################################
# Copyright (C) 1998 D. Richard Hipp
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Library General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Library General Public License for more details.
#
# You should have received a copy of the GNU Library General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA  02111-1307, USA.
#
# Author contact information:
#   drh@acm.org
#   http://www.hwaci.com/drh/
###############################################################################
proc labeledFrame {w args} {
   frame $w -bd 0
   label $w.l
   frame $w.f -bd 2 -relief groove
   frame $w.f.f
   pack $w.f.f
   set text {}
   set font {}
   set padx 3
   set pady 7
   set ipadx 2
   set ipady 9
   foreach {tag value} $args {
      switch -- $tag {
         -font  {set font $value}
         -text  {set text $value}
         -padx  {set padx $value}
         -pady  {set pady $value}
         -ipadx {set ipadx $value}
         -ipady {set ipady $value}
         -bd     {$w.f config -bd $value}
         -relief {$w.f config -relief $value}
      }
   }
   if {"$font"!=""} {
      $w.l config -font $font
   }
   $w.l config -text $text
   pack $w.f -padx $padx -pady $pady -fill both -expand 1
   place $w.l -x [expr $padx+10] -y $pady -anchor w
   pack $w.f.f -padx $ipadx -pady $ipady -fill both -expand 1
   raise $w.l
   return $w.f.f
}
;##
;# Implement a scrolled frame (Y Axis) that auto sizes
;# for the X Axis.
;# To use this, treat just like a regular frame.
;# The returned widget is the hull which you stuff data into
;# You have to pack/grid the $W.data you pass in
;# e.g.   set hull [scrolledFrame::Frame .data -initHeight 5i]
;#        for {set i 0} {$i < 50} {incr i} {
;#           pack [label $hull.lab$i -text "Number $i"]
;#        } ;# end for
;#        pack .data
;#
namespace eval scrolledFrame {
   ;##
   ;# Scrolled Canvas
   ;#
   proc Frame {base args} {
      set height 1i  ;# Initially
      set myArgs [list]
      ;# Pull off my unique arguments
      foreach {config setting} $args {
         if {[string match -nocase -inithei* $config]} {
            set height $setting
         } else {
            lappend myArgs $config $setting
         }
      } ;# end foreach

      frame $base
      frame $base.fm -bd 2 -relief groove

      canvas $base.fm.cv -yscrollcommand \
         [list scrollCmd $base.fm.cv_vertscrollbar [list pack $base.fm.cv_vertscrollbar -side right -fill y]]
      #[list $base.fm.cv_vertscrollbar set]
      scrollbar $base.fm.cv_vertscrollbar -orient vertical \
            -command  [list $base.fm.cv yview]
      pack $base.fm -side top -fill both -expand true
      pack $base.fm.cv_vertscrollbar -side right -fill y
      pack $base.fm.cv -side left -fill both -expand true

      set hull [eval frame $base.fm.cv.hull $myArgs]
      set wid [winfo width $base.fm]
      $base.fm.cv create window 0 0 -anchor nw -window $hull \
            -width $wid -height $height -tag window
      ;# On a canvas resize/configure
      bind $base.fm.cv <Configure> \
            [list scrolledFrame::canvasResize %W %w %h $hull]
      ;##
      mscroll $base.fm.cv
      ;# Insert an after idle callback that fires after the
      ;# stuff in the hull is drawn so we can query the size
      ;# and adjust the canvas width to accomodate the data.
      after idle [list scrolledFrame::propagate $base $hull $height]
      return $hull
   }
   ;##
   ;# Called on a canvas resize or configure event, need to
   ;# The canvas size and the scroll region
   proc canvasResize {win W H hull} {
      $win itemconfigure all -width $W -height [winfo reqheight $hull]  ;# Canvas fits the data
      set bbox [$win bbox all]
      $win configure -scrollregion $bbox -yscrollincrement 0.1i
   }
   ;# This forces the canvas container to the initial width
   ;# It is called once before the initial canvasResize event
   proc propagate {base hull height} {
      ;# Compute internal geometry then
      update idletasks
      ;# Configure the width of the canvas under the frame to the
      ;# required width of the frame.
      [winfo parent $hull] configure -width [winfo reqwidth $hull] -height $height
   }
}
###############################################################################
# Function   : scrollCmd
# Description: Generic callback to show/hide scrollbars when needed
#              scroll bars if they aren't needed
# Author     : Tom Wilkason
###############################################################################
proc scrollCmd {scrollbar geoCmd offset size} {
   catch {
      if {$offset > 0.02 || $size != 1.0} {
         eval $geoCmd
         $scrollbar set $offset $size
      } else {
         set manager [lindex $geoCmd 0]
         $manager forget $scrollbar
      } ;# end if
   }
}
namespace eval tabbedOptionFrame {
   variable TOF
}
###############################################################################
# Function   : tabbedOptionFrame::create
# Description: Create a tabbed option frame
# Author     : Tom Wilkason
# Date       : 5/15/2002
###############################################################################
proc tabbedOptionFrame::Create {base args} {
   variable TOF
   set myArgs [list]
   set options [list]
   set help [list]
   foreach {config setting} $args {
      switch -glob -- $config {
         -opt* {set options $setting}
         -help* {set help $setting}
         -tit* {}
         default {lappend myArgs $config $setting}
      } ;# End Switch
   } ;# end foreach
   eval frame $base $myArgs
   ;##
   ;# Listbox, register snackAmp colors. Contains the options
   ;# allowing user to select them
   ;#
   set plb [listbox $base.lb -exportselection no \
      -height [llength $options] -background $::snackAmpSettings(bg) \
      -foreground $::snackAmpSettings(fg) \
      -selectbackground $::snackAmpSettings(highlight) \
      -font listbox]
   #settings::regColor $plb -foreground ::snackAmpSettings(fg) -background ::snackAmpSettings(bg)
   ;##
   ;# For each option passed in with -options, insert it into the listbox
   ;# and bind the select on the listbox to show that frame
   ;#
   set i 0
   foreach {opt} $options {
      set hull [labeledFrame $base.fr$i -text $opt -bd 2 -relief groove -font menu]  ;#TFW: Place to add help
      set TOF($plb,$i,opt) $opt
      set TOF($plb,$i,frame) $base.fr$i
      set TOF($plb,$i,hull) $hull
      incr i
      $plb insert end $opt
   } ;# end foreach
   ;# Insert the options
   ;# When user selects an option, pack it
   bind $plb <ButtonPress-1> {
      focus %W
      tabbedOptionFrame::select %W [%W nearest %y]
   }
   bind $base.lb <Down> {
      tabbedOptionFrame::select %W [%W index active]
   }
   bind $base.lb <Up> {
      tabbedOptionFrame::select %W [%W index active]
   }
   ;# Put this new binding at the end so the up/down fires after
   ;# the active line is active.
   set tags [bindtags $base.lb]
   set tags [concat [lrange $tags 1 end] [lindex $tags 0]]
   bindtags $base.lb $tags
   ;# Pack the selector list box
   grid $plb -row 0 -column 0 -sticky nsew
   ;# Expand data only on resize
   grid columnconfigure $base 1 -weight 1
   grid rowconfigure $base 0 -weight 1
   set TOF($base,lb) $plb
   ;# Insert the initial frame if it exists
   if {$i > 0} {
      $plb selection set 0
      $plb activate 0
      select $plb 0
   } ;# end if
   return $base
}
###############################################################################
# Function   : tabbedOptionFrame::select
# Description: Callback when user selects a option
# Author     : Tom Wilkason
# Date       : 5/15/2002
###############################################################################
proc tabbedOptionFrame::select {lb index} {
   variable TOF
   if {[info exists TOF($lb,currentFrame)]} {
      grid forget $TOF($lb,currentFrame)
   } ;# end if
   grid $TOF($lb,$index,frame) -row 0 -column 1 -sticky nsew
   set TOF($lb,currentFrame) $TOF($lb,$index,frame)
}
###############################################################################
# Function   : tabbedOptionFrame::getHull
# Description: Return the hull for some index in an option frame
# Author     : Tom Wilkason
# Date       : 5/15/2002
###############################################################################
proc tabbedOptionFrame::getHull {base index} {
   variable TOF
   set lb $TOF($base,lb)
   if {[info exists TOF($lb,$index,hull)]} {
      return $TOF($lb,$index,hull)
   } else {
      return -code error "$base with index $index does not exist"
   } ;# end if
}
###############################################################################
# Function   : tabbedOptionFrame::getFrame
# Description: Return the frame for some index in an option frame
# Author     : Tom Wilkason
# Date       : 5/15/2002
###############################################################################
proc tabbedOptionFrame::getFrame {base index} {
   variable TOF
   set lb $TOF($base,lb)
   if {[info exists TOF($lb,$index,frame)]} {
      return $TOF($lb,$index,frame)
   } else {
      return -code error "$base with index $index does not exist"
   } ;# end if
}
###############################################################################
# Function   : tabbedOptionFrame::lock
# Description: Lock the minimimum size of each tab to the size of the largest
#              tab.
# Author     : Tom Wilkason
# Date       : 5/15/2002
###############################################################################
proc tabbedOptionFrame::lock {base} {
   variable TOF
   set i 0
   update idletasks
   set maxHeight 0
   set maxWidth 0
   foreach {tofIndex} [array names TOF "$base*,frame"] {
      set tof $TOF($tofIndex)
      set height [winfo reqheight $tof]
      set width [winfo reqwidth $tof]
      #puts "$height $width $tof"
      if {$height > $maxHeight} {set maxHeight $height} ;# end if
      if {$width > $maxWidth} {set maxWidth $width} ;# end if
   } ;# end foreach
   #puts "$base Max Height $maxHeight, Max Width $maxWidth"
   grid columnconfigure $base 1 -minsize $maxWidth
   grid rowconfigure $base 0 -minsize $maxHeight
}
###############################################################################
# Function   : waitWindow namespace
# Description: Functions to pop-up a window while stuff is going on
# Author     : Tom Wilkason
###############################################################################
namespace eval waitWindow {
   variable W .waitWindow
   variable linkedText "Please Wait"
   variable width 60
}
proc waitWindow::init {args} {
   variable W
   variable width
   variable linkedText
   set splash [toplevel $W -relief raised -bd 4]
   wm withdraw $W
   wm overrideredirect $W 1
   set i 0
   pack [label $W.[incr i] -image saLogo] -side left
   pack [frame $W.[incr i] -width 2 -borderwidth 2 -relief sunken] -expand true -fill y -side left
   pack [label $W.[incr i] -anchor w -textvariable waitWindow::linkedText -width [expr {$width+3}] \
      -font {helvetica 12 bold}] \
      -side left -expand 1 -fill x
   update idletasks
   # center it
   wm geometry $W +[expr {([winfo screenwidth .]-[winfo reqwidth $W])/2}]+[expr {([winfo screenheight .]-[winfo reqheight $W])/2}]
}
proc waitWindow::show {Text} {
   variable W
   variable linkedText
   variable width
   if {[wm state $W] != "normal"} {
      wm deiconify $W
   } else {
      raise $W
   }
#   if {[string length $Text] > $width} {
#      set linkedText "[string range $Text 0 $width]..."
#   } else {
      set linkedText $Text
#   }
}
proc waitWindow::close {args} {
   variable W
   wm withdraw $W
}
;##
;# Lock the minimum toplevel size to required size
;#
proc lockSize {B} {
   update idletasks
   wm minsize $B [winfo reqwidth $B] [winfo reqheight $B]
}
###############################################################################
# Function   : killAfter
# Description: Kill some lingering after loops
# Author     : Tom Wilkason
# Date       : 8/4/2001
###############################################################################
proc killAfter {name} {
   foreach one [after info] {
      if {[lindex [after info $one] 0] == $name} {
         after cancel $one
      } ;# end if
   }
}
###############################################################################
# Function   : amAfterExists
# Description: Determine if an after loop is running for some process
# Author     : Tom Wilkason
# Date       : 6/21/1999
###############################################################################
proc afterExists {Proc {arg ""}} {
   foreach aft [after info] {
      set afinfo [lindex [after info $aft] 0]
      if {[string length $arg] == 0} {
         if {[string equal [lindex $afinfo 0] $Proc]} {
            return 1
         }
      } else {
         if {[string equal [lindex $afinfo 0] $Proc] && [string equal [lindex $afinfo 1] $arg]} {
            return 1
         }
      }
   }
   return 0
}
###############################################################################
# Function   : centerWindow
# Description: Center a window on the screen
# Author     : Tom Wilkason
# Date       : 11/3/2001
###############################################################################
proc centerWindow {w} {
   wm withdraw $w
   update idletasks
   wm geometry $w "+[expr {([winfo screenwidth .]-[winfo reqwidth $w])/2}]+[expr {([winfo screenheight .]-[winfo reqheight $w])/2}]"
   wm deiconify $w
}
;# Windows RMB for scales, this will eventually
;# make its way into 8.4. Allows RMB to jump
;# to a scale location as it does on *nix already
###############################################################################
# Function   : fixScaleBinding
# Description: On pre 8.4 systems on windows, the RMB didn't jump the scale
#              make it so here
# Author     : Tom Wilkason
# Date       : 6/9/2002
###############################################################################
proc fixScaleBinding {args} {
   if {$::tcl_platform(platform) == "windows"} {
      if {$::tcl_version <= 8.3} {
         bind Scale <3> {
             tkScaleButton2Down %W %x %y
         }
         bind Scale <B3-Motion> {
             tkScaleDrag %W %x %y
         }
         bind Scale <B3-Leave> { }
         bind Scale <B3-Enter> { }
         bind Scale <ButtonRelease-3> {
             tkCancelRepeat
             tkScaleEndDrag %W
             tkScaleActivate %W %x %y
         }
      }
   } ;# end if
}
###############################################################################
# Function   : APLExists
# Description: Return a bool to denote the state of the Autoplaylist
# Author     : Tom Wilkason
# Date       : 6/20/2002
###############################################################################
proc APLExists {args} {
   return [expr {[info exists ::PLB] && [winfo exists $::PLB]}]
}
###############################################################################
# Function   : ontop namespace
# Description: Provide windows functions to pin a window on top
# Author     : Tom Wilkason
# Date       : 7/26/2002
###############################################################################
namespace eval ontop {
   variable enabled 0
   foreach {left top right bottom} {0 0 0 0} {
      set ::TA(screen_left) 0
      set ::TA(screen_top) 0
      set ::TA(screen_right) [winfo screenwidth .]
      set ::TA(screen_bottom) [expr {[winfo screenheight .]-50}]
   }
   if {$::tcl_platform(platform)=="windows"} {
      if {[catch {package require TkWinTools} result]} {
         puts $result
         proc control {win how} {
            saLog "tkwintools not found"
         }
         proc state {win} {
            saLog "tkwintools not found"
         }
      } else {
         proc control {win how} {
            tkwintools::topmost [winfo toplevel $win] $how
         }
         proc state {win} {
            return [tkwintools::topmost [winfo toplevel $win]]
         }
         foreach {left top right bottom} [tkwintools::sysinfo -workarea] {
            set ::TA(screen_left) $left
            set ::TA(screen_top) $top
            set ::TA(screen_right) $right
            set ::TA(screen_bottom) $bottom
         }
         set enabled 1
      }
   } ;# end if
   ;# Put a pin on top menu on the system menu bar under windows
   proc pinOnSystemMenu {M W} {
      variable enabled
      if {$enabled} {
         if {$M=={}} {
            set M $W.menubar
            menu $M -tearoff 0
            $W config -menu $M
         }
         $M add cascade -label System -font menu -menu $M.system
         menu $M.system -tearoff 0
         set ::TA($W,pinned) 0
         $M.system add separator
         $M.system add checkbutton -label "Always on Top" \
            -variable ::TA($W,pinned) \
            -command "ontop::control $W \$::TA($W,pinned)"
      }
   }
}
;##
;# This fixes a bug in 8.4 that posted windows menus incorrectly
;#
if {$::tcl_version==8.4} {
   proc ::tk::PostOverPoint {menu x y {entry {}}}  {
      global tcl_platform

      if {[string compare $entry {}]} {
         if {$entry == [$menu index last]} {
            incr y [expr {-([$menu yposition $entry] \
               + [winfo reqheight $menu])/2}]
         } else {
            incr y [expr {-([$menu yposition $entry] \
               + [$menu yposition [expr {$entry+1}]])/2}]
         }
         incr x [expr {-[winfo reqwidth $menu]/2}]
      }
      $menu post $x $y
      if {$entry ne "" && [$menu entrycget $entry -state] ne "disabled"} {
         $menu activate $entry
         GenerateMenuSelect $menu
      }
   }
} ;# end if
namespace eval levelmeter {
   variable left {}
   variable right {}
}
;##
;# Left/Right channel meter
;#
;# -width 20 -length 200 -orient vertical -oncolor green
proc levelmeter::create {widget args} {
   variable left $widget.left
   variable right $widget.right
   frame $widget
   grid [eval snack::levelMeter $left $args] \
        [eval snack::levelMeter $right $args]
   return $widget
}
;##
;# Compute the current sample levels
;#
proc levelmeter::sampleLevel {s _left _right} {
   upvar $_left left
   upvar $_right right
   if {[snack::audio active]} {
      if {[catch {
         set level $::TA(line,$::snackAmpSettings(lineSelect))
         set here [$s current_position]
         set slevel [$s sample $here]
         switch -- [llength $slevel] {
            1 {
                  set left [expr {abs($slevel*$level/100.)}]
                  set right [expr {abs($slevel*$level/100.)}]
            }
            2 {
               foreach {s1 s2} $slevel {
                  set left [expr {abs($s1*$level/100.)}]
                  set right [expr {abs($s2*$level/100.)}]
               }
            }
            default {}
         }
      } result]} {
         set left 0
         set right 0
      }
   } else {
      set left 0
      set right 0
   }
}
;##
;# Update the meter with current sample levels
;#
proc levelmeter::cmUpdate {s} {
   variable left
   variable right
   if {[winfo exists $left] && [winfo ismapped $left]} {
      sampleLevel $s l r
      $left  configure -level $l
      $right configure -level $r
      after $::snackAmpSettings(meterUpdateRate) levelmeter::cmUpdate $s
   } else {
      after 3000 levelmeter::cmUpdate $s
   }
}

package provide gpWidgets 1.0
