##################################################################################
# snackAmp - saFiltering package 1.0 for snackAmp
# provides music filtering utilities
#
# Music File Player in Tcl/Tk
#
# Copyright (C) 2001 Tom Wilkason
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
#
# Please send any comments/bug reports to
# tom.wilkason@cox.net  (Tom Wilkason)
##################################################################################
set [file root [file tail [info script]]]_History {
   $Header: /cvsroot/snackamp/snackamp/lib/saFiltering.tcl,v 1.19 2003/04/16 01:49:59 wilkason Exp $
}
namespace eval filtering {}
###############################################################################
# Function   : filtering::filterHistory
# Description: postcommand callback to Display the history filter, called from
#              button next to filter box.
# Author     : Tom Wilkason
###############################################################################
proc filtering::filterHistory {menu} {
   global snackAmpSettings
   ;# Only keep up to last to filters
   set snackAmpSettings(History) [lrange $snackAmpSettings(History) 0 20]
   $menu delete 0 end
   destroy $menu.replace
   destroy $menu.append
   destroy $menu.select
   destroy $menu.insert
   $menu add cascade -label "Append To Playlist..." -menu [menu $menu.append -tearoff 0 -font menu]
   $menu add cascade -label "Insert Into Playlist..." -menu [menu $menu.insert -tearoff 0 -font menu]
   $menu add cascade -label "Replace Playlist..." -menu [menu $menu.replace -tearoff 0 -font menu]
   $menu add cascade -label "Select In Playlist..." -menu [menu $menu.select -tearoff 0 -font menu]

   $menu.append add command -label "Filtering '$::TA(filter)' on all known files" -command {
      filtering::pushFilter
      macros::filterMusic append 1
   }
   $menu.append add command -label "Filtering '$::TA(filter)' on Auto-playlist files" -command {
      filtering::pushFilter
      macros::filterMusic append 0
   }

   $menu.select add command -label "Filtering '$::TA(filter)' on current Playlist" -command {
      filtering::pushFilter
      macros::filterMusic select playList
   }

   $menu.replace add command -label "Filtering '$::TA(filter)' on all known files" -command {
      filtering::pushFilter
      macros::filterMusic replace 1
   }
   $menu.replace add command -label "Filtering '$::TA(filter)' on Auto-playlist files" -command {
      filtering::pushFilter
      macros::filterMusic replace 0
   }
   $menu.replace add command -label "Filtering '$::TA(filter)' on current Playlist" -command {
      filtering::pushFilter
      macros::filterMusic replace playList
   }

   $menu.insert add command -label "Filtering '$::TA(filter)' on Auto-playlist files" -command {
      filtering::pushFilter
      macros::filterMusic insert 0
   }
   $menu.insert add command -label "Filtering '$::TA(filter)' on current Playlist" -command {
      filtering::pushFilter
      macros::filterMusic insert playList
   }

   $menu add separator
   $menu add command -label "Reset Filter to Auto-playlist" -command {
      set ::TA(filter) {}
      macros::filterMusic replace $::TA(matchAgainstAll)
   }
   $menu add separator
   foreach {item} $snackAmpSettings(History) {
      $menu add command -label "$item" -command "[list set ::TA(filter) $item]"
   }
}
###############################################################################
# Function   : filtering::pushFilter
# Description: Save the latest filter entry if it is not already in the filter history
# Author     : Tom Wilkason
###############################################################################
proc filtering::pushFilter {args} {
   if {$::TA(filter) != {} && [lsearch -exact $::snackAmpSettings(History) $::TA(filter)] < 0} {
      set ::snackAmpSettings(History) [linsert $::snackAmpSettings(History) 0 $::TA(filter)]
   }
}
###############################################################################
# Function   : filtering::checkIfIncluded
# Description: Determine if a file is included in the folder path, check
#              down to the root.
# Author     : Tom Wilkason
###############################################################################
proc filtering::checkIfIncluded {File} {
   global TF teFolders
   ;##
   ;# Remove entries that are excluded
   if {[info exists teFolders($File)] && $teFolders($File)==1} {
      return 1
   } else {
      return 0
   } ;# end if
}
################################################################################
# Function   : filtering::filterMusic
# Description:
# Author     : Tom Wilkason
# Date       : 7/1/2002
###############################################################################
proc filtering::filterMusic {how what} {
   global TF TA LB TFList snackAmpSettings CF
   soundControl::latency long
   update
   ;##
   ;# Filter on entire list or just the playlist
   ;#
   set useList [list]
   switch -- $what {
      "playList" {
         ;# Use only playlist
         foreach {Entry} [LB::getAllNames] {
            lappend useList $TF($Entry)
         } ;# end foreach
         set MA 1
      }
      1 {
         ;# Use all files (sorted)
         set useList [catalog::popList]
         set MA 1
      }
      0 {
         ;# use APL files (MA will force an APL check)
         set useList [catalog::popList]
         set MA 0
      }
      def {
         ;# use Match against All flag
         set useList [catalog::popList]
         set MA $::TA(matchAgainstAll)
      }
   } ;# End Switch
   ;##
   ;# Where to put the filtered data
   ;#
   switch -- $how {
      "replace" {
         LB::deleteAll
         catch {unset TF}
         set select 0
         set from "end"
      }
      "insert" {
         set from [expr {[LB::active]+1}]
         set select 0
      }
      "append" {
         set from "end"
         set select 0
      }
      "select" {
         set from "end"
         set select 1
      }
      default {
         set from "end"
         set select 0
      }
   } ;# End Switch
   set insertList [list]
   ;##
   ;# Support boolean searches with 'or' clause
   ;#
   if {[string length $::TA(filter)]==0} {
      set fast 1
      set regPattern ""
   } else {
      set fast 0
      set regPattern [list]                                ;#retain wildcards
      set pattern [split [string map {" or " "|" " OR " "|" "*" ".*" "?" "."} $TA(filter)] |]
      foreach part $pattern {
         lappend regPattern "($part)+"
      }
      set regPattern [join $regPattern |]
   }

   ;# Below will prevent duplicates due to hashing
   ;# TFList is is a complete list of all known songs
   ;# TF is a hashed array of playlisted songs pointed to full name
   set RF $snackAmpSettings(UseRemovableFolders)
   set rmtFilter [file join $::RMT *]
   ;##
   ;# TFList is using all files,
   ;#
   if {[catch {
      set i -1
      foreach {File} $useList {
         incr i
         ;# Skip removable files if disabled
         if {!$RF && [string match $rmtFilter $File]} {
            continue
         }
         ;# Pattern match (of if no pattern ignore regexp)
         if {$fast || [regexp -nocase $regPattern $File]} {
            ;# Verify this file is in a checked folder
            ;# See if this file matches folder menu entry
            if {$MA || [filtering::checkIfIncluded [file dirname $File]]} {
               if {$select} {
                  LB::selectionSet $i
               } else {
                  set keyName [LB::getListName $File]
                  lappend insertList $keyName
                  ;# Create new TF entry for fast access to full file name
                  set TF($keyName) $File
               }
            } ;# end if
         } ;# end if
      } ;# end foreach
   } result]} {
      saLog "macros::filterMusic: '$regPattern' $result"
   }
   if {[llength $insertList] > 0} {
      LB::insertList $from $insertList
   } ;# end if
   LB::tryRestorePlaying
   ;# Show number of files Parent listbox
   updateFileDisplay
   soundControl::latency normal
}
###############################################################################
# Function   : toggleFilter
# Description: Toggle the filter button (checkbutton won't lie flat!)
# Author     : Tom Wilkason
# Date       : 11/11/2001
###############################################################################
proc filtering::toggleFilter {args} {
   if {$::TA(matchAgainstAll)} {
      $::TA(filterAllButton) configure -relief flat
      $::TA(filterAllButton) configure -image filterOff
      set ::TA(matchAgainstAll) 0
   } else {
      $::TA(filterAllButton) configure -relief sunken
      $::TA(filterAllButton) configure -image filterImage
      set ::TA(matchAgainstAll) 1
   } ;# end if
}

###############################################################################
# Function   : filtering::sortMusic
# Description: Sort or shuffle the master playlist
# Author     : Tom Wilkason
# Date       : 12/26/2000
###############################################################################
proc filtering::sortMasterList {args} {
   switch -- $::snackAmpSettings(sortMode) {
      0       {set tflist [catalog::sortedList]}
      1       {set tflist [shuffle $::TFList]}
      2       {set tflist [shuffleByFolder $::TFList ldsort]}
      3       {set tflist [shuffleByFolder $::TFList shuffle]}
      default {set tflist [list]}
   } ;# End Switch
   ;# Need to save order for filtering
   catalog::pushList $tflist
   return $tflist
}
###############################################################################
# Function   : filtering::sortMusic
# Description: Sort or shuffle the playlist files
# Author     : Tom Wilkason
# Date       : 12/26/2000
###############################################################################
proc filtering::sortMusic {args} {
   global TF TA LB snackAmpSettings CF TFList
   ;# 1st, shuffle or sort the overally list
   soundControl::latency long
   set tflist [filtering::sortMasterList]
   ;##
   ;# Need to restore the current playlist, in the
   ;# order of the newly sorted master list
   ;#
   set shortList [list]
   set playList [LB::getAllNames 1]
   if {[llength $playList] > 0} {
      foreach item $playList {
         lappend shortList $item
         set playlistHash($item) [LB::getListName $item]
      } ;# end foreach
      ;# Now build up new playlist in the order that they appear
      ;# in the newly sorted list (pushedList). This maintains ordering between filter
      ;# operations and retains manually added items to the list.
      ;# This takes a while for large playlists and large master lists
      set insertList [list]
      foreach item $tflist {
         if {[info exists playlistHash($item)]} {
            lappend insertList $playlistHash($item)
         } ;# end if
      } ;# end foreach
      ;# Replace the current playlist
      LB::deleteAll
      LB::insertList end $insertList
   } ;# end if
   LB::tryRestorePlaying
   updateFileDisplay
   soundControl::latency normal
}
;##
;# Apend a file to the bottom of the play list
;#
proc filtering::Append {file} {
   global TF TA LB
   if {[string length $file] > 0} {
      set keyName [LB::getListName $file]
      LB::insert end $keyName
      set TF($keyName) $file
      ;# If a new incoming file, make sure it is also in the master list
      catalog::addToDbIfNeeded $file
   } ;# end if
}
;##
;# Insert a song at current position of the playlist and then play it
;#
proc filtering::InsertPlay {file} {
   global TF TA LB
   if {$file != {}} {
      ;##
      ;# If file exists in playlist, jump to it instead
      ;# of inserting it
      ;#
      set curPost [LB::indexOfFile $file]
      if {$curPost < 0} {
         set curPost [LB::active]
         incr curPost
         ;# Make sure it is in the master list
         set keyName [LB::getListName $file]
         LB::insert $curPost $keyName
         set TF($keyName) $file
         ;# If a new incoming file, make sure it is also in the master list
         catalog::addToDbIfNeeded $file
         updateFileDisplay
      }
      LB::activate $curPost
      Play active
   } ;# end if
}
;##
;# Insert a file into the playlist
;#
proc filtering::Insert {curPost file} {
   global TF TA LB
   if {$file != {}} {
      ;# Make sure it is in the master list
      set keyName [LB::getListName $file]
      LB::insert $curPost $keyName
      set TF($keyName) $file
   } ;# end if
}

;##
;# LB Namespace wrapper for the ListBox
;#
namespace eval LB {
   variable indexPlaying -1
}
proc LB::see {index} {
   if {[catch {$::LB see $index} result] } then {
      saLog $result
   }
}
###############################################################################
# Function   : LB::move
# Description: Move an entry restore any highlighting it may have
# Author     : Tom Wilkason
# Date       : 6/3/2002
###############################################################################
proc LB::move {old new} {
   variable indexPlaying
   global LB
   set item [$LB get $old]
   set fg  [$LB itemcget $old -foreground]
   set bg  [$LB itemcget $old -background]
   set sfg [$LB itemcget $old -selectforeground]
   set sbg [$LB itemcget $old -selectbackground]
   $LB delete $old
   $LB insert $new $item
   $LB activate $new
   if {[string length $fg]>0} {
      $LB itemconfigure $new -background $bg
      $LB itemconfigure $new -foreground $fg
      set indexPlaying $new
   } ;# end if
   if {[string length $sfg]>0} {
      $LB itemconfigure $new -selectbackground $sbg
      $LB itemconfigure $new -selectforeground $sfg
      set indexPlaying $new
   } ;# end if
}
;##
;# Make sure some file is viewable
;#
proc LB::seeFile {File} {
   global LB TF
   if {$File == {}} {
      return 0
   } ;# end if
   set index [LB::indexOfFile $File]
   if {$index > -1} {
      $LB see $index
      return 1
   } else {
      return 0
   }
}
;##
;# Make sure the playing file is viewable
;#
proc LB::seePlaying {args} {
   global LB TF
   set index $LB::indexPlaying
   if {$index > -1} {
      $LB see $index
      return 1
   } else {
      $LB see [active]
      return 0
   }
}
proc LB::end {args} {
   return [$::LB index end]
}
proc LB::get {index} {
   return [$::LB get $index]
}
proc LB::index {index} {
   return [$::LB index $index]
}
proc LB::activate {index} {
   if {[catch {$::LB activate $index} result] } then {
      saLog $result
   } else {
      $::LB see $index
   }
}
proc LB::selectionSet {index} {
   $::LB selection set $index
}
###############################################################################
# Function   : LB::getListName
# Description: Return a song name with the appropriate folders prepended
#              Remove the file extension if needed
# Author     : Tom Wilkason
# Date       : 1/3/2001
###############################################################################
proc LB::getListName {File} {
   global snackAmpSettings
   if {$snackAmpSettings(ListDepth) > 0} {
      ;# File joing doesnt work due to single list
      set File [string map {// /} $File]      ;# remove UNC stuff
      set name [join [lrange [file split $File] end-$snackAmpSettings(ListDepth) end] /]
   } else {
      set name [file tail $File]
   }
   if {$::TA(noTail)} {
      set name [file root $name]
   }
   return $name
}
;##
;# After the playlist has been altered, try and restore
;# the playing song with highlighting.
;# This assumes the current highlighted file
;# was removed first.
;# TFW: If duplicates exist in the playlist
;#      this will find only the first one.
;#
proc LB::tryRestorePlaying {args} {
   global TA LB TF
   if {![LB::seeFile $TA(currentMF)]} {
      LB::activate 0
   } else {
      if {[soundControl::isPlaying]} {
         LB::inverseVideoFile $TA(currentMF) 1
      } ;# end if
   }
}
;##
;# Inverse video a file
;#
proc LB::inverseVideoFile {File {On 1}} {
   Trace
   variable indexPlaying
   global LB
   if {$On} {
      set index [LB::indexOfFile $File]
      if {$index >= 0} {
         LB::inverseVideoIndex $index $On
         LB::activate $index
      }
   } else {
      ;# Don't care about file, always assume playing
      LB::inverseVideoIndex $indexPlaying $On
   } ;# end if
}
;##
;# Inverse video a selection
;#
proc LB::inverseVideoIndex {index {On 1}} {
   Trace
   variable indexPlaying
   global LB
   if {$index > -1} {
      set bg  [$LB cget -background]
      set fg  [$LB cget -foreground]
      set sbg [$LB cget -selectbackground]
      set sfg [$LB cget -selectforeground]
      ;##
      ;# Color or uncolor the file
      ;#
      if {$On} {
         $LB itemconfigure $index -background $fg
         $LB itemconfigure $index -foreground $bg
         $LB itemconfigure $index -selectforeground $sbg
         $LB itemconfigure $index -selectbackground $sfg
         set indexPlaying $index
      } else {
         $LB itemconfigure $index -background $bg
         $LB itemconfigure $index -foreground $fg
         $LB itemconfigure $index -selectforeground $sfg
         $LB itemconfigure $index -selectbackground $sbg
      }
   }
}
;##
;# Locate a file in the playlist from the anchor position
;#
proc LB::indexOfFile {File} {
   global LB TF
   if {$File == {}} {
      return -1
   } ;# end if
   set lnFile [LB::getListName $File]
   set anchor [$LB index active]
   set index [lsearch -exact [$::LB get $anchor end] $lnFile]
   if {$index == -1} {
      set index [lsearch -exact [$::LB get 0 $anchor] $lnFile]
      if {$index == -1} {
         return -1
      }
   } else {
      ;# Search must be offset from anchor
      incr index $anchor
   }
   return $index
}
proc LB::getAllNames {{long 0}} {
   set all [$::LB get 0 end]
   set files [list]
   if {$long} {
      foreach {item} $all {
         lappend files $::TF($item)
      } ;# end foreach
      return $files
   } else {
      return $all
   }
}
proc LB::selected {args} {
   return [$::LB curselection]
}
proc LB::all {args} {
   set result [list]
   set end [$::LB index end]
   for {set i 0} {$i < $end } {incr i} {
      lappend result $i
   } ;# end for
   return $result
}
proc LB::anchor {args} {
   return [$::LB index anchor]
}
proc LB::active {args} {
   return [$::LB index active]
}
proc LB::playing {args} {
   variable indexPlaying
   return $indexPlaying
}
proc LB::replace {i newKey} {
   $::LB delete $i
   $::LB insert $i $newKey
   ;# TFW: Inverse if it was
}
proc LB::getFullName {Index {translate 0}} {
   if {[catch {
      if {$translate} {
         set res [catalog::TranslateMount $::TF([$::LB get $Index])]
      } else {
         set res $::TF([$::LB get $Index])
      }
   }]} then {
      set res ""
   }
   return $res
}
proc LB::getFullNameFromHash {hash {translate 0}} {
   if {[catch {
      if {$translate} {
         set res [catalog::TranslateMount $::TF($hash)]
      } else {
         set res $::TF($hash)
      }
   }]} then {
      set res $hash
   }
   return $res
}
proc LB::getActiveFullName {args} {
   if {[catch {
      set active [$::LB index active]
      set res $::TF([$::LB get $active])    ;# lookup long name
   }]} then {
      set res ""
   }
   return $res
}
proc LB::getSelectedFullNames {{translate 1}} {
   set fullNames [list]
   if {$translate} {
      foreach {Index} [selected] {
         lappend fullNames [catalog::TranslateMount $::TF([$::LB get $Index])]
      } ;# end foreach
   } else {
      foreach {Index} [selected] {
         lappend fullNames [$::LB get $Index]
      } ;# end foreach
   }
   return $fullNames
}
proc LB::deleteAll {args} {
   variable indexPlaying
   $::LB delete 0 end
   set indexPlaying -1
}
proc LB::delete {position} {
   variable indexPlaying
   $::LB delete $position
   if {[string is double $position] && ($indexPlaying >= $position)} {
      incr indexPlaying -1
   } ;# end if
   return $position
}
proc LB::insertList {position items} {
   variable indexPlaying
   ;# TFW: adjust play position here if needed
   eval $::LB insert $position $items
   if {[string is double $position] && ($indexPlaying >= $position)} {
      incr indexPlaying [llength $items]
   } ;# end if
}
proc LB::insert {position item} {
   variable indexPlaying
   $::LB insert $position $item
   ;# TFW: adjust play position here needed
   if {[string is double $position] && ($indexPlaying >= $position)} {
      incr indexPlaying
   } ;# end if
}
;##
;# Select/Drag, set on mouse click
;#
proc LB::Select {y} {
   global TF TA LB
   set TA(old) [$LB nearest $y]
   set Value [$LB get $TA(old)]
   if {$Value != {}} {

      set TA(entryName) $TF($Value)
      if {$::snackAmpSettings(updateAPL)} {
         catch {APL::SelectFile $TF($Value)}
      } ;# end if
   } ;# end if
}
;##
;# Drag an entry up/down
;#
proc LB::Drag {y} {
   global TF TA LB
   set new [$LB nearest $y]
   if {$new == -1} return
   set to $TA(old)
   LB::move $TA(old) $new
   LB::selectionSet $new
   $LB selection anchor $new
   set TA(old) $new
}

package provide saFiltering 1.0

