##################################################################################
# snackAmp - saSettings package 1.0 for snackAmp
# provides user preferences and GUI
#
# Music File Player in Tcl/Tk
#
# Copyright (C) 2001 Tom Wilkason
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
#
# Please send any comments/bug reports to
# tom.wilkason@cox.net  (Tom Wilkason)
##################################################################################
set [file root [file tail [info script]]]_History {
   $Header: /cvsroot/snackamp/snackamp/lib/saSettings.tcl,v 1.65 2003/05/17 21:33:58 wilkason Exp $
}

namespace eval settings {
   variable reg
   variable FONTS
   variable Colors
   ;##
   ;# Create the initial fonts we will need, they may be scaled or changed later
   ;#
   set FONTS {
      entry      {Helvetica 11 bold   roman} "Entry Boxes"
      listbox    {Helvetica 10 bold   roman} "List Boxes"
      balloon    {Helvetica 8  normal roman} "Balloon Help"
      smallLabel {Helvetica 9 bold   italic}  "Small Track Label"
      songInfo   {Helvetica 10 bold   roman} "Small Track Info"
      songLabel  {Helvetica 13 bold   roman} "Track Title"
      timeStatus {Helvetica 16 bold   roman} "Time Counters"
      timeScale  {Helvetica 8  normal roman} "Time Scale"
      treeFont   {Helvetica 10 normal roman} "APL Tree Text"
   }
   ;##
   ;# Set some defaults for colors, overridden later
   ;#
   set ::snackAmpSettings(ebg)  "lightyellow"     ;# entry box background
   set ::snackAmpSettings(efg)  "black"           ;# entry text color
   set ::snackAmpSettings(bg)   "MidnightBlue"   ;# text background
   set ::snackAmpSettings(tfg)  "LightCyan"      ;# Title Text Color
   set ::snackAmpSettings(fg)   "cornsilk"        ;# text foreground
   set ::snackAmpSettings(sfg)  "yellow"          ;# Time Counter forground
   set ::snackAmpSettings(lfg)  "gold3"          ;# song extra info label forground
   set ::snackAmpSettings(tsbg) "lightblue"       ;# table strip background
   set ::snackAmpSettings(highlight) "blue"       ;# highlight color
   #   default    {Helvetica 10 normal roman} "Default"
   foreach {Font Specs Description} $FONTS {
      foreach {family size weight slant}  $Specs {break}
      font create $Font -family $family -weight $weight -size $size -slant $slant
   } ;# end foreach
}
###############################################################################
# Function   : settings::tabbedSettingsGui
# Description: Main Settings GUI
# Author     : Tom Wilkason
###############################################################################
proc settings::tabbedSettingsGui {args} {
   variable reg
   package require snID3
   global TA snackAmpSettings
   set B .tsg
   destroy $B
   toplevel $B  -cursor $::TA(arrow)
   wm title $B "SnackAmp Settings"
   set options {
      "Play Settings"   "playSettings"      "- Settings that are common or affect "
      "User Interface"  "uiSettings"        "- Settings that are common or affect "
      "Auto Playlist"   "APLSettings"     "- (Applies to multi-column table only)"
      "Auto Level"      "AutoLevelSettings" ""
      "ID3 Tags"        "id3Settings"       ""
      "Folder Settings" "folderSettings"    ""
      "Fonts"           "fontSettings"      ""
      "Colors"          "colorSettings"     ""
      "Web Server"      "webServer"      ""
      "Streaming Audio" "streamingSettings" ""
      "Diagnostic"      "AdvancedSettings"  ""
   }
   ;# Windows only
   if {$::tcl_platform(platform) == "windows"} {
      lappend options "File Associations" "associationSettings" ""
   } ;# end if

   foreach {item callback help} $options {
      lappend entries $item
   } ;# end foreach
   ;# Save current settings for possible restore
   array set ::oldSnackAmpSettings [array get snackAmpSettings]
   ;##
   ;# Create a tabbed option frame
   ;#
   set tof [tabbedOptionFrame::Create $B.tf -options $entries -title "Select settings"]
   pack $tof -expand true -fill both
   set i 0
   foreach {item callback help} $options {
      set hull [tabbedOptionFrame::getHull $tof $i]
      ;# Create metawidget to pack into hull for this tab
      set tab [$callback $hull]
      pack $hull
      incr i
   } ;# end foreach

   ;# Line
   pack [frame $B.[incr i] -height 2 -borderwidth 2 -relief sunken] -expand true -fill x
   ;# Button frame
   set bf [frame $B.[incr i]]
   pack $bf -side bottom -fill none -expand false
   #  ----------------------------------------------------------------------------
   #  Buttons at the bottom
   #  ----------------------------------------------------------------------------
   ;# Apply and Close
   set okBut [button $bf.[incr i] -text "OK" -width 8 -font menu -borderwidth 1 -command {
      if {[winfo exists .saTop] && ($::oldSnackAmpSettings(ListDepth) != $snackAmpSettings(ListDepth))} {
         macros::filterMusic replace $::TA(matchAgainstAll)
      }
      settings::saFontSize $snackAmpSettings(FontSizeFactor)
      settings::saveSettings
      destroy .tsg
   } -relief raised]
   bind $B <Return> [list $bf.$i invoke]
   balloon::define $okBut "Push this to apply and save the current settings then close this dialog"
   ;# Apply
   set app [button $bf.[incr i] -width 8 -font menu -text "Apply" -borderwidth 1 -command {
      settings::saFontSize $snackAmpSettings(FontSizeFactor)
      settings::saveSettings
      array set ::oldSnackAmpSettings [array get ::snackAmpSettings]
      catch {set snackAmpSettings(HttpLocal) [file normalize $snackAmpSettings(HttpLocal)]}
      set snackAmpSettings(HttpLocal) [string trim $snackAmpSettings(HttpLocal)]
   }   -relief raised]
   balloon::define $app "Push this apply and save the current settings"

   ;# Cancel, restore settings
   set canBut [button $bf.[incr i] -text "Cancel" -width 8 -font menu -borderwidth 1 -command {
         array set snackAmpSettings [array get ::oldSnackAmpSettings]
         catch {set snackAmpSettings(HttpLocal) [file normalize $snackAmpSettings(HttpLocal)]}
         settings::installFontCallbacks
         settings::applyColor
         destroy .tsg
   } -relief raised]
   bind $B <Escape> [list $bf.$i invoke]
   pack $okBut $app $canBut -side left -padx 20 -pady 5
   balloon::define $canBut "Push this to ignore the changes made and restore the prior settings."
   catch {wm transient $B .saTop}
   centerWindow $B
   ;# Lock down the minimum size of each tab (do after center to prevent flash)
   tabbedOptionFrame::lock $tof
   myIcon $B
   ;# Lock the overall minimum size
   lockSize $B
   ;# Block for closure
   tkwait window $B
   ;# If font size changed, reapply it
   settings::saFontSize $snackAmpSettings(FontSizeFactor)
}
###############################################################################
# Function   : settings::playSettings
# Description: Settings page Misc functions
# Author     : Tom Wilkason
###############################################################################
proc settings::playSettings {W} {
   set optionData {
      checkbutton ContinuousPlay          "Default of Continuous Play"
       "Check if you want to continuously play the playlist, uncheck if you want to stop after last track in playlist"
      checkbutton NextAfterCurrnt         "Next Track always after Playing Track"
       "Check this to play the next song after the playing song rather than after the current selection."

      checkbutton restorePlayList         "Restore Prior Playlist"
       "Check this if you want to automatically restore the entire playlist and location from the previous session."
      checkbutton resumePlay              "Resume Play on Restart"
       "Check this to have SnackAmp resume play at the point the application exited (if a track was playing and restore prior playlist is set)."

      checkbutton fadeFast               "Fade on Stop"
       "Check this fade the sound level down rather than abruptly stop it on a stop or next/previous track play. See Fade Factors setting above."
      checkbutton fadePause               "Fade on Pause"
       "Check this fade the sound level down rather than abruptly stop it on a pause track. See 'Fade Factors' slider to set the fade time."

      checkbutton autoLevel               "Enable Auto Leveling"
       "Check this to have SnackAmp adjust the audio level based on how loud the track is. Useful for mp3/ogg files encoded using different encoders. Recommend you enable this if different rippers encoded your tracks."
      checkbutton AutoDelete              "Auto Delete Invalid Files"
       "Check this to automatically remove a file if it detected as invalid or corrupt (be careful with this setting, the file will be permanently removed!)"

      checkbutton restorePlaying              "Restore Play Location on Playlists"
       "Check this to be prompted to resume playing at the track when the playlist was last saved"

      scale2 fadeFactor                    {"Fade Factor" .1 1.0 .1}
       "Adjust this to change how rapidly the fade occurs if 'Fade on Pause' or 'Fade on Stop' is enabled, smaller is faster. Factor should be < 1."
      scale2 normalPlayLatency             {"Buffer Time (ms)" 500 5000 200}
       "Adjusts the snack audio buffer time, increase if you find audio choppy. 2000 ms is typical."
      scale2 meterUpdateRate              {"Meter Update Rate (ms)" 10 500 10}
       "Set the time the audio meter is updated."


   }
   #   checkbutton quietEnd                "Auto Fade out"
   #    "Check this to reduce the sound level at the end of the track, good for the occasional truncated file or abrupt ending. See Fade Time and Fade Factors setting above."
   #   entry2 fadeTime          "Fade Time from End"
   #    "Enter a time in seconds to start fading the song volume if or 'Fade on Stop' is enabled."


   stuffSettingData $optionData $W
   ;##
   ;# Play Mode Settings (note the order)
   ;#
   set hull [labeledFrame $W.pm -text "Play Mode" -bd 2 -relief groove -font menu]
   set i 0
   foreach {opt text} {
      "Next Track Plays"     "Play the next track in the playlist (normal play)"
      "Play Track Once"      "Play current track once and then stop"
      "Repeat Play of Track" "Repeat play of current track"
   } {
      pack [radiobutton $hull.$i -text $opt -value $i -font menu -variable ::snackAmpSettings(SongPlayMode)] \
         -side left
      balloon::define $hull.$i $text
      incr i
   } ;# end foreach
   grid $W.pm -columnspan 4 -sticky ew

   ;##
   ;# Sort Mode Settings (note the order)
   ;#
   set hull [labeledFrame $W.sm -text "Sort Mode" -bd 2 -relief groove -font menu]
   set i 0
   foreach {opt text} {
     "Sort All"                        "Sort playlist by all track names"
     "Shuffle All"                     "Shuffle all tracks in playlist"
     "Shuffle Folders, Sort Tracks"    "Shuffle by folder names but keep tracks sorted within the folder"
     "Shuffle Folders, Shuffle Tracks" "Shuffle by folder names and keep tracks shuffled within the folder"
   } {
      pack [radiobutton $hull.$i -text $opt -value $i -font menu -variable ::snackAmpSettings(sortMode)] \
         -side left
      balloon::define $hull.$i $text
      incr i
   } ;# end foreach
   grid $W.sm -columnspan 4 -sticky ew

}
###############################################################################
# Function   : settings::uiSettings
# Description: User Interface Settings
# Author     : Tom Wilkason
###############################################################################
proc settings::uiSettings {W} {
   set optionData {
      checkbutton ID3ballon               "Balloon Help Track Details"
       "Check this to have detailed ID3 and file information pop-up when hovering the mouse cursor over a track in a playlist."
      checkbutton showTicks               "Show Time Ticks"
       "Check this to display tick marks on the time adjust slider (requires a restart)."

      checkbutton noTail              "Hide File Extension in Playlist"
       "Check this to hide (or uncheck to show) the file extension in the main Playlist (requires a restart)"
      checkbutton restoreMiniPlayer          "Restore Mini-player Position"
       "Check if you want restore the Mini-player position and state upon startup."

      checkbutton levelMeter          "Display Sound Level Meter"
       "Check if you want to show a sound level meter on the front (requires restart)."

   }
   pack [subFrame $optionData "Visual Settings" $W.visual] -side top -anchor w -fill x

   set optionData {
      scale2 ListDepth                    {"Subfolders Depth" 0 5 1}
       "Adjusts the number of folder names that are prepended to the track names in the playlist. Useful if your files are named with only track title, or possibly track title and album name and you organize them by artist. Set to 0 if you only want the file name to show up"
      scale2 playListLines                 {"Playlist Window Size (lines)" 5 50 1}
       "Adjust this to the number of lines you want to initially show in the playlist window when SnackAmp starts."

      scale2 FontSizeFactor                {"Font Size Factor" .5 1.5 .05}
       "Adjust the overall font size scaling to your system. Recommend you leave at 1 and adjust the fonts individually for your system."
      scale2 balloonDelay                  {"Balloon Help Delay (ms)" 100 5000 100}
       "Adjust this to set the amount of delay before a pop-up balloon is shown."
   }
   pack [subFrame $optionData "Visual Adjustments" $W.adj] -side top -anchor w -fill x

   ;##
   ;# Display Info (note the order)
   ;#
   set hull [labeledFrame $W.info -text "Track Display Details" -bd 2 -relief groove -font menu]
   set i 1  ;# starts at 1
   foreach {opt text} {
     "Title Only"              "Show only the Track Title on the main display, RMB on title to show menus (compact form)"
     "Tit., Art. & Alb."       "Show only the Track Title, Artist and Album"
     "Tit., Art., Alb., Trk., Rate & Gen." "Show the Track Title, Artist, Album, Rate, Date and Track."
     "All Info"                "Show all of the information on the main display (full form)"
   } {
      pack [radiobutton $hull.$i -text $opt -value $i -font menu -variable ::snackAmpSettings(showDetail)] \
         -side left
      balloon::define $hull.$i $text
      incr i
   } ;# end foreach
   pack $W.info -side top -anchor w -fill x
}
###############################################################################
# Function   : settings::folderSettings
# Description: Settings for the Folders
# Author     : Tom Wilkason
###############################################################################
#      entrydir uskDir                     "Folder to hold default Streaming Shortcuts"
#       "Select a folder to place Streaming Shortcuts when playing streams from the web (e.g. shoutcast). You should place it within your current media folders to manage it from the Auto Playlist Manager"
proc settings::folderSettings {W} {
   set i 0

   set optionData {
      entryList mediaDirs                     "Media Folders"
       "Enter a list of music library folders separated by spaces to catalog. Each folder and its sub folders will be cataloged. Each base folder will contain a catalog file.\ne.g. d:/music \"//george/music library\""
      entryList IncomingDirs                  "Monitor Folders"
       "Enter a list of music library folders separated by spaces to monitor for changes (e.g. CD ripping location). New files will be automatically appended to the playlist. Sub-folders are not monitored."
      checkbutton MonitorCacheFolders     "Monitor Catalogs for Changes"
       "Check this to continuously monitor for changes to the catalog files (i.e. from other users) and read them in if they change."
      checkbutton MonitorScanFolders      "Monitor Folders for New Files"
       "Check this to continuously monitor for new files in the above monitor folders, good if you are downloading files there."
      checkbutton AutoRescan              "Catalog Media Folders on Startup"
       "Check if you want to automatically re-catalog your folders on startup. Not recommended for large media sets."
      checkbutton dontMoveFiles              "Inhibit File Modifications"
       "Check if you want to inhibit moving, removing or renaming files or folders."
   }
   pack [subFrame $optionData "Media Folders" $W.[incr i]] -side top -anchor w -fill x
   #---------
   set optionData {
      entryList RemovableFolders           "Removable Folders"
       "Enter a list of drive letters or volume names you want considered as removable media."
      checkbutton UseRemovableFolders     "Allow Removable Media"
       "Check if you want to include tracks contained on removable media in the playlist."
      checkbutton PromptRemovableFolders  "Prompt to Mount Removable Media"
       "Check if you want to be prompted to insert removable media when a track in the playlist requires it. If deselected then an automatic mount will be tried. If the track is still not found, then it is skipped. This is only valid if 'Allow Removable Media' is checked."
   }
   pack [subFrame $optionData "Removable Media" $W.[incr i]] -side top -anchor w -fill x
   #---------
   set optionData {
      entry fileExtensions                "Supported File Extensions"
       "Enter a list of file extenstions (with the . before) to support when cataloging files. Upper or lowercase is acceptable."
      entrybrowse  hotkeyFile             "Hotkey Assignment File"
       "Enter the name of file containing hotkey assignments (you must restart after changing)."

   }
   pack [subFrame $optionData "Other Settings" $W.[incr i]] -side top -anchor w -fill x
}
###############################################################################
# Function   : settings::id3Settings
# Description: Settings for the Folders
# Author     : Tom Wilkason
###############################################################################
proc settings::id3Settings {W} {
   set optionData {
      entry2 smartName                 "Format for Smart Naming"
      "Enter a naming scheme for converting tags to file names:\n%%A - Artist, %%C - Album, %%N - Track Number, %%T - Title"
      entry2 ID3editLimit                 "Group Edit Track Limit"
      "Enter the maximum number of tracks that be edited with the ID3 Tag editors."
      entry2 splitChar                 "Name Separator"
      "Enter character used when cleaning up or building a file name from tags or cleaning up file names, recommend using -"
      entry2 scrolledCanvasHeight      "Editor Window Initial Height"
      "Enter a number (in inches) followed by an i e.g. 4i to denote how high the tag editor window should be when editing a large number of files. Any additional space needed will be scrollable."
      checkbutton ID3OKCreate         "Group Editing will Create Tags"
       "Check this to allow the ID3 Group Editor to create tags when editing multiple files, with this off files without tags are ignored."
      checkbutton preserveTime         "Preserve Time Stamp on Edit"
       "Check this to not change the file time stamp when changing the ID3 tag. Recommended you left this un-checked."
      checkbutton ID3ReadOnly          "Disallow ID3 Tag Modification"
       "Check this inhibit modification of ID3 Tag information."
      checkbutton ID3V12               "Allow Writing ID3 V1.2 Tags"
       "Check this to allow V1.2 Tags to be written (Compatible extension to V1 Tags)."
   }
   stuffSettingData $optionData $W
}
###############################################################################
# Function   : settings::fontSettings
# Description: Settings page for Fonts
# Author     : Tom Wilkason
###############################################################################
proc settings::fontSettings {W} {
   variable FONTS
   global snackAmpSettings
   set i 0
   set span 5
   foreach {Name Specs Description} $FONTS {
      label $W.lb$Name -text "$Description" -relief flat -font menu -anchor e
      set size [lindex [$W.lb$Name cget -font] 1]
      balloon::define $W.lb$Name "Select the font attributes for $Description to the right. Apply or OK will save them."
      foreach {family size weight}  $Specs {break}
      ;# Cascaded Font List
      menubutton $W.mnu$Name -width 12 \
         -textvariable ::snackAmpSettings($Name,family) -relief raised -indicatoron 1 \
         -font $Name -menu $W.mnu$Name.menu -direction flush -highlightthickness 2
      balloon::define $W.mnu$Name "Click to select which font to use from a menu"
      menu $W.mnu$Name.menu \
         -postcommand [list settings::postMenuNames $W.mnu$Name.menu $Name] \
         -tearoff 0
      ;# Font sizes, A  callback (trace) is fired when these change
      ;# so the changes are made visible.
      tk_optionMenu $W.siz$Name   ::fontMetrics($Name,size) \
         6 7 8 9 10 11 12 13 14 15 16 17 18 19 20 22 24 26 28 30 32 34 36
      tk_optionMenu $W.wgt$Name   ::fontMetrics($Name,weight) normal bold
      tk_optionMenu $W.slant$Name ::fontMetrics($Name,slant)  roman  italic
      grid $W.lb$Name $W.mnu$Name $W.siz$Name $W.wgt$Name $W.slant$Name -sticky nsew
   } ;# end foreach
   return
}
###############################################################################
# Function   : settings::colorSettings
# Description: Settings page for Colors
# Author     : Tom Wilkason
###############################################################################
proc settings::colorSettings {W} {
   set changeColors {
      "Text Background"          bg
      "Text Foreground"          fg
      "Selected Text Background" highlight
      "Table Stripe Background"  tsbg
      "Entry Background"         ebg
      "Entry Foreground"         efg
      "Title Text Foreground"    tfg
      "Info Label Foreground"    lfg
      "Time Counter Foreground"  sfg
   }
   foreach {Text Index} $changeColors {
      button $W.$Index -text $Text -font menu -width 25 \
         -relief raised -borderwidth 1 -command "settings::colorPicker $Index"
      balloon::define $W.$Index "Push this to set the $Text Color"
      label $W.lbl$Index -width 30 -bd 2 -relief raised -bg $::snackAmpSettings($Index)
      regColor $W.lbl$Index -background ::snackAmpSettings($Index)
      grid $W.$Index $W.lbl$Index -padx 10 -pady 2
   }
}
###############################################################################
# Function   : settings::webServer
# Description: Settings page Misc functions
# Author     : Tom Wilkason
###############################################################################
proc settings::webServer {W} {
   set i 0
   set optionData {
      entry2 port                     "Initial Http Server Port Number"
       "Enter a port number to use for the integrated web server. This should be a free port on your system."
      label2 usedPort                  "Used Http Server Port Number"
       "Actual port used by internal web server, will be next free port at or after above port."
      entry2 httpMaxused             "Max Number of Requests"
       "Specify the maximum number of simultaneous requests per client (~200 works well)."
      label2 HttpHits                 "Number of HTTP connects this session"
       "Enter the maximum number of simultaneous clients."
      entry2 httpNextLength           "Number Of Queued Songs"
       "Specify the number of future playlist songs to show in the remote control web page."
      entry2 refreshRate              "Browser Refresh Rate"
       "Enter time in seconds the play control browser should refresh itself."
      checkbutton EnableWebServer "Enable Internal Web Server"
       "Check this to allow a web browser to remote control this player or show internal help (requires restart)."
      checkbutton EnableRemoteControl "Enable Remote Control via Browser"
       "Check this to allow a web browser to remote control this player or show internal help (requires restart)."
   }
   pack [subFrame $optionData "Server Settings" $W.[incr i]] -side top -anchor w -fill x
   set optionData {
      entry subNetMask               "Subnet Mask for Local Addresses"
       "Enter your Subnet Mask to validate local vs. remote clients, used for checking local only access."
      label ipAddress               "Your detected IP Address"
       "IP Address of this node (if 0 no ports were available to check)."
      checkbutton localOnly           "Local Web Access Only"
       "Check this to only allow access from clients on your local network (those within your subnet). Requires Internal Web Server to be enabled."
   }
   pack [subFrame $optionData "Local Access Settings" $W.[incr i]] -side top -anchor w -fill x
#      entry HttpDefault               "Browser Control File Name"
#       "File name to use for main play control page. Do not specify a folder, it must be located in a 'docs' folder under the location of SnackAmp."
   set optionData {
      entry HttpLocal                 "Local Folder to map /local to"
       "If desired, you can enter a folder here on your lan that the url /local will map to. This allows you to extend SnackAmp with your own HTML pages."
      label internalWebBase           "URL of Internal Web Server"
       "URL of the Internal Web Server. Enter this value in your browser for remote control."
      label webRoot                   "Root Folder of Internal Web Server"
       "Root Folder of the Internal Web Server."
   }
   if {$::tcl_platform(platform)=="unix"} {
      lappend optionData \
      entrybrowse HelpBrowser        "Help Browser" \
       "Program to use for browsing help pages. Leave empty to use system default"
   } ;# end if
   pack [subFrame $optionData "URL Locations" $W.[incr i]] -side top -anchor w -fill x
#   set furtureOptionData {
#      entry allmusicBase             "Internet Query Location"
#       "Web URL to use for querying web data"
#   }
#      entry helpFile                 "Help File Name"
#       "File name to use for help."
}
###############################################################################
# Function   : settings::AdvancedSettings
# Description: Settings page Advanced functions
# Author     : Tom Wilkason
###############################################################################
proc settings::AutoLevelSettings {W} {
   set ::snackAmpSettings(xxx) ""

   set optionData {
      button {soundControl::calibrateAudio}        "Compute Auto Levels"
       "Iterate over each track and compute the level setting if needed."
      button {settings::resetGainData}        "Reset Auto Levels"
       "Press this to reset all auto level values."
      button {gainControl::cleanStaleEqualizations}     "Remove Stale Auto Levels"
       "Press this to remove Auto Level values for tracks that no longer exist."

   }
   pack [subFrame $optionData "Auto Level Functions" $W.f1] -side top -anchor w -fill x

   set optionData {
      scale2 maxGain                 {"Max Gain (in dB)" 0 6 .1}
       "Adjust to limit the amount of gain applied during playback to quiet songs. 3 dB is typical."
      scale2 minGain                 {"Max Atten (in dB)" -15 0 .1}
       "Adjust to limit the amount of attenuation applied to loud songs during playback. -10 dB is typical."

      scale2 autoLevelReference      {"Reference Offset (dB)" -6 6 1}
       "Adjust the internal reference level used during computation. If number is too large, clipping may occur. If unsure leave at 0. If you change this setting, you should Auto Level your entire collection."
      scale2 autoLevelSeconds        {"Sample Period (sec)" 5 50 1}
       "Adjusts period between samples for the Auto Level setting, smaller is slower but more accurate. 25 is typical, try 10 if you have a fast CPU."
      scale2 autoLevelMaxSamples     {"Max Samples" 5 100 5}
       "Enter a number to limit the maximum number of samples taken, larger numbers are more accurate but slower to compute. For example if Sample Period is 10 sec, and the track is 10 minutes long, 60 samples or this setting, whichever is less, will be taken."

      scale2 autoLevelWindow         {"Window Size" 512 18000 1024}
       "Adjusts the Window Size of the FFT for computing the Auto Level setting, larger is slower but more accurate. 4096 is typical, try 8192 or greater if you have a fast CPU."
      scale2 gainFactor              {"Level Control Factor" 15 50 1}
       "Adjusts the amount of level control applied during computation, 20-40 is typical, larger numbers produces a smaller effect (i.e. load tracks stay somewhat loud, quiet tracks stay somewhat quiet). 20 is Theory but may not work as well as 33."
   }
   pack [subFrame $optionData "Auto Level Settings" $W.f2] -side top -anchor w -fill x
}
;##
;# Diagnostic Settings
;#
proc settings::AdvancedSettings {W} {
   ;# Auto Level
   ;# Diagnostics
   set optionData {
      checkbutton Trace               "Enable Diagnostic Outputs"
       "Check this to print out diagnostic output, not normally needed."
      checkbutton logFile             "Log Activity to [file tail $TA(logFile)]"
       "Check this to direct all console output to a file"

      checkbutton autoSavePlaylist    "Save Playlist on Track Change"
       "If checked, the play list will be saved after every track, only useful if you are experiencing problems. The playlist is always saved on exit. Set the Restore Prior Playlist option to restore the playlist at startup"
      checkbutton enableConsole       "Enable use of Console"
       "If not checked, the console option will not be shown in the settings menu (requires restart)"

      checkbutton Version2Settings     "Use Clear Text Settings File"
       "Check this to save your settings in a user editable file format"
      checkbutton fastscan        "Fast Catalog Scan"
       "Press this to scan for files 2X faster, but won't see folders ending in same file extensions as sound files."

      checkbutton webLog              "Log Web Access"
       "Check this to log all web access to YYMMDD.log files in your home folder (requires restart)."
   }
   pack [subFrame $optionData "Diagnostic Settings" $W.tr] -side top -anchor w -fill x

}
###############################################################################
# Function   : settings::APLSettings
# Description: Settings page Misc functions
# Author     : Tom Wilkason
###############################################################################
proc settings::APLSettings {W} {
   ;##
   ;# Other options
   ;#
   set e 0
   set gr [list]
   set optionData {
     checkbutton updateAPL        "Sync on Selection"  "Sync Auto Playlist Tree on track selection"
     checkbutton syncAPL          "Sync on Change"     "Sync Auto Playlist Tree on track change"
     checkbutton treeLine         "Node Lines"         "Show Node-Lines on Auto Playlist Tree View (requires restart)"
     checkbutton useTableStripes  "Row Stripes"        "Add Stripe to Alternate Table Rows"
     checkbutton useTableBars     "Column Bars"        "Add Vertical Bar Between Columns"
     checkbutton autoTableStretch "Auto Expand"        "Allow Table Columns to Expand to Fill Window"
     checkbutton useAPLColors     "Use Colors"         "Use SnackAmp Colors for Text and Background (needs restart)"
     checkbutton dndExpand        "Expand on Drop"     "Expand Target Folder after Moving A Folder to it"
     checkbutton savePlonApply    "Auto-Save Playlist" "Save current Playlist before changing to a new Auto-Playlist"
   }
   pack [subFrame $optionData "General Settings" $W.apl] -side top -anchor w -fill x
   foreach {xxxx Setting Text Help} $optionData {
      trace variable ::snackAmpSettings($Setting) w settings::setTableOption
   }
   ;# Columns to Show
   set hull [labeledFrame $W.pm -text "Columns To Show in Auto Playlist Table" -bd 2 -relief groove -font menu]
   foreach {item} [lrange $id3Tag::TableCols 1 end] {
      lappend gr [checkbutton $hull.[incr e] \
         -pady 0 -highlightthickness 0 \
         -variable ::snackAmpSettings(column,$item) \
         -anchor w -text $item -font menu]
      balloon::define $hull.$e "Check to show $item in playlist table"
      if {[llength $gr]>=6} {
         eval grid $gr -sticky w
         set gr [list]
      }
   } ;# end foreach
   if {[llength $gr]>=1} {
      eval grid $gr -sticky w
   }
   pack $W.pm -side top -anchor w -fill x
}
##
# Streaming Media options
#
proc settings::streamingSettings {W} {
   set optionData {
      entrydir uskDir                     "Streaming shortcut folder"
       "Select a folder to place Streaming Shortcuts when playing streams from the web (e.g. shoutcast). You should place it within your current media folders to manage it from the Auto Playlist Manager"
   }
   pack [subFrame $optionData "Play Settings" $W.play] -side top -anchor w -fill x

   set optionData {
      entry proxyhost                 "Host Name of Proxy"
       "If you need a proxy to connect to the internet, enter the host name here. Don't forget to enable it with checkbox below."
      entry proxyport                 "Port Number for Proxy"
       "Enter port number of proxy. Don't forget to enable it with checkbox below."
      entry localurl                 "Exclude patterns for local hosts"
       "Enter matching patterns to indicate local (non--proxy) machines separated by semi-colons. E.g. *.mydomain.com;*george*"
      checkbutton useproxy            "Enable Use of Proxy"
       "Check this if your internet connection requires a proxy."
   }
   pack [subFrame $optionData "Proxy Settings" $W.pxy] -side top -anchor w -fill x

   set optionData {
      checkbutton EnableStreaming "Enable Internal Streaming Server"
       "Check this to allow a web browser to remote control this player or show internal help (requires restart)."
   }
   pack [subFrame $optionData "Streaming Server" $W.adj] -side top -anchor w -fill x
}
##
# Helper for subframes
#
proc settings::subFrame {optionData Title W} {
   set hull [labeledFrame $W -text $Title -bd 2 -relief groove -font menu]
   stuffSettingData $optionData $hull
   return $W
}
;##
;# Web helper to return settings state
;#
proc settings::webStatus {index} {
   global snackAmpSettings
   if {[info exists snackAmpSettings($index)]} {
      if {$snackAmpSettings($index)} {
         return "Enabled"
      } else {
         return "Disabled"
      }
   } else {
      return "Unknown"
   }
}
###############################################################################
# Function   : settings::resetGainData
# Description: Reset all gain data to default values
# Author     : Tom Wilkason
# Date       : 7/7/2002
###############################################################################
proc settings::resetGainData {args} {
   gainControl::resetAll
}
###############################################################################
# Function   : settings::setTableOption
# Description: Callback with table options get modified
# Author     : Tom Wilkason
###############################################################################
proc settings::setTableOption {_glob option args} {
   upvar $_glob glob
   set Var $glob($option)
   if {[APLExists]} {
      switch -- $option {
         useTableStripes {
            if {$Var} {
               $::PLB config -stripebackground $::snackAmpSettings(tsbg)
            } else {
               $::PLB config -stripebackground ""
            }
         }
         useTableBars {
            if {$Var} {
               $::PLB config -showseparators yes
            } else {
               $::PLB config -showseparators no
            }
         }
         autoTableStretch {
            if {$Var} {
               $::PLB config -stretch all
            } else {
               $::PLB config -stretch [list]
            }
         }
      } ;# End Switch
   }
}
###############################################################################
# Function   : settings::colorSettings
# Description: Settings page for File Associations
# Author     : Tom Wilkason
###############################################################################
proc settings::associationSettings {W} {
   variable reg
   variable FONTS
   global snackAmpSettings
   package require fileAssociation
   set exts $snackAmpSettings(fileExtensions)
   ;# Support playlists
   lappend exts pls m3u
   foreach type [lsort $exts] {
      fileAssociation::isRegisteredMF .$type settings::reg($type)
      radiobutton $W.cb$type -text "Associate $type with SnackAmp" \
         -command "[list fileAssociation::registerMF .$type];[list fileAssociation::isRegisteredMF .$type settings::reg($type)]" \
         -variable settings::reg($type) -value 1 -font menu
      radiobutton $W.res$type -text "Restore $type to original" \
         -command "[list fileAssociation::unRegisterMF .$type];[list fileAssociation::isRegisteredMF .$type settings::reg($type)]" \
         -variable settings::reg($type) -value 0 -font menu
      grid $W.cb$type $W.res$type -sticky w
   } ;# end foreach

}
###############################################################################
# Function   : settings::postMenuNames
# Description: On pop-up, post the fonts, if already built then post menu
#              Menu is shared for all buttons to save memory
# Author     : Tom Wilkason
###############################################################################
proc settings::postMenuNames {Menu Name} {
   ;# If menu has data then use it,
   ;# otherwise build it up with the
   ;# font information.
   #puts "posted $Menu"
   if {[$Menu index end] == 0 || [$Menu index end]=="none"} {
      set families [lsort -dictionary [font families]]
      foreach {family} $families {
         set first [string index $family 0]
         lappend fontNames($first) $family
      } ;# end foreach
      $Menu delete 0 end
      foreach {Value} [lsort [array names fontNames]] {
         $Menu add cascade -label $Value -menu $Menu.f$Value
         menu $Menu.f$Value -tearoff 0
         foreach {item} $fontNames($Value) {
           $Menu.f$Value add command -label "$item" \
            -command [list settings::setFontFamily $item $Name] -font [list $item]
         }
      }
   } ;# end if
}
###############################################################################
# Function   : settings::installFontCallbacks
# Description: Install callbacks to immediately change the visible fonts when
#              the user changes them
# Author     : Tom Wilkason
# Date       : 2/6/2002
###############################################################################
proc settings::installFontCallbacks {args} {
   ;##
   ;# Callbacks to immediatly set the fonts when the user changes them
   ;#
   trace vdelete ::fontMetrics w settings::fontCallback
   array set ::fontMetrics [array get ::snackAmpSettings *,size]
   array set ::fontMetrics [array get ::snackAmpSettings *,weight]
   array set ::fontMetrics [array get ::snackAmpSettings *,slant]
   trace variable ::fontMetrics w settings::fontCallback
}
###############################################################################
# Function   : settings::setFontFamily
# Description: On callback, set the appropriate widget type to a font value passed in
# Author     : Tom Wilkason
###############################################################################
proc settings::setFontFamily {family Name} {
   set ::snackAmpSettings($Name,family) $family
   ;# This will make the effect immediate
   settings::saFontSize $::snackAmpSettings(FontSizeFactor)
}
###############################################################################
# Function   : settings::fontCallback
# Description: When radio buttons change, change the fonts immediatly
# Author     : Tom Wilkason
###############################################################################
proc settings::fontCallback {_glob index mode} {
   upvar $_glob Settings
   set ::snackAmpSettings($index) $Settings($index)
   settings::saFontSize $::snackAmpSettings(FontSizeFactor)
}
###############################################################################
# Function   : settings::colorPicker
# Description: Bring up color picker to change colors
# Author     : Tom Wilkason
# Date       : 10/30/2001
###############################################################################
proc settings::colorPicker {index} {
   global snackAmpSettings
   set result [tk_chooseColor -initialcolor $snackAmpSettings($index)]
   if {[string length $result]} {
      set snackAmpSettings($index) $result
      settings::applyColor
   } ;# end if
}
###############################################################################
# Function   : settings::fileChooser
# Description: Bring up file selector to select a file. Assign it to the var
# Author     : Tom Wilkason
# Date       : 10/30/2001
###############################################################################
proc settings::fileChooser {varname {parentWin .}} {
   upvar $varname filename
   set file [chooseFile [file dirname $filename] $parentWin]
   if {[string length $file] > 0} {
      set filename $file
   }
}
proc settings::folderChooser {varname {parentWin .}} {
   upvar $varname filename
   set file [chooseDirectory [file dirname $filename] $parentWin]
   if {[string length $file] > 0} {
      set filename $file
   }
}
###############################################################################
# Function   : settings::regColor
# Description: register a color for a widget
# Author     : Tom Wilkason
# Date       : 10/30/2001
###############################################################################
proc settings::regColor {widget args} {
   variable Colors
   foreach {option color} $args {
      set Colors($widget,$option) $color
   } ;# end foreach
}
###############################################################################
# Function   : settings::applyColor
# Description: Go through each option and apply the latest values if they
#              are different.
# Author     : Tom Wilkason
# Date       : 10/30/2001
###############################################################################
proc settings::applyColor {args} {
   variable Colors
   foreach {index} [array names Colors] {
      foreach {widget option} [split $index ,] {break}
      catch {
         set val [$widget cget $option]
         set newVal [set $Colors($index)]
         if {![string equal $val $newVal]} {
            $widget config $option $newVal
         } ;# end if
      }
   } ;# end foreach
}
###############################################################################
# Function   : settings::stuffSettingData
# Description: General purpose routine to grid a number of items
#              for use in settings GUIs.
# Author     : Tom Wilkason
# Date       : 5/15/2002
###############################################################################
proc settings::stuffSettingData {optionData B {i 100}} {
   #
   ####################################################################
   #
   global TA
   set e [expr {$i+100}]
   set ew [expr {$e+100}]
   set cbData [list]
   set enData [list]
   foreach {Widget LinkedVariable Label Help} $optionData {
      switch -- $Widget {
         label {
            grid [label $B.[incr i] -text [subst $Label] -anchor e -font menu] \
               [$Widget $B.[incr e] -textvariable snackAmpSettings($LinkedVariable) -width 35 \
               -relief sunken \
               -anchor w \
               -foreground $::snackAmpSettings(efg)] \
               - - \
               -sticky ew -padx 2
            balloon::define $B.$e $Help
            settings::regColor $B.$e -background ::snackAmpSettings(ebg) -foreground ::snackAmpSettings(efg)
         }
         label2 {
            if {[string is double -strict $::snackAmpSettings($LinkedVariable)]} {
               set anc e
            } else {
               set anc w
            }
            lappend enData [label $B.[incr i] -text [subst $Label] -anchor e -font menu]
            lappend enData [label $B.[incr e] \
               -textvariable snackAmpSettings($LinkedVariable) \
               -width 6 \
               -relief sunken \
               -anchor $anc \
               -foreground $::snackAmpSettings(efg)]
            balloon::define $B.$e $Help
            settings::regColor $B.$e -background ::snackAmpSettings(ebg) -foreground ::snackAmpSettings(efg)
         }
         scale2 {
            foreach {Label min max res} $Label {break}
            set interval [expr {($max-$min)/10.0}]
            # -tickinterval $interval
            lappend enData [scale $B.[incr e] -label [subst $Label] \
               -from $min -to $max  -orient horiz \
               -variable snackAmpSettings($LinkedVariable) -width 7 -bd 2  \
               -sliderlength 15 -resolution $res -font balloon -relief groove \
               -activebackground blue -troughcolor $::snackAmpSettings(bg)]
            balloon::define $B.$e $Help
            settings::regColor $B.$e -troughcolor ::snackAmpSettings(bg)
         }
         entry2 {
            if {[string is double -strict $::snackAmpSettings($LinkedVariable)]} {
               set anc right
            } else {
               set anc left
            }
            lappend enData [label $B.[incr i] -text [subst $Label] -anchor e -font menu]
            lappend enData [entry $B.[incr e] -textvariable snackAmpSettings($LinkedVariable) -width 6 \
               -background $::snackAmpSettings(ebg) \
               -foreground $::snackAmpSettings(efg) \
               -justify $anc]
            balloon::define $B.$e $Help
            settings::regColor $B.$e -background ::snackAmpSettings(ebg) -foreground ::snackAmpSettings(efg)
         }
         entry {
            set anc left
            grid [label $B.[incr i] -text [subst $Label] -anchor e -font menu] \
               [entry $B.[incr e] -textvariable snackAmpSettings($LinkedVariable) -width 35 \
               -background $::snackAmpSettings(ebg) -justify $anc \
               -foreground $::snackAmpSettings(efg)]  \
               - - \
               -sticky ew -padx 2
            balloon::define $B.$e $Help
            settings::regColor $B.$e -background ::snackAmpSettings(ebg) -foreground ::snackAmpSettings(efg)
         }
         entrybrowse {
            ;# Only want to set the browser for systems that use env(BROWSER), e.g. *nix
            ;# TFW: Check this logic
            grid [label $B.[incr i] -text [subst $Label] -anchor e -font menu] \
                 [entry $B.[incr e] -textvariable snackAmpSettings($LinkedVariable) -width 35 \
                   -background $::snackAmpSettings(ebg) -foreground $::snackAmpSettings(efg)] \
                 - -\
                 [button $B.[incr ew] -text "Browse..." -command [list settings::fileChooser snackAmpSettings($LinkedVariable) $B] \
                 -font menu -bd 1 -pady 0 -relief raised] -sticky ew -padx 2
            balloon::define $B.$e $Help
            settings::regColor $B.$e -background ::snackAmpSettings(ebg) -foreground ::snackAmpSettings(efg)
         }
         entrydir {
            grid [label $B.[incr i] -text [subst $Label] -anchor e -font menu] \
                 [entry $B.[incr e] -textvariable snackAmpSettings($LinkedVariable) -width 35 \
                   -background $::snackAmpSettings(ebg) -foreground $::snackAmpSettings(efg)] \
                 - -\
                 [button $B.[incr ew] -text "Browse..." -command [list settings::folderChooser snackAmpSettings($LinkedVariable) $B] \
                 -font menu -bd 1 -pady 0 -relief raised] -sticky ew -padx 2
            balloon::define $B.$e $Help
            settings::regColor $B.$e -background ::snackAmpSettings(ebg) -foreground ::snackAmpSettings(efg)
         }
         entryList {
            grid [label $B.[incr i] -text [subst $Label] -anchor e -width 10 -font menu] \
                 [entry $B.[incr e] -textvariable snackAmpSettings($LinkedVariable) \
                   -background $::snackAmpSettings(ebg) -foreground $::snackAmpSettings(efg)] \
                    - - \
                 [button $B.[incr ew] -text "Manage..." -command [list folderManager::launch $LinkedVariable] \
                 -font menu -bd 1 -pady 0 -relief raised] -sticky ew -padx 2
            balloon::define $B.$e $Help
            settings::regColor $B.$e -background ::snackAmpSettings(ebg) -foreground ::snackAmpSettings(efg)
         }
         checkbutton {
            lappend cbData [checkbutton $B.[incr e] -pady 0 -highlightthickness 0 -width 33 -variable snackAmpSettings($LinkedVariable) \
            -anchor w -text [subst $Label] -font menu]
            balloon::define $B.$e $Help
         }
         button {
            lappend enData [button $B.[incr e] -relief raised -bd 2 -anchor center -command $LinkedVariable \
            -anchor w -text [subst $Label] -font menu]
            balloon::define $B.$e $Help
         }
         default {}
      } ;# End Switch
      ;# Grid in checkbutton or button if we are two across
      if {[llength $cbData] == 2} {
         eval grid $cbData -columnspan 2 -pady 0 -sticky ew
         set cbData [list]
      } ;# end if
      ;# Grid in small entry boxes if we are four across
      if {[llength $enData] == 4} {
         eval grid $enData -sticky ew -padx 2 -pady 0
         set enData [list]
      } ;# end if
   } ;# end foreach
   ;# Grid remaining checkbuttons
   if {[llength $cbData] > 0} {
      eval grid $cbData  -columnspan 2 -pady 0 -sticky ew
      set cbData [list]
   } ;# end if
   ;# Grid remaining small entries
   if {[llength $enData] > 0} {
      eval grid $enData -sticky ew -padx 2 -pady 0
      set enData [list]
   } ;# end if
   ;# Allow entrys and other RH stuff to expand
   for {set i 0} {$i < 5} {incr i} {
      grid columnconfigure $B $i -weight 1
   } ;# end for
}
###############################################################################
# Function   : settings::saveSettings
# Description: Save the settings from the settings::settingsGui
# Author     : Tom Wilkason
# Date       : 12/30/2000
###############################################################################
proc settings::saveSettings {args} {
   global snackAmpSettings TA
   catalog::fixMediaPath
   settings::buildFileExtenstions
   if {[catch {
      set fid [open $TA(settingsFile) w]
   } result] \
   } then {
      tk_messageBox -icon warning -type ok -title "Open Error" -message $result
   } else {
      if {[info exists APL::folderList]} {
         set ::snackAmpSettings(folderList) [lsort -unique -dictionary [lrange $APL::folderList end-20 end]]
      }
      fconfigure $fid -buffersize 10000
      if {$snackAmpSettings(Version2Settings)} {
         set result "#######\n"
         append result "# SnackAmp Settings File Saved [clock format [clock seconds]]\n"
         append result "#\n"
         append result "array set ::snackAmpSettings \{\n"
         foreach {index} [lsort -dictionary [array names ::snackAmpSettings *]] {
            set value [string trim $::snackAmpSettings($index)]
            ;# Format the line real pretty
            set line [format "   %-25s %s\n" [list $index] [list $value]]
            append result $line
         } ;# end foreach
         append result "\}\n"
      } else {
         set result [array get ::snackAmpSettings]
      } ;# end if
      puts $fid $result
      close $fid
   }
}
;##
;# Set the default fonts
;#
proc settings::setDefaultFont {args} {
   variable FONTS
   global snackAmpSettings
   foreach {Font Specs Description} $FONTS {
      foreach {family size weight slant}  $Specs {break}
      foreach item {family size weight slant} {
         set ::snackAmpSettings($Font,$item) [set $item]
      }
   } ;# end foreach
}
;##
;# Scale all of the fonts by some factor
;#
proc settings::saFontSize {factor} {
   variable FONTS
   global snackAmpSettings
   set families [lsort -dictionary [font families]]
   lappend families Helvetica Times
   foreach {Name Specs Description} $FONTS {
      if {[lsearch -exact $families $::snackAmpSettings($Name,family)] < 0} {
         tk_messageBox -icon warning -type ok -title "Invalid Font" \
            -message "The font family '$Description' you have selected, '$::snackAmpSettings($Name,family)' is invalid "
         return
      } ;# end if
      set size  $::snackAmpSettings($Name,size)
      font configure $Name -size [expr {round($size*$factor)}] \
         -family $::snackAmpSettings($Name,family) \
         -slant $::snackAmpSettings($Name,slant) \
         -weight $::snackAmpSettings($Name,weight)
   } ;# end foreach
   return
}
###############################################################################
# Function   : settings::readSettings
# Description: Read the default settings from the user home dir, first assign
#              defaults then read the user file to change them.
# Author     : Tom Wilkason
###############################################################################
proc settings::readSettings {args} {
   global snackAmpSettings TA
   ;# Defaults below
   settings::setDefaultFont
   set snackAmpSettings(version)               2.0
   set snackAmpSettings(gainFactor)            33.  ;#
   set snackAmpSettings(allmusicBase)           {http://allmusic.com/cg/amg.dll?p=amg&sql=} ;#1Eva|Cassidy
   set snackAmpSettings(autoLevelSeconds)       25
   set snackAmpSettings(autoLevelMaxSamples)   50
   set snackAmpSettings(autoLevelReference)     0
   set snackAmpSettings(autoLevelWindow)       4096
   set snackAmpSettings(AutoDelete)             0
   set snackAmpSettings(autoLevel)              1
   set snackAmpSettings(AutoRescan)             0
   set snackAmpSettings(autoSavePlaylist)       0
   set snackAmpSettings(enableConsole)          1
   set snackAmpSettings(balloonDelay)           1000
   set snackAmpSettings(ContinuousPlay)         1
   set snackAmpSettings(EnableWebServer)        0
   set snackAmpSettings(EnableRemoteControl)    0
   set snackAmpSettings(EnableStreaming)        0
   set snackAmpSettings(webLog)                 0
   set snackAmpSettings(fileExtensions)         [list wav ogg mp3 au aif usk]
   set snackAmpSettings(uskDir)                 ""
   set snackAmpSettings(FileListSize)           0.75
   set snackAmpSettings(FontSizeFactor)         1
   set snackAmpSettings(HelpBrowser)            {}
   set snackAmpSettings(History)                [list]
   set snackAmpSettings(hotkeyFile)             $::HOME/snackAmpHotKeys.tcl
#   set snackAmpSettings(HttpDefault)            remote/snackamp.html
   set snackAmpSettings(ID3ballon)              1
   set snackAmpSettings(ID3ReadOnly)            0
   set snackAmpSettings(ID3V12)                 1
   set snackAmpSettings(ID3OKCreate)            0
   set snackAmpSettings(ID3editLimit)           200
   set snackAmpSettings(dontMoveFiles)          0
   set snackAmpSettings(IncomingDirs)           {}
   set snackAmpSettings(fadeTime)               2
   set snackAmpSettings(fadeFast)               0
   set snackAmpSettings(fadePause)              0
   set snackAmpSettings(fadeFactor)             0.6
   set snackAmpSettings(lastFolderFile)         $TA(defaultFolderFile)
   set snackAmpSettings(lineSelect)             [lindex [snack::mixer lines]  0]
   set snackAmpSettings(ListDepth)              0                           ;# Will show all folder
   set snackAmpSettings(noTail)                 0                           ;# Will show all folder
   set snackAmpSettings(logFile)                0
   set snackAmpSettings(minWidth)               450                         ;# Initial width of the track info area
   set snackAmpSettings(MonitorCacheFolders)    1
   set snackAmpSettings(MonitorScanFolders)     0
   set snackAmpSettings(NextAfterCurrnt)        1
   set snackAmpSettings(playListLines)          15
   set snackAmpSettings(port)                   8127
   set snackAmpSettings(preserveTime)           0
   set snackAmpSettings(PromptRemovableFolders) 0
   set snackAmpSettings(quietEnd)               0
   set snackAmpSettings(sortMode)               0
   set snackAmpSettings(refreshRate)            20
   set snackAmpSettings(RemovableFolders)       {}
   set snackAmpSettings(restorePlayList)        1
   set snackAmpSettings(resumePlay)             1
   set snackAmpSettings(scrolledCanvasHeight)   4i
   set snackAmpSettings(hideFileList)           0
   set snackAmpSettings(showTicks)              1
   set snackAmpSettings(smartName)              "%A - %C - %N - %T"
   set snackAmpSettings(SongPlayMode)           0
   set snackAmpSettings(splitChar)              "-"
   set snackAmpSettings(Trace)                  0
   set snackAmpSettings(treeSize)               "640x480"
   set snackAmpSettings(updateAPL)              1
   set snackAmpSettings(syncAPL)                0
   set snackAmpSettings(savePlonApply)                0
   set snackAmpSettings(restorePlaying)                0
   set snackAmpSettings(UseRemovableFolders)    0
   set snackAmpSettings(httpNextLength) 50
   set snackAmpSettings(ipAddress)  $iptools::myIP
   set snackAmpSettings(subNetMask) $iptools::defMask
   set snackAmpSettings(localOnly) 0
   set snackAmpSettings(httpMaxused) 200
   set snackAmpSettings(maxGain) 3
   set snackAmpSettings(minGain) -10
   set snackAmpSettings(priority) normal
   set snackAmpSettings(Version2Settings) 0
   set snackAmpSettings(fastscan) 1
   set snackAmpSettings(sortMode) 0
   set snackAmpSettings(restoreMiniPlayer) 0
   set snackAmpSettings(levelMeter) 0
   set snackAmpSettings(positionMiniPlayer) [list 0 0]
   set snackAmpSettings(activeMiniPlayer) 0
   set snackAmpSettings(showDetail) 4
   set snackAmpSettings(treeDepth) 0
   set snackAmpSettings(HttpLocal) ""
   set snackAmpSettings(useproxy) 0
   set snackAmpSettings(proxyport) ""
   set snackAmpSettings(localurl) ""
   set snackAmpSettings(proxyhost) ""
   set snackAmpSettings(treeLine) 0
   set snackAmpSettings(usedPort) "disabled"
   set snackAmpSettings(HttpHits) 0

   if {[info exists ::env(PLAYLATENCY)]} {
      set snackAmpSettings(normalPlayLatency) $::env(PLAYLATENCY)
   } else {
      set snackAmpSettings(normalPlayLatency) 2000.0
   }
   set snackAmpSettings(meterUpdateRate) 150
   set snackAmpSettings(longPlayLatency) 5000.0

   ;# Tablelist columns to use initially
   set snackAmpSettings(table,sortcol) 1
   set snackAmpSettings(table,sortorder) "increasing"
   set snackAmpSettings(column,Key) 1     ;# never shown but needed
   set snackAmpSettings(column,File) 1
   set snackAmpSettings(column,Title) 1
   set snackAmpSettings(column,Rate) 1
   set snackAmpSettings(column,Duration) 1
   set snackAmpSettings(column,Album) 1
   set snackAmpSettings(useTableStripes) 0
   set snackAmpSettings(useTableBars) 1
   set snackAmpSettings(autoTableStretch) 1
   set snackAmpSettings(dndExpand)    1
   set snackAmpSettings(useAPLColors) 1
   set snackAmpSettings(outputDevice) [lindex [snack::audio outputDevices] 0]
   if {[catch {
      # Try User settings file
      settings::readSettingsFile $::TA(settingsFile)
      settings::saFontSize $snackAmpSettings(FontSizeFactor)
   } result] \
   } then {
set msg {
Please select a folder containing all of your media folders. You
may add additional folders later from the settings menu.

Note: You can change your home folder from $::HOME by setting the
SNACKAMPHOME environment variable or by specifying it on the command
line with a "-home folder" option.

If you no longer use SnackAmp you can safely remove the $::HOME folder
and its contents.

Do you want to proceed?
}
      settings::saFontSize $snackAmpSettings(FontSizeFactor)
      package require saSettings
      if {[catch {
         ;# Try a common file in same folder as script/executable
         settings::readSettingsFile $::TA(defaultSettingsFile)
      } result] \
      } then {
         # no settings file
         waitWindow::close
         set result [tk_messageBox -type yesno -title "SnackAmp - Please Respond" \
            -icon info -message [subst $msg]]
         if {$result == "no"} {
            catch {file delete $::HOME}
            exit
         } ;# end if
         set snackAmpSettings(mediaDirs) "\"[chooseDirectory [file dirname $::HOME] . {Select a base Media Folder}]\""
            ;# Try default settings file
#         set result [tk_messageBox -title "Select Music Folders" \
#            -icon info -message \
#            "You may now optionally set other preferences or add additional media folders.\nPress OK on the next dialog when complete."]
#         settings::tabbedSettingsGui
      }
   }
   ;# Restart Settings
   set ::TA(noTail)       $snackAmpSettings(noTail)
   set ::TA(showTicks)    $snackAmpSettings(showTicks)
   settings::buildFileExtenstions
}
;##
;# Read the snackAmp settings file, handling both old
;# and new formats.
;#
proc settings::readSettingsFile {fileName} {
   global TA snackAmpSettings
   set fid [open $fileName r]
   set data [read $fid]
   close $fid
   ;# Handle new or old formats
   if {[string range $data 0 2] == "###"} {
      ;# New style
      if {[catch {eval $data} result]} {
         tk_messageBox -icon warning -type ok -title "Open Error" \
         -message "Error interpreting $fileName\n$::errorInfo"
         return -code error $result
      }
   } else {
      ;# Old style
      array set snackAmpSettings $data
   }
}
###############################################################################
# Function   : settings::buildFileExtenstions
# Description: Create the lists of file extensions needed by snackAmp
# Author     : Tom Wilkason
# Date       : 11/15/2001
###############################################################################
proc settings::buildFileExtenstions {args} {
   catch {unset ::taExtensions}
   catch {unset ::TA(upcaseExtensions)}
   foreach {extension} $::snackAmpSettings(fileExtensions) {
      # force an extension letter
      set ext ".[string trimleft $extension .]"
      set glb ".[string trimleft $extension .]"
      lappend ::TA(upcaseExtensions) [string toupper [string trimleft $extension .]]
      set  ::taExtensions([string tolower $ext]) 1
      set  ::taExtensions([string toupper $ext]) 1
      set  ::taExtensions([stringTitle $ext]) 1
   } ;# end foreach
   ;# Make sure usk is there
   if {![info exists ::taExtensions(.usk)]} {
      lappend ::snackAmpSettings(fileExtensions) usk
      settings::buildFileExtenstions
   }
}
###############################################################################
# Function   : bindings namespace
# Description: Handle key binding functions for displaying shortcuts
# Author     : Tom Wilkason
# Date       : 6/15/2002
###############################################################################
namespace eval bindings {
   ;##
   ;# save the shortcut key bindings for all the widgets
   ;#
   proc showBindings {args} {
      foreach window [list .saTop .wFC $::FI $::LB $::TREE [winfo parent $::TREE]] {
         puts "$window"
         foreach key [bind $window] {
            set command [string trim [bind $window $key]]
            if {[llength $command] < 3} {
               puts "  [cleanKeyName $key] = '$command'"
               set ::TA(binding,$window,$command) [cleanKeyName $key]
            }
         } ;# end foreach
      } ;# end foreach
   }
   ;##
   ;# Internal function to lookup a binding
   ;#
   proc Accelerator {W command} {
      variable Data
      set command [string trim $command]
      if {[info exists Data($command)]} {
         return $Data($command)
      } else {
         return ""
      }
   }
   ;##
   ;# Init the lookup table for some menu
   ;#
   proc init {win _Data} {
      variable Data
      variable W $win
      catch {unset Data}
      #puts "$W----------"
      foreach key [bind $W] {
         set boundCommand [string trim [bind $W $key]]
         if {[llength $boundCommand] < 5} {
            #puts "  [cleanKeyName $key] = '$boundCommand'"
            set Data($boundCommand) [cleanKeyName $key]
         }
      } ;# end foreach
      return ""
   }
   ;##
   ;# Add a command menu with the appropriate accelerator shortcut appended
   ;#
   proc addCommand {Menu Label Command} {
      variable W
      $Menu add command -label $Label \
         -accelerator [bindings::Accelerator $W [string trim $Command]] \
         -command $Command
   }
}
###############################################################################
# Function   : settings::assignHotKeys
# Description: Init routine to process the user hot keys file.
# Author     : Tom Wilkason
# Date       : 2/23/2001
###############################################################################
proc bindings::assignHotKeys {args} {
   global snackAmpSettings TA
   ;##
   ;# If this is a wrapped file, create the hotkey file if it doesn't already exist
   ;#
   if {$TA(isWrapped) && ![file exists $snackAmpSettings(hotkeyFile)]} {
      set vfsHotkeyFile [file join [file dirname $TA(snackRoot)] snackAmphotKeys.tcl]
      if {![file exists [file dirname $snackAmpSettings(hotkeyFile)]]} {
         set snackAmpSettings(hotkeyFile) [file join $::HOME snackAmphotKeys.tcl]
      } ;# end if
      if {[catch {
         set fin [open $vfsHotkeyFile r]
         set fout [open $snackAmpSettings(hotkeyFile) w]
         puts -nonewline $fout [read $fin]
         close $fin
         close $fout
         set result [tk_messageBox -title "Create Hotkey File" \
            -icon info -message "I Created a new hotkey file [file nativename $snackAmpSettings(hotkeyFile)]\n you can edit it in a text editor if so desired."]
      } result] } then {
         saLog $::errorInfo
      }
   } ;# end if
   if {([winfo exists .saTop]) && ([file exists $snackAmpSettings(hotkeyFile)])} {
      if {[catch {source $snackAmpSettings(hotkeyFile)} result]} {
         set result [tk_messageBox -title "Hotkey Error" \
            -icon info -message "$::errorInfo"]
      } else {
         saLog "Using hotkey file of $snackAmpSettings(hotkeyFile)"
      }
   } ;# end if
}
###############################################################################
# Function   : settings::defaultHotKeys
# Description: Default hot keys, user can override
# Author     : Tom Wilkason
# Date       : 2/23/2001
###############################################################################
proc bindings::defaultHotKeys {args} {
##########################################################
#  Playlist Assignments
##########################################################
   PlaylistAssign <BackSpace> {Cut 1}
   PlaylistAssign <Delete> {Cut 1}
   PlaylistAssign <Control-Key-Up> {
      Swap up
      break
   }
   PlaylistAssign <Control-Key-Down> {
      Swap down
      break
   }
   PlaylistAssign <Control-Key-a> {selectAll}
   PlaylistAssign <Return> {Play}
   PlaylistAssign <<Copy>> {clipboardCopy $::LB 1}
   PlaylistAssign <<Cut>> {clipboardCut $::LB}
   PlaylistAssign <<Paste>> {clipboardPaste $::LB}
   ;##
   ;# Default Assignments for Playlist
   ;# Note: Add break as as last command so
   ;# prevent rolling up to top level.
##########################################################
#  TOP Level Assignments
##########################################################
   Assign <F1> {::browser::help}
   Assign <Control-Key-f> {findNextSong 1}
   Assign <Control-Key-s> {sortSelection}
   Assign <Control-Key-g> {findNextSong 0}
   Assign <Alt-Key-Up>    {bumpLevel 1;break}
   Assign <Alt-Key-Down>  {bumpLevel -1;break}
   Assign <Key-Home>   locatePlaying
   Assign <F2>         {RenameFile [LB::getActiveFullName]}
   Assign <Control-F2> {RenameFile [soundControl::fileName]}
   # Keys for editing ID3
   Assign <F3>         {EditPLID3}
   Assign <Control-F3> {EditPlaying}
   Assign <F4>         {PLNameToTag 0}
   Assign <Alt-F4>     {taExit}   ;# If not added the F4 binding fires, windows unique?
   Assign <F5>         {SelectFile 0}
   Assign <Control-F5> {SelectFile 1}
   Assign <F6>         {PLTagToName}
##########################################################
#  File Box Assignments
##########################################################
#  FilterAssign <Double-1> macros::clearFilter
   FilterAssign <Escape> clearFilter
   FilterAssign <Return> {filterMusic replace}
   FilterAssign <Double-1> {
      clearFilter
      filterMusic replace
   }
   FilterAssign <Control-Return> {
      filterMusic replace
   }
   FilterAssign <Alt-Return> {
      filterMusic append
   }
}
###############################################################################
# Function   : folderManager
# Description: Pop-up box to manage the folders
# Author     : Tom Wilkason
# Date       : 12/17/2001
###############################################################################
namespace eval folderManager {
   variable fmLB
   variable Index
}
###############################################################################
# Function   : folderManager::launch
# Description: Launch the folder
# Author     : Tom Wilkason
###############################################################################
proc folderManager::launch {_Index} {
   variable fmLB
   variable Index
   set Index $_Index
   catch {destroy .fm}
   toplevel .fm -cursor $::TA(arrow)
   wm withdraw .fm
   grid [frame .fm.list] [frame .fm.button]
   wm title .fm "Folder Manager"
   set fmLB [Scrolled_Listbox .fm.list.lb -exportselection no -height 8 -width 40 \
      -foreground $::snackAmpSettings(fg) \
      -selectbackground $::snackAmpSettings(highlight) \
      -background $::snackAmpSettings(bg)]
   # Listbox popup menu
   settings::regColor $fmLB  \
      -foreground ::snackAmpSettings(fg) \
      -selectbackground ::snackAmpSettings(highlight) \
      -background ::snackAmpSettings(bg)
   set m $fmLB.popup
   menu $m -tearoff 0 -font menu
   $m add command -label "Remove Folder"  -command {}
   $m add command -label "Add Folder"  -command {}
   pack .fm.list.lb
   if {$Index == "mediaDirs"} {
      lappend buttons "Catalog Folder"    folderManager::Catalog  "Catalog selected folder"
   } ;# end if
   lappend buttons \
     "Add Folder"    folderManager::Add  "Add a new folder to the list on the left" \
     "Change Folder" folderManager::Edit "Modify the selected folder on the left" \
     "Remove Folder" folderManager::Remove  "Remove the selected folder from the list" \
     "OK"            folderManager::OK   "Push this to apply and save the current settings then close this dialog" \
     "Cancel"        {destroy .fm}       "Push this to ignore changes made and restore the previous settings"
   ;# Buttons
   set i 0
   foreach {Text Command Help} $buttons {
      button .fm.button.but[incr i] -text $Text -command $Command -relief raised -pady 0 -width 12
      balloon::define .fm.button.but$i $Help
      lappend toPack .fm.button.but$i
   } ;# end foreach
   ;# Insert current data into box
   eval pack $toPack
   if {[llength $::snackAmpSettings($Index)] > 0} {
      eval $fmLB insert end $::snackAmpSettings($Index)
   } ;# end if
   $fmLB selection set 0
   $fmLB activate 0

   catch {wm transient .fm .saTop}

   centerWindow .fm
   myIcon .fm
   lockSize .fm
   ;# Block on window closure
   tkwait window .fm
}
###############################################################################
# Function   : folderManager::OK
# Description: OK Button
# Author     : Tom Wilkason
###############################################################################
proc folderManager::OK {args} {
   variable fmLB
   variable Index
   ;# TFW Note: This causes blanks in elements to be surrounded by {}
   ;# Which previously they may have been surronded by blanks.
   set ::snackAmpSettings($Index) [$fmLB get 0 end]
   destroy .fm
}
###############################################################################
# Function   : folderManager::Add
# Description: Add a folder
# Author     : Tom Wilkason
###############################################################################
proc folderManager::Add {args} {
   variable fmLB
   variable Index
   set result [chooseDirectory . .]
   if {[file exists $result]} {
      $fmLB insert end $result
      $fmLB selection clear 0 end
      $fmLB selection set end
      $fmLB activate end
   } ;# end if
}
###############################################################################
# Function   : folderManager::Catalog
# Description: Catalog a folder
# Author     : Tom Wilkason
###############################################################################
proc folderManager::Catalog {args} {
   variable fmLB
   variable Index
   set result [$fmLB get [$fmLB index active]]
   if {$result != {}} {
      catalog::Rescan $result
   } ;# end if
}
###############################################################################
# Function   : folderManager::Edit
# Description: Change an existing folder
# Author     : Tom Wilkason
###############################################################################
proc folderManager::Edit {args} {
   variable fmLB
   variable Index
   set pos [$fmLB index active]
   set result [chooseDirectory [$fmLB get active] .]
   if {[file exists $result]} {
      $fmLB delete $pos
      $fmLB insert $pos $result
      $fmLB selection clear 0 end
      $fmLB selection set $pos
      $fmLB activate $pos
   } ;# end if
}
###############################################################################
# Function   : folderManager::Remove
# Description: Remove an existing folder
# Author     : Tom Wilkason
###############################################################################
proc folderManager::Remove {args} {
   variable fmLB
   variable Index
   set pos [$fmLB index active]
   $fmLB delete $pos
   ;# TFW: We should now remove all knowledge of these folders in TFList & Playlist
}

package provide saSettings 1.0

