##################################################################################
# snackAmp - urlRedirect package 1.0 for snackAmp
# provides dynamic http applications for SnackAmp
# web server.
#
# Music File Player in Tcl/Tk
#
# Copyright (C) 2001 Tom Wilkason
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
#
# Please send any comments/bug reports to
# tom.wilkason@cox.net  (Tom Wilkason)
##################################################################################
#
# To create a new dynamic url create a namespace with the name of the url
# e.g. namespace eval playlist -> /playlist
# to enable it, create an init routine that returns a 1 (for security)
# you can have the url return either an existing file (e.g. an mp3 file), in
# that case create a namespace command ::process, that returns the name
# of the file.
#
# If you want to create dynamic html with the namespace, create a namespace
# command ::process, that returns content. You will need an extra parameter
# passed to it that you set to the file extension required.
#
set [file root [file tail [info script]]]_History {
   $Header: /cvsroot/snackamp/snackamp/lib/urlRedirect.tcl,v 1.9 2003/05/17 14:20:16 wilkason Exp $
}
##################################################################################
#                              j u s t c o n t r o l
##################################################################################

namespace eval remote/justcontrol {
   proc init { } {
     return 1
   }
}
proc remote/justcontrol::process {sock {params ""}} {
   Trace
   foreach param [split $params &]  {
      foreach {CmdType Cmd} [split $param =] {break}
      set Cmd [HttpD::Decode $Cmd]
      switch -- $CmdType {
         bt_Level {
            ;# Would be nice not to have to push data back
            ;# maybe hold to end to push it out if needed.
            if {[catch {::macros::bumpLevel $Cmd} result] } then {
            }
         }
         default {}
      } ;# End Switch
   } ;# end foreach
   return [list dhtml html 0]
}

##################################################################################
#                               c o n t r o l
##################################################################################
##
# control namespace to control the SnackAmp player from
# a web browser. When user enters /control ... in to the browser
# The process function of this namespace will be called passing any
# user parameters. To add a new command then create another namespace
# and provide an init and process function.
#
namespace eval remote/control {
   variable Details
   variable LastFile {xxxxx}
   variable TagIDs {
      Active        Active
      "File Name"   Name
      Folder        Folder
      Artist        Artist
      Title         Title
      Album         Album
      Track         Track
      Media         Media
      Year          Year
      Genre         Genre
      Note          Note
      Comp          Comp
      Encoding      Enc
      Desc          Desc
      Size          fSize
      Date          fDate
      Duration      duration
      Position      Position
      "Sample Rate" rate
      Gain          levelFactor
   }
     ;# <meta http-equiv="Content-Type" content="text/html; charset=windows-1252">

   variable head {
<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN">
<html>
   <head>
      <meta http-equiv="Content-Language" content="en-us">
      <meta http-equiv="pragma"           content="no-cache">
      <meta name="GENERATOR"              content="SnackAmp Web Server">
      <meta name="ProgId"                 content="SnackAmp.Document">
      <LINK REL=StyleSheet                HREF="$css" TYPE="text/css" MEDIA=screen>
      <title>$title</title>
   }
   variable songTemplate {
      <table class="detailsTable" border="1" width="100%" height="50" cellpadding="0" cellspacing="0">
        <tr>
          <td width="10%" class="songTemplateLabel">Title</td>
          <td width="90%" colspan="5" class="songTitle"><b>$Details(Title)</td>
        </tr>
        <tr>
          <td class="songTemplateLabel" width="10%">Artist</td>
          <td class="songTemplateInfo"  width="90%" colspan="5">$Details(Artist)</td>
        </tr>
        <tr>
          <td class="songTemplateLabel" class="songTemplateLabel" width="10%">Album</td>
          <td class="songTemplateInfo" width="90%" colspan="5">$Details(Album)</td>
        </tr>
        <tr>
          <td class="songTemplateLabel" width="10%">File</td>
          <td class="songTemplateInfo" width="90%" colspan="5">$Details(Name)</td>
        </tr>
        <tr>
          <td class="songTemplateLabel" width="10%">Folder</td>
          <td class="songTemplateInfo" width="90%" colspan="5">$Details(Folder)</td>
        </tr>
        <tr>
          <td width="9%"   class="songTemplateLabel">Size</td>
          <td width="30%"  class="songTemplateInfo">$Details(fSize)</td>
          <td width="13%"  class="songTemplateLabel">Genre</td>
          <td width="20%"  class="songTemplateInfo">$Details(Genre)</td>
          <td width="10%"  class="songTemplateLabel">Gain</font></i></b></td>
          <td width="100%" class="songTemplateInfo">[gainControl::levelText $Details(levelFactor)]</td>
        </tr>
        <tr>
          <td width="9%"   class="songTemplateLabel">Date</td>
          <td width="30%"  class="songTemplateInfo">$Details(fDate)</td>
          <td width="13%"  class="songTemplateLabel">Rate</td>
          <td width="20%"  class="songTemplateInfo">$Details(rate)</td>
          <td width="10%"  class="songTemplateLabel">Track</td>
          <td width="100%" class="songTemplateInfo">$Details(Track)</td>
        </tr>
        <tr>
          <td width="9%"   class="songTemplateLabel">Year</td>
          <td width="30%"  class="songTemplateInfo">$Details(Year)</td>
          <td width="13%"  class="songTemplateLabel">Duration</td>
          <td width="20%"  class="songTemplateInfo">$Details(duration)</td>
          <td width="10%"  class="songTemplateLabel">Position</td>
          <td width="100%" class="songTemplateInfo">$Details(Position)</td>
        </tr>
      </table>
   }
   variable row_styles
   array set row_styles {
      0 {"lightRows"}
      1 {"darkRows"}
   }
   proc init { } {
     return 1
   }
}
###############################################################################
# Function   : remote/control::process state machine
# Description: Main function that is called by web server when /remote/control
#              is entered. Inside is a loop that iterates over each parameter
#              on the POST/GET and handles each one.
#              e.g. /remote/control?a=2&b=c  <= two params a= and b=
# Author     : Tom Wilkason
# Date       : 4/24/2002
###############################################################################
proc remote/control::process {sock {params ""}} {
   Trace
   ;##
   ;# Replace puts with appending to string then doing as a batch if needed
   ;#
   append Result [remote/control::leader "SnackAmp Control" "/snackamp.css" "/remote/control?Status=Refresh+Data"]
   append Result "<body>\n"
   append Result ${remote/jukebox::javascript}
   #append Result "$params"
   foreach param [split $params &]  {
      foreach {CmdType Cmd} [split $param =] {break}
      set Cmd [HttpD::Decode $Cmd]
      switch -- $CmdType {
         bt_Control {
            if {[catch {::macros::$Cmd} result] } then {
               append Result "<p>$::errorInfo</p>\n"
            }
            after 50 set ::remote/control::flag 1
            vwait ::remote/control::flag
         }
         bt_Level {
            if {[catch {::macros::bumpLevel $Cmd} result] } then {
               append Result "<p>$::errorInfo</p>\n"
            }
         }
         play {
            ;# Play a specific song
            ;# /control?play=some+name.mp3
            if {[catch {playSpecificSong $Cmd} error] } then {
               saLog $result
               append Result $result
            }
            after 50 set ::remote/control::flag 1
            vwait ::remote/control::flag
         }
         delete {
            LB::delete $Cmd
            updateFileDisplay
            ;# Remove from playlist
         }
         cmb_APL {
            ;# Select an auto-playlist file
            ;# TFW: If a real + is in file name, this may fail
            macros::readAPLFile [file join $::HOME $Cmd.tafs] 1
         }
         cmb_Sort {
            ;# Sort/Shuffle the files
            switch -glob -- [string tolower $Cmd] {
               sort*tracks     {set ::snackAmpSettings(sortMode) 0}
               shuffle*tracks  {set ::snackAmpSettings(sortMode) 1}
               shuffle*sort    {set ::snackAmpSettings(sortMode) 2}
               shuffle*shuffle {set ::snackAmpSettings(sortMode) 3}
            } ;# End Switch
            filtering::sortMusic
         }
         filter_text {
            ;# stage 1
            set ::TA(filter) $Cmd
            ;# Note: These will happen in this order, search_replace can process
            ;# Button=Filter&filter_text=ab&filter_what=All+Files&search_replace=Replace
         }
         filter_what {
            ;# stage 2
            switch -glob -- [string tolower $Cmd] {
               "all*files" {set Search_What 1}
               "play*" {set Search_What playList}
               "auto*" {set Search_What 0}
               default {set Search_What 1}
            } ;# End Switch
         }
         search_replace {
            ;# stage 3, do the filtering
            if {![info exists Search_What]} {set Search_What 1}
            set Search_Replace [string tolower $Cmd]
            filtering::pushFilter
            macros::filterMusic $Search_Replace $Search_What
            ;# stage 3
         }
         cmb_PL {
            ;# Stage 1
            ;# Set the playlist file
            foreach ext {m3u taf} {
               set Playlist_File [file join $::HOME $Cmd.$ext]
               if {[file exists $Playlist_File]} {
                  break
               }
            } ;# end foreach
         }
         PL_search_replace {
            ;# Stage 2
            ;# And then apply it
            macros::ProcessPlaylistFile $Playlist_File [string tolower $Cmd] 0
         }
         default {}
      } ;# End Switch
   } ;# end foreach
   ;# Wait to let player take hold
   append Result [sndInfo]
   append Result [nextToPlay 1 $::snackAmpSettings(httpNextLength)]
   append Result "</body>\n</html>"
   return [list dhtml html $Result]
}
###############################################################################
# Function   : leader
# Description: Page leader for dymanic page
# Author     : Tom Wilkason
# Date       : 4/24/2002
###############################################################################
proc remote/control::leader {title css {url "xx"}} {
   Trace
   variable head
   ;# title is used in subst
   append result [subst $head]
   if {![string equal $url "xx"]} {
     append result \n
     append result [subst {    <META HTTP-EQUIV="Refresh" CONTENT="$::snackAmpSettings(refreshRate);URL=$url">}]
     append result \n
   } ;# end if
   append result "  </head>\n"
}
###############################################################################
# Function   : trackInfo
# Description: Return, in a table, the details on some file
#
# Author     : Tom Wilkason
###############################################################################
proc remote/control::trackInfo {fileName} {
   variable TagIDs
   foreach {name tag} $TagIDs {
      set Details($tag) ""
   }
   set tempSound [snack::sound -debug 0 -file $fileName]
   if {[catch {id3Tag::songDetails $fileName $tempSound Details} error] } then {
      HttpD::Log "sndInfo:$error"
      $tempSound destroy
      return $error
   }
   $tempSound destroy
   set Details(Folder) [file dirname $fileName]
   set Details(Name) [file tail $fileName]
   set Details(Position) "N/A"
   ;# TFW:This forces table borders on blank cells, there may be a way
   ;# to do this in the table definition.
   foreach field [array names Details] {
      if {[string length $Details($field)]==0} {
         set Details($field) "&nbsp;"
      }
   } ;# end foreach
   set result [remote/control::songTable Details]
}
###############################################################################
# Function   : sndInfo
# Description: Return, in a table, the details on the current song
#
# Author     : Tom Wilkason
###############################################################################
proc remote/control::sndInfo {{playWhere 1}} {
   Trace
   variable LastFile
   variable TagIDs
   variable Details
   global snackAmpSettings
   set fileName [soundControl::fileName]
   ;##
   ;# If the file name hasn't changed, don't
   ;# bother requerying the static song data.
   ;#
   if {![string equal $fileName $LastFile]} {
      foreach {name tag} $TagIDs {
         set Details($tag) ""
      }
      if {[catch {id3Tag::songDetails $fileName s Details} error] } then {
         HttpD::Log "sndInfo:$error"
         return
      }
      set Details(Folder) [file dirname $fileName]
      set Details(Name) [file tail $fileName]
      set LastFile $fileName
   }
   set Active [string tolower [soundControl::playState]]

   ;# Fill in non tagged data based on file name
   if {$Details(Title)==""} {
       set Details(Title) [file root [file tail $fileName]]
   }
   if {$Details(Artist)==""} {
       set Details(Artist) [lindex [lrange [file split [file root $fileName]] end-$snackAmpSettings(ListDepth) end] 0]
   }
   if {$Details(Album)==""} {
       set Details(Album) [lindex [lrange [file split [file root $fileName]] end-$snackAmpSettings(ListDepth) end] end-1]
   }

   set Details(Position) [clock format [expr {round([soundControl::position])}]  -format %M:%S]

   if {$playWhere} {
      append result "<h2>SnackAmp currently $Active on [info hostname]:$::HttpD::Httpd(port)</h2>\n"
   } else {
      append result "<h2>SnackAmp currently $Active</h2>\n"
   }
   ;# TFW:This forces table borders on blank cells, there may be a way
   ;# to do this in the table definition.
   foreach field [array names Details] {
      if {[string length $Details($field)]==0} {
         set Details($field) "&nbsp;"
      }
   } ;# end foreach
   append result [remote/control::songTable Details]
   append result "<br>\n"
   return $result
}
###############################################################################
# Function   : nextToPlay
# Description: Print out the next N tracks in the playlist
#
# Author     : Tom Wilkason
###############################################################################
proc remote/control::nextToPlay {playButton {N 20}} {
   Trace
   global LB
   variable row_styles
   set offset 5
   set result {}
   if {[catch {
      set top [expr {$LB::indexPlaying-$offset}]
      set end [LB::index end]
      set total [expr {$end - [LB::active]}]
      set newEnd [expr {$top+$N}]
      if {$newEnd > $end} {set newEnd $end} ;# end if
      if {$top < 0} {set top 0} ;# end if
      #set fnt1 {<font face="Helvetica">}
      #set fnt2 {</font>}
      set current [file tail $::TA(currentMF)]
      append result "<h2>Next [expr {$newEnd-$top}] queued tracks (of [commify $end] tracks in playlist)</h2>\n"
      append result {<table border="1" width ="100%" cellpadding="0" cellspacing="0" class="songPlaylist">}
      if {$playButton} {
         append result {<tr class="headerRow"><td>Num</td><td>I</td><td>P</td><td>D</td><td>Folder</td><td>Track Name</td></tr>}
      } else {
         append result {<tr class="headerRow"><td>Num</td><td>Folder</td><td>Track Name</td></tr>}
      }
      for {set i $top} {$i < $newEnd} {incr i} {
         ;# <a href="remote//control?play=Artie+Shaw+-+Begin+The+Beguine.mp3">Artie Shaw - Begin The Beguine.mp3</a>
         set bgc $row_styles([expr {($i/5)&0x1}])
         set song [LB::get $i]
         set file [LB::getFullName $i 0]
         set track [file root [file tail $file]]
         set folder [httpHelpers::displayFolder $file]
         if {$folder==""} {
            set folder "&nbsp;"
         } ;# end if
         set trackInfo "<td>$folder</td><td>$track</td>"
         if {$playButton} {
            set imRef [remote/jukebox::questionButton $song 200 "info"]
            append imRef [subst {<td><a href="/remote/control?play=[HttpD::Encode $song]"><img border="0" src="/images/play.png"></a></td>}]
           # append imRef [subst {<td><a href="/remote/control?play=[HttpD::Encode $song]"><img border="0" src="/remote/images/playM.gif"></a></td>}]
            append imRef [subst {<td><a href="/remote/control?delete=$i"><img border="0" src="/images/delete.png"></a></td>}]
          #  append imRef [subst {<td><a href="/remote/control?delete=$i"><img border="0" src="/remote/images/deleteM.gif"></a></td>}]
         } else {
            set imRef ""
         }
         if {$i == $LB::indexPlaying} {
            set bgc {"activeRow"}
         }
         append result "<tr class=$bgc><td>$i</td>${imRef}${trackInfo}</tr>\n"
      } ;# end for
      append result </table>
   } error]} {
      return $error
   } else {
      return $result
   }
}

###############################################################################
# Function   : ::remote/control::aplFiles
# Description: Callback from TML file to return a list of all auto-playlist files
# Author     : Tom Wilkason
# Date       : 4/28/2002
###############################################################################
proc remote/control::aplFiles {args} {
   Trace
   set result ""
   ;# Create option list of all auto-playlist files
   foreach {file} [catalog::getAutoPlaylists] {
      set choice [file root [file tail $file]]
      if {[string equal $choice $::TA(AutoPlayListSelection)]} {
         append result "<option selected>$choice</option>\n"
      } else {
         append result "<option>$choice</option>\n"
      }
   }
   return [string trim $result]
}
###############################################################################
# Function   : ::remote/control::plFiles
# Description: Callback from TML file to return a list of all playlist files
# Author     : Tom Wilkason
# Date       : 4/28/2002
###############################################################################
proc remote/control::plFiles {args} {
   Trace
   set result ""
   ;# Create option list of all auto-playlist files
   foreach {file} [catalog::getPlaylists] {
      set choice [file root [file tail $file]]
      if {[string equal $choice $::TA(PlayListSelection)]} {
         append result "<option selected>$choice</option>\n"
      } else {
         append result "<option>$choice</option>\n"
      }
   } ;# end foreach
   return [string trim $result]
}
###############################################################################
# Function   : ::remote/control::songTable
# Description: Return HTML for the song details
# Author     : Tom Wilkason
# Date       : 4/29/2002
###############################################################################
proc remote/control::songTable {_Details} {
   upvar $_Details Details
   variable songTemplate
   set result [subst $songTemplate]
   return $result
}


##################################################################################
#                              d i s p l a y
##################################################################################
namespace eval display {
   proc init { } {
     return 1
   }
}
###############################################################################
# Function   : display
# Description: Main function that is called by web server when /display
#              is entered.
#              e.g. /display
# Author     : Tom Wilkason
# Date       : 12/15/2002
###############################################################################
proc display::process {sock {params ""}} {
   Trace
   append Result [remote/control::leader "SnackAmp Playlist" "/snackamp.css" ""]
   append Result "<body>\n"
   append Result ${remote/jukebox::javascript}
   ;# Wait to let player take hold
   #append Result ${remote/jukebox::close_window}
   append Result [remote/control::sndInfo 0]
   append Result [remote/control::nextToPlay 0 $::snackAmpSettings(httpNextLength)]
   ##append Result "<br><a href=/remote/jukebox>Request Tracks to Add To Playlist</a>"
   append Result "</body>\n</html>"
   return [list dhtml html $Result]
}

##################################################################################
#                              j u k e b o x
##################################################################################
###############################################################################
# Function   : remote/jukebox namespace
# Description: Handles interface to request songs to be added to playlist
#
# Author     : Tom Wilkason
###############################################################################
namespace eval remote/jukebox {
   variable search_form {
    <!-- Search Database -->
    <table border="1" width="100%" bordercolor="darkblue" cellspacing="0" cellpadding="0">
      <tr>
        <td width="100%" bgcolor="darkblue" align="center"><font color="white" face="Helvetica">
        <b>Search All Tracks for Matching Text</b></font></td>
      </tr>
      <tr>
        <td width="100%" align="left">
          <form method="POST" action="/remote/jukebox" TITLE="Enter text then hit Search to find all songs that contain the text">
            <input type="submit" value="Search" name="Button">
            <font face="Helvetica">
              <input type="text" name="jb_search" size="30">
            </font>
          </form>
        </td>
      </tr>
    </table>
    <p>
   }
   variable close_window {<a href="" onClick="window.close()">Click this to close the window</A>}
   ;##
   ;# Javascript helper routines
   ;#
   variable javascript {
     <script language="JavaScript">
     //<!--
        function popupInfo(url,height,width,tag) {
           var prefs = "scrollbars=yes,menubars=no,toolbars=no,"
           + "resizable=yes,left=0,top=0,height="+height+",width="+width;
           //alert(url + ' ' + escape(url) +' ' + encodeURI(url) + ' ' + encodeURIComponent(url))
           window.open(url, tag, prefs);
        }
        function toggleAll_jb_add(checked) {
           len = document.jb_add_form.elements.length;
           var i;
           for (i = 0; i < len; i++) {
              document.jb_add_form.elements[i].checked = checked;
           }
        }
        function toggleAll_jb_search_pl(checked) {
           len = document.jb_search_pl_form.elements.length;
           var i;
           for (i = 0; i < len; i++) {
              document.jb_search_pl_form.elements[i].checked = checked;
           }
        }
     //-->
     </script>
   }
   proc init { } {
     return 1
   }
}
###############################################################################
# Function   : remote/jukebox::process state machine
# Description: Main function that is called by web server when /remote/jukebox
#              is entered.
#              e.g. /remote/jukebox?jb_search_pl=*abc*
# Author     : Tom Wilkason
# Date       : 12/15/2002
###############################################################################
proc remote/jukebox::process {sock {params ""}} {
   Trace
   variable search_form
   variable javascript
   upvar #0 Httpd$sock data
   set title "SnackAmp Jukebox"
   set css "/snackamp.css"
   append Result [subst ${remote/control::head}]
   append Result [subst {<a href="/remote/jukebox?$params">Refresh</a>|}]
   ##append Result {<a href="/display" target=saplaylist>Display Playlist</a>|}
   #append Result {<a href="javascript:popupInfo('/display',600,600,'playlist')">Display Playlist</a>|}
   append Result ${remote/jukebox::close_window}
   ;##
   ;# Need to have search_form optional
   ;#
   #append Result [subst $search_form]
   append Result $javascript
   set Search_What pl
   set Tracks [list]
   foreach param [split $params &]  {
      foreach {CmdType Cmd} [split $param =] {break}
      set Cmd [HttpD::Decode $Cmd]
      switch -- $CmdType {
         jb_search {
            ;# stage 1
            set filter $Cmd
         }
         search_what {
            ;# stage 2
            switch -glob -- [string tolower $Cmd] {
               "all*files" {
                  ;# User specified text to search catalog
                  append Result "<h2>The following tracks matched your search criteria '$filter'</h2>"
                  append Result "<h2>Please check desired tracks and press <i>Append</i> or <i>Insert</i> at the bottom to add them to the playlist</h2>"
                  ;# Pass params for putting in back button on next form
                  append Result [remote/jukebox::searchCatalog $filter [expr {$::snackAmpSettings(httpNextLength)*4}] $params]
               }
               "play*" {
                  ;# User specifed some text to search within the playlist
                  ;# This goes in a pop-up
                  append Result "<h2>The following tracks from the Play List matched your search criteria '$filter'</h2>"
                  append Result [remote/jukebox::searchPlaylist $filter 0]
               }
               "auto*" {
                  ;# User specifed some text to search within the playlist
                  ;# This goes in a pop-up
                  append Result "<h2>The following tracks from the Auto-Playlist matched your search criteria '$filter'</h2>"
                  append Result [remote/jukebox::searchPlaylist $filter 1]
               }
            } ;# End Switch
         }
         play {
            ;# Play a specific song
            ;# /control?play=some+name.mp3
            if {[catch {filtering::InsertPlay $Cmd} error] } then {
               saLog $result
               append Result $result
            }
         }
         delete {
            LB::delete $Cmd
            updateFileDisplay
            ;# Remove from playlist
         }
         question {
            ;# This goes in a pop-up
            ;# javascript pop-up track details (like playinfo)
            if {![file exists $Cmd]} {
               ;# TFW: TF Lookup needs to be wrapped in LB
               set Cmd [LB::getFullNameFromHash $Cmd 1]
            } ;# end if
            ;# TFW: would be nice to make the fonts smaller
            append Result [remote/control::trackInfo $Cmd]
         }
         jb_add_start {
            ;# When processing a playlist add, this is the first param
            set playing [LB::active]
            set old_end [LB::end]
            ;# The back reference uses the jb_add_start param (which is the params from the
            ;# request that originated the search) which was passed in here. It is saved
            ;# In a hidden field from the prior form.
            append Result [subst {|<a href="/remote/jukebox?$Cmd">Back</a>\n}]
            append Result "<h2>Added the following tracks to the playlist</h2>\n"
            append Result {<table border="1" width="100%" cellpadding="0" cellspacing="0"\n}
         }
         jb_add {
            ;# When processing a playlist add, these are the middle params, one per item to add
            set track [file root [file tail $Cmd]]
            set folder [httpHelpers::displayFolder $Cmd]
            append Result "<tr><td>$folder</td><td>$track</td></tr></p>\n"
            lappend Tracks $Cmd
         }
         jb_add_process {
            ;# When processing a playlist add, this is the final param, signals end of list
            append Result "</table>\n"
            append Result "<h2>Your requested tracks were added at the end, [expr {$old_end - $playing}] tracks to go</h1>\n"
            foreach {track} $Tracks {
               macros::Append $track
            } ;# end foreach
            updateFileDisplay
         }
         jb_insert_process {
            ;# When processing a playlist add, this is the final param, signals end of list
            append Result "</table>\n"
            append Result "<h2>Your requested tracks are now queued to play next</h1>\n"
            set cp [LB::playing]
            foreach {track} $Tracks {
               filtering::Insert [incr cp] $track
            } ;# end foreach
            updateFileDisplay
         }
      }
   }
   ;# Insert Display Playlist link
   append Result "</body>\n</html>"
   return [list dhtml html $Result]
   return
}
###############################################################################
# Function   : getAPLFileList
# Description: Return a list of all APL files
# Author     : Tom Wilkason
# Date       : 12/24/2002
###############################################################################
proc remote/jukebox::getAPLFileList {args} {
   #set RF $snackAmpSettings(UseRemovableFolders)
   #set rmtFilter [file join $::RMT *]
   ;# TFW: Add proper support for removable media here
   ;##
   set result [list]
   foreach {File} [catalog::sortedList] {
      if {[filtering::checkIfIncluded [file dirname $File]]} {
         lappend result $File
      }
   }
   return $result
}
###############################################################################
# Function   : question
# Description: Question Button, pop-up info on a song
#              Handling this two encoding is a real pain in the ass
#              still won't handle a + in file name
# Author     : Tom Wilkason
# Date       : 4/24/2002
###############################################################################
proc remote/jukebox::questionButton {ref height tag} {
   ######set ref [HttpD::Encode $ref]
   #####return [subst {<td><a href="/remote/jukebox?question=$ref" target=sainfo><img border="0" src="/remote/images/questionM.gif"></a></td>}]

   ;# Escape single quotes for javascript popup window
   set ref [string map {' \\'} $ref]
   set ref "/remote/jukebox?question=$ref"
   return [subst {<td><a href="javascript:popupInfo('$ref',$height,600,'$tag')"><img border="0" src="/images/info.png"></a></td>}]
   #return [subst {<td><a href="javascript:popupInfo('$ref',$height,600,'$tag')"><img border="0" src="/remote/images/questionM.gif"></a></td>}]
}
###############################################################################
# Function   : searchPlaylist
# Description: Based on a pattern from the web entry box, filter the current
#              playlist and return the matching results
#
# Author     : Tom Wilkason
###############################################################################
proc remote/jukebox::searchPlaylist {filter {aplOnly}} {
   set i 0
   upvar #0 remote/control::row_styles row_styles
   set result {<table border="1" width="100%" cellpadding="0" cellspacing="0">}
   if {$aplOnly} {
      set files [remote/jukebox::getAPLFileList]
   } else {
      set files [LB::getAllNames 1]
   } ;# end if
   set j 0
   append result {<tr class="headerRow"><td>Num</td><td>I</td><td>P</td><td>D</td><td>Folder</td><td>Track Name</td></tr>}
   foreach {file} $files {
      if {[string match -nocase *$filter* $file]} {
         set bgc $row_styles([expr {($j/5)&0x1}])
         #set ref [HttpD::Encode $file]
         #append result "<tr><td><a href=\"/control?play=$ref\">$file</a></td></tr>\n"
         set track [file root [file tail $file]]
         set folder [httpHelpers::displayFolder $file]
         if {$folder==""} {
            set folder "&nbsp;"
         } ;# end if
         set btn [remote/jukebox::questionButton $file 200 "info"]
         ;# The second param just performs another search using the same filter
         append btn [subst {<td><a href="/remote/jukebox?play=[HttpD::Encode $file]&jb_search=$filter&search_what=Playlist"><img border="0" src="/images/play.png"></a></td>}]
         #append btn [subst {<td><a href="/remote/jukebox?play=[HttpD::Encode $file]&jb_search=$filter&search_what=Playlist"><img border="0" src="/remote/images/playM.gif"></a></td>}]
         append btn [subst {<td><a href="/remote/jukebox?delete=$i&jb_search=$filter&search_what=Playlist"><img border="0" src="/images/delete.png"></a></td>}]
         #append btn [subst {<td><a href="/remote/jukebox?delete=$i&jb_search=$filter&search_what=Playlist"><img border="0" src="/remote/images/deleteM.gif"></a></td>}]
         append result "<tr class=$bgc><td>$i</td>$btn<td>$folder</td><td>$track</td></tr>\n"
         incr j
      } ;# end if
      incr i
   } ;# end foreach
   append result "</table>"
   #append result "<a href=\"javascript:toggleAll_jb_search_pl(true)\">Select All</a> | <a href=\"javascript:toggleAll_jb_search_pl(false)\">Unselect All"
   return $result
}

###############################################################################
# Function   : remote/jukebox::searchCatalog
# Description: Search all songs (up to N) for results
#              Returns HTML shown all the tracks that matched the filter
#              Info and checkboxes are included
# Author     : Tom Wilkason
# Date       : 12/22/2002
###############################################################################
proc remote/jukebox::searchCatalog {filter max parms} {
   set i 0
   upvar #0 remote/control::row_styles row_styles
   append result {<FORM METHOD="POST" ACTION="/remote/jukebox" NAME="jb_add_form">}
   append result "<input type=hidden name=jb_add_start value=$parms>"
   ;# Initial param will be jb_add_start, followed by each checked item as jb_add, ending with jb_add_process=Submit
   append result {<table border="1" width="100%" cellpadding="0" cellspacing="0">}
   append result {<tr class="headerRow"><td>Chk</td><td>I</td><td>P</td><td>Folder</td><td>Track Name</td></tr>}
   foreach {file} [catalog::sortedList] {
      if {[string match -nocase *$filter* $file]} {
         set bgc $row_styles([expr {($i/5)&0x1}])
         set track [file root [file tail $file]]
         set folder [httpHelpers::displayFolder $file]
         if {$folder==""} {
            set folder "&nbsp;"
         } ;# end if
         set btn [remote/jukebox::questionButton $file 200 "info"]
         ;# The second param just performs another search using the same filter
         append btn [subst {<td><a href="/remote/jukebox?play=[HttpD::Encode $file]&jb_search=$filter&search_what=AllFiles"><img border="0" src="/images/play.png"></a></td>}]
         #append btn [subst {<td><a href="/remote/jukebox?play=[HttpD::Encode $file]&jb_search=$filter&search_what=AllFiles"><img border="0" src="/remote/images/playM.gif"></a></td>}]
         append result "<tr class=$bgc><td><INPUT TYPE=checkbox NAME=\"jb_add\" VALUE=\"$file\"></td>$btn<td>$folder</td><td>$track</td></tr>\n"
         incr i
      } ;# end if
      if {$i>$max} {
         break
      } ;# end if
   } ;# end foreach
   append result "</table>\n"
   if {$i>$max} {
      append result "<p>Results limited to $max matches</p>"
   } ;# end if
   append result "<INPUT TYPE=submit VALUE=Append NAME=jb_add_process>\n"
   append result "<INPUT TYPE=submit VALUE=Insert NAME=jb_insert_process>\n"
   append result "</FORM>\n"
   append result "<a href=\"javascript:toggleAll_jb_add(true)\">Select All</a> | <a href=\"javascript:toggleAll_jb_add(false)\">Unselect All<br>"
   return $result
}
##################################################################################
#                             g e t p l a y l i s t
##################################################################################
##
# playlist namespace, returns a nicly formatted list of playlist and other
# to build up streaming playlists.
#
namespace eval streaming/getplaylist.pls {
   proc init { } {
     return 1
   }
}
##
# Playlist code, from a web browser
# Processes a request to play a track, playlist, active playlist or auto-playlist
# Creates a winamp type pls dynamic file that when processed by the /streaming/live command
# will create a queue of songs then play them.
# Note: Add ability to send the entire stream length down as icy data. Good for tracks
##
# Note: TFW: Add another option to sort, shuffle or shufflesort
#            pass it on to the /streaming/live parameter, the sort param
#            should be specified first, otherwise it will be
#            ignored.
proc streaming/getplaylist.pls::process {sock {params ""}} {
   Trace
   upvar #0 Httpd$sock data
   set result ""
   set sortMethod "shuffle"
   set how "stream"
   set param ""
   set method "unknown"
   set max 50
   set refer $::snackAmpSettings(internalWebBase)
   if {[string length $params]} {
      foreach pr [split $params &]  {
         foreach {CmdType Cmd} [split $pr =] {break}
#puts "streaming/getplaylist.pls::process $CmdType=[HttpD::Decode $Cmd]"
         ;# Create a dynamic playlist
         switch -- $CmdType {
            active -
            playlist -
            track -
            folder -
            apl {
               set method $CmdType
               set param $Cmd
            }
            tree {
               switch -- $Cmd {
                  no {set fmethod folder}
                  yes {set fmethod treefolder}
               } ;# End Switch
            }
            sort {set sortMethod [HttpD::Decode $Cmd]}
            how  {set how [HttpD::Decode $Cmd]}
            submit  {}
            default {saLog "streaming/getplaylist.pls::process: Invalid subcommand - '$CmdType'"}
         }
      }
      ##
      # We have all the params, now process the options
      #
      switch -- [string tolower $how] {
         stream {
            ;# Create a stream of multiple tracks (except track option)
            switch -- $method {
               active {
                  set    result "File1=$refer/streaming/live?sort=$sortMethod&active=$param\n"
                  append result "Title1=Active Playlist"
               }
               apl {
                  set    result "File1=$refer/streaming/live?sort=$sortMethod&apl=$param\n"
                  append result "Title1=Auto Playlist [HttpD::Decode $param]"
               }
               playlist {
                  set    result "File1=$refer/streaming/live?sort=$sortMethod&playlist=$param\n"
                  append result "Title1=Playlist [HttpD::Decode $param]"
               }
               treefolder -
               folder {
                  set    result "File1=$refer/streaming/live?sort=$sortMethod&$fmethod=$param\n"
                  append result "Title1=Track [HttpD::Decode $param]"
               }
               track {
                  set    result "File1=$refer/streaming/live?track=$param\n"
                  append result "Title1=Track [HttpD::Decode $param]"
               }
               default {saLog "streaming/getplaylist.pls::process: Invalid subcommand method=$method"}
            }
            #puts "stream:$result"
         }
         tracks {
            ;# Create individual tracks
            switch -- $method {
               active {
                  if {[catch {streaming/live::activeplaylist $sortMethod} tracks]} then {
                     saLog $tracks
                  } else {
                     set i 1
                     set result ""
                     foreach index $tracks {
                        set track [lindex [catalog::sortedList] $index]
                        append result "File$i=$refer/streaming/live?track=[HttpD::Encode $track]\n"
                        append result "Title$i=[LB::getListName $track]\n"
                        incr i;if {$i > $max} {break}
                     }
                  }
               }
               apl {
                  if {[catch {streaming/live::apl [HttpD::Decode $param] $sortMethod} tracks] } then {
                     saLog $tracks
                  } else {
                     set i 1
                     set result ""
                     foreach index $tracks {
                        set track [lindex [catalog::sortedList] $index]
                        append result "File$i=$refer/streaming/live?track=[HttpD::Encode $track]\n"
                        append result "Title$i=[LB::getListName $track]\n"
                        incr i;if {$i > $max} {break}
                     }
                  }
               }
               playlist {
                  if {[catch {streaming/live::playlist [HttpD::Decode $param] $sortMethod} tracks]} then {
                     saLog $tracks
                  } else {
                     set i 1
                     set result ""
                     ;# shuffle
                     foreach index $tracks {
                        set track [lindex [catalog::sortedList] $index]
                        append result "File$i=$refer/streaming/live?track=[HttpD::Encode $track]\n"
                        append result "Title$i=[LB::getListName $track]\n"
                        incr i;if {$i > $max} {break}
                     }
                  }
               }
               folder {
                  set i 1
                  set result ""
                  set path [HttpD::Decode $param]
                  set files [lsort -dictionary [glob -nocomplain -directory $path *]]
                  foreach track $files {
#puts "get playlist folder $track"
                     if {[info exists ::taExtensions([file extension $track])]} {
                        append result "File$i=$refer/streaming/live?track=[HttpD::Encode $track]\n"
                        append result "Title$i=[LB::getListName $track]\n"
                        incr i;if {$i > $max} {break}
                     }
                  } ;# end foreach
               }
               treefolder {
                  set i 1
                  set result ""
                  set path [HttpD::Decode $param]
                  set tracks [catalog::filesMatchingPattern [file join $path *]]
                  foreach track $tracks {
#puts "treefolder $track"
                     append result "File$i=$refer/streaming/live?track=[HttpD::Encode $track]\n"
                     append result "Title$i=[LB::getListName $track]\n"
                     incr i;if {$i > $max} {break}
                  }
               }
               track {
                  set    result "File1=$refer/streaming/live?track=$param\n"
                  append result "Title1=Track [LB::getListName [HttpD::Decode $param]]"
               }
               default {saLog "streaming/getplaylist.pls::process: Invalid subcommand method=$method"}
            }
         }
         default {saLog "streaming/getplaylist.pls::process: Invalid subcommand how=$how"}
      }
#puts "playlist=$result"
      return [list dhtml pls [string trim $result]]
   } else {
      set title "SnackAmp Jukebox"
      set css "/snackamp.css"
      set result [remote/control::leader $title $css]
      append result "<h1>Please Specify a parameter</h1>\n"
      append result "</body>\n</html>"
      return [list dhtml html [string trim $result]]
   }
}
##################################################################################
#                              l i v e
##################################################################################
###############################################################################
# Function   : streaming/live namespace
# Description: Process streaming files
# Author     : Tom Wilkason
# Date       : 3/13/2003
###############################################################################
namespace eval streaming/live {
   proc init {args} {
      return 1
   }
}
##
# the /streaming/live process method just retuns a file name from an existing file.
# used to initiate a stream to a client. This is run from shortcuts
# AFTER the playlist has been built and selected.
# Note: TFW: Add another option to sort, shuffle or sort/shuffle
#
proc streaming/live::process {sock {params ""}} {
   Trace
   upvar #0 Httpd$sock data
   set Tracks [list]
   set sortMethod "shuffle"
   if {[string length $params]} {
      foreach param [split $params &]  {
         foreach {CmdType Cmd} [split $param =] {break}
         set Cmd [HttpD::Decode $Cmd]
#puts "streaming/live::process $CmdType=$Cmd"

         switch -- $CmdType {
            apl {
               ;# Get some apl file list of indicies
               if {[catch {streaming/live::apl $Cmd $sortMethod} result] } then {
                  return -code error $result
               } else {
                  if {[llength $result] > 0} {
                     set data(streaming) $result
                     set data(streamingLength) 44000000
                     set data(Title) "AutoPlaylist From $data(mime,host)"
                     set data(Album) "$Cmd ([llength $result] trks)"
                     return [list file html [streamBuffer::popFile $sock]]
                  } else {
                     return -code error "No apl files for '$Cmd'"
                  }
               }
            }
            playlist {
               ;# Get some tracks from a playlist
               if {[catch {streaming/live::playlist $Cmd $sortMethod} result]} then {
                  return -code error $result
               } else {
                  ;##
                  ;# Don't return track, return a m3u file
                  ;#
                  if {[llength $result] > 0} {
                     set data(streaming) $result
                     set data(streamingLength) 44000000
                     set data(Title) "Playlist From $data(mime,host)"
                     set data(Album) "$Cmd ([llength $result] trks)"
                     return [list file html [streamBuffer::popFile $sock]]
                  } else {
                     return -code error "No playlist files for '$Cmd'"
                  }
               }
            }
            active {
               ;# Get some tracks from a playlist
               if {[catch {streaming/live::activeplaylist $sortMethod} result]} then {
                  return -code error $result
               } else {
                  ;##
                  ;# Don't return track, return a m3u file
                  ;#
                  if {[llength $result] > 0} {
                     set data(streaming) $result
                     set data(streamingLength) 44000000
                     set data(Title) "Active Playlist From $data(mime,host)"
                     set data(Album) "$Cmd ([llength $result] trks)"
                     return [list file html [streamBuffer::popFile $sock]]
                  } else {
                     return -code error "No active playlist files for '$Cmd'"
                  }
               }
            }
            folder {
#puts "streaming/live folder $Cmd"
               set result [list]
               set files [sortByMethod [glob -nocomplain -directory $Cmd *] $sortMethod]
               foreach file $files {
                  if {[info exists ::taExtensions([file extension $file])]} {
                     set loc [catalog::whereInDb $file]
                     if {$loc >= 0} {
                        lappend result $loc
                     }
                  }
               } ;# end foreach

               if {[llength $result] > 0} {
                  set data(streaming) $result
                  set data(streamingLength) 44000000
                  set data(Title) "Folder $Cmd From $data(mime,host)"
                  set data(Album) "$Cmd ([llength $result] trks)"
                  return [list file html [streamBuffer::popFile $sock]]
               } else {
                  return -code error "No folder files found for '$Cmd'"
               }
            }
            treefolder {
#puts "treefolder $Cmd"
               set files [sortByMethod [catalog::filesMatchingPattern [file join $Cmd *]] $sortMethod]
               set result [list]
               foreach file $files {
                  set loc [catalog::whereInDb $file]
                  if {$loc >= 0} {
                     lappend result $loc
                  }
               } ;# end foreach

               ;##
               ;# Don't return track, return a m3u file
               ;#
               if {[llength $result] > 0} {
                  set data(streaming) $result
                  set data(streamingLength) 44000000
                  set data(Title) "Treefolder $Cmd From $data(mime,host)"
                  set data(Album) "$Cmd ([llength $result] trks)"
                  return [list file html [streamBuffer::popFile $sock]]
               } else {
                  return -code error "No folder files found for '$Cmd'"
               }
            }
            track {
               ;# Process a single track in a playlist
               set loc [catalog::whereInDb $Cmd]
               if {$loc >= 0} {
                  set file [lindex [catalog::sortedList] $loc]
                  set data(streaming) $loc
                  set data(streamingLength) [file size $file]

                  ;##
                  ;# How do we configure time to this?
                  ;#

                  set tempSound [snack::sound -debug 0 -file $file]
                  if {[catch {id3Tag::songDetails $file s Details} error] } then {
                     saLog $error
                  } else {
                     ;# Fill in data that will go into icy-meta tags
                     foreach {index} {Title Album Artist Genre rate} {
                        if {[info exists Details($index)]} {
                           set data($index) $Details($index)
                        } else {
                           set data($index) ""
                        }
                     }
                     # add bitrate here
                  }
                  $tempSound destroy
                  return [list file html [streamBuffer::popFile $sock]]
               }
            }
            sort        {set sortMethod $Cmd}
            default {
               return -code error "streaming/live::process: Invalid subcommand - '$CmdType'"
            }
         } ;# End Switch
      }
   } else {
      return -code error "streaming/live::process: No Parameter '$params'"
   }
}
##
#  Return a list of indices to auto playlist files
#
proc streaming/live::apl {name sortMethod} {
   Trace
   set fName [file join $::HOME $name.tafs]
   if {![file exists $fName]} {
      return -code error "$fName does not exist"
   } else {
      if {[catch {open $fName r} fid]} then {
         return -code error $fid
      }
      fconfigure $fid -buffersize 100000
      set full [split [read $fid] \n]
      close $fid
      ;# Get folder list
      set biglist [list]
      array set myte [array get ::teFolders]
      foreach line $full {
         foreach {Index Value} $line {
            set myte($Index) $Value
         }
      } ;# end foreach
      set i 0
      ;# Grab locations in an array and shuffle it
      set tracks [sortByMethod [catalog::sortedList] $sortMethod]
      foreach track $tracks {
         set trackDir [file dirname $track]
         if {[info exists myte($trackDir)]} {
            if {$myte($trackDir)==1} {
               lappend biglist $i
            }
         } else {
            #saLog "Couldn't find '$trackDir' as an index"
         }
         incr i
      }
   }
   return $biglist
}
##
# Return a list of indicies for a playlist
#
proc streaming/live::playlist {name sortMethod} {
   Trace
   set name [string trim $name \"] ;#"
   foreach ext {m3u taf} {
      set fName [file join $::HOME $name.$ext]
      if {[file exists $fName]} {
         break
      }
   } ;# end foreach
   if {![file exists $fName]} {
      saLog "streaming/live::playlist: $fName does not exist"
      return -code error "$fName does not exist"
   } else {

      ;# Here we have a file case issue
      if {[catch {open $fName r} fid]} then {
         return -code error $fid
      }
      fconfigure $fid -buffersize 100000
      set full [split [read $fid] \n]

      close $fid
      set tracks [sortByMethod $full $sortMethod]
      ;# Get folder list
      set biglist [list]
      foreach line $tracks {
         set line [file join $line] ;# make tcl name
         set loc [catalog::whereInDb $line]
         if {$loc >= 0} {
            lappend biglist $loc
         } else {
 #           saLog "$line not in database"
         }
      }
   }
   return $biglist
}
;##
;# Return a list of indicies to current playlist on GUI
;#
proc streaming/live::activeplaylist {sortMethod} {
   set biglist [list]
   ;##
   ;# Note: May want to get from cursor foward.
   ;#
   set tracks [sortByMethod [LB::getAllNames 1] $sortMethod]
   foreach file $tracks {
      set loc [catalog::whereInDb $file]
      if {$loc >= 0} {
         lappend biglist $loc
      }
   }
   return $biglist
}
proc streaming/live::folder {name} {
}
proc streaming/live::recurfolder {name} {
}
###############################################################################
# Function   : sortByMethod
# Description: Sort a list by the appropriate method
# Author     : Tom Wilkason
# Date       : 3/24/2003
###############################################################################
proc streaming/live::sortByMethod {list how} {
   switch -glob -- [string tolower $how] {
      sort*tracks     {return [lsort -dictionary $list]}
      shuffle*tracks  {return [shuffle $list]}
      shuffle*sort    {return [shuffleByFolder $list ldsort]}
      shuffle*shuffle {return [shuffleByFolder $list shuffle]}
      none {return $list}
   } ;# End Switch
   return $list
}

##################################################################################
#                              s t r e a m i n g  f i l e  b r o w s e r
##################################################################################
namespace eval streaming/browser {
   proc init { } {
     return $::snackAmpSettings(EnableStreaming)
   }
   variable folderopen {<img border=0 src=/images/folderopen.gif>}
   variable folderclosed {<img border=0 src=/images/folderclosed.gif>}
   variable playbutton {<img border=0 src=/images/play.png>}
   variable brform {
  <H2>Enter Play Folder Option</H2>
  <TABLE>
    <TR>
      <TD>Sorting</TD>
      <TD>Streaming</TD>
      <TD>Subfolders</TD>
    </TR>
    <TR>
      <form name=option_form>
        <TD>
          <select size="1" name="cmb_Sort" class="formText">
            <option>Sort-Tracks</option> <option selected>Shuffle-Tracks</option> <option>Shuffle->Sort</option>
            <option>Shuffle->Shuffle</option>
          </select>
        </TD>
        <TD>
          <select size="1" name="cmb_How" class="formText">
            <option selected>stream</option> <option>tracks</option>
          </select>
        </TD>
        <TD>
          <select size="1" name="cmb_Sub" class="formText">
            <option selected>no</option> <option>yes</option>
          </select>
        </TD>
      </form>
    </TR>
  </TABLE>
   }
   variable foldersel {
  <script language="JavaScript">
   function popupInfo(url,height,width,tag) {
     var prefs = "scrollbars=yes,menubars=no,toolbars=no,"
     + "resizable=yes,left=0,top=0,height="+height+",width="+width;
     //alert(url + ' ' + escape(url) +' ' + encodeURI(url) + ' ' + encodeURIComponent(url))
     window.open(url, tag, prefs);
   }
   // getselected: Return the selected item from a combo box
   // TFW: element as object doesn't work due to ?? no type ??
   function getselected(element,len) {
      var i;
      for (i==0; i<len; i++) {
         alert(element[i].selected + ' ' + element[i].text);
         if (element[i].selected) {
            return element[i].text;
         }
      }
      return 'none';
   }
   function foldersel(url,folder) {
     var useurl='none';
     var selsort='no sort';
     var selhow='no stream';
     var selsub='no';
     var i=0;
     //stream
     //selhow = getselected(document.option_form.cmb_How,2);
     for (i==0;i<2;i++) {
        if (document.option_form.cmb_How[i].selected) {
           selhow=document.option_form.cmb_How[i].text;
           break;
        }
     }

     //subfolders
     for (i==0;i<2;i++) {
        if (document.option_form.cmb_Sub[i].selected) {
           selsub=document.option_form.cmb_Sub[i].text;
           break;
        }
     }
     //sort
     for (i==0;i<4;i++) {
        if (document.option_form.cmb_Sort[i].selected) {
           selsort=document.option_form.cmb_Sort[i].text;
           break;
        }
     }
     useurl = url+
       '/?sort='  + selsort +
       '&tree='   + selsub  +
       '&how='    + selhow  +
       '&folder=' + folder;
     //alert(useurl);
     window.open(useurl,"tinyWindow", 'width=100,height=100');
     //window.open(useurl);
   }
</script>
  }
}
proc streaming/browser::leader {title css {url "xx"}} {
   Trace
   variable foldersel
   ;# title is used in subst
   append result [subst ${remote/control::head}]
   if {![string equal $url "xx"]} {
     append result \n
     append result [subst {    <META HTTP-EQUIV="Refresh" CONTENT="$::snackAmpSettings(refreshRate);URL=$url">}]
     append result \n
   } ;# end if
   append result $foldersel
   append result "  </head>\n"
}

##
# url to show a directory listing to allow browsing
# of files by folder.
# todo:
# button to play folder
# button in insert file/folder into playlist
# todo: Add form controls to define how streaming/tracks/sorting is to be used.
proc streaming/browser::process {sock {params ""}} {
   Trace
   variable folderopen
   variable folderclosed
   variable playbutton
   variable brform
   set myns [string trim [namespace current] :]
   upvar #0 Httpd$sock data
   set title "SnackAmp Jukebox"
   set css "/snackamp.css"
   set result [leader $title $css]
   append result <body>
   set sortMethod "shuffle"
   if {[string length $params]} {
      foreach param [split $params &]  {
         foreach {CmdType Cmd} [split $param =] {break}
         switch -- $CmdType {
            sort {
               set sortMethod $Cmd
            }
            folder {
               append result "<h1>Media Folders</h1>\n"
               set folders [list]
               set files [list]
               set folder [HttpD::Decode $Cmd]
               set contents [lsort -dictionary [glob -nocomplain -directory $folder *]]
               foreach entry $contents {
                  if {[file isdirectory $entry]} {
                     lappend folders $entry
                  } elseif {[info exists ::taExtensions([file extension $entry])]} {
                     lappend files $entry
                  }
               } ;# end foreach
               #----------------------------
               ;##
               ;# Create a list of all folders, includling toplevel ones
               ;#
               set fldr $folder
               while {1} {
                  ;# Get list of all parent folders
                  set under [file dirname $fldr]
                  if {[string equal $under $fldr]} {
                     break
                  } else {
                     lappend folders $fldr ;########
                     set fldr $under
                  }
               }
               ;# List of toplevel folder
               foreach item $::snackAmpSettings(mediaDirs) {
                  lappend folders $item
                  if {[string match "$item*" $folder]} {
                     set icon $folderopen
                  } else {
                     set icon $folderclosed
                  }
                  append result "$icon <a href=/$myns?folder=[HttpD::Encode $item]>$item</a><br>\n"
               } ;# end foreach
               #----------------------------
               set folders [lsort -unique $folders]
               ;# Show the folders
               if {[llength $folders]} {
                  append result "<hr>\n"
                  append result $brform
                  append result "<h1>Current Folder $folder</h1>\n"
                  append result {<table border="1" cellpadding="0" cellspacing="0" class="songPlaylist">}
                  ;# Find minimum folder length for indent offsets
                  foreach item $folders {
                     lappend reps [llength [file split $item]]
                  }
                  set minrep [lindex [lsort -integer $reps] 0]
                  foreach item $folders {
                     set icon $folderclosed
                     if {[string equal $item $folder]} {
                        set icon $folderopen
                     }
                     set rep [expr {[llength [file split $item]]-$minrep}]
                     set link "[string repeat &nbsp $rep]$icon <a href=/$myns?sort=\"$sortMethod\"&folder=[HttpD::Encode $item]>$item</a><br>\n"
                     ;# S = Stream folder contents
                     ;# T = Tracks for folder contents
                     set url [string map {' \\'} $item]
                     set base "/streaming/getplaylist.pls"
                     set F [subst {<a href="javascript:foldersel('$base','$url')"><img border=0 src=/images/play.png></A>}]
                     append result [httpHelpers::tableRow [list $F $link] darkRows]
                  } ;# end foreach
                  append result "</table>\n"
               }
               ;# Show the files
               if {[llength $files]} {
                  append result "<hr>\n"
                  append result "<h1>Tracks within $folder</h1>\n"
                  append result {<table border="1" cellpadding="0" cellspacing="0" class="songPlaylist">}
                  foreach file $files {
                     set ref [string map {' \\'} $file]
                     set I [subst {<a href="javascript:popupInfo('/remote/jukebox/?question=$ref',200,600,'Info')"><img border=0 src=/images/info.png></A>}]
                     set link "<a href=/streaming/getplaylist.pls?sort=\"$sortMethod\"&track=[HttpD::Encode $file]>$playbutton</a><br>\n"
                     set size "[commify [expr {[file size $file]/1024}]] KB"
                     append result [httpHelpers::tableRow [list $I $link [file tail $file] $size] lightRows]
                  } ;# end foreach
                  append result "</table>\n"
               } else {
                  append result "<hr>\n"
                  append result "<h1>No Tracks within $folder</h1>\n"
               }
            }
         } ;# End Switch
      }
   } else {
      append result "<h1>SnackAmp Folder Browser</h1>\n"
      append result "<h2>Please Select a Folder Below to Browse</h2>\n"
      foreach folder $::snackAmpSettings(mediaDirs) {
         append result "$folderclosed <a href=/$myns?folder=[HttpD::Encode $folder]>$folder</a><br>\n"
      } ;# end foreach
   }
   append result "</body>\n</html>\n"
   return [list dhtml html [string trim $result]]
}
##################################################################################
#                 s t r e a m i n g   p l a y l i s t  e x a m p l e
##################################################################################
##
# Example code to display a selection of sources on this server
#
namespace eval streaming/playlists {
   proc init { } {
     return 1
   }
}
proc streaming/playlists::process {sock {params ""}} {
   Trace
   upvar #0 Httpd$sock data
   ;##
   ;# This can be prettied up to select the various
   ;# playlists that are created.
   ;# Can add options to use sorting options
   ;#
   set title "SnackAmp Jukebox"
   set css "/snackamp.css"
   set result [remote/control::leader $title $css]
   append result <body>
   ##
   # Need to add code to show folder trees
   #
   append result {<h1><a href=/streaming/getplaylist.pls?active=active>Current Playlist</a></h1>}
   set Files [catalog::getAutoPlaylists]
   ;# Create option list of all auto-playlist files
   if {[llength $Files] > 0} {
      append result "<h1>Auto Playlist Files</h1>\n"
      foreach {file} $Files {
         set choice [file root [file tail $file]]
         set ref "/streaming/getplaylist.pls?apl=[HttpD::Encode $choice]"
         append result [subst {<a href="$ref">$choice</a><br>\n}]
      } ;# end foreach
   }
   ;# Create option list of all playlist files
   set Files [catalog::getPlaylists]
   if {[llength $Files] > 0} {
      append result "<h1>Playlist Files</h1>\n"
      foreach {file} $Files {
         set choice [file root [file tail $file]]
         set ref "/streaming/getplaylist.pls?playlist=[HttpD::Encode $choice]"
         append result [subst {<a href="$ref">$choice</a><br>\n}]
      } ;# end foreach
   }
   append result "</body>\n</html>\n"
   return [list dhtml html [string trim $result]]
}
##
# Http helper routines
#
namespace eval httpHelpers {
   proc tableRow {cols {class ""}} {
      if {$class==""} {
         set result <tr>
      } else {
         set result "<tr class=\"$class\">"
      }
      foreach {element} $cols {
         append result "<td>$element</td>"
      }
      append result </tr>
      return $result
   }
   proc streamInfo {{class ""}} {
      set result ""
      foreach {file ip} [streamBuffer::allStreamInfo] {
         append result "[tableRow [list $file $ip] $class]\n"
      } ;# end foreach
      return $result
   }
   proc displayFolder {File} {
      global snackAmpSettings
      if {$snackAmpSettings(ListDepth) > 0} {
         ;# File join doesnt work due to single list
         set File [string map {// /} $File]      ;# remove UNC stuff
         set name [file dirname [join [lrange [file split $File] end-$snackAmpSettings(ListDepth) end] /]]
      } else {
         set name [file tail [file dirname $File]]
      }
      return $name
   }
}
package provide urlRedirect 1.0
##
# Utility routine
#
proc dumpSocket {} {
   foreach glb [info globals Httpdsock*] {
      global $glb
      puts "Array $glb--------------"
      catch {parray $glb}
   } ;# end foreach
}

set httpToDo {
- Playlists on play_control
- Ability to change the queued print out size
- FF/FR to play control
- Figure a way to push data to browser without it requesting it
  Have knowlege when a socket is open.
- Clicking on a song should bring up a pop-up with song details, use
  play button to play a song (can use javascript like my pop-up)
- Animated gif to denote if a song is playing

Unix bugs
- KDE/Netscape the meta refresh doesn't seem to work
- Pausing a song doesn't get an indication on the browser,
  it only knows if it is stopped or playing.
}

