#!/bin/sh
# the next line restarts using wish \
exec wish "$0" ${1+"$@"}
set SAVersion "2.2"
##################################################################################
# snackAmp 2.1.2
#
# Music File Player in Tcl/Tk
#
# Copyright (C) 2001 Tom Wilkason
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
#
# Please send any comments/bug reports to
# tom.wilkason@cox.net  (Tom Wilkason)
#
# Originally derived from Snack snamp.tcl demo player by Tom Wilkason
##################################################################################
set [file root [file tail [info script]]]_History {
   $Header: /cvsroot/snackamp/snackamp/snackAmp.tcl,v 1.101 2003/05/18 17:07:46 wilkason Exp $
}
package require Tk
wm withdraw .
# Check if this is the only invocation; According to the docs, a second
# invocation will contain #2 after SnackAmp (on unix)
if {![string equal SnackAmp [tk appname SnackAmp]]} {
   if {[llength $argv]==0} {
      tk_messageBox -message "Sorry! You can't run two versions of SnackAmp simultaneously!"
      send SnackAmp {raise .saTop}
   } else {
      send SnackAmp PlayFile $argv
   }
   exit
}
proc scriptDir {args} {
   global src_dir auto_path argv0
   set script $argv0
   while {[file type $script] == "link"} {
      set link [file readlink $script]
      if {[file pathtype $link] == "relative"} {
         set link [file join [file dirname $script] $link]
      }
      set script $link
   }
   set src_dir [file dirname $script]
   set here [pwd]
   cd $src_dir
   set src_dir [pwd]
   cd $here
   return $src_dir
}
set TA(PlayListSelection) {}
set TA(AutoPlayListSelection) {}
set TA(PlayListFile) {}
;# compatability: namespace exists is part of 8.4
proc hasNamespace {namespace} {
   if {[lsearch [namespace children ::] $namespace] >= 0} {
      return 1
   } else {
      return 0
   }
}
;##
;# If this is not wrapped, add libs off snackAmp to auto_path
;# The wrapped version uses virtual folders which are inserted
;# by the wrap program. This method works with all freewrap
;# or tclkit versions.
;#
if {[hasNamespace ::freewrap]} {
   set TA(isWrapped) 1
   set TA(snackRoot) [file dirname [info script]]
   set TA(appname) "\"[file nativename [info nameofexecutable]]\""
} elseif {[hasNamespace ::starkit]} {
   ;# Note: Dual dirname since SnackAmp/wrap is location
   ;# of info script
   set TA(appname) "\"[file nativename [info nameofexecutable]]\""
   set TA(isWrapped) 1
   set TA(snackRoot) $::starkit::topdir
   set TA(realRoot) [file dirname $::starkit::topdir]
   ###############################################################################
   # Function   : load
   # Description: If files can be loaded locally, load them if possible
   # Author     : Tom Wilkason
   # Date       : 1/15/2003
   ###############################################################################
   rename load _load
   proc load {args} {
      set packageName {}
      foreach {filename packageName} $args {break} ;# end foreach
      set newfilename [file join $::TA(realRoot) [file tail $filename]]
      if {[file exists $newfilename] && [string length $filename]} {
         #puts "1 Loading '$newfilename' '$packageName'"
         eval [list _load $newfilename $packageName]
      } else {
         #puts "2 Loading '$filename' '$packageName'"
         eval [list _load $filename $packageName]
      }
   }
} else {
   set TA(appname) "\"[file nativename [info nameofexecutable]]\" \"[info script]\""
   ;# Not wrapped in any way
   set TA(realRoot) [scriptDir]
#   puts "Running as scripts"
   set TA(isWrapped) 0
   set TA(snackRoot) [scriptDir]
   ;# Need to build our own auto_path
   lappend auto_path $TA(snackRoot) [file join $TA(snackRoot) lib]
} ;# end if

set TA(webRoot) [file join $TA(snackRoot) docs]
switch -- $::tcl_platform(platform) {
   "unix" {
      set ::TA(arrow) "arrow"
      set ::TA(TEMP) /tmp
   }
   default {
      set ::TA(arrow) "arrow"
      set ::TA(TEMP) $::env(TEMP)
   }
} ;# End Switch

###############################################################################
# Function   : findMyHome
# Description: Location the appropirate location of the HOME folder
# Author     : Tom Wilkason
# Date       : 1/23/2002
###############################################################################
proc findMyHome {args} {
   global env
   global argv
   ;##
   ;# Default home folder names
   ;#
   ;# First check the command line, use it if specified
   set sHome ".snackAmp"
   if {[info exists argv] && [llength $argv] > 0} {
      foreach {opt param} $argv {
         if {[string equal $opt "-home"]} {
            set home $param
            ;##
            ;# Check if home needs to be created
            ;#
            set ::argv [lrange $argv 2 end]
            return $home
         } ;# end if
      } ;# end foreach
   }
   if {[info exists env(SNACKAMPHOME)]} {
      ;# Then check the SNACKAMPHOME env var for home
      ;# and use it if it exists
      return [string trim $env(SNACKAMPHOME) \"]  ;#"
   } else {
      ;# Determine appropriate home location
      switch -- $::tcl_platform(platform) {
         "unix" {
            return [file join ~ $sHome]
         }
         "windows" {
            ;# Only use HOME if setup by a prior SnackAmp
            if {[info exists env(HOME)] && [file exists [file join $env(HOME) $sHome]]} {
               return [file join [string trim $env(HOME) \"] $sHome]  ;#"
            } ;# end if
            ;# Otherwise try to use USERPROFILE
            if {[info exists env(USERPROFILE)]} {
               return [file join $env(USERPROFILE) $sHome]
            } ;# end if
            ;# Some older windows with profiles, just use folder
            ;# where snackamp is stored.
            return [file join [scriptDir] $sHome]
         }
         default {
            return [file join ~ $sHome]
         }
      } ;# End Switch
   }
}
;# Now we know where the home folder is
set HOME                    [findMyHome]
set TA(settingsFile)        [file join $::HOME .snackAmpDefaults]
set TA(playCacheFile)       [file join $::HOME .snackAmpPlayCache]
set TA(defaultFolderFile)   [file join $::HOME default.tafs]
set TA(logFile)             [file join $::HOME snackAmp.log]
set TA(scriptHome)          [scriptDir]
set TA(defaultSettingsFile) [file join $TA(scriptHome) ".snackAmpDefaults"]
set TA(cacheFile)           ".snackAmpCache2"
set TA(STOP) 0
if {[file exists $TA(logFile)]} {
   catch {file delete $TA(logFile)}
} ;# end if
;##
;# Some initial values
;#
set TA(Console) 0
set TA(songDuration) 0
set TA(p) 0
set TA(matchAgainstAll) 0     ;# don't match against all
set TA(currentMF) {}
set TA(useSAFilter) 0
set TFList [list]
;##
;# Removable media tag
;#
set RMT "/RMC"
;##
;# Trick to create a menu font that matches
;# the native windows
;#
menu .temp
if {[string match "*-*-*-*" [.temp cget -font]]} {
   ;# Handle X style fonts as defaults
   .temp configure -font {Helvetica -16}
}
foreach {font size} [.temp cget -font] {break}
font create menu -size $size -family $font
unset font
unset size
destroy .temp
;# To do this winico should be loaded first
;# Load the snack package
#<<magic code>> <- Needed by wrap function, don't change comment
;# Other packages required immediatly
package require saImages
createSAImages

package require cframe
package require gpWidgets
waitWindow::init
waitWindow::show "SnackAmp $::SAVersion Initializing...Please Wait"
update
package require gpUtils
package require saFiltering
package require catalog
package require saSettings
package require playurl
if {[info command console] == {}} {
   catch {package require saConsole}
}
if {[catch {package require snack} result]} {
   tk_messageBox -icon warning -type ok -title "Snack Load Error\$result" \
      -message "$result"
   exit
}
;##
;# Windows platforms get taskbar icons
;#
if {$tcl_platform(platform) == "windows"} {
   if {[catch {package require IconSupport} result]} {
      puts stderr $result
   }
   package require dde
   dde servername snackAmp
}

###############################################################################
# Define the procedures and continue processing at the end of the file
###############################################################################

###############################################################################
# Function   : PlayFile
# Description: Play a file at the top of the list (linked to DDE command on Windows)
#              Set this up early so a file association doesn't time out when launching
#              SnackAmp on a slow computer.
# Author     : Tom Wilkason
###############################################################################
proc PlayFile {file} {
   #puts "PlayFile $file"
   set file [string map {\\ /} $file]
   set ext [string tolower [file extension $file]]
   switch -- $ext {
      .m3u
      -
      .pls {
         if {[string match -nocase "http://*" $file]} {
            playurl::checkuskfolder
            set contents [split [playurl::getUrlBody $file] \n]
         } else {
            if {[file readable $file]} {
              set fid [open $file r]
              set contents [split [string trim [read $fid]] \n]
              close $fid
            } else {
               tk_messageBox -icon warning -type ok -message "File '$file' cannot be read"
               return
            }
         }
         set fileTail [file root [file tail $file]]
         switch -- $ext {
            .m3u {
                set toappend [list]
                set i 0
                foreach {row} $contents {
                   set drow [HttpD::Decode $row]  ;# not sure if this is needed
                   if {![string match "\#*" $drow]} {
                      set fullFile [file join [file dirname $file] $drow]
                      if {[string match -nocase "http://*" $drow]} {
                         # Create a usk (url) shortcut, then append it
                         set name [file tail $drow]
                         # Note: trying tail of file name, works if a file if a stream
                         # may not work as well.
                         set fname [playurl::addUrl $::snackAmpSettings(uskDir) "$name" $row]
                         lappend toappend $fname
                         incr i
                      } elseif {[file exists $fullFile]} {
                         lappend toappend $fullFile
                      } else {
                         saLog "File:'$row' does not exist"
                      }
                   }
                }
                set added 0
                set end [LB::end]
                foreach {file} $toappend {
                   incr added
                   filtering::Append $file
                } ;# end foreach
                if {$added > 0} {
                   LB::inverseVideoFile xx 0
                   LB::activate $end
                   updateFileDisplay
                   Play active
                }
            }
            .pls {
               set title $fileTail
               set i 0
               set Fnum 0
               set Tnum 0
               foreach {row} $contents {
                  if {[regexp {File([0-9]+)=(.*)} $row -> Fnum url]} {
                     set pls($Fnum,url) $url
                  }
                  if {[regexp {Title([0-9]+)=(.*)} $row -> Tnum title]} {
                     set title [string trim [file root $title] .]
                     set pls($Tnum,title) $title
                  }
               } ;# end foreach
               set i 1
               foreach {index} [lsort -dictionary [array names pls *,url]] {
                  foreach {num type} [split $index ,] {break} ;# end foreach
                  if {[info exists pls($num,title)]} {
                     set title $pls($num,title)
                  } else {
                     set title $fileTail.$num
                  }
                  set fname [playurl::addUrl $::snackAmpSettings(uskDir) $title $pls($index)]
                  if {$i==1} {
                     set end [LB::end]
                  }
                  filtering::Append $fname
                  incr i
               }
               if {[info exists end]} {
                  LB::inverseVideoFile xx 0
                  LB::activate $end
                  updateFileDisplay
                  Play active
               }

            }
         }
      }
      default {
         ;# usk shortcut files also handled here
         if {[file readable $file]} {
           filtering::InsertPlay $file
         } else {
            tk_messageBox -icon warning -type ok -message "File '$file' cannot be read"
         }
      }
   } ;# End Switch
}
###############################################################################
# Function   : buildGui
# Description: Build the player Gui
# Author     : Tom Wilkason
# Date       : 12/27/2000
###############################################################################
proc buildGui {args} {
   global TA TF LB SS CF
   set ::songArray(Title) "SnackAmp Music Player"
   set TA(songTitle) "SnackAmp Music Player"
   set TA(songBalloon) "Track Details"

   ;##
   ;# Hide the toplevel and move it off the visible screen
   ;#
   toplevel .saTop -class SnackAmp -cursor $::TA(arrow) -borderwidth 0
   set ::SA .saTop
   wm title $::SA "SnackAmp Music File Player"
   ;##
   ;# clean up nicely
   ;#
   wm protocol $::SA WM_DELETE_WINDOW {taExit}
   ###InstallMenu
   ;# Init the volumn linkages
   foreach {line} [snack::mixer lines] {
      set ::TA(line,$line) [snack::mixer volume $line]
      if {[snack::mixer channels $line] == "Mono"} {
         snack::mixer volume $line ::TA(line,$line)
      } else {
         snack::mixer volume $line ::TA(line,$line) ::TA(line,$line)
      }

   } ;# end foreach
   wm withdraw $::SA
   ;##
   ;# Control Panel
   ;#
   pack [frame $::SA.panel -borderwidth 2 -relief groove] -side top -fill x -expand false -anchor w
   frame $::SA.panel.stat -relief sunken -borderwidth 2

   ;##
   ;# Main area for music details
   ;#
   frame $::SA.panel.stat.l -background $::snackAmpSettings(bg) \
      -borderwidth 0 -width 70
   settings::regColor $::SA.panel.stat.l -background ::snackAmpSettings(bg)
   layoutPlayInfo $::snackAmpSettings(showDetail) ::songArray $::SA.panel.stat.l
   ;# Trace the showDetail settings to update the display detail immediatly from the settings menu
   trace variable ::snackAmpSettings(showDetail) w [list reLayoutPlayInfo $::SA.panel.stat.l ::songArray]


   ;##
   ;# Time counter
   ;#
   label $::SA.panel.stat.t -background $::snackAmpSettings(bg) -foreground $::snackAmpSettings(sfg) -borderwidth 0 \
   -font timeStatus -height 1 -textvar TA(timeStatus) -anchor e
   settings::regColor $::SA.panel.stat.t -background ::snackAmpSettings(bg) -foreground ::snackAmpSettings(sfg)
   balloon::define $::SA.panel.stat.t "Time Playing\nTime Remaining\nTotal Time"

   ;##
   ;# Click in title area toggles muting
   ;#
   set TA(muteFactor) 1.0
   bind $::SA.panel.stat.l.dtTitle <ButtonPress-3> {EditPlaying}

   ;##
   ;# Level Meter, on the left
   ;#
   frame $::SA.panel.stat.meter -background $::snackAmpSettings(bg) \
      -borderwidth 0
   settings::regColor $::SA.panel.stat.meter -background ::snackAmpSettings(bg)
   if {$::snackAmpSettings(levelMeter)} {
      after idle showLevelMeter
   }

   ;# Place the labels next to each other
   grid $::SA.panel.stat.meter -column 0 -row 0 -rowspan 1 -sticky  news    ;# Meter
   grid $::SA.panel.stat.l -column 1 -row 0 -sticky nsew                ;# Song Title
   grid $::SA.panel.stat.t -column 2 -row 0 -rowspan 1 -sticky nsew     ;# Time Status

   grid $::SA.panel.stat -row 0 -column 1 -columnspan 2 -sticky nsew;# Inner Grid
   ;# Player controls
   ;#
   snack::createIcons
   grid [frame $::SA.panel.player] -row 1 -column 1 -sticky news

   ;##
   ;# Button data
   ;#
   set buttonData {
      bitmap snackPlayPrev   Prev  "Play previous in list"
      bitmap snackPlay       {set ::TA(STOP) 0; PlayResume}  "Play Selection"
      bitmap snackPause      Pause "Pause/Resume playback"
      bitmap snackStop       {set ::TA(STOP) 1;Stop}  "Stop playing"
      bitmap snackPlayNext   {set ::TA(STOP) 0;Next}  "Play next in list"
      image  openFile        OpenFile  "Add a file to play list and play it immediately"
      image  treeImage       AutoPlaylist "Manage your music collection and define Auto-playlists"
   }
#      image  sortImage       {set snackAmpSettings(randomPlay) 0;filtering::sortMusic} "Sort the tracks by title"
#      image  shuffleImage    {set snackAmpSettings(randomPlay) 1;filtering::sortMusic} "Shuffle the tracks randomly"
   ;# build each one
   set tf $::SA.panel.player
   foreach {bmType File Command Help} $buttonData {
      button  $tf.$File -$bmType $File -command $Command -borderwidth 1
      pack $tf.$File -side left -padx 0 -pady 0
      balloon::button $tf.$File $Help
   } ;# end foreach
   ;# Put this button before tree button
   pack [menubutton $tf.sortMenu -image shuffleImage -menu $tf.sortMenu.drop] -before $tf.$File -side left
   balloon::button $tf.sortMenu  "Select sort mode"
   set mb $tf.sortMenu.drop
   ;# History menu, one for button and one for RMB
   menu $mb -tearoff 0 -font menu
     # 0 sortImage    "Sort by all Tracks"               {set snackAmpSettings(sortMode) 0;filtering::sortMusic}
   foreach {val img txt cmd} {
     0 sortImage    "Sort by all Tracks"               {filtering::sortMusic}
     1 shuffleImage "Shuffle by all Tracks"            {filtering::sortMusic}
     2 shuffleImage "Shuffle by Folder, Sort by Track"    {filtering::sortMusic}
     3 shuffleImage "Shuffle by Folder, Shuffle by Track" {filtering::sortMusic}
   } {
      $mb add radiobutton -label $txt \
         -value $val -variable ::snackAmpSettings(sortMode) \
        -command $cmd ;#-image $img -compound left
   } ;# end foreach

   set TA(treeButton) $tf.treeImage
   set TA(sortButton) $tf.sortImage
   set TA(shuffleButton) $tf.shuffleImage
   set TA(pauseButton) $tf.snackPause
   ;##
   ;# Scale to show/control Offset within file
   ;#
   if {$::TA(showTicks)} {
      set ti 20
   } else {
      set ti 0
   }
   set SS [scale $::SA.panel.ss -show no -orient horiz -width 8 -len 200 -from 0. -to 60. \
      -var TA(p) -tickinterval $ti -sliderlength 15 -resolution 1 \
      -highlightcolor $::snackAmpSettings(bg) \
      -troughcolor $::snackAmpSettings(bg) -command updateTime] ;#-cursor sb_h_double_arrow
   settings::regColor $::SS -troughcolor ::snackAmpSettings(bg) \
      -highlightcolor $::snackAmpSettings(bg)

   balloon::define $SS "Show/Adjust playback location within file"
   grid $SS -column 2 -row 1 -sticky ew

   ;# Allow time to be adjusted for all the buttons
   set TA(setdrag) 1
   bind $SS <ButtonPress-1> {set TA(setdrag) 0}
   bind $SS <ButtonRelease-1> {set TA(setdrag) 1; PlayOffset}
   bind $SS <ButtonPress-2> {set TA(setdrag) 0}
   bind $SS <ButtonRelease-2> {set TA(setdrag) 1; PlayOffset}
   bind $SS <ButtonPress-3> {set TA(setdrag) 0}
   bind $SS <ButtonRelease-3> {set TA(setdrag) 1; PlayOffset}
   ;##
   ;# Scale show/adjust Sound level
   ;#
   set TA(gain) [macros::getPlayGain]

   scale $::SA.panel.sv -show no -orient vert -width 10 -command {PlayGain}\
         -len 25 -var TA(gain) -from 64. -to 0. -sliderlength 15 \
         -highlightcolor $::snackAmpSettings(bg) \
         -troughcolor $::snackAmpSettings(bg)  ;#-cursor sb_v_double_arrow
   settings::regColor $::SA.panel.sv -troughcolor ::snackAmpSettings(bg) \
      -highlightcolor $::snackAmpSettings(bg)
   grid $::SA.panel.sv -column 0 -row 0 -rowspan 3 -sticky ns
   balloon::define $::SA.panel.sv "Adjust the playback level"
   ;##
   ;# Mousewheel adjust volume from toplevel
   ;#
   bind $::SA.panel.sv <Any-Enter> {
      set TA(pushFocus) [focus]
      focus %W
   }
   bind $::SA.panel.sv <Any-Leave> {
      if {[string equal %W $TA(pushFocus)]} {
         focus [winfo toplevel %W]
      } else {
         if {[catch {focus $TA(pushFocus)}]} {
            focus [winfo toplevel %W]
         }
      } ;# end if
   }
   if {$::tcl_platform(platform) != "windows"} {
      bind $::SA.panel.sv <Button-5> {bumpLevel -2}
      bind $::SA.panel.sv <Button-4> {bumpLevel 2}
   } else {
      bind $::SA.panel.sv <MouseWheel> {
         set amount [expr {%D/abs(%D)*1}] ;# get +/- 2
         bumpLevel $amount
         break
      }
   }


   # Adjust grid resize behavior and make things span properly
   grid columnconfigure $::SA.panel.stat 0 -weight 0
   grid columnconfigure $::SA.panel.stat 1 -weight 1            ;# For the label to stretch
   grid columnconfigure $::SA.panel 0 -weight 0
   grid columnconfigure $::SA.panel 1 -weight 0
   grid columnconfigure $::SA.panel 2 -weight 1


   ;# Collapsable frame for playlist
   set CF [cframe $::SA.hull "Music Files"]
   pack $::SA.hull -side top -fill both -expand true
   cframe.close $CF

   ;##
   ;# Filter box
   pack [frame $CF.filter -borderwidth 2 -relief groove] -side top -expand no -fill x
   pack [button $CF.filter.label -text "Apply Filter" -command {
      filtering::pushFilter
      macros::filterMusic replace $::TA(matchAgainstAll)
   } -borderwidth 1] \
      -side left -expand no -fill x
   balloon::button $CF.filter.label "Replace files matching filter to current playlist"

   ;##
   ;# Filter entry
   pack [entry $CF.filter.entry -textvariable TA(filter) -width 20] -side left -expand yes -fill x
   set ::FI $CF.filter.entry
   $::FI config -background $::snackAmpSettings(ebg) -foreground $::snackAmpSettings(efg)
   settings::regColor $::FI -background ::snackAmpSettings(ebg) -foreground ::snackAmpSettings(efg)
   balloon::define $::FI "Enter text to restrict files to show in playlist, Buttons to left or right to apply filter. Double click to clear. You can use 'or' for boolean matches. Use * or ? for wildcards."
   ;##
   ;# Dropdown button
   pack [menubutton $CF.filter.history -image dropdownImage -menu $CF.filter.history.list] -side left
   balloon::button $CF.filter.history  "Options to retrieve prior filter entries or apply current entries against the current auto-playlist files (APL) or against all of the known files"
   ;# History menu, one for button and one for RMB
   menu $CF.filter.history.list -postcommand [list filtering::filterHistory $CF.filter.history.list]
   menu $::FI.popup -postcommand [list filtering::filterHistory $CF.filter.entry.popup]
   bind $::FI <ButtonPress-3> {
      tk_popup %W.popup %X %Y
   }

   ;##
   ;# Filter against all or just auto playlist
   set TA(filterAllButton) $CF.filter.doall
   pack [button $CF.filter.doall -image filterOff \
      -command filtering::toggleFilter -borderwidth 1] \
      -side left -expand no -fill x
   balloon::button $CF.filter.doall "Push this to have filter applied to all files (by default), not just ones in the current auto-playlist"

   ;##
   ;# filtering::Append/Replace buttons for filter
   pack [button $CF.filter.btApp -image appendImage -command {
      filtering::pushFilter
      macros::filterMusic append $::TA(matchAgainstAll)
   } -borderwidth 1] -side left
   balloon::button $CF.filter.btApp "Append files matching filter to current playlist"

   pack [button $CF.filter.btRep -image replaceImage -command {
      filtering::pushFilter
      macros::filterMusic replace $::TA(matchAgainstAll)
   } -borderwidth 1] -side left
   balloon::button $CF.filter.btRep "Replace files matching filter to current playlist"
   ;##
   ;# Find button
   ;#
   pack [button $CF.filter.btFind -image findImage -command {
      findNextSong
   } -borderwidth 1] -side left
   balloon::button $CF.filter.btFind "Find the next song in the playlist that matches the Filter"

   ;##
   ;# Scrolled Listbox
   ;#
   set LB [Scrolled_Listbox $CF.frame -selectmode extended -exportselection no \
      -height $::snackAmpSettings(playListLines) \
      -foreground $::snackAmpSettings(fg) \
      -selectforeground $::snackAmpSettings(fg) \
      -selectbackground $::snackAmpSettings(highlight) \
      -background $::snackAmpSettings(bg)]
   # Listbox popup menu
   settings::regColor $LB  -foreground ::snackAmpSettings(fg) \
   -selectbackground ::snackAmpSettings(highlight) \
   -background ::snackAmpSettings(bg)

   ;##
   ;# Default then User Key Assigments and menus
   ;#
   bindings::defaultHotKeys
   bindings::assignHotKeys
   InstallMenu
   InstallLBMenu $LB
   ;# The songlist packs last, so it is the one that contracts the most
   pack $CF.frame -expand yes -fill both -side top
   fixMenuDetail $::snackAmpSettings(showDetail)
   ;##
   ;# Callbacks to immediatly set the fonts when the user changes them
   ;#
   settings::installFontCallbacks
}
###############################################################################
# Function   : showLevelMeter
# Description: Insert the level meter at the appropriate height
# Author     : Tom Wilkason
# Date       : 3/6/2003
###############################################################################
proc showLevelMeter {args} {
   update
   set height [winfo reqheight $::SA.panel.stat]
   levelmeter::create $::SA.panel.stat.meter.lm -width 8 -length $height -orient vertical -oncolor lightblue
   pack $::SA.panel.stat.meter.lm -side bottom
   ;# Sound object in use is called s
   after 1000 [list levelmeter::cmUpdate s]
}

###############################################################################
# Function   : reLayoutPlayInfo
# Description: On a write to the showDetail variable, update the amount
#              of displayed info on the main gui
# Author     : Tom Wilkason
# Date       : 12/18/2003
###############################################################################
proc reLayoutPlayInfo {hull sArray _G I mode} {
   upvar #0 $_G G
   set level $G($I)
   foreach child [winfo children $hull] {
      destroy $child
   } ;# end foreach
   layoutPlayInfo $level $sArray $hull
   fixMenuDetail $level
   updateTime
   wm geometry .saTop {}
}
;##
;# Place the top level menu in the appropriate place
;#
proc fixMenuDetail {level} {
   ;##
   ;# Move menu bar to pop-up for very small layout
   ;#
   set tlm .saTop.bRmenubar ;# hack
   set tm $::SA.panel.stat.l.dtTitle ;# hack
   if {$level > 1} {
      if {[string length [$::SA cget -menu]]==0} {
         $::SA configure -menu $tlm
      }
      bind $tm <ButtonPress-3> {
         EditPlaying
      }
   } else {
      if {[string length [$::SA cget -menu]]!=0} {
         $::SA config -menu {}
      }
      bind $tm <ButtonPress-3> {
         tk_popup .saTop.bRmenubar %X %Y
      }
   }
}
###############################################################################
# Function   : layoutPlayInfo
# Description: Layout the contents of the main window
# Author     : Tom Wilkason
# Date       : 1/12/2002
###############################################################################
proc layoutPlayInfo {threshold _Var hull} {
   # Label Variable Row Col ColSpan Font
   set color(songLabel) tfg
   set color(songInfo) fg
   set gridData {
      1 Title Title     0  0  6  songLabel
      2 Artist Artist   1  0  6  songInfo
      2 Album Album     2  0  6  songInfo
      4 Size  fSize     3  0  1  songInfo
      4 Freq  Sample    3  2  1  songInfo
      4 Genre Genre     3  4  1  songInfo
      3 Date  fDate     4  0  1  songInfo
      3 Rate  rate      4  2  1  songInfo
      3 Track Track     4  4  1  songInfo
   }
   foreach {Use Title Field sRow sCol cSpan font} $gridData {
      ;# label
      if {$threshold >= $Use} {
         label $hull.lb$Title -text $Title  -font smallLabel -pady 0 \
            -background $::snackAmpSettings(bg) \
            -foreground $::snackAmpSettings(lfg) \
            -anchor e
         settings::regColor $hull.lb$Title \
            -background ::snackAmpSettings(bg) \
            -foreground ::snackAmpSettings(lfg)
         ;# Song info
         label $hull.dt$Title -textvariable ${_Var}($Field) -font $font -pady 0  -bd 0 \
            -background $::snackAmpSettings(bg) \
            -foreground $::snackAmpSettings($color($font)) \
            -anchor w -bd 0
         settings::regColor $hull.dt$Title \
            -background ::snackAmpSettings(bg) \
            -foreground ::snackAmpSettings($color($font))
         ;# If only title, then don't bother with label
         if {$threshold > 1} {
            grid $hull.lb$Title -row $sRow -column $sCol -sticky e -pady 0
         } ;# end if
         grid $hull.dt$Title -row $sRow -column [incr sCol] -columnspan $cSpan -sticky w -pady 0
      }
   } ;# end foreach
   grid columnconfigure $hull 1 -weight 1
}

###############################################################################
# Function   : InstallMenu
# Description: Build the menu bar at the top of the GUI
# Author     : Tom Wilkason
###############################################################################
proc InstallMenu {} {
   global TF TA LB snackAmpSettings tcl_platform
   bindings::init $::SA menuBindings
   set M $::SA.bRmenubar
   ;# Make menus as small as possible
   catch {menu $M -tearoff 0}
   # attach it to the main window
   $::SA config -menu $M
   # Create more cascade menus
   foreach m {File Edit Playlists "AutoPlaylists" Settings Help} {
      if {$m == "Playlists"} {
         set $m [menu $M.m$m -tearoff 0 -postcommand [list playListFiles $M.m$m] -font menu]
         set TA(plPopUp) $M.m$m
         $M.m$m add command -label "Clear" -command {Clear}
         $M.m$m add command -label "Open" -command {macros::ReadPlaylistFile {} replace}
         $M.m$m add command -label "Save all" -command {macros::SavePlaylistFile $::LB replace all $::TA(PlayListFile)}
         $M.m$m add command -label "Save all as" -command {macros::SavePlaylistFile $::LB replace all}
         $M.m$m add command -label "Save selected as" -command {macros::SavePlaylistFile $::LB replace selected}
         $M.m$m add command -label "Append all as" -command {macros::SavePlaylistFile $::LB append all}
         $M.m$m add separator
         $M.m$m add cascade -label "Save selected to..." -menu [menu $M.m$m.uReplace -tearoff 0 -font menu]
         $M.m$m add cascade -label "Append selected to..." -menu [menu $M.m$m.uAppend -tearoff 1 -font menu]
         $M.m$m add separator
         $M.m$m add cascade -label "Append from..." -menu [menu $M.m$m.mAppend -tearoff 0 -font menu]
         $M.m$m add cascade -label "Insert from..." -menu [menu $M.m$m.mInsert -tearoff 0 -font menu]
         $M.m$m add cascade -label "Replace from..." -menu [menu $M.m$m.mReplace -tearoff 0 -font menu]
         $M.m$m add separator
         set plrem [menu $M.m$m.remove -tearoff 0 -postcommand "[list PLFilesRemove $M.m$m.remove]"  -font menu]
         $M.m$m add cascade -label "Remove..." -menu $plrem

      } elseif {$m == "AutoPlaylists"} {
         set TA(FolderMenu) $m
         set $m [menu $M.m$m -tearoff 1 -postcommand [list APLFiles $M.m$m] -font menu]
         set Removal [menu $M.m$m.remove -tearoff 0 -postcommand "[list APLFilesRemove $M.m$m.remove]"  -font menu]

      } else {
         set $m [menu $M.m$m -tearoff 0 -font menu]
      }
      $M add cascade -label $m -menu $M.m$m
   }
   ;##
   ;# Here is how to add a menu to the "system" menu
   ;# Use it to put "pinned" command
   ;#
   ontop::pinOnSystemMenu $M $::SA
   # FILE

   bindings::addCommand $File "Open File" {OpenFile}
   bindings::addCommand $File "Open URL" {OpenURL}
   $File add separator
   bindings::addCommand $File "Catalog Folders" {catalog::RescanAskFirst}
   bindings::addCommand $File "Read Catalogs" {catalog::ReReadCatalogs}
   $File add separator
   bindings::addCommand $File "Catalog Removable Media" catalog::CatalogMedia
   bindings::addCommand $File "Mount Removable Media"  {catalog::MountMedia "Removable"}
   $File add separator
   bindings::addCommand $File "Manage Media Folders..."  {catalog::manageMediaFolders}
   bindings::addCommand $File "Manage Removable Folders..." {
      package require saSettings
      folderManager::launch RemovableFolders
   }
   bindings::addCommand $File "Manage Folders to Monitor for Changes..." {
      package require saSettings
      folderManager::launch IncomingDirs
   }
   $File add separator
   bindings::addCommand $File "Create Music Listing" {catalog::documentCache}
   bindings::addCommand $File "Open a file for playing" {OpenFile}
   bindings::addCommand $File "Save Track As ..." {catalog::saveAs}
   $File add separator
   bindings::addCommand $File "Exit" {taExit}

   # EDIT

   bindings::addCommand $Edit "Select All" {$::LB selection set 0 end;focus $LB}
   $Edit add separator
   bindings::addCommand $Edit "Sync APL to Selected Track" {SelectFile 0}
   bindings::addCommand $Edit "Sync APL to Playing Track" {SelectFile 1}
   bindings::addCommand $Edit "Locate Playing Track" {locatePlaying}
   $Edit add separator
   bindings::addCommand $Edit "Edit ID3 for Selected Track" {EditPLID3}
   bindings::addCommand $Edit "Edit ID3 for Playing Track" {EditPlaying}
   $Edit add separator
   bindings::addCommand $Edit "Rename Selected Track" {RenameFile [LB::getActiveFullName]}
   bindings::addCommand $Edit "Rename Playing Track" {RenameFileDialog [soundControl::fileName]}
   $Edit add separator
   $Edit add command -label "Copy Selected to Clipboard" \
      -accelerator $::TA(key,copy) \
      -command {clipboardCopy $::LB 0}
   $Edit add command -label "Cut Selected to Clipboard" \
      -accelerator $::TA(key,cut) \
      -command {clipboardCut $::LB 0}
   $Edit add command -label "Paste Clipboard" \
      -accelerator $::TA(key,paste) \
      -command {clipboardPaste $::LB 0}
   $Edit add separator
#   $Edit add command -label "Reset Playing Track Auto Gain"  -command {
#      gainControl::resetSome [list [soundControl::fileName]]
#      snack::audio scaling 1.0
#   }
   $Edit add command -label "Delete Selection" -command {Delete}


   bindings::addCommand $AutoPlaylists "Manage Auto-Playlists" {AutoPlaylist}
   $AutoPlaylists add cascade -label "Remove..." -menu $Removal
   ;# bug, below is deleted by callback (even though it should not be)
   $AutoPlaylists add separator

   # SETTINGS
   if {$snackAmpSettings(enableConsole)} {
      $Settings add check -label "Show Console" -variable TA(Console) -command {
         if {[info command console] == {}} {
            catch {package require saConsole}
         }
         Console
      }
   } ;# end if
   $Settings add check -label "Show Mini Player" -variable ::snackAmpSettings(activeMiniPlayer) -command {
      package require miniPlayer
      toggleMiniPlayer
   }
   $Settings add separator
   if {[llength [snack::mixer lines]] > 1} {
      $Settings add cascade -label "Volume Controls..." -menu [menu $Settings.volume -tearoff 0 -font menu]
         foreach {line} [snack::mixer lines] {
            $Settings.volume add radiobutton -label $line -variable snackAmpSettings(lineSelect) -value $line
         } ;# end foreach
   } ;# end if
   ;# Force select the first (default) output device, fixes bug
   ;# on Linux where write errors occur randomly
   if {! [info exists ::snackAmpSettings(outputDevice)]} {
      set snackAmpSettings(outputDevice) [lindex [snack::audio outputDevices] 0]
   } ;# end if
   snack::audio selectOutput $::snackAmpSettings(outputDevice)
   $Settings add cascade -label "Output Device..." -menu [menu $Settings.output -tearoff 0 -font menu]
      foreach {device} [snack::audio outputDevices] {
         $Settings.output add radiobutton -label $device -variable snackAmpSettings(outputDevice) -value $device \
            -command [list snack::audio selectOutput $::snackAmpSettings(outputDevice)]
      } ;# end foreach
   $Settings add cascade -label "Minimum Bitrate..." -menu [menu $Settings.bitrate -tearoff 0 -font menu]
      set ::TA(minRate) 0
      foreach {rate} [list 0 24 64 96 128 140 160 180 192 256 320] {
         $Settings.bitrate add radiobutton -label $rate -variable ::TA(minRate) -value $rate
      } ;# end foreach
   if {[string length [info command ntSetPriority]] > 0} {
      $Settings add cascade -label "Runtime Priority..." -menu [menu $Settings.priority -tearoff 0 -font menu]
      ntSetPriority $::snackAmpSettings(priority)
      foreach {priority} [list high normal low] {
         $Settings.priority add radiobutton -label $priority \
            -variable ::snackAmpSettings(priority) -value $priority \
            -command {after 0 ntSetPriority $::snackAmpSettings(priority)}
      } ;# end foreach
   } ;# end if
   $Settings add separator
   #$Settings add check -label "Continuous Play" -variable snackAmpSettings(ContinuousPlay)
   #$Settings add check -label "Shuffle Play" -variable snackAmpSettings(randomPlay) \
   #   -command filtering::sortMusic
   #$Settings add separator
   $Settings add command -label "Manage Media Folders..."  -command {catalog::manageMediaFolders}
   $Settings add command -label "Manage Removable Folders..."  -command {
      package require saSettings
      folderManager::launch RemovableFolders
   }
   $Settings add command -label "Manage Folders to Monitor for Changes..."  -command {
      package require saSettings
      folderManager::launch IncomingDirs
   }
   $Settings add separator
   $Settings add command -label "Other Settings ..." -command {package require saSettings;settings::tabbedSettingsGui}

   ;# HELP
   if {$::snackAmpSettings(EnableWebServer)==1} {
      bindings::addCommand $Help "User Manual" {::browser::help}
      bindings::addCommand $Help "Internal Web Server Home" {::browser::urlOpen $snackAmpSettings(internalWebBase)/}
   } else {
      bindings::addCommand $Help "User Manual" {::browser::urlOpen "http://snackamp.sourceforge.net/user_manual/index.html"}
   }
   bindings::addCommand $Help "SnackAmp Home Page at SourceForge" {::browser::urlOpen "http://snackamp.sourceforge.net/"}
   $Help add separator
   bindings::addCommand $Help "Versions"  cvsVersions
   bindings::addCommand $Help "About"  about
}
###############################################################################
# Function   : InstallLBMenu
# Description: Install Listbox Menus and hard coded bindings
# Author     : Tom Wilkason
# Date       : 6/18/2002
###############################################################################
proc InstallLBMenu {LB} {
   set m $LB.popup
   bindings::init $LB menuBindings
   menu $m -tearoff 0 -font menu
   bindings::addCommand $m "Play Selection" {Play}
   $m add separator
   $m add command -label "Copy"  -accelerator $::TA(key,copy) -command {clipboardCopy $::LB 0}
   $m add command -label "Cut"  -accelerator $::TA(key,cut) -command {clipboardCut $::LB 0}
   $m add command -label "Paste"  -accelerator $::TA(key,paste) -command {clipboardPaste $::LB 0}
   $m add separator
   $m add command -label "Sort Selection"  -command {sortSelection ::LB}
   $m add command -label "Shuffle Selection"  -command {shuffleSelection ::LB}
   $m add separator
   $m add command -label "Remove Selection"  -accelerator "Delete" -command {Cut 1}
   $m add command -label "Clear Entire List"  -command {
      LB::deleteAll
      updateFileDisplay
   }
   $m add separator
   $m add command -label "Delete File(s)"  -command {macros::Delete}
   $m add command -label "Rename File"  -accelerator "F2" -command {macros::RenameFile}
   $m add separator
   $m add command -label "Edit ID3 Info"  -accelerator "F3" \
      -command {macros::EditPLID3}
   $m add command -label "ID3 Table Editor" -accelerator "F4" \
      -command {macros::PLNameToTag 0}
   $m add command -label "ID3 Tag To Name" -accelerator "F6" \
      -command {macros::PLTagToName}
#   $m add command -label "Remove ID3 Tags"  -command {macros::removeTag}
   $m add separator
   $m add command -label "Reset Auto Level"  -command {gainControl::resetSome}
   $m add separator
   $m add cascade -label "Playlist Tracks" -menu $::TA(plPopUp)
#   $m add command -label "Save Selected to Playlist File"  -command {macros::SavePlaylistFile $::LB replace selected}
#   $m add command -label "Append Selected to Playlist File"  -command {macros::SavePlaylistFile $::LB append selected}
   bind $LB <ButtonPress-3> {
      tk_popup %W.popup %X %Y
   }
   # Bindings
   balloon::plDefine
   bind $LB <Double-ButtonPress-1> {
      set here [$LB nearest %y]
      LB::activate $here
      Play active
   }

   bind $LB <B1-Motion> {
      LB::Drag %y
   }
   bind $LB <ButtonPress-1> {
      focus %W
      LB::Select %y
   }
}
###############################################################################
# Function   : playListFiles
# Description: postcommand callback to show play list files on menu, called
#              from play list menu postcommand
# Author     : Tom Wilkason
###############################################################################
proc playListFiles {menu} {
   global TF TA LB
   set Files [catalog::getPlaylists]
   if {[llength $Files] == 0} {
      return
   }
   $menu.mAppend delete 0 end
   $menu.mReplace delete 0 end
   $menu.mInsert delete 0 end
   $menu.uAppend delete 0 end
   $menu.uReplace delete 0 end
   ;##
   ;# Need to sort the list by decending date
   ;#
   foreach {File} $Files {
      set Value [file rootname [file tail $File]]
      $menu.mAppend add radiobutton -label $Value -value $Value -variable ::TA(PlayListSelection) \
         -command [list macros::ReadPlaylistFile $File "append"]
      $menu.mInsert add radiobutton -label $Value -value $Value -variable ::TA(PlayListSelection) \
         -command [list macros::ReadPlaylistFile $File "insert"]
      $menu.mReplace add radiobutton -label $Value -value $Value -variable ::TA(PlayListSelection) \
         -command [list macros::ReadPlaylistFile $File "replace"]

      $menu.uAppend add radiobutton -label $Value -value $Value -variable ::TA(PlayListSelection) \
         -command [list macros::SavePlaylistFile $::LB "append" "selected" $File]
      $menu.uReplace add radiobutton -label $Value -value $Value -variable ::TA(PlayListSelection) \
         -command [list macros::SavePlaylistFile $::LB "replace" "selected" $File]
   }
}
###############################################################################
# Function   : APLFiles
# Description: post command Callback to show auto playlist setting files, called
#              from auto-play list menu postcommand
# Author     : Tom Wilkason
###############################################################################
proc APLFiles {menu} {
   global TF TA LB
   set Files [catalog::getAutoPlaylists]
   if {[llength $Files] == 0} {
      return
   }
   $menu delete 3 end
   $menu add separator
   foreach {File} $Files {
      set Value [file rootname [file tail $File]]
      $menu add radiobutton -label $Value -value $Value \
         -variable TA(AutoPlayListSelection) -command [list readAPLFile $File 1]
   }
}
###############################################################################
# Function   : APLFilesRemove
# Description: post command Callback to allow removal of auto playlist setting files
#              from auto-play list menu postcommand
# Author     : Tom Wilkason
###############################################################################
proc APLFilesRemove {menu} {
   global TF TA LB
   set Files [catalog::getAutoPlaylists]
   if {[llength $Files] == 0} {
      return
   }
   $menu delete 0 end
   foreach {File} $Files {
      set Value [file rootname [file tail $File]]
      $menu add radiobutton -label $Value -value $Value \
         -command [list removeAPLFile $File]
   }
}

###############################################################################
# Function   : PLFilesRemove
# Description: post command Callback to allow removal of auto playlist setting files
#              from auto-play list menu postcommand
# Author     : Tom Wilkason
###############################################################################
proc PLFilesRemove {menu} {
   global TF TA LB
   set Files [catalog::getPlaylists]
   if {[llength $Files]==0} {
      return
   }
   $menu delete 0 end
   foreach {File} $Files {
      set Value [file rootname [file tail $File]]
      $menu add radiobutton -label $Value -value $Value \
         -command [list removePLFile $File]
   }
}
###############################################################################
# Function   : setTeDefault
# Description: Fixup a tree of teFolders that may be undetermined
#              Work backwards and set the childrens value to that
#              of the parent all the way to the root. Default to 0
# Author     : Tom Wilkason
# Date       : 3/24/2001
###############################################################################
proc setTeDefault {Index} {
   global teFolders
   switch -- $teFolders($Index) {
      -1 {
         set parent [file dirname $Index]
         if {$parent != $Index} {
            if {[info exists teFolders($parent)]} {
               set teFolders($Index) [setTeDefault $parent]
            } else {
              set teFolders($Index) 0
            }
         } else {
            set teFolders($Index) 0
         }
      }
      1 {
         set parent [file dirname $Index]
         if {($parent != $Index) && [info exists teFolders($parent)] && ($teFolders($parent)==0)} {
            set teFolders($parent) 2
         }
      }
      2 {
        set teFolders($Index) 0
      }
      default {}
   } ;# End Switch
   return $teFolders($Index)
}
###############################################################################
# Function   : getNextCommand
# Description: Return the approprite play command based on settings
# Author     : Tom Wilkason
# Date       : 4/6/2002
###############################################################################
proc getNextCommand {args} {
   ;# See if we want to repeat this song
   switch -- $::snackAmpSettings(SongPlayMode) {
      0 {set nextCommand [list after 0 Play next]}
      1 {set nextCommand {}}
      2 {set nextCommand [list after 0 Play current]}
      default {{}}
   } ;# End Switch
   return $nextCommand
}
;##
;# Play a music file at some offset, likely bound to slider control
;#
proc PlayOffset {{manual 0}} {
   Trace
   global TF TA LB snackAmpSettings
   ;# Make sure we have an open file
   set soundLength [soundControl::length]
   if {$soundLength > 0} {
      ;# Some files are fast and some are slow to respond to -start = 0
      if {$manual == 0} {
         set offset [expr {int(double($TA(p))/$TA(songDuration)*$soundLength)}]
      } else {
         set offset $manual
      }
      if {[catch {
         soundControl::stop
         #debug "Playing at $offset for [soundControl::position] $soundLength"
         LB::inverseVideoFile $TA(currentMF) 1
         ;# Workaround for bug on linux/ogg where file plays from last pos, not from -start
         s config -file [s cget -file]
         soundControl::play -start $offset -command [getNextCommand]
         #soundControl::play -start $offset -end [expr {[soundControl::length]-1}] -command [getNextCommand]
      } result] \
      } then {
         saLog "PlayOffset: $result"
         return
      }
      if {![afterExists Timer]} {
         Timer
      } ;# end if
   } ;# end if
}
###############################################################################
# Function   : updateTime
# Description: When Slider is updated, update time display, this is called
#              automatically when the time slider is updated during play.
# Author     : Tom Wilkason
# Date       : 1/9/2001
###############################################################################
proc updateTime {args} {
   global TA
   set elapsed   [expr {round($TA(p))}]
   set remaining [expr {round($TA(songDuration) - $elapsed)}]
   switch -- $::snackAmpSettings(showDetail) {
      1 {
         if {$TA(songDuration) < 3600} {
            set TA(timeStatus) "[clock format $elapsed -format %M:%S]/[clock format $remaining -format %M:%S]"
         } else {
            set TA(timeStatus) "[clock format $elapsed -format %T -gmt 1]/[clock format $remaining -format %T -gmt 1]"
         }
      }
      2 {
         if {$TA(songDuration) < 3600} {
            set TA(timeStatus) "[clock format $elapsed -format %M:%S]\n[clock format $remaining -format %M:%S]"
         } else {
            set TA(timeStatus) "[clock format $elapsed -format %T -gmt 1]\n[clock format $remaining -format %T -gmt 1]"
         }
      }
      default {
         if {$TA(songDuration) < 3600} {
            set TA(timeStatus) "[clock format $elapsed -format %M:%S]\n[clock format $remaining -format %M:%S]\n[clock format [expr {round($TA(songDuration))}] -format %M:%S]"
         } else {
            set TA(timeStatus) "[clock format $elapsed -format %T -gmt 1]\n[clock format $remaining -format %T -gmt 1]\n[clock format [expr {round($TA(songDuration))}] -format %T -gmt 1]"
         }
      }
   } ;# End Switch
}
###############################################################################
# Function   : Timer
# Description: Timer to update the display data, active while file playing
#              Ramp the level down the final two seconds to handle truncated songs.
###############################################################################
proc Timer {args} {
   global TF TA LB SS
   # Time is curren pos/total len * total time
   if {$TA(setdrag)} {
      set elapsed   [soundControl::position]
#      if {$::snackAmpSettings(quietEnd)} {
#         set line $::snackAmpSettings(lineSelect)
#         if {($TA(songDuration) - $elapsed) < $::snackAmpSettings(fadeTime)} {
#            soundControl::fadeFast
#            $SS set $elapsed
#            ;# Song will continue at no volume, timer is not retriggered
#            return
#         }
#      } ;# end if
      ;# If user dragging slider, don't also try to set slider
      catch {
         $SS set $elapsed
      }
   }
   after 500 Timer
}
###############################################################################
# Function   : sortSelection
# Description: Sort the selection
# Author     : Tom Wilkason
# Date       : 12/4/2001
###############################################################################
proc sortSelection {{_listBox ::LB}} {
   upvar $_listBox lb
   global TF TA
   set selection [lsort -decreasing -integer [$lb curselection]]
   set current [$lb index anchor]
   if {[llength $selection] > 1} {
      ;##
      ;# Cut the data out and sort it
      ;#
      foreach {ind} $selection {
         lappend data [$lb get $ind]
      } ;# end foreach
      $lb delete [lindex $selection end] [lindex $selection 0]
      set data [lsort -decreasing -dictionary $data]
      ;##
      ;# Now paste it back in
      ;#
      foreach {file} $data {
         $lb insert $current $file
      } ;# end foreach
      $lb selection anchor $current
      $lb selection set [lindex $selection end] [lindex $selection 0]
   }
}
###############################################################################
# Function   : shuffleSelection
# Description: Shuffle the selection
# Author     : Tom Wilkason
# Date       : 12/4/2001
###############################################################################
proc shuffleSelection {{_listBox ::LB}} {
   upvar $_listBox lb
   global TF TA
   set selection [lsort -decreasing -integer [$lb curselection]]
   set current [$lb index anchor]
   if {[llength $selection] > 1} {
      ;##
      ;# Cut the data out and sort it
      ;#
      foreach {ind} $selection {
         lappend data [$lb get $ind]
      } ;# end foreach
      $lb delete [lindex $selection end] [lindex $selection 0]
      set data [shuffle $data]
      ;##
      ;# Now paste it back in
      ;#
      foreach {file} $data {
         $lb insert $current $file
      } ;# end foreach
      $lb selection anchor $current
      $lb selection set [lindex $selection end] [lindex $selection 0]
   }
}
;##
;# Update the number of files in the playlist
;#
proc updateFileDisplay {} {
   global snackAmpSettings LB TA
   array set lookup {
      0  "Sorted"
      1  "Shuffled"
      2  "Shuffled Folders, Sorted Tracks"
      3  "Shuffled Folders, Shuffled Tracks"
   }
   $::SA.hull.label configure -text "[LB::end] Music Files ($lookup($snackAmpSettings(sortMode)))"
}
###############################################################################
# Function   : Wait
# Description: Change the cursor to hourglass if needed
# Author     : Tom Wilkason
# Date       : 1/5/2001
###############################################################################
proc Wait {{how 0}} {
   if {$how} {
      $::SA config -cursor watch
      update
   } else {
      $::SA config -cursor $::TA(arrow)
   }
}
###############################################################################
# Function   : saLog
# Description: Log and event, either to a file or the screen
# Author     : Tom Wilkason
# Date       : 2/8/2001
###############################################################################
proc saLog {Message} {
   set now  [timestamp]
   if {$::snackAmpSettings(logFile)} {
      if {[catch {set fid [open $::TA(logFile) a]} result]} {
         puts $result
         puts "$now $Message"
         return
      }
      puts $fid "$now $Message"
      flush $fid
      close $fid
   } else {
      puts stdout "$now $Message"
   }
}
###############################################################################
# Function   : getTime
# Description:
# Author     : Tom Wilkason
# Date       : 2/1/2003
###############################################################################
proc timestamp {args} {
   set secs [clock seconds]
   set ms [clock clicks -milliseconds]
   set base [expr { $secs * 1000 }]
   set fract [expr { $ms - $base }]
   if { $fract > 1000 } {
      set diff [expr { $fract / 1000 }]
      incr secs $diff
      incr fract [expr { -1000 * $diff }]
   }
   return "[clock format $secs -format %T].[format %03ld $fract]"
}
###############################################################################
# Function   : Exit
# Description: Exit gracefully
# Author     : Tom Wilkason
# Date       : 1/9/2001
###############################################################################
proc taExit {args} {
   global TA
   catch {
      winico taskbar delete $TA(Icon)
   }
   ;##
   ;# Ramp the volume down first if needed
   ;#
   catalog::SavePlaylistCache
   HttpD::Shutdown
   set ::snackAmpSettings(minWidth) [winfo width $::SA]
   catalog::SaveDirtyCache
   catch {APL::close}
   settings::saveSettings
   gainControl::saveSettings
   soundControl::stop
   ;# close any stale tcl temp files
   #set tclTemp [glob -nocomplain -directory $::env(TEMP) TCL*.tmp]
   #foreach file $tclTemp {
   #   catch {file delete -force $file}
   #} ;# end foreach
   ;# close any remaining channels
   foreach {channel} [file channels] {
      catch {close $channel}
   } ;# end foreach
   exit
}
###############################################################################
# Function   : about
# Description: About dialog box
# Author     : Tom Wilkason
###############################################################################
proc snackVersion {args} {
   set results "Snack engine version $::snack::patchLevel"
   if {[catch {package present snackogg} result]} {
      append results "\n   Snack ogg extension not present"
   } else {
      append results "\n   Snack ogg extension $result"
   }
   return $results
}
proc about {args} {
   set Message {
   SnackAmp Music Player Version $::SAVersion
   [snackVersion]
   Tcl/Tk Version [set ::tcl_patchLevel]
   [commify [llength $::TFList]] tracks in [commify [array size ::teFolders]] folders indexed
   Snackamp user folder located at '$::HOME'

   Features:
   - Plays wav, mp3 and ogg files and streams
   - VCR Player controls including sound level
   - Saving and retrieving specific songs in playlists including searching
   - Saving and retrieving music grouped in folders as auto-playlists,
     folders may be checked/unchecked to include/exclude files contained
     within them. Used to group music by taste and play groups as "radio stations"
   - Reorganize your music files, supports drag-n-drop, clipboard. Changes are
     immediately reflected in playlists.
   - Edit ID3 tags, including smart-renaming and smart-tagging.
   - Integrated web server for remote control and serving streaming music
   - Flexible filtering of playlists.
   - snackAmp works well for large number of music files, just organize
     them by folder under one or more directories and select the desired
     folders and save different settings under auto-playlist files.
   - Catalogs and supports removable media.

    (c)Tom Wilkason - enjoy
       tom.wilkason@cox.net
       http://snackamp.sourceforge.net/
   }
   clipboard clear
   clipboard append [subst $Message]
   tk_messageBox -title "SnackAmp Information" -icon info -message [subst $Message]
}
;##
;# Dialog showing the CVS Headers
;#
proc cvsVersions {args} {
   append vinfo "CVS Version Info:\n"
   foreach {_module} [lsort -dictionary [info globals *_History]] {
      global $_module
      upvar $_module module
      foreach {xx file rest} [split $module ":,"] {break}
      foreach {v ver date time user xx yy} $rest {break}
      append vinfo [format "%-5s\t%-10s\t%-60s" $ver $date $file]\n
   } ;# end foreach
   clipboard clear
   clipboard append $vinfo
   tk_messageBox -title "OK to Continue" -icon info -message $vinfo
}
###############################################################################
# Function   : updateBrowser
# Description: Changes the env(BROWSER) entry when the help browser is
# changed. Deletes it if the Value is set to an empty string.
###############################################################################
proc updateBrowser {arrName index op} {
   # Normally we should worry about the array name but since we *know* its
   # snackAmpSettings, we won't.
   global snackAmpSettings

   if {[string length $snackAmpSettings(HelpBrowser)] == 0} {
      # do nothing
   } elseif {[string equal $op u]} {
      if {[info exists ::env(BROWSER)]} {
         unset ::env(BROWSER)
      }
   } else {
      set ::env(BROWSER) $snackAmpSettings(HelpBrowser)
   }
}
###############################################################################
###############################################################################
###############################################################################
###############################################################################
namespace eval macros {
   namespace export *
}
###############################################################################
# Function   : bumpScaling
# Description: Adjust the current scaling for the playing song for some amount
# Author     : Tom Wilkason
# Date       : 1/9/2002
###############################################################################
proc macros::bumpScaling {percent} {
   global TA
   set gain [snack::audio scaling]
   gainControl::add $TA(currentMF) [expr {$gain*(1.0+$percent)}]
   snack::audio scaling [expr {$gain*(1.0+$percent)}]
}
###############################################################################
# Function   : Assign
# Description: Routine to assign global hotkeys, called from hotkeys file
# Author     : Tom Wilkason
###############################################################################
proc macros::Assign {key Command} {
   if {[catch {bind .saTop $key $Command} result]} {
      saLog "FilterAssign Assignment Error: Invalid hotkey '$key' for command $Command\n$::errorInfo"
   }
}
###############################################################################
# Function   : FilterAssign
# Description: Routine to assign filter box hotkeys, called from hotkeys file
# Author     : Tom Wilkason
###############################################################################
proc macros::FilterAssign {key Command} {
   if {[catch {bind $::FI $key $Command} result]} {
      saLog "FilterAssign Assignment Error: Invalid hotkey '$key' for command $Command\n$::errorInfo"
   }
}
###############################################################################
# Function   : PlaylistAssign
# Description: Routine to assign play list hotkeys, called from hotkeys file
# Author     : Tom Wilkason
###############################################################################
proc macros::PlaylistAssign {key Command} {
   if {[catch {bind $::LB $key $Command} result]} {
      saLog "PlaylistAssign Assignment Error: Invalid hotkey '$key' for command $Command\n$::errorInfo"
   }
}
###############################################################################
# Function   : APL_ListAssign
# Description: Routine to assign APL play list hotkeys, called from hotkeys file
# Author     : Tom Wilkason
###############################################################################
proc macros::APL_ListAssign {key Command} {
   if {[APLExists]} {
      set plb [$::PLB bodypath]
      if {[catch {bind $plb $key $Command} result]} {
         saLog "APL_ListAssign Assignment Error: Invalid hotkey '$key' for command $Command\n$::errorInfo"
      }
   }
}
###############################################################################
# Function   : APL_TreeAssign
# Description: Routine to assign APL tree hotkeys, called from hotkeys file
# Author     : Tom Wilkason
###############################################################################
proc macros::APL_TreeAssign {key Command} {
   if {[APLExists]} {
      if {[catch {bind [winfo parent $::TREE] $key $Command} result]} {
         saLog "APL_TreeAssign Assignment Error: Invalid hotkey '$key' for command $Command\n$::errorInfo"
      }
   }
}
###############################################################################
# Function   : APL_TreeAssign
# Description: Routine to assign APL tree hotkeys, called from hotkeys file
# Author     : Tom Wilkason
###############################################################################
proc macros::APL_TreeNodeAssign {key Command} {
   if {[APLExists]} {
      if {[catch {$::TREE bind x $key $Command} result]} {
         saLog "APL_TreeNodeAssign Assignment Error: Invalid hotkey '$key' for command $Command\n$::errorInfo"
      }
   }
}
###############################################################################
# Function   : APLFilterAssign
# Description: Routine to assign APL filter box hotkeys, called from hotkeys file
# Author     : Tom Wilkason
###############################################################################
proc macros::APLFilterAssign {key Command} {
   if {[APLExists]} {
      if {[catch {bind $::SAFind $key $Command} result]} {
         saLog "APLFilterAssign Assignment Error: Invalid hotkey '$key' for command $Command\n$::errorInfo"
      }
   }
}

###############################################################################
# Function   : editPLID3
# Description: Edit the highligted files ID3 info
# Author     : Tom Wilkason
# Date       : 3/18/2002
###############################################################################
proc macros::SelectFile {{playing 0}} {
   if {!([winfo exists .wFC] && [wm state .wFC] == "normal")} {
      AutoPlaylist
   }
   #APL::SelectFile [soundControl::fileName]
   if {$playing} {
      APL::SelectFile [soundControl::fileName]
   } else {
      APL::SelectFile [LB::getActiveFullName]
   }
}
proc macros::EditPLID3 {args} {
   set files [LB::getSelectedFullNames 1]
   if {[llength $files] > $::snackAmpSettings(ID3editLimit)} {
      tk_messageBox -type ok -icon warning \
         -message "Limiting operation to first $::snackAmpSettings(ID3editLimit) files"
   }
   id3Tag::id3Editor [lrange $files 0 $::snackAmpSettings(ID3editLimit)]
}
proc macros::PLNameToTag {smart} {
   id3Tag::id3NameToTag [LB::getSelectedFullNames 1] $smart
}
proc macros::PLTagToName {args} {
   id3Tag::id3TagToName [LB::getSelectedFullNames 1]
}
proc macros::PLTagToName {args} {
   id3Tag::id3TagToName [LB::getSelectedFullNames 1]
}
proc macros::removeTag {args} {
   id3Tag::removeTag [LB::getSelectedFullNames 1]
}
proc macros::EditPlaying {args} {
   id3Tag::id3Editor [list $::TA(currentMF)]
}
###############################################################################
# Function   : macros::Highlight
# Description: macros::Highlight one and only one selection
# Author     : Tom Wilkason
# Date       : 12/27/2000
###############################################################################
proc macros::Highlight {i {inverse 0}} {
   Trace
   global TF TA LB
   $LB selection clear 0 end
   set i [expr {$i < 0 ? 0 : $i}]
   set i [expr {$i >= [LB::end] ? [expr {[LB::end] - 1}] : $i}]
   LB::selectionSet $i
   $LB activate $i
   $LB see $i
   $LB selection anchor $i
   if {$inverse} {
      LB::inverseVideoIndex $i
   } ;# end if
   set TA(entryName) [LB::getFullName $i 0]
   catch {
      if {$::snackAmpSettings(syncAPL)} {
         APL::SelectFile $TA(entryName)
      }
   }
}
###############################################################################
# Function   : Swap
# Description: Move a selection up/down the list (bind keys to this)
# Author     : Tom Wilkason
# Date       : 10/29/2001
###############################################################################
proc macros::Swap {how} {
   global LB
   if {[llength [LB::selected]] > 0} {
      set selected [lsort -integer [LB::selected]]
      set first [lindex $selected 0]
      set last [lindex $selected end]
      if {$how=="up"} {
         ;##
         ;# Move the item above the selection below the selection
         ;#
         set moveIndex [expr {$first-1}]
         if {$moveIndex >= 0} {
            LB::move $moveIndex $last
         } ;# end if
      } else {
         ;##
         ;# Move the item below the selection above the selection
         ;#
         set moveIndex [expr {$last+1}]
         if {$moveIndex < [LB::end]} {
            LB::move $moveIndex $first
         } ;# end if
      }
   }
}
###############################################################################
# Function   : PlayGain/getPlayGain
# Description: Provide x2 gain control to better match audio experience
# Author     : Tom Wilkason
# Date       : 10/19/2001
###############################################################################
proc macros::PlayGain {value} {
   set newValue [expr {round(pow($value,2)/40.0)}]
   set ::TA(line,$::snackAmpSettings(lineSelect)) $newValue
}

###############################################################################
# Function   : macros::getPlayGain
# Description: Return the current gain setting log scaled
# Author     : Tom Wilkason
###############################################################################
proc macros::getPlayGain {args} {
   set audioGain $::TA(line,$::snackAmpSettings(lineSelect))
   return [expr {pow($audioGain*40.,0.5)}]
}
###############################################################################
# Function   : bumpLevel
# Description: Adjust the current play level by some amount
# Author     : Tom Wilkason
# Date       : 11/9/2001
###############################################################################
proc macros::bumpLevel {factor} {
   set ::TA(gain) [expr {$::TA(gain)+$factor}]
   PlayGain $::TA(gain)
}
###############################################################################
# Function   : Console
# Description: Toggle the console for debugging purposes (WIN32 only)
# Author     : Tom Wilkason
# Date       : 12/27/2000
###############################################################################
proc macros::Console {args} {
   global TF TA LB
   if {$TA(Console)==0} {
      console hide
   } else {
      console show
   }
}
###############################################################################
# Function   : applyCurrentAPL
# Description: Apply the latest Auto Playlist settings, ignoring any filtering
#              in effect.
# Author     : Tom Wilkason
# Date       : 11/8/2001
###############################################################################
proc macros::applyCurrentAPL {{resort 1}} {
   ;# Re-sort if required, otherwise use the existing sorting order
   soundControl::latency long
   if {$resort} {
      filtering::sortMasterList        ;# create a new master sorted/shuffled list
   } ;# end if
   update
   set restore $::TA(filter)
   set ::TA(filter) {}
   filtering::filterMusic replace 0 ;# This applies the current APL (resetting the playlist)
   set ::TA(filter) $restore
   soundControl::latency normal
}
###############################################################################################
# Function   : RenameFileIP
# Description: Rename a file directly from the listbox
# Author     : Tom Wilkason
# Date       : 6/19/2002
###############################################################################
proc macros::RenameFile {{fullName {}}} {
   set W $::LB
   if {[string length $fullName] == 0} {
      set fullName [LB::getActiveFullName]    ;# lookup long name
   }
   set oldName [file tail $fullName]
   set dirName [file dirname $fullName]
   set newName $oldName
   ;##
   ;# Block on dialog
   ;#
   set name [catalog::renameInPlaceDialog $W 0 $oldName]
   ;##
   ;# Back from window closure
   ;#
   if {[string length $::TA(rename)]>0 && $::TA(rename) != $oldName} {
      #puts "Renaming '$oldName' to '$::TA(rename)'"
      set newFullName [file join $dirName $::TA(rename)]
      if {[catch {catalog::RenameFile $fullName $newFullName} result]} {
         tk_messageBox -icon warning -message "Could not rename\n$fullName to \n$newFullName\n$result"
      } else {
         catalog::RenameEntry $fullName $newFullName
      }
   }
}
####################################################################
# Function   : macros::RenameFile
# Description: Rename a file using a dialog, called from key binding or menu
# Author     : Tom Wilkason
# Date       : 10/30/2001
###############################################################################
proc macros::RenameFileDialog {{fullName {}}} {
   if {[string length $fullName] == 0} {
      set fullName [LB::getActiveFullName]    ;# lookup long name
   }
   set oldName [file tail $fullName]
   set dirName [file dirname $fullName]
   set newName $oldName

   set answer [getEntryDialog "$oldName" newName]
   if {$answer == "ok" && $newName != $oldName} {

      set newFullName [file join $dirName $newName]
      if {[catch {catalog::RenameFile $fullName $newFullName} result]} {
         tk_messageBox -icon warning -message "Could not rename\n$fullName to \n$newFullName\n$result"
      } else {
         catalog::RenameEntry $fullName $newFullName   ;# update full sorted list
      }
   }
}

###############################################################################
# Function   : SaveFolderFile
# Description: Save the current folder settings to a file
# Author     : Tom Wilkason
###############################################################################
proc macros::SaveFolderFile {{fName ""}} {
   global TF TA LB teFolders snackAmpSettings
   ;##
   ;# If a list for GlobData is passed in, the second item should be the name to save as (if different)
   ;#
   if {$fName == ""} {
      set types [list [list "Auto-Playlists Files" .tafs] [list "All Files" "*"]]
      set fName [tk_getSaveFile -title "Save Auto-Playlists File" -filetypes $types \
         -initialdir $::HOME -defaultextension .tafs]
      catch {focus $ThisWin}
      if { $fName == "" } {
         return  ""
      }
   }
   if {[catch {open $fName w} InFile]} {
      set result [tk_messageBox -type ok -title "SnackAmp - Please Respond" -icon warning \
         -message "Save Error: Unable to open '$fName' for writing"]
      return ""
   }
   fconfigure $InFile -buffersize 100000

   set snackAmpSettings(lastFolderFile) $fName
   set TA(AutoPlayListSelection) [file rootname [file tail $fName]]
   if {$InFile!=0} {
      foreach {File} [lsort [array names teFolders]] {
         puts $InFile [list $File $teFolders($File)]
      } ;# end foreach
      close $InFile
      set ::TA(APLDirty) 0
      set snackAmpSettings(lastFolderFile) $fName
      return $fName
   } else {
      return -1
   }
}

###############################################################################
# Function   : removeAPLFile
# Description: Remove an APL file, called from menu callback
# Author     : Tom Wilkason
###############################################################################
proc macros::removeAPLFile {fName} {
   set result [tk_messageBox -type yesno -title "Please Answer" \
      -icon question -message "Are you sure you want to remove APL setting '[file root [file tail $fName]]'?"]
   if {[string equal $result "yes"]} {
      if {[catch {file delete -force $fName} result] } then {
         tk_messageBox -type ok -title "Error" \
            -icon warning -message "Failed to remove '[file root [file tail $fName]]'\n$result"
      }

   } ;# end if
}
###############################################################################
# Function   : removePLFile
# Description: Remove an playlist file, called from menu callback
# Author     : Tom Wilkason
###############################################################################
proc macros::removePLFile {fName} {
   set result [tk_messageBox -type yesno -title "Please Answer" \
      -icon question -message "Are you sure you want to playlist '[file root [file tail $fName]]'?"]
   if {[string equal $result "yes"]} {
      if {[catch {file delete -force $fName} result] } then {
         tk_messageBox -type ok -title "Error" \
            -icon warning -message "Failed to remove '[file root [file tail $fName]]'\n$result"
      }

   } ;# end if
}
###############################################################################
# Function   : readAPLFile
# Description: Read a file and change the favorate folders
# Author     : Tom Wilkason
###############################################################################
proc macros::readAPLFile {fName {filter 0}} {
   Trace
   global TA TF TFList snackAmpSettings
   global teFolders
   ;# Save a corresponding playlist file
   set lf $::snackAmpSettings(lastFolderFile)
   if {$::snackAmpSettings(savePlonApply) && [file exists $lf]} {
      set plfile "[file root $lf].m3u"
      file delete -force $plfile
      SavePlaylistFile $::LB replace all $plfile
   }
   update
   if {[file exists $fName]} {
      if {[catch {
         set fid [open $fName r]
      } result] \
      } then {
         tk_messageBox -icon warning -type ok -title "Open Error" -message $result
         return
      }
      fconfigure $fid -buffersize 100000
      set full [split [read $fid] \n]
      close $fid
      ;# Invalidate all current settings
      foreach {Index} [array names teFolders] {
         set teFolders($Index) -1
      }
      ;# Only update values if an entry already exists from prior scan
      ;# This prevents garbage if directories were moved
      foreach line $full {
         foreach {Index Value} $line {
            if {[info exists teFolders($Index)]} {
               set teFolders($Index) $Value
            } ;# endif
         } ;# end foreach
      } ;# end foreach
      ;# Some folders will still be tagged with -1 from
      ;# above invalidation, meaning and they weren't in the autoplaylist
      ;# file. If so, then set them to the value of their parent if possible.
      ;# Set subfolders to a 2 if they have children with checkes
      foreach {Index} [lsort [array names teFolders]] {
         setTeDefault $Index
      } ;# end foreach
      catch {APL::refreshTree}
      if {$filter} {
         applyCurrentAPL
      } else {
         LB::tryRestorePlaying
      }
   } else {
      ;# If no default file create one with everything one
      ;##
      ;# Create the default Folder list based on files
      ;# New ones are turned off by default. teFolders is
      ;# not cached separately.
      ;#
      foreach {Index} [catalog::sortedList] {
         while {1} {
            set Index [file dirname $Index]
            set teFolders($Index) 1
            if {$Index == [file dirname $Index]} {
               break
            }
         }
      } ;# end foreach


      SaveFolderFile $fName
      applyCurrentAPL
      waitWindow::close
      set result [tk_messageBox -type yesno -title "SnackAmp - Please Respond" -icon question \
         -message "All of your Auto-playlist folders containing music are set true \nby default, would you like to change them now?"]
      if {$result == "yes"} {
         AutoPlaylist
      } ;# end if

   }
   set snackAmpSettings(lastFolderFile) $fName
   set TA(APLDirty) 0
   set TA(AutoPlayListSelection) [file rootname [file tail $fName]]

}
###############################################################################
# Function   : saveCurrentAPL
# Description: Save and then apply the current Auto-playlist
# Author     : Tom Wilkason
# Date       : 11/8/2001
###############################################################################
proc macros::saveCurrentAPL {{apply 1}} {
   global snackAmpSettings
   SaveFolderFile $snackAmpSettings(lastFolderFile)
   if {$apply} {
      applyCurrentAPL 0
   } ;# end if
   set ::TA(APLDirty) 0
}
###############################################################################
# Function   : Clear
# Description: Clear the playlist
# Author     : Tom Wilkason
# Date       : 12/27/2000
###############################################################################
proc macros::Clear {args} {
   if {[tk_messageBox -icon question -type yesno -title "Please Respond" \
      -message "Are you sure you want to clear the playlist"]=="yes"} {
      global TF TA LB
      LB::deleteAll
      updateFileDisplay
   } ;# end if
}
###############################################################################
# Function   : OpenFile
# Description: Open a single file for playing, add it to play list and play it
# Author     : Tom Wilkason
###############################################################################
proc macros::OpenFile {args} {
   global TF TA LB
   if {[catch {
      ;# This isn't in earlier snack versions. Also needs
      ;# Tcl 8.43+ to work, therefore the catch.
      set files [snack::getOpenFile -format MP3 -multiple 1]
   } result] } then {
      set files [snack::getOpenFile -format MP3]
   }
   ;# Here is a funky way to tell if a list of files was returned
   ;# Look at the string rep.
   if {![string match "\{*" $files]} {
      filtering::InsertPlay $files
   } else {
      if {[llength $files] > 0} {
         foreach {file} $files {
            filtering::Append $file
         } ;# end foreach
         ;# And play the first one
         filtering::InsertPlay [lindex $files 0]
      } ;# end if
   }
}
;##
;# Open up a URL, save a shortcut and play it
;#
proc macros::OpenURL {args} {
   if {[playurl::checkuskfolder]} {
      set shortcut [playurl::addPromptUrl $::snackAmpSettings(uskDir)]
      filtering::InsertPlay $shortcut
   }
}
###############################################################################
# Function   : macros::Append
# Description: Wrapper for filtering::Append for macros & registry
# Author     : Tom Wilkason
###############################################################################
proc macros::Append {file} {
   filtering::Append $file
}
###############################################################################
# Function   : macros::PlayResume
# Description: Play button binding, resume play if paused otherwise
#              play active
# Author     : Tom Wilkason
###############################################################################
proc macros::PlayResume {args} {
   switch -- [soundControl::playState] {
      "Playing" {
         ;# Restart
         Play
      }
      "Paused" {
         ;# Resume play
         Pause
      }
      "Stopped" {
         ;# Play
         Play
      }
   } ;# End Switch
}
###############################################################################
# Function   : Play
# Description: Play a list item. Bound to Play key but used elsewhere.
#              Updates the GUI with the song info. Allows an offset within
#              the song to be passed in.
# Author     : Tom Wilkason
###############################################################################
proc macros::Play {{What active} {offset 0}} {
   Trace
   global TF TA LB SS snackAmpSettings
   ;##
   ;# Determine if the current active index
   ;# is a valid song. If so play it. Otherwise, flag it and jump to the next one.
   ;#
   if {$snackAmpSettings(NextAfterCurrnt)} {
      set i $LB::indexPlaying
      if {$i < 0} {
         set i [$LB index active]
      }
   } else {
      set i [$LB index active]
   }
   ;##
   ;# Based on prev, current or next position cursor to proper
   ;# file
   ;#
   switch -- $What {
      current {
         Stop
         if {$i == ""} {
            set i 0
         }
      }
      active {
         ;# Always use active (e.g double click)
         Stop
         set i [$LB index active]
         if {$i == ""} {
            set i 0
         }
      }
      next {
         ;# Only called, on automatic transition to next song
         ;# Can be called just before other song complete.
         ;# Don't issue a stop to keep the snack queue flowing.
         LB::inverseVideoFile xx 0
         set soundControl::playState "Stopped"
         killAfter Timer
         $::SS set 0
         if {$i == ""} {
            set i 0
         } else {
            incr i
         }
         ;# check for continuous play
         if {$i >= [LB::end]} {
            if {$snackAmpSettings(ContinuousPlay)} {
               set i 0
               ;# Resort on a wrap if some form of random play
               if {$snackAmpSettings(sortMode) > 0} {
                  filtering::sortMusic
               } ;# end if
            }
         } ;# end if
      }
      prev {
         if {$i == ""} {
            set i 0
         } elseif {$i==0} {
            set i [LB::end]
         } else {
            if {$i > 0} {
               incr i -1
            } ;# end if
         }
      }
      default {
         saLog "Invalid subcommand :Play $What"
      }
   } ;# End Switch
   set Selection [$LB get $i]
   ;# Check file existance
   if {$Selection=={}} {
      return
   }
   ;# currentMF is song selected for play
   set TA(currentMF) $TF($Selection)
   set SelectionName $TF($Selection)
   ;##
   ;# If file does not exist then assume it needs to be mounted
   ;#
   if {! [file exists $TF($Selection)]} {
      ;# Try to translate/mount removable media
      set tryName [catalog::TranslateMount $TF($Selection)]
      if {[string length $tryName] == 0} {
         ;# macros::Highlight this file then skip to the next one
         ;# if we haven't been commanded to stop
         saLog "Play: Could not validate '$TF($Selection)' as valid file"
         $LB activate $i
         set LB::indexPlaying $i
         $LB itemconfigure $i -bg darkgray
         if {$::TA(STOP) == 0} {
            after idle Next
            ;# Next will clear background so delay
            after 200 [list $LB itemconfigure $i -bg darkgray]
         }
         return
      } else {
         ;# Translate was sucessfull
         #set TA(currentMF) $tryName
         set SelectionName $tryName
      }
   }
   set ::errorInfo ""
   ;# File exists, inverse video it showing it is playing
   LB::inverseVideoFile $TA(currentMF) 1
   ;##
   ;# Reset the sound object, swapping between channels/files can
   ;# cause it to crash
   ;#
   soundControl::reset
   ;##
   ;# If this is a url to play, then play it directly
   ;#
   if {[playurl::isusk $SelectionName]} {
      foreach {index} [array names ::songArray] {
         set ::songArray($index) ""
      } ;# end foreach
      id3Tag::id3Label $SelectionName ::songArray

      set ::songArray(Title) "Connecting ... $::songArray(Title)"
      set ::songArray(Album) "$SelectionName"
      if {[catch {playurl::playusk [soundControl::sound] $SelectionName [getNextCommand] ::songArray} result] } then {
         saLog $::errorInfo
         set ::songArray(Title) "Error: $result"
         after 200 [list eval [getNextCommand]]
         return
      } else {
         soundControl::pauseRelief 0
         set max [expr {3600*24}]      ;# 24 hours
         set ::TA(songDuration) $max
         if {$::TA(showTicks)} {
            $SS configure -to $max -tickinterval [expr {3600*4}]
         } else {
            $SS configure -to $max
         }
         balloon::define $::SA.panel.stat.l.dtTitle $SelectionName
         if {[winfo exists .miniPlayer]} {
            balloon::define .miniPlayer.tt.title $SelectionName
         } ;# end if
         catch {$::mpSS configure -to $max}
         set songTitle [string map {// /} [file rootname [join [lrange [file split $TA(currentMF)] end-1 end] /]]]
         wm title $::SA "$songTitle"
         set TA(songTitle) $songTitle

         if {![afterExists Timer]} {
            Timer
         } ;# end if
      }
      return
   }
   ;##
   ;# It seems faster on play -start to have the short file name and cd to that directory
   ;# Also, UNC names are really slow (on Win2K anyway)!
   ;# Strangly, network drive mappings are also fast (if not using UNC)
   ;#
   cd [file dirname $SelectionName]

   if {[catch {soundControl::fileName $SelectionName} result]} {
      tk_messageBox -icon warning -type ok -title "sound error" \
         -message "Error creating sound object using $SelectionName\n$result\n$::errorInfo"
      return
   }

   set Info [soundControl::sinfo]
   ;##
   ;# Check for an invalid header
   ;#
   if {[lindex $Info 0] == -1 || [lsearch -exact $::TA(upcaseExtensions) [lindex $Info 6]] < 0} {
      soundControl::fileName ""
      saLog "Invalid file: $SelectionName\n $Info"
      set ::songArray(Title) "Invalid header:$Selection"
      update
      if {$snackAmpSettings(AutoDelete)} {
         Delete
      } ;# end if
      after 1000 [list eval [getNextCommand]]
      return
   }
   ;##
   ;# See if user changed the slider to adjust where to play
   ;# will reset any fading that happened
   ;#
   PlayGain $TA(gain)
   ;##
   ;# Adjust gain if the song has an auto level setting (and it is enabled)
   ;#
   if {$snackAmpSettings(autoLevel)} {

      set value [gainControl::createIfNeeded $SelectionName s]

      ;# Clamp db Gain at runtime
      if {[catch {toDb $value} db] } then {
         set db 0.0
      } else {
         if {$db > $::snackAmpSettings(maxGain)} {
            set db $::snackAmpSettings(maxGain)
         } elseif {$db < $::snackAmpSettings(minGain)} {
            set db $::snackAmpSettings(minGain)
         }
      }
      snack::audio scaling [fromDb $db]
   } else {
      snack::audio scaling 1.0
   }
   ;##
   ;# Pull out the ID3 info for the titles then launch the play
   ;#
   if {[catch {
      ;# #
      ;# Get ID3 and detailed song info and display it, if ID3 not available
      ;# Use file name as song title
      set data [id3Tag::songDetails $SelectionName [soundControl::sound] ::songArray]
      if {$::songArray(ID3)=={}} {
         set ::songArray(Title) [stringTitle [file rootname [file tail $TA(currentMF)]]]
      }
      set rate [lindex $::songArray(rate) 0]
      if {$rate < $::TA(minRate)} {
         set ::songArray(Title) "Skipping $SelectionName"
         if {$::TA(STOP)==0} {
            after idle Next
         }
      } else {
         ;# Register balloon help
         set TA(songBalloon) $data
         balloon::define $::SA.panel.stat.l.dtTitle $data
         if {[winfo exists .miniPlayer]} {
            balloon::define .miniPlayer.tt.title $data
         } ;# end if
         ;# Fires Next command at end of this file
         soundControl::play -start $offset -command [getNextCommand]
      }
   } result] \
   } then {
      set ::songArray(Title) "Error: $SelectionName"
      snack::audio selectOutput $::snackAmpSettings(outputDevice)
      saLog "Play: $result\n$::errorInfo"
      after 1000 [list eval [getNextCommand]]
      return
   }
   ;# Only log this if going to a file
   if {$::snackAmpSettings(logFile)} {
      saLog "Playing $TA(currentMF)"
   }
   ;##
   ;# What ~10 ticks on 5 or multiple sec intervals
   ;#
   set trackTime  [soundControl::duration]
   if {$TA(showTicks)} {
      set Interval [expr {(round($trackTime/3.3)/10)*10}]
      if {$Interval == 0} {
         set Interval 10
      } ;# end if
      $SS configure -to $trackTime -tickinterval $Interval
   } else {
      $SS configure -to $trackTime
   }
   ;# Miniplayer (catch if not up)
   if {[info exists ::mpSS] && [winfo exists $::mpSS]} {
      $::mpSS configure -to $trackTime
   }
   ;##
   ;# Song Title is shown in the title bar
   ;#
   set songTitle [string map {// /} [file rootname [join [lrange [file split $TA(currentMF)] end-1 end] /]]]
   ;# If the folder name is the same as the first part of file name,
   ;# Then we can ignore the folder name
   if {[string match -nocase "[file dirname $songTitle]*" [file tail $songTitle]]} {
      set songTitle [file tail $songTitle]
   }
   wm title $::SA "$songTitle"
   set TA(songTitle) $songTitle

   #set TA(songInfo) "$Details(fSize) - $Details(fDate) - $Details(rate)"
   set TA(songDuration) $::songArray(length)
   if {$snackAmpSettings(autoSavePlaylist)} {
      catalog::SavePlaylistCache
   } ;# end if
   if {![afterExists Timer]} {
      Timer
   } ;# end if
}
;##
;# Play next sound file in list
;#
proc macros::Next {} {
   Trace
   Stop
   after 0 Play next
}
;##
;# Play previous sound file in list
;#
proc macros::Prev {} {
   Trace
   Stop
   after 0 Play prev
}
;##
;# Stop playback of a file and cancle the update timer
;#
proc macros::Stop {args} {
   Trace
   global TF TA LB
   killAfter Timer
   if {[snack::audio active]} {
      if {$::snackAmpSettings(fadeFast)} {
         soundControl::fadeFast 1
      } else {
         soundControl::stop
      }
      $::SS set 0
   } ;# end if
   soundControl::pauseRelief 2
   playurl::cleanupLast
   LB::inverseVideoFile xx 0
}
;##
;# Move the cursor to the currently playing file,
;# assign to <Home> key
;#
proc macros::locatePlaying {args} {
   LB::seePlaying
}

;##
;# Pause/resume playback
;#
proc macros::Pause {args} {
   Trace
   global TF TA LB
   if {[snack::audio active]} {
      if {$soundControl::paused==1} {
         soundControl::pause 0
         Timer
      } else {
         killAfter Timer
         if {$::snackAmpSettings(fadePause)} {
            soundControl::fadeFast 0 1
         } ;# end if
         soundControl::pause 1
      }
   } ;# end if
}
;##
;# Callback to remove the current selections from the display
;#
proc macros::Cut {{playNext 0}} {
   Trace
   global TF TA LB
   set playActive 0
   if {[llength [LB::selected]] > 0} {
      set Index [LB::anchor]
      foreach {ind} [lsort -decreasing -integer [LB::selected]] {
         set file $TF([$LB get $ind])
         if {$file == $TA(currentMF)} {
            set playActive [soundControl::isPlaying]
            if {$playActive} {
               Stop
            } ;# end if
         } ;# end if
         LB::delete $ind
      } ;# end foreach
      ;# Stay on 0 if that is what we were on
      ##macros::Highlight [expr {$Index-0}]
      updateFileDisplay
      ;# Set selection to current active
      $LB selection set [$LB index active]
      ;# This will play the next song
      if {$playNext && $playActive} {
         Play
      } ;# end if
   }
}
;##
;# Cut the listbox to the clibboard
;# (only used for main playlist)
proc macros::clipboardCut {lb {append 1}} {
   Trace
   global TF TA
   if {[llength [$lb curselection]] > 0} {
      clipboardCopy $lb $append
      Cut 1
      updateFileDisplay
   }
}
;##
;# Copy the listbox files to the clipboard
;#
proc macros::clipboardCopy {lb {append 1}} {
   Trace
   global TF TA
   if {[llength [$lb curselection]] > 0} {
      if {$append == 0} {
         clipboard clear
      } else {
         clipboard append " "
      }
      foreach {ind} [lsort -decreasing -integer [$lb curselection]] {
         switch -glob -- $lb {
            ".saTop*"  {lappend data $TF([$lb get $ind])}
            ".wFC*"    {lappend data [aplLB::rowIndex2file $ind]}
         } ;# End Switch
      } ;# end foreach
      clipboard append $data
   }
}
;##
;# Paste the listbox files to the playlist
;#
proc macros::clipboardPaste {lb {clear 1}} {
   Trace
   global TF TA TFList
   if {[catch {selection get -selection CLIPBOARD} files]} {
      return
   }
   set current [$lb index anchor]
   set processed [list]
   foreach {file} $files {
      ;# Only insert files that are in the master list
      ;# And not already processed (i.e. no duplicates)
      ;# TFW: move to shared
      if {[lsearch -exact $processed $file] < 0 && [catalog::whereInDb $file] >= 0} {
         set keyName [LB::getListName $file]
         $lb insert $current $keyName
         set TF($keyName) $file
         lappend processed $file
      }  else {
         #puts "skipping $file"
      }
   }
   if {$clear} {
      clipboard clear
   } ;# end if
   updateFileDisplay
}
;##
;# Callback to remove the current selections from the disk
;#
proc macros::Delete {{Next 1}} {
   Trace
   global TF TA LB
   set goNext 0
   if {[llength [LB::selected]] > 0} {
      if {$::snackAmpSettings(dontMoveFiles)} {
         tk_messageBox -icon warning -type ok \
            -message "You don't have permission to remove files"
         return
      }
      set Index [LB::anchor]

      set selection [LB::selected]
      foreach {ind} $selection {
         lappend Files [$LB get $ind]
      }

      set result [tk_messageBox -type yesno -title "Please Answer" \
         -icon question -message "Are you sure you want to delete:\n[join $Files \n]?"]
      if {$result == "yes"} {
         foreach {Entry} $Files {
            set File $TF($Entry)
            set goNext [catalog::DeleteFile $File]
         } ;# end foreach
         ##macros::Highlight $Index
      }
   }
   ;##
   ;# Play the next song if we deleted the current one and we want to
   ;#
   if {$Next && $goNext} {
      Play
   } ;# end if
   updateFileDisplay
   return $goNext
}
###############################################################################
# Function   : macros::selectExclustions
# Description: Bring up the GUI to select the exclusion tree
# Author     : Tom Wilkason
# Date       : 1/9/2001
###############################################################################
proc macros::AutoPlaylist {args} {
   if {![winfo exists .wFC]} {
      package require saTree
      package require autoPlayList
      $::TA(treeButton) configure -relief sunken
      ;# First make sure we have the proper roots for the tree viewer
      APL::FolderTree
      update  ;# let it draw
      ;# May not have a selection
      catch {APL::SelectFile $::TA(currentMF)}
   } elseif {[wm state .wFC] != "normal"} {
      $::TA(treeButton) configure -relief sunken
      APL::FolderTree
      update  ;# let it draw
      catch {APL::SelectFile $::TA(currentMF)}
   } else {
      wm withdraw .wFC
      $::TA(treeButton) configure -relief flat
   }
}
###############################################################################
# Function   : macros::findNextSong
# Description: Locate the next occurance of a song in the playlist based on the
#              current filter value
# Author     : Tom Wilkason
# Date       : 11/08/2001
###############################################################################
proc macros::findNextSong {{top 0}} {
   global TF TA LB TFList snackAmpSettings CF
   ;# Get the current playlist
   set regPattern [list]                                ;#retain wildcards
   set pattern [split [string map {" or " "|" " OR " "|" "*" ".*" "?" "."} $TA(filter)] |]
   foreach part $pattern {
      lappend regPattern "($part)+"
   }
   set regPattern [join $regPattern |]
   if {$top} {
      set i -1
   } else {
      set i [lindex [LB::selected] 0]
      if {$i=={}} {
         set i -1
      } ;# end if
   } ;# end if
   set ok 2
   ;##
   ;# Allow one wrap around if at first not found
   ;#
   if {[catch {
      while {$ok} {
         incr i
         foreach {Entry} [$LB get $i end] {
            if {[regexp -nocase $regPattern $TF($Entry)]} {
               macros::Highlight $i
               return
            } ;# end if
            incr i
         } ;# end foreach
         set i -1
         incr ok -1
      }
   } result]} {
      #saLog "findNextSong $result"
   }
}
###############################################################################
# Function   : macros::playSpecificSong
# Description: Locate and Play a specific song in the playist, ignore if not in playlist
# Author     : Tom Wilkason
# Date       : 4/28/2002
###############################################################################
proc macros::playSpecificSong {song} {
   Trace
   set i 0
   set i [lsearch -exact [LB::getAllNames] $song]
   if {$i >= 0} {
      LB::activate $i
      Play active
   } ;# end if
}
###############################################################################
# Function   : macros::clearFilter
# Description:
# Author     : Tom Wilkason
# Date       : 11/9/2001
###############################################################################
proc macros::clearFilter {args} {
   set ::TA(filter) {}
}
proc macros::selectAll {args} {
   $::LB selection set 0 end
}
###############################################################################
# Function   : macros::filterMusic
# Description: Restrict files shown based on the filter & folder settings
# Author     : Tom Wilkason
# Date       : 12/26/2000
###############################################################################
proc macros::filterMusic {how {what def}} {
   filtering::filterMusic $how $what
}
###############################################################################
# Function   : macros::SavePlaylistFile
# Description: Save current files in list
# Author     : Tom Wilkason
###############################################################################
proc macros::SavePlaylistFile {ListBox {how "replace"} {what "selected"} {fName {}}} {
   Trace
   global TF TA LB
   ;##
   ;# If a list for GlobData is passed in, the second item should be the name to save as (if different)
   ;#
   if {$how=="replace"} {
      set access "w"
   } else {
      set access "a+"
   } ;# end if
   ;# Use different dialog depending on what you want to do
   array set tk_dialog {
      replace tk_getSaveFile
      append tk_getOpenFile
   }
   if {$what == "selected"} {
      set Indicies [LB::selected]
   } else {
      set Indicies [LB::all]
   }
   if {[string length $fName]==0} {
      set types [list [list "Playlist files" {.taf .m3u}] [list "All Files" "*"]]

      set fName [$tk_dialog($how) -title "Save [llength $Indicies] tracks to Playlist File" -filetypes $types \
         -initialdir $::HOME -defaultextension .m3u]
      catch {focus $ThisWin}
      if { $fName == "" } {
         return  ""
      }
   } elseif {[file exists $fName]} {
      if {$how=="replace" && [tk_messageBox -icon question -type yesno -title "Overwrite File" -message "Is it ok to $how $fName with [llength $Indicies] tracks?"]!="yes"} {
         return {}
      } ;# end if
   }
   if {[catch {open $fName $access} InFile]} {
      tk_messageBox -icon warning -type ok -title "File Error" -message "Save Error: Unable to open '$fName' for writing"
      return
   }
   set TA(PlayListSelection) [file rootname [file tail $fName]]
   set TA(PlayListFile) $fName
   set ext [file extension $fName]
   if {[string equal $ext .m3u]} {
      puts $InFile "#\n# Playlist file: [file nativename $fName]"
      puts $InFile "# Saved on [clock format [clock seconds]] by user $::tcl_platform(user) on host [info hostname]\n#"
   }
   if {$InFile!=0} {
      ;##
      ;# Process either selected or entire list
      ;#
      set i 1
      set names [list]
      if {$ListBox == $::LB} {
         foreach {Index} $Indicies {
            set File $TF([$LB get $Index])
            lappend names $File
            if {[string equal $ext .m3u]} {
               set File [file nativename $File]
            }
            puts $InFile $File
         } ;# end foreach
      } elseif {$ListBox == $::PLB} {
         if {$what == "selected"} {
            set names [aplLB::getSelectedFullNames 0]
         } else {
            set names [aplLB::getAllNames 1]
         }
         foreach {File} $names {
            if {[string equal $ext .m3u]} {
               set File [file nativename $File]
            }
            incr i
            puts $InFile $File
         } ;# end foreach
      } else {
         saLog "unknown listbox '$ListBox'"
      } ;# end if
      # TFW: Only save if playing is in list
      if {[lsearch -exact $names $::TA(currentMF)]>= 0} {
         puts $InFile "#Playing=$::TA(currentMF)"
      }
      close $InFile
      return $fName
   } else {
      return -1
   }
}
###############################################################################
# Function   : macros::ReadPlaylistFile
# Description: Verify playlist file and notify user of any problems
# Author     : Tom Wilkason
###############################################################################
proc macros::ReadPlaylistFile {fName {how "replace"}} {
   ;# If file doesn't exist, then prompt for it
   if {$fName == {} || ![file exists $fName]} {
      set types [list [list "Playlist files" {.taf .m3u}] [list "All Files" "*"]]
      set fName [tk_getOpenFile -title "Open Playlist File" -filetypes $types \
         -initialdir $::HOME]
      if { $fName == "" } {
         return  ""
      }
   }
   set failed [ProcessPlaylistFile $fName $how 1]
   set numFailed [llength $failed]
   if {$numFailed > 0} {
      ;##
      ;# Tell user how many failed (if any)
      ;#
      if {$numFailed > 10} {
         set failed [lrange $failed 0 9]
         lappend failed "...."
      } ;# end if
      tk_messageBox -icon warning -type ok -title "Invalid Files" \
         -message "A total of $numFailed files do no exist, and could not be added to the playlist\n[join $failed \n]"
   } ;# end if
}
###############################################################################
# Function   : macros::ProcessPlaylistFile
# Description: Read a file and append/replace current list
# Author     : Tom Wilkason
###############################################################################
proc macros::ProcessPlaylistFile {fName {how "replace"} {resume 0}} {
   Trace
   global TF TA LB snackAmpSettings
   if {[catch {
      set fid [open $fName r]
   } result] \
   } then {
      tk_messageBox -icon warning -type ok -title "Open Error" -message $result
      return
   }
   set full [split [string trim [read $fid]] \n]
   close $fid
   if {$how == "replace"} {
      LB::deleteAll
   } ;# end if
   ;# Insert the files if they are not already in the list
   ;# Translate Removable media if necessary.
   ;# Also support winamp format playlists, this means no files can
   ;# start with # or else they will be ignored (that is a comment in m3u)
   set here [file dirname $fName]
   set push [pwd]
   cd $here
   set comment "\#*"
   set failed [list]
   Wait 1
   set i 0
   set cp [LB::active]
   foreach file $full {
      incr i
      if {! [string match $comment $file]} {
         ;# This will be ignored if abs path names
         set file [file join $here $file]
         if {[set rfile [catalog::findIfExists $file]] != {} } {
            if {$how=="insert"} {
               filtering::Insert [incr cp] $rfile
            } else {
               filtering::Append $rfile
            }
         } else {
            lappend failed $file
         }
         if {$i%200==0} {update}   ;# let the music play
      } else {
         if {$resume && [string match "#Playing*" $file]} {
            set resumeFile [lindex [split $file =] 1]
            if {$::snackAmpSettings(restorePlaying) && [file exists $resumeFile]} {
               set ans [tk_messageBox -type yesno -icon question -message "Do you want to resume playing of \n$resumeFile"]
               if {$ans=="yes"} {
                  after 100 [list PlayFile $resumeFile]
               }
            }
         }
      } ;# end if
   } ;# end foreach
   Wait 0
   set TA(PlayListFile) $fName
   set TA(PlayListSelection) [file rootname [file tail $fName]]
   cd $push
   updateFileDisplay
   return $failed
}
###############################################################################
# Function   : soundControl namespace
# Description: All snack related calls are in here
# Author     : Tom Wilkason
# Date       : 6/28/2002
###############################################################################
namespace eval soundControl {
   variable paused  0
   variable butRelief
   variable butState
   variable startOffset 0.0
   variable playState "Stopped"
   variable fadeIn
   variable fadeOut
   array set butRelief {0 flat   1 sunken 2 flat}
   array set butState {0 normal 1 normal 2 disabled}
   ;# Try and load ogg vorbis support
   if {![catch {package require snackogg} result]} {
      lappend extTypes  {OGG .ogg}
      lappend loadTypes {{Ogg Vorbis Files} {.ogg}}
      lappend loadKeys  OGG
      lappend saveTypes {{Ogg Vorbis Files} {.ogg}}
      lappend saveKeys  OGG
      snack::addExtTypes $extTypes
      snack::addLoadTypes $loadTypes $loadKeys
      snack::addSaveTypes $saveTypes $saveKeys
   }
   ;# Create the snack sound objects
   snack::sound s
   ;# Some filters for fading
   #set fadeIn [snack::filter fade in linear 1000]
   #set fadeOut [snack::filter fade out linear 2000]

   if {[llength [snack::mixer lines]] == 0} {
      tk_messageBox -icon warning -type ok -message "There are no available output devices"
      exit
   } ;# end if
}
###############################################################################
# Function   : soundControl::latency
# Description: Adjust latency for long operations, normally a big latency
#              makes the cpu usage jump due to the rapid decode that fills
#              the output buffer with audio.
# Author     : Tom Wilkason
# Date       : 10/29/2001
###############################################################################
proc soundControl::latency {{how normal}} {
   if {$how == "normal"} {
      snack::audio playLatency $::snackAmpSettings(normalPlayLatency)
   } else {
      snack::audio playLatency $::snackAmpSettings(longPlayLatency)
   }
   update
}
###############################################################################
# Function   : soundControl::length
# Description: Return sample length of sound object
# Author     : Tom Wilkason
###############################################################################
proc soundControl::length {args} {
   return [s length]
}
###############################################################################
# Function   : soundControl::sinfo
# Description: Return info for sound object
# Author     : Tom Wilkason
###############################################################################
proc soundControl::sinfo {args} {
   return [s info]
}
###############################################################################
# Function   : soundControl::duration
# Description: Return duration of sound object
# Author     : Tom Wilkason
###############################################################################
proc soundControl::duration {args} {
   return [s length -units seconds]
}
###############################################################################
# Function   : soundControl::reset
# Description: Recreate the sound object
# Author     : Tom Wilkason
###############################################################################
proc soundControl::reset {args} {
   Trace
   if {[info command s]=="s"} {
      s stop
      s destroy
   }
   snack::sound s
}
###############################################################################
# Function   : playState
# Description: Return the play state
# Author     : Tom Wilkason
###############################################################################
proc soundControl::playState {args} {
   variable playState
   return $playState
}
###############################################################################
# Function   : calibrateAudio
# Description: Command line utility function to calibrate a bunch of songs.
# Author     : Tom Wilkason
# Date       : 11/10/2001
###############################################################################
proc soundControl::calibrateAudio {args} {
   set done 0
   set total 0
   ;##
   ;# Compute the task
   ;#
   foreach {file} [catalog::sortedList] {
      if {[gainControl::exists $file]} {
         incr done
      }
      incr total
   }
   if {($total-$done) == 0} {
      return
   }
   if {[tk_messageBox \
-message "Are you sure you want to auto-level [commify [expr {$total-$done}]]  tracks?
No track files will be modified.\n\nThis may take a while..." \
-icon question -type yesno] != "yes"} {
      return
   }
   soundControl::latency long
   waitWindow::show "Checking Existing Files"
   set ::BREAK 0
   set ww ${waitWindow::W}.stop
   if {![winfo exists $ww]} {
      pack [button $ww -text "Cancel" -relief raised -bd 2 -font "Helvetica 12 bold" -command {set ::BREAK 1}] \
      -side left
   } ;# end if
   update
   ;##
   ;# Now do the work
   ;#
   set i 0
   update
   foreach {file} [catalog::sortedList] {
      if {$::BREAK==1} {
         gainControl::saveSettings
         destroy $ww
         waitWindow::close
         return
      } ;# end if
      set baseFileName [file tail $file]
      if {![gainControl::exists $file]} {
         if {![file exists $file]} {
            #saLog "Skipping $file"
            continue
         } ;# end if
         waitWindow::show "[commify $done] of [commify $total] : $baseFileName"
         update
         gainControl::create $file
         if {$i%10==0} {
            gainControl::saveSettings
         } ;# end if
         incr done
         update ;# Let music play
      }
      incr i
   } ;# end foreach
   destroy $ww
   soundControl::latency
   waitWindow::close
}
;##
;# Return the current play position
;#
proc soundControl::position {args} {
   variable startOffset
   set stream [expr {$startOffset + [snack::audio elapsedTime]}]
   #set stream [expr {$startOffset + [s current_position -units sec]}]
   return $stream
}
proc soundControl::samplePosition {args} {
   set rate [s cget -rate]
   set stream [expr {round([position]*$rate)}]
   return $stream
}
proc soundControl::sound {args} {
   return "s"
}
proc soundControl::fileName {{name "xxx"}} {
   if {![string equal $name "xxx"]} {
      s configure -file $name -debug $::snackAmpSettings(Trace)
   }
   return [s cget -file]
}
###############################################################################
# Function   : isPlaying
# Description: Determine if a song is playing or not
# Author     : Tom Wilkason
# Date       : 10/30/2001
###############################################################################
proc soundControl::isPlaying {args} {
   variable playState
   ;# Note audo active is true till the latency period expires
   ;# which means the buffer is flushed. We want to know when
   ;# the buffer has been filled, not flushed.
   if {[string equal $playState "Playing"]} {
      return 1
   } else {
      return 0
   }
}
###############################################################################
# Function   : soundControl::fadeFast
# Description: Routine to quickly ramp the volume down
# Author     : Tom Wilkason
###############################################################################
proc soundControl::fadeFast {{stop 0} {restore 0}} {
   Trace
   set line $::snackAmpSettings(lineSelect)
   set level $::TA(line,$line)
   set ff $::snackAmpSettings(fadeFactor)
   ;# Don't let it blow up
   if {$ff >=1.0} {
      set $ff 0.85
   }
   if {[snack::audio active]} {
      set fname [soundControl::fileName]
      ;# While this song is still playing fade the level down.
      while {$::TA(line,$line) > 0 && [snack::audio active] && [string equal $fname [soundControl::fileName]]} {
         set ::TA(line,$line) [expr {int($::TA(line,$line)*$ff)}]
         after 20
         update ;# keep playing song, note a new song could have transitioned here
      } ;# end while
      ;##
      ;# If stop requested, only do so if still on same song after fade
      ;#
      if {$stop && [string equal $fname [soundControl::fileName]]} {
         ;# restore level and stop
         soundControl::stop
         set  ::TA(line,$line) $level
      } elseif {$restore} {
         set  ::TA(line,$line) $level
      }
   }
}
###############################################################################
# Function   : soundControl::pauseRelief
# Description: If paused, push the pause button up/down. Called from pause/stop
#              If stopped, then disable the button
# Author     : Tom Wilkason
###############################################################################
proc soundControl::pauseRelief {down} {
   Trace
   global TA
   variable butState
   variable butRelief
   $TA(pauseButton) configure -state $butState($down) -relief $butRelief($down)
   if {[winfo exists .miniPlayer]} {
      .miniPlayer.frame.snackPause configure \
         -state $butState($down) -relief $butRelief($down)
   }
}
###############################################################################
# Function   : soundControl::pause
# Description: Toggle pausing of a sound object
# Author     : Tom Wilkason
###############################################################################
proc soundControl::pause {how} {
   Trace
   variable paused
   variable playState
   if {[snack::audio active]} {
      pauseRelief $how
      snack::audio pause
      #s pause
      set paused [expr {!$paused}]
      if {$paused} {
         set playState "Paused"
      } else {
         set playState "Playing"
      }
   }
}
;##
;# Function to stop sound and return after it is verifyied stopped
;#
proc soundControl::stop {args} {
   Trace
   variable playState
   variable paused
   if {[snack::audio active]} {
      pauseRelief 2
      s stop
   }
   set paused 0
   set playState "Stopped"
}
###############################################################################
# Function   : soundControl::play
# Description: Encapsulate the play
#              Called from the command line.
# Author     : Tom Wilkason
# Date       : 11/18/2001
###############################################################################
proc soundControl::play {xx start args} {
   Trace
   variable playState
   variable paused
   variable fadeIn
   variable fadeOut
   variable startOffset
   ;##
   ;# In snack 2.2, the output buffer is kept full,
   ;# however at about 6 hours expr pow(2,31)/(2*44100*3600)=6.7
   ;# an internal int counter rolls over and causes the stream
   ;# to stop. So we simply block until the
   ;# output buffer flushes to cleanly restart
   ;# the counter.
   ;# What we do here is force the queue to flush
   ;# It more that an hours worth of music has
   ;# played.
   ;#  3600
   while {[snack::audio elapsedTime] > 3600.0} {
      #saLog "current_pos=[format %5.3f [s current_position -units sec]], elapsedTime=[format %5.3f [snack::audio elapsedTime]], position=[format %5.3f [soundControl::position]]"
      update
      #saLog "Bleeding Queue"
      after 100
   }
   # -filter $fadeIn
   if {[catch {eval s play -start $start $args} result] } then {
      saLog "$result\n$::errorInfo"
   } else {
      pauseRelief 0
      set paused 0
      set playState "Playing"
      #saLog "Playing at [s cget -file] at [snack::audio elapsedTime]"
   }
   ;# Save position in seconds of start of song, offset it by any elapsedTime
   ;# that may be active from a queued play. Typically elapsedTime will be 0.
   set startOffset [expr {round($start/[s cget -rate]-[snack::audio elapsedTime])}]
   #saLog "startOffset = [snack::audio elapsedTime]"     ;# debug
   #set startOffset [expr {round($start/[s cget -rate]-[s current_position -units sec])}]
}
###############################################################################
# Function   : soundControl::logPower
# Description: Compute the averge powerlevel of the song for level equalization
# Author     : Tom Wilkason
# Date       : 11/10/2001
###############################################################################
proc soundControl::logPower {snd {seconds 10}} {
   Trace
   ;# Power command works like this:
   ;# windowlengh is area within each block to keep a level^2 total
   ;# for each sample in each window, power is squared and accumlated
   ;# Number of blocks to do is the end-start samples divded by the frame
   ;# length. The frame lengh is specificed in seconds, but internally
   ;# converted to samples. For example
   ;# start = 100,000, end = 320,000 (in samples), framelength = 1 (44.1k samples)
   ;# A total of 220,000/44100 5 blocks will be spaced in region. Each block
   ;# will be computed on starting at the beginning of each block for the
   ;# window length (e.g. initial 8192 samples of a 44000 sample block)
   ;# Start after any possible header (framelen is in seconds, window is
   ;# put at start of each frame. num frames is total width/frame len (in samples)
   ;# at 10 seconds, or 441,000 samples. A window length of 8192 will sample ~ 2%
   set trackLen [$snd length -units seconds]
   set samples [expr {$trackLen/double($seconds)}]
   ;# Make 60 a setting, also the 68 below (along with snack debug level)
   set max $::snackAmpSettings(autoLevelMaxSamples)
   if {$samples > $max} {
      ;# Scale back time if over 60 samples
      set seconds [expr {$seconds*double($samples)/$max}]
   }
   set Totals [list]
   if {[catch {
      set levels [$snd power -start 100000 -end [expr {[$snd length]-10000}] \
         -framelength $seconds \
         -windowlength $::snackAmpSettings(autoLevelWindow)]
      foreach level $levels {
         if {[string is double $level]} {
            lappend Totals $level
         } ;# end if
      } ;# end foreach
      ;# Sort so we can toss out low samples, then take the average of the rest
      set llTotals [llength $Totals]
      if {$llTotals > 0} {
         set Totals [lsort -decreasing -dictionary $Totals]
         #puts -nonewline "\n[soundControl::fileName] : $Totals "
         # Grab the top ones for consideration (we are concerned with max power)
         #debug "[$snd cget -file]\n  $Totals"
         #debug "[$snd cget -file]\n  [llength $Totals] samples taken"
         if {$llTotals > 12} {
            set end 4 ;#5
         } elseif {$llTotals > 6} {
            set end 2
         } else {
            set end [expr {$llTotals/2}]
         }
         set Totals [lrange $Totals 0 $end]
         set average [expr ([join $Totals +]) / double([llength $Totals])]
         ;# Compute offset, 6dB is a double. 68dB is a reference for song that won't clip
         ;# The absolute peak is about 80 but will only find it if entire file is scanned.
         set offset [expr {pow(10.,(68.0 + $::snackAmpSettings(autoLevelReference) - $average)/$::snackAmpSettings(gainFactor))}]   ;# 20 is theory, 33 doesn't expand as much
      } else {
         set offset 1.0
      } ;# end if
      ;# Round to 2 decimal places
      set value [expr {round($offset*1000.)/1000.}]
   } result]} {
      tk_messageBox -icon warning -type ok -message "Error: $result\n$::errorInfo"
      set value 1.0
   }
   return $value
}
###############################################################################
# Function   : toDb/fromDb
# Description: Convert a gain factor from and to Db
# Author     : Tom Wilkason
# Date       : 7/29/2002
###############################################################################
proc toDb {value} {
   return [expr {log10($value)*$::snackAmpSettings(gainFactor)}]
}
proc fromDb {value} {
   return [expr {pow(10.,$value/$::snackAmpSettings(gainFactor))}]
}
##
# Set some widget defaults before we build any guis
#
set tk_strictMotif 1
option add *Button.activeBackground       yellow
option add *Button.borderWidth            1
option add *Button.font                   menu
option add *Button.pady                   0
option add *Button.relief                 flat
option add *Checkbutton.borderWidth       1
option add *Checkbutton.font              menu
option add *Dialog.msg.font menu           ;# For non-windows systems
option add *Dialog.msg.wrapLength 9i       ;# For non-windows systems
option add *Entry.borderWidth             1
option add *Entry.font                    entry
option add *Entry.relief                  sunken
option add *Label.borderWidth             1
option add *Label.font                    entry
option add *Label.relief                  flat
option add *Listbox.borderWidth           1
option add *Listbox.font                  listbox
option add *Menu.font                     menu
option add *Menu.tearoff                  0
if {$tcl_platform(platform)=="unix"} {
   option add *Menu.borderWidth              1
   option add *Menu.activeBorderWidth        1
} ;# end if
option add *Menubutton.borderWidth        1
option add *Menubutton.font               menu
option add *Menubutton.highlightThickness 2
option add *Menubutton.pady               1
option add *Menubutton.relief             flat
option add *Radiobutton.font              menu
option add *Scale.borderWidth             1
option add *Scale.activeBackground        blue
option add *Scale.font                    timeScale
option add *highlightThickness            0

###############################################################################
# Fire it up, do init stuff with settings and make smart decisions based on
# exsistence of various settings files. Also do Windows specific Icon stuff.
###############################################################################
# Make sure init files are in correct place
if {![file exists $::HOME]} {
   if {[catch {
      file mkdir $::HOME
   } result] \
   } then {
      tk_messageBox -message "Sorry! I could not create the folder '$::HOME'
to save your settings file. You can do one of three things:
1. Set the environment variable SNACKAMPHOME to some folder you can write to
2. Run SnackAmp with a '-home location' option where you can write to location
3. Gain write access to $::HOME

Restart SnackAmp when you are ready" -type ok -title "Press OK when done" -icon warning
   exit
      #error
   }
} ;# end if
;# We really don't have to call upgradeMoveHome anymore
;# Set/read the rest of the defaults, prompt user if they don't exist
namespace import macros::*
settings::readSettings
;# Some of these may have changed
option add *Entry.background              $::snackAmpSettings(ebg)
option add *Entry.foreground              $::snackAmpSettings(efg)
option add *Scale.troughColor             $::snackAmpSettings(bg)
option add *Listbox.background            $::snackAmpSettings(bg)
option add *Listbox.foreground            $::snackAmpSettings(fg)
;# Do special stuff for help browser
if {[string length $snackAmpSettings(HelpBrowser)] > 0} {
   set env(BROWSER) $snackAmpSettings(HelpBrowser)
}
trace variable snackAmpSettings(HelpBrowser) wu updateBrowser
set TA(key,copy)  [cleanKeyName [lindex [event info <<Copy>>]  0]]
set TA(key,paste) [cleanKeyName [lindex [event info <<Paste>>] 0]]
set TA(key,cut)   [cleanKeyName [lindex [event info <<Cut>>]   0]]
gainControl::restoreSettings
;# Build the GUI
buildGui
soundControl::stop
fixScaleBinding
;##
;# Bring up the GUI now and set the width to the prior saved width
;#
;# Read the various cache files and restore playlist if needed.
update idletasks
wm geometry $::SA $::snackAmpSettings(minWidth)x[winfo reqheight $::SA]
snack::debug $::snackAmpSettings(Trace)
s config -debug $::snackAmpSettings(Trace)

wm deiconify $::SA
catalog::initCacheFiles
;# Monitor stuff if required
after 5000 catalog::MonitorScanFolders
after 10000 catalog::MonitorCacheFolders
;# quietly mount any removable media
if {$snackAmpSettings(UseRemovableFolders)} {
   after 1000 catalog::MountMedia "xx" 1
} ;# end if
soundControl::latency normal
;##
;# Restore mini-player if required
;#
if {$::snackAmpSettings(restoreMiniPlayer)} {
   if {$::snackAmpSettings(activeMiniPlayer)} {
      foreach {x y} $::snackAmpSettings(positionMiniPlayer) {break}
      package require miniPlayer
      toggleMiniPlayer $x $y
      unset x
      unset y
   }
} else {
   set ::snackAmpSettings(activeMiniPlayer) 0
}

;# Fire up the web server
package require HttpD
package require urlRedirect

;##
;# If windows then install the winico icons
;# command line arguments
if {$tcl_platform(platform) == "windows"} {
   iconSupport
} else {
   ;##
   ;# Play any files that are passed in (on Windows
   ;# these are handled by the dde server)
   ;#
   myIcon $::SA
   if {[info exists argv] && [llength $argv] > 0} {
      set last [LB::end]
      foreach {file} $argv {
         if {[file extension $file]=="" || [regexp -nocase {\.m3u|\.pls} $file]} {
            PlayFile $file
            return
         }
         if {[file exists $file]} {
            # puts "Appending $file"
            filtering::Append $file
         }
      } ;# end foreach
      ;# If some valid files were added, then play the first one
      set here [LB::end]
      if {$here > $last} {
         incr last
         macros::Highlight $last
         after 200 Play
      } ;# end if
   } ;# end if
}
waitWindow::close
